#include "ch.h"

#include "hal/led.h"
#include "hal/usart1.h"
#include "hal/eth.h"

#include "util/boot.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/terminal.h"
#include "util/sys_time.h"

#include "cli.h"
#include "network.h"
#include "wifi.h"
#include "vision.h"
#include "rst.h"

static WORKING_AREA(_ccm_waCLI, 8192);
static WORKING_AREA(_ccm_waNetwork, 4096);
static WORKING_AREA(_ccm_waWifi, 4096);
static WORKING_AREA(_ccm_waTerminal, 4096);
static WORKING_AREA(_ccm_waRST, 2048);

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	// Enable important clocks
	RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN | RCC_AHB1ENR_DMA2EN | RCC_AHB1ENR_GPIOAEN
			| RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_GPIODEN
			| RCC_AHB1ENR_GPIOEEN;
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;

	// Enable compensation cell to reduce noise on power lines with 50MHz/100MHz I/Os
	SYSCFG->CMPCR = SYSCFG_CMPCR_CMP_PD;

	// Enable usage fault, bus fault, and mem fault
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	SystemClockInitData clkInit;
	clkInit.pll.M = 25;
	clkInit.pll.N = 336;
	clkInit.pll.P = 2;
	clkInit.pll.Q = 7;
	clkInit.HSEBypass = 1;
	clkInit.APB1Div = 4;	// = 42MHz
	clkInit.APB2Div = 4;	// = 42MHz
	clkInit.RTCDiv = 12;	// = 1MHz
	clkInit.flashLatency = 5;
	clkInit.sysTickFreq = CH_FREQUENCY;
	SystemClockInit(&clkInit, 168000000UL);

	chSysInit();

	// set vector table offset
#ifdef ENV_BOOT
	SCB->VTOR = 0;
#else
	SCB->VTOR = BOOT_APP_OFFSET;
#endif

	CPULoadCalibrate(10);
	chThdSleep(20);

	// --- HAL ---
	LEDInit();		// init LEDs to signal RTOS startup failures
	FlashFSInit(0x08060000, 0x20000);	// 128k
	USART1Init();
	CRCInit();

	// --- High-Level Systems / PART 1 ---
	TerminalInit(&usart1.inQueue, &usart1.outQueue);
	CLIInit();

#ifdef ENV_BOOT
	ConsolePrint("\f--- Base Station v2 (Bootloader) ---\r\n");
#else
	ConsolePrint("\f--- Base Station v2 ---\r\n");
#endif

	// --- High-Level Systems / PART 2 ---
	RSTInit();
	VisionInit();
	EthInit();
	NetworkInit();
	WifiInit();

	chThdCreateStatic(_ccm_waWifi, sizeof(_ccm_waWifi), NORMALPRIO+10, WifiTask, 0);
	chThdCreateStatic(_ccm_waNetwork, sizeof(_ccm_waNetwork), NORMALPRIO+5, NetworkTask, 0);
	chThdCreateStatic(_ccm_waTerminal, sizeof(_ccm_waTerminal), NORMALPRIO-2, TerminalTask, 0);
	chThdCreateStatic(_ccm_waCLI, sizeof(_ccm_waCLI), NORMALPRIO-5, CLITask, 0);
	chThdCreateStatic(_ccm_waRST, sizeof(_ccm_waRST), NORMALPRIO+1, CLITask, 0);
/*
	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	while(1)
	{
#ifdef ENV_RUN
		LEDToggle(LED_GREEN);
#else
		LEDToggle(LED_RED);
#endif

		chThdSleep(MS2ST(500));

		if(SysTimeUSec() - wifi.lastCall > 200000)
			BootReset();
	}

	return 0;
}
