/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"
#include "util/boot.h"
#include "intercom_constants.h"
#include "hal/led.h"
#include "util/intercom.h"
#include "hal/kicker.h"
#include "supply.h"
#include "hal/motor.h"
#include "util/sys_time.h"
#include "util/flash.h"
#include "util/flash_fs.h"
#include "util/crc.h"
#include "kick.h"
#include "cpu_load.h"
#include "util/network_print.h"
#include "util/log.h"

#include <string.h>

CLI cli;

static void processCommand();

void CLIInit()
{
	chMtxInit(&cli.procBufMutex);

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

	ConsoleIntercomAdapterInit(&cli.adapter, INTERCOM_ADDRESS_KD, INTERCOM_ADDRESS_MAIN);
}

msg_t CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

#ifdef ENV_BOOT
	ConsolePrint("--- Bootloader (KD) ---\r\n");
#else
	ConsolePrint("--- KD ---\r\n");
#endif

	while(1)
	{
		chThdSleepMilliseconds(20);

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}
	}

	return 0;
}

static void processCommand()
{
	uint16_t u16;
	uint32_t u32[3];
	uint16_t vref;
	int32_t i32;

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("test2"))
	{
		ConsolePrint("0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\r\n");
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("Name            rPrio\tTime\r\n");

		for(Thread* pThread = chRegFirstThread(); pThread != 0; pThread = chRegNextThread(pThread))
		{
			ConsolePrint("%-16s%u\t%u\r\n", pThread->p_name, pThread->p_realprio, pThread->p_time);
		}
	}

	if(ConsoleCmpCmd("cpu"))
	{
		CPULoad load = cpuLoad;

		ConsolePrint("CPU Load: %3u.%1u%%\r\n", load.load/10, load.load%10);
		ConsolePrint("Max: %3u.%1u%%\r\n", load.maxLoad/10, load.maxLoad%10);
		ConsolePrint("incPerTick: %10u/%u\r\n", load.incPerSysTick, load.maxIncPerSysTick);

		chThdSleep(MS2ST(10));
		cpuLoad.maxLoad = 0;
	}

	if(ConsoleScanCmd("vmax %hu", &u16) == 1)
	{
		ConsolePrint("vMax: %hu\r\n", u16);

		KickerSetConfig(u16, kicker.config.cooldown);
	}

	if(ConsoleScanCmd("chg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable charging\r\n");
			KickerEnableCharge(1);
		}
		else
		{
			ConsolePrint("Disable charging\r\n");
			KickerEnableCharge(0);
		}
	}

	if(ConsoleScanCmd("autochg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable autocharging\r\n");
			KickerEnableAutoRecharge(1);
		}
		else
		{
			ConsolePrint("Disable autocharging\r\n");
			KickerEnableAutoRecharge(0);
		}
	}

	if(ConsoleCmpCmd("autodischg"))
	{
		ConsolePrint("Enabling auto discharging\r\n");
		KickerAutoDischarge();
	}

	if(ConsoleScanCmd("kstress %hu", &u16) == 1)
	{
		ConsolePrint("Setting auto-fire stress test: %hu\r\n", u16);

		kicker.stress = ((float)u16)/1000.0f;
	}

	if(ConsoleScanCmd("dis1 %hu", &u16) == 1)
	{
		ConsolePrint("Fire1: %hu\r\n", u16);

		KickerFire1(((float)u16)/1000.0f);
	}

	if(ConsoleScanCmd("arm1 %hu", &u16) == 1)
	{
		ConsolePrint("Arm1: %hu\r\n", u16);

		KickArmIr(((float)u16)/1000.0f, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleScanCmd("dis2 %hu", &u16) == 1)
	{
		ConsolePrint("Fire2: %hu\r\n", u16);

		KickerFire2(((float)u16)/1000.0f);
	}

	if(ConsoleScanCmd("arm2 %hu", &u16) == 1)
	{
		ConsolePrint("Arm2: %hu\r\n", u16);

		KickArmIr(((float)u16)/1000.0f, KICKER_DEVICE_CHIP);
	}

	if(ConsoleCmpCmd("boot"))
	{
		BootSetBootloaderSelected(1);
		BootReset();
	}

	if(ConsoleCmpCmd("kstat"))
	{
		ConsolePrint("vCap: %d, iChg: %d\r\n", (int32_t)kicker.vCap, (int32_t)kicker.iChg);
		ConsolePrint("State: %hu, Auto: %hu\r\n", (uint16_t)kicker.state, (uint16_t)kicker.autoCharge);
	}

	if(ConsoleCmpCmd("kadc"))
	{
		for(uint8_t i = 0; i < SPI1_ADC_SAMPLES*2; i++)
			ConsolePrint("%hu ", kicker.rxData[i]);

		ConsolePrint("\r\n");
	}

	if(ConsoleScanCmd("cooldown %u", &u32[0]) == 1)
	{
		KickerSetConfig(kicker.config.maxVoltage, u32[0]);

		ConsolePrint("Kicker Cooldown set to %ums\r\n", kicker.config.cooldown);
	}

	if(ConsoleCmpCmd("iradc"))
	{
		uint16_t irData[NUM_IR_SAMPLES];
		memcpy(irData, supply.irData, sizeof(uint16_t)*NUM_IR_SAMPLES);
		ConsolePrint("IRADC: ");
		for(uint16_t i = 0; i < NUM_IR_SAMPLES; i++)
			ConsolePrint("%hu ", irData[i]);
		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("ir"))
	{
		ConsolePrint("IR: %hu/%hu\r\n", supply.irResult[0], supply.irResult[1]);
	}

	if(ConsoleCmpCmd("supply"))
	{
		ConsolePrint("Supply: %hu/%hu\r\n", (uint16_t)(supply.bat*1000), (uint16_t)(supply.shunt*1000));
	}

	if(ConsoleScanCmd("auto com %hu", &u16) == 1)
	{
		ConsolePrint("Auto-Commutation: %hu\r\n", u16);

		MotorEnableAutoCommutation(u16);
	}

	if(ConsoleScanCmd("speed ctrl %hu", &u16) == 1)
	{
		ConsolePrint("Speed control: %hu\r\n", u16);

		MotorEnableSpeedPID(u16);
	}

	if(ConsoleScanCmd("vol %hu", &vref) == 1)
	{
		ConsolePrint("Setting dribbler voltage to: %humV\r\n", vref);

		MotorSetVoltage(vref*0.001f);
	}

	if(ConsoleScanCmd("rpm %i", &i32) == 1)
	{
		ConsolePrint("Setting dribbler to %i RPM\r\n", i32);

		MotorSetSpeed((float)i32);
	}

	if(ConsoleScanCmd("pwm %i", &i32) == 1)
	{
		ConsolePrint("Setting motor PWM value: %i\r\n", i32);

		pMotor->pwmVal = i32;
	}

	if(ConsoleCmpCmd("adc"))
	{
		for(uint8_t a = 0; a < 2; a++)
		{
			ConsolePrint("ADC%hhu: ", a);
			for(uint8_t i = 0; i < DRIVE_ADC_SAMPLES; i++)
			{
				ConsolePrint("%5hu ", pMotor->adc[a][i]);
			}
			ConsolePrint("\r\n");
		}
	}

	if(ConsoleCmpCmd("dstat"))
	{
		ConsolePrint("Avg: %dmA\r\n", (int32_t)(motor.avgCurrent.value*1000.0f));
		ConsolePrint("Temp: %d\r\n", (int32_t)(motor.temp*1000.0f));
		ConsolePrint("RPM: %d\r\nOverload: %hu\r\ndT: %d\r\n", (int32_t)pMotor->curVelocity, (uint16_t)motor.pidSpeed.overload, pMotor->comDt);
		ConsolePrint("PID output: %d\r\n", (int32_t)(motor.pidSpeed.output*1e3f));
		ConsolePrint("Bias: %d\r\n", (int32_t)(motor.biasCurrent.value*1000.0f));
//		ConsolePrint("iA: %d, iB: %d, iC: %d\r\n", (int32_t)(pMotor->i[0]*1000), (int32_t)(pMotor->i[1]*1000), (int32_t)(pMotor->i[2]*1000));
//		ConsolePrint("iAlpha: %d, iBeta: %d\r\n", (int32_t)(pMotor->iAlpha*1000), (int32_t)(pMotor->iBeta*1000));
//		ConsolePrint("iD: %d, iQ: %d\r\n", (int32_t)(drib.pOut->iD*1000), (int32_t)(drib.pOut->iQ*1000));
	}

	if(ConsoleScanCmd("ema cur %d", &i32) == 1)
	{
		ConsolePrint("Current EMA Filter alpha: 0.%06d\r\n", i32);

		motor.avgCurrent.alpha = i32*1e-6f;
	}

	if(ConsoleCmpCmd("hall"))
	{
		ConsolePrint("Hall: %hu\r\n", pMotor->halPos);
	}

	if(ConsoleCmpCmd("drib"))
	{
		ConsolePrint("RPM: %d\r\nOverload: %hu\r\ndT: %d\r\n", (int32_t)pMotor->curVelocity, (uint16_t)motor.pidSpeed.overload, pMotor->comDt);
	}

	if(ConsoleCmpCmd("dt"))
	{
		ConsolePrint("comDt: %i\r\n", pMotor->comDt);
	}

	if(ConsoleScanCmd("pid %u %u %u", u32, u32+1, u32+2) == 3)
	{
		ConsolePrint("%u, %u, %u\r\n", u32[0], u32[1], u32[2]);

		PIDSetParams(&motor.pidSpeed, u32[0]*1e-3f, u32[1]*1e-3f, u32[2]*1e-3f);
	}

	if(ConsoleScanCmd("mtest %hu", &vref) == 1)
	{
		ConsolePrint("Motor hall test\r\n");

		float hallAngles[7][2];

		MotorEnableSpeedPID(0);
		MotorEnableAutoCommutation(0);
		MotorEnablePhase(CCER_ABC);

		uint16_t hall = pMotor->halPos;
		float v = ((float)vref)/1000.0f;

		for(float a = 0; a < 360; a += 1.0f)
		{
			MotorSetVoltageAngle(v, a);
			chThdSleep(MS2ST(20));

			if(hall != pMotor->halPos)
			{
				hall = pMotor->halPos;
				ConsolePrint(".");

				hallAngles[hall][0] = a;
			}
		}

		for(float a = 360; a > 0; a -= 1.0f)
		{
			MotorSetVoltageAngle(v, a);
			chThdSleep(MS2ST(20));

			if(hall != pMotor->halPos)
			{
				hall = pMotor->halPos;
				ConsolePrint(".");

				hallAngles[hall][1] = a;
			}
		}

		MotorEnableSpeedPID(1);
		MotorEnableAutoCommutation(1);

		const uint16_t index2Angle[] = {0, 300, 60, 0, 180, 240, 120};

		ConsolePrint("\r\nResults:\r\n");

		for(uint16_t i = 1; i < 7; i++)
		{
			ConsolePrint("%3hu - %3hu - %3hu\r\n", (uint16_t)hallAngles[i][0], index2Angle[i], (uint16_t)hallAngles[i][1]);
		}
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	if(ConsoleCmpCmd("crc32"))
	{
		uint8_t data[12];
		uint32_t crc;
		uint32_t check;

		for(uint8_t i = 0; i < 12; i++)
			data[i] = (i+1)*4;

		crc = CRC32CalcChecksum(data, 8);
		ConsolePrint("8: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 8, crc);
		ConsolePrint("8: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 7);
		ConsolePrint("7: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 7, crc);
		ConsolePrint("7: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 6);
		ConsolePrint("6: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 6, crc);
		ConsolePrint("6: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 5);
		ConsolePrint("5: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 5, crc);
		ConsolePrint("5: 0x%08X\r\n", check);
	}

	if(ConsoleCmpCmd("netprint"))
	{
		NetworkPrint("Hello Network - KD\n");
	}

	if(ConsoleScanCmd("log %hu", &u16) == 1)
	{
		LogPrintAllRecent(u16);
	}
}
