/*
 * kicker.c
 *
 *  Created on: 20.02.2014
 *      Author: AndreR
 */

#include "kicker.h"
#include "../constants.h"
#include "util/console.h"
#include "util/init_hal.h"

#include "intercom_constants.h"
#include "../intercom1.h"

#define STATE_IDLE	0
#define STATE_CHG	1
#define STATE_HYST	2
#define STATE_FIRE	3

#define EVENT_CHG_STOP		0
#define EVENT_CHG_START		1
#define EVENT_VMAX_REACHED	3
#define EVENT_BELOW_HYST	4
#define EVENT_FIRE_DONE		5
#define EVENT_KICK_ERR 		6
#define EVENT_FIRE_START1	7
#define EVENT_FIRE_START2	8

#define LOW3 (TIM_CCMR2_OC3M_2 | TIM_CCMR2_OC3M_0)
#define LOW4 (TIM_CCMR2_OC4M_2 | TIM_CCMR2_OC4M_0)
#define HIGH3 (TIM_CCMR2_OC3M_2)
#define HIGH4 (TIM_CCMR2_OC4M_2)
#define PWM3 (TIM_CCMR2_OC3M_2 | TIM_CCMR2_OC3M_1)
#define PWM4 (TIM_CCMR2_OC4M_2 | TIM_CCMR2_OC4M_1)

typedef union _u32u16
{
	uint32_t u32;
	uint16_t u16[2];
} u32u16_t;

Kicker kicker = { .config = { .maxVoltage = 0, .cooldown = 100 } };

static uint16_t getChgVal()
{
	float chg = (730.0f*kicker.vCap+2350.0f)/(kicker.vCap+47.0f);
//	float chg = (kicker.vCap*2.5f+50.0f);
	if(chg < 50.0f)
		chg = 50.0f;
	if(chg > 600.0f)
		chg = 600.0f;

	return (uint16_t)chg;
}

static void enableChg(uint8_t enable)
{
	if(enable)
		TIM17->CCR1 = getChgVal();
	else
		TIM17->CCR1 = 0;
}

static void engagePhaser(uint32_t eventIn)
{
	uint32_t event = (eventIn & 0xFF);

	if(chTimeElapsedSince(kicker.lastKickDone) < kicker.config.cooldown)
		return; 	// prevent double kick

	switch(event)
	{
		case EVENT_FIRE_START1:
		{
			enableChg(0);
			kicker.state = STATE_FIRE;

			TIM4->CCMR2 = LOW4 | PWM3;	// configure outputs
			TIM4->ARR = (eventIn & 0xFFFF0000) >> 16;	// set auto-reload value
			TIM4->CR1 |= TIM_CR1_CEN;
		}
		break;
		case EVENT_FIRE_START2:
		{
			enableChg(0);
			kicker.state = STATE_FIRE;

			TIM4->CCMR2 = LOW3 | PWM4;
			TIM4->ARR = (eventIn & 0xFFFF0000) >> 16;	// set auto-reload value
			TIM4->CR1 |= TIM_CR1_CEN;
		}
		break;
		default: break;
	}

	++kicker.kickCounter;
}

static void processEvent(uint32_t eventIn)
{
	uint32_t event = (eventIn & 0xFF);

	switch(kicker.state)
	{
		case STATE_IDLE:
		{
			switch(event)
			{
				case EVENT_CHG_START:
				{
					enableChg(1);
					kicker.state = STATE_CHG;
				}
				break;
				case EVENT_FIRE_START1:
				case EVENT_FIRE_START2:
					engagePhaser(eventIn);
				break;
				default: break;
			}
		}
		break;
		case STATE_CHG:
		{
			switch(event)
			{
				case EVENT_CHG_STOP:
				{
					enableChg(0);
					kicker.state = STATE_IDLE;
				}
				break;
				case EVENT_FIRE_START1:
				case EVENT_FIRE_START2:
					engagePhaser(eventIn);
				break;
				case EVENT_VMAX_REACHED:
				{
					enableChg(0);
					kicker.state = STATE_HYST;
				}
				break;
				default: break;
			}
		}
		break;
		case STATE_HYST:
		{
			switch(event)
			{
				case EVENT_CHG_STOP:
				{
					kicker.state = STATE_IDLE;
				}
				break;
				case EVENT_BELOW_HYST:
				{
					enableChg(1);
					kicker.state = STATE_CHG;
				}
				break;
				case EVENT_FIRE_START1:
				case EVENT_FIRE_START2:
					engagePhaser(eventIn);
				break;
				default: break;
			}
		}
		break;
		case STATE_FIRE:
		{
			switch(event)
			{
				case EVENT_FIRE_DONE:
				{
					kicker.lastKickDone = chTimeNow();

					if(kicker.autoCharge)
					{
						enableChg(1);
						kicker.state = STATE_CHG;
					}
					else
						kicker.state = STATE_IDLE;
				}
				break;
				default: break;
			}
		}
		break;
		default: break;
	}
}

void EXTI0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR0;

	chSysLockFromIsr();
	chMBPostI(&kicker.eventQueue, EVENT_KICK_ERR);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void TIM4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM4->SR = 0;

	chSysLockFromIsr();
	chMBPostI(&kicker.eventQueue, EVENT_FIRE_DONE);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void TIM1_TRG_COM_TIM17_IRQHandler()
{
	const float alphaVCap = 0.95f;
	const float alphaChgOff = 0.95f;

	CH_IRQ_PROLOGUE();

	TIM17->SR = 0;

	u32u16_t sum;
	sum.u32 = kicker.rxData32[0];

	for(uint8_t i = 1; i < SPI1_ADC_SAMPLES; i++)
		sum.u32 = __UQADD16(sum.u32, kicker.rxData32[i]);

	float vCap = (sum.u16[1]*0.049560546875f)/((float)SPI1_ADC_SAMPLES);
	float iChg = (sum.u16[0]*0.00732421875f)/((float)SPI1_ADC_SAMPLES) ;

	kicker.vCap = (alphaVCap*kicker.vCap + (1.0f-alphaVCap)*vCap);

	if(kicker.state != STATE_CHG)
	{
		kicker.chgOffset = (alphaChgOff*kicker.chgOffset + (1.0f-alphaChgOff)*iChg);
	}

	kicker.iChg = iChg - kicker.chgOffset;

	switch(kicker.state)
	{
	case STATE_CHG:
		if(kicker.vCap > kicker.config.maxVoltage)
		{
			chSysLockFromIsr();
			chMBPostI(&kicker.eventQueue, EVENT_VMAX_REACHED);
			chSysUnlockFromIsr();
		}
		else
		{
			TIM17->CCR1 = getChgVal();
		}
		break;
	case STATE_HYST:
		if(kicker.vCap < kicker.config.maxVoltage-kicker.hystVoltage)
		{
			chSysLockFromIsr();
			chMBPostI(&kicker.eventQueue, EVENT_BELOW_HYST);
			chSysUnlockFromIsr();
		}
		break;
	}

	CH_IRQ_EPILOGUE();
}

void KickerInit()
{
	GPIOInitData gpioInit;

	// ### Init variables
	kicker.state = STATE_IDLE;
	kicker.autoCharge = 0;
	kicker.autoDischarge = 0;
	kicker.hystVoltage = 5.0f;

	FlashFSOpenOrCreate("kicker/cfg", 0, &kicker.config, sizeof(KickerInternalConfig), &kicker.pConfigFile);

	chMBInit(&kicker.eventQueue, kicker.eventQueueData, KICKER_EVENT_QUEUE_SIZE);

	for(uint16_t i = 0; i < SPI1_ADC_SAMPLES; i++)
	{
		kicker.txData[i*2+0] = 0;		// select IN1
		kicker.txData[i*2+1] = 0x800;	// select IN2
	}

	// ### KICK-ERR external interrupt
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.pupd = GPIO_PUPD_UP;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;
	GPIOInit(GPIOA, GPIO_PIN_0, &gpioInit);

//	NVICEnableIRQ(EXTI0_IRQn, IRQL_KICK_ERR);

	// ### ADC via SPI1
	RCC->APB2ENR |= RCC_APB2ENR_SPI1EN;

	// init GPIO
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;	// SPI1
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_50MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7, &gpioInit);	// MISO, MOSI, SCK, NSS

	// init SPI1
	// prescaler 16 = 250kHz conversion rate, master mode
	SPI1->CR1 = SPI_CR1_BRDIV16 | SPI_CR1_MSTR;
	// 16bit datasize, NSS pulse, NSS output enabled, RX/TX DMA enabled
	SPI1->CR2 = 0x0F00 | SPI_CR2_NSSP | SPI_CR2_SSOE | SPI_CR2_RXDMAEN | SPI_CR2_TXDMAEN;
	SPI1->SR = 0;

	// init DMA
	// TX
	DMA1_Channel3->CCR = DMA_PL_LOW | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_DIR_MEM2PERIPH | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA1_Channel3->CNDTR = SPI1_ADC_SAMPLES*2;
	DMA1_Channel3->CMAR = (uint32_t)&kicker.txData;
	DMA1_Channel3->CPAR = (uint32_t)&SPI1->DR;

	// RX
	DMA1_Channel2->CCR = DMA_PL_LOW | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_DIR_PERIPH2MEM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA1_Channel2->CNDTR = SPI1_ADC_SAMPLES*2;
	DMA1_Channel2->CMAR = (uint32_t)&kicker.rxData;
	DMA1_Channel2->CPAR = (uint32_t)&SPI1->DR;

    // Enable the SPI peripheral
	SPI1->CR1 |= SPI_CR1_SPE;

    // Enable the DMA channels
	DMA1_Channel3->CCR |= DMA_CCR_EN;
	DMA1_Channel2->CCR |= DMA_CCR_EN;

	// ### CHG output
	RCC->APB2ENR |= RCC_APB2ENR_TIM17EN;

	TIM17->CR1 = 0;
	TIM17->CR2 = TIM_CR2_OIS1;	// output idle state high
	TIM17->PSC = 0;
	TIM17->ARR = 719;	// = 100kHz
	TIM17->RCR = 9;	// = 10kHz
	TIM17->CCMR1 = 0x70 | TIM_CCMR1_OC1PE;	// PWM mode 2, CCR1 preload enable
	TIM17->CCER = TIM_CCER_CC1E;	// OC1 output enable
	TIM17->DIER = TIM_DIER_UIE;
	TIM17->CCR1 = 0;

	NVICEnableIRQ(TIM1_TRG_COM_TIM17_IRQn, IRQL_KICK_CTRL);

	TIM17->CR1 |= TIM_CR1_CEN;	// enable timer
	TIM17->BDTR = TIM_BDTR_MOE;	// enable main output enable

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 10;	// TIM17_1
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_50MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOB, GPIO_PIN_5, &gpioInit);

	// ### DIS outputs
	RCC->APB1ENR |= RCC_APB1ENR_TIM4EN;

	TIM4->CR1 = TIM_CR1_OPM;	// one pulse mode
	TIM4->CR2 = 0;
	TIM4->PSC = 71;	// max pulse width 65536us
	TIM4->ARR = 0xFFFF;
	TIM4->RCR = 0;
	TIM4->CCMR2 = LOW3 | LOW4;
	TIM4->CCR3 = 1;
	TIM4->CCR4 = 1;
	TIM4->CCER = TIM_CCER_CC3E | TIM_CCER_CC4E;
	TIM4->EGR = TIM_EGR_UG;
	TIM4->SR = 0;
	TIM4->DIER = TIM_DIER_UIE;

	NVICEnableIRQ(TIM4_IRQn, IRQL_KICK_CTRL);

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 2;	// TIM4
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_10MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOB, GPIO_PIN_8 | GPIO_PIN_9, &gpioInit);
}

void KickerFire1(float duration)
{
	duration *= 1e6f;	// convert to us
	if(duration > 65535.0f)
		duration = 65535.0f;

	uint32_t event = duration;
	event = (event << 16) | EVENT_FIRE_START1;

	chMBPost(&kicker.eventQueue, event, MS2ST(10));
}

void KickerFire2(float duration)
{
	duration *= 1e6f;	// convert to us
	if(duration > 65535.0f)
		duration = 65535.0f;

	uint32_t event = duration;
	event = (event << 16) | EVENT_FIRE_START2;

	chMBPost(&kicker.eventQueue, event, MS2ST(10));
}

void KickerEnableAutoRecharge(uint8_t enable)
{
	kicker.autoDischarge = 0;
	kicker.autoCharge = enable;
}

void KickerSetConfig(float v, uint32_t cooldown)
{
	if(v > 190.0f)
		v = 190.0f;

	kicker.config.maxVoltage = v;
	kicker.config.cooldown = cooldown;

	FlashFSWrite(&kicker.pConfigFile, &kicker.config, sizeof(KickerInternalConfig));
}

void KickerSetVoltageHysteresis(float vHyst)
{
	kicker.hystVoltage = vHyst;
}

void KickerEnableCharge(uint8_t enable)
{
	if(enable)
		chMBPost(&kicker.eventQueue, EVENT_CHG_START, MS2ST(10));
	else
		chMBPost(&kicker.eventQueue, EVENT_CHG_STOP, MS2ST(20));
}

void KickerAutoDischarge()
{
	KickerEnableCharge(0);
	KickerEnableAutoRecharge(0);
	kicker.autoDischarge = 1;
}

msg_t KickerTask(void* params)
{
	(void)params;
	msg_t event;

	chRegSetThreadName("Kicker");

	uint32_t last100Run = chTimeNow();

	while(1)
	{
		if(chMBFetch(&kicker.eventQueue, &event, MS2ST(20)) == RDY_OK)
		{
			processEvent(event);

			if(event == EVENT_KICK_ERR)
			{
		//		ConsolePrint("Kicker error!\r\n");
			}
		}

		if(kicker.stress != 0.0f && kicker.state == STATE_HYST)
		{
			KickerFire1(kicker.stress);
		}

		if(chTimeElapsedSince(last100Run) > MS2ST(100))
		{
			last100Run = chTimeNow();

			if(kicker.autoDischarge)
			{
				if(kicker.vCap < 2.0f)
					kicker.autoDischarge = 0;

				KickerFire1(0.1f/(kicker.vCap+20.0f));
			}
		}
	}

	return 0;
}
