/*
 * motor.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

/**	Timer/ADC/Motor assignments
 *
 * 			Dribbler:
 * 	TIM1:	M5
 * 	ADC3:	M5B
 * 	ADC4:	M5A
 */


#pragma GCC optimize ("Ofast")

#include "motor.h"
#include "arm_math.h"

#include "../constants.h"
#include "ch.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include "util/console.h"
#include "util/init_hal.h"
#include "../supply.h"
#include <math.h>

#define PWM_MAX_VAL 3599
#define PI_3RD 1.047197551196598f
#define MOT_ADC_TO_CURRENT 0.0029296875f
#define MOTOR_RESISTANCE 0.2015f	// [Ohm]
#define H5Mask ((1 << 13) | (1 << 14) | (1 << 15))
#define M5Pos ((GPIOC->IDR & H5Mask) >> 13)

Motor motor;

Motor* pMotor = &motor;

// W3  W2  W1
//  5   5   5
IN_CCM_VAR uint32_t ccerLUT[8] = {0x555, 0x505, 0x055, 0x550, 0x550, 0x055, 0x505, 0x000 };

// W1, W2, W3 modifier
IN_CCM_VAR uint16_t hallMod[8][3] = {
	{0, 0, 0},
	{1, 0, 0},
	{0, 1, 0},
	{0, 1, 0},
	{0, 0, 1},
	{1, 0, 0},
	{0, 0, 1},
	{0, 0, 0}
};

static float min3(float a, float b, float c);
static float max3(float a, float b, float c);
static void updateADC(Motor* pDrive);
static void setPWM(uint32_t pwmVal, uint8_t halPos);

/**
 * IRQs
 */
IN_CCM void TIM1_UP_TIM16_IRQHandler()
{
	TIM1->SR = 0;

	updateADC(pMotor);
}

// 100Hz speed update
IN_CCM void TIM3_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	IN_CCM_VAR static uint16_t noHallCounter = 0;

	TIM3->SR = 0;

	float rpm = pMotor->curVelocity;

	if(TIM2->SR & TIM_SR_CC3IF)
	{
		TIM2->CCR3;

		float comDt = (float)pMotor->comDt;

		if(comDt > 15000 || comDt < -15000)
			rpm = 720e6f/comDt;	// [rpm]
	}
	else
	{
		++noHallCounter;

		if(noHallCounter > 25)
		{
			rpm = 0;
			noHallCounter = 0;
		}
	}

	pMotor->curVelocity = rpm;

	PIDUpdate(&motor.pidSpeed, pMotor->curVelocity);

	float cur = motor.avgCurrent.value;
	if(motor.pidSpeed.setpoint < 100.0f)
	{
		EMAFilterUpdate(&motor.biasCurrent, cur);
	}

	cur -= motor.biasCurrent.value;

	motor.temp = LagElementPT1Process(&motor.tempPT1, cur)+25.0f;

	if(motor.temp > 90.0f)
	{
		motor.overheated = 1;
	}

	if(motor.overheated)
	{
		MotorSetSpeed(0);
	}

	if(motor.temp < 70.0f)
	{
		motor.overheated = 0;
	}

	if(motor.pidSpeed.enabled)
	{
		float setVoltage = motor.pidSpeed.output*MOTOR_KE;

		if(motor.pidSpeed.setpoint < 1000)
		{
			setVoltage = 0;
			motor.pidSpeed.iTerm = 0;
		}

		MotorSetVoltage(setVoltage);
	}

	CH_IRQ_EPILOGUE();
}

#define AVG_HALL 6

IN_CCM void EXTI15_10_IRQHandler()
{
	IN_CCM_VAR static uint32_t lastCOMTime;
	IN_CCM_VAR static uint32_t comSamples[AVG_HALL];
	IN_CCM_VAR static uint16_t samplePos;

	EXTI->PR = EXTI_PR_PR13 | EXTI_PR_PR14 | EXTI_PR_PR15;

	// store commutation timestep
	TIM2->EGR = TIM_EGR_CC4G | TIM_EGR_CC3G;

	// commutation logic
	uint8_t halPos = M5Pos;

	int32_t pwmVal = pMotor->pwmVal;
	if(pMotor->enableAutoCommutation)
		setPWM(pwmVal, halPos);

	// delta time measurement between hall events
	uint32_t now = TIM2->CCR4;
	uint32_t dT = now-lastCOMTime;
	lastCOMTime = now;

	comSamples[samplePos++] = dT;
	samplePos %= AVG_HALL;

	uint32_t dtSum = 0;
	for(uint16_t i = 0; i < AVG_HALL; i++)
		dtSum += comSamples[i];

	dtSum /= AVG_HALL;

	pMotor->comDt = dtSum;
	pMotor->halPos = halPos;
}

static void initHw()
{
	GPIOInitData gpioInit;

	// ### DMA configuration (ADC)
	// peripheral-to-memory, half-words, medium priority, memory increment, circular mode
	DMA2_Channel5->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel5->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel5->CPAR = (uint32_t)&ADC3->DR;
	DMA2_Channel5->CMAR = (uint32_t)pMotor->adc[1];

	DMA2_Channel2->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel2->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel2->CPAR = (uint32_t)&ADC4->DR;
	DMA2_Channel2->CMAR = (uint32_t)pMotor->adc[0];

	// ### ADC
	RCC->AHBENR |= RCC_AHBENR_ADC34EN;

	gpioInit.mode = GPIO_MODE_ANALOG;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOB, GPIO_PIN_13 | GPIO_PIN_14, &gpioInit); // B2, B13, B14

	// independent, synchronous clock from AHB
	ADC3_4->CCR = ADC34_CCR_CKMODE_0;

	// enable voltage regulator for each ADC
	ADC3->CR = ADC_CR_ADVREGEN_0;
	ADC4->CR = ADC_CR_ADVREGEN_0;
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	// start single-ended calibration
	ADC3->CR |= ADC_CR_ADCAL;
	ADC4->CR |= ADC_CR_ADCAL;
	chThdSleep(MS2ST(4));	// calibration delay

	pMotor->adcCal[0]  = ADC3->CALFACT;
	pMotor->adcCal[1]  = ADC4->CALFACT;

	// continuous conversion, 12Bit, trigger TIM1_TRGO2
	ADC3->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC3->SQR1 = DRIVE_ADC_SAMPLES-1;

	ADC4->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC4->SQR1 = DRIVE_ADC_SAMPLES-1;

	for(uint8_t i = 1; i <= DRIVE_ADC_SAMPLES; i++)
	{
		ADCRegularChannelConfig(ADC3, 5,  i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC4, 4,  i, ADC_SampleTime_181Cycles5);
	}

	// Enable the ADC DMA
	ADC3->CFGR |= ADC_CFGR_DMAEN;
	ADC4->CFGR |= ADC_CFGR_DMAEN;

	ADC3->CR |= ADC_CR_ADEN;
	ADC4->CR |= ADC_CR_ADEN;
	chThdSleep(MS2ST(4));	// ADC ready delay

	// Enable the DMA channel
	DMA2_Channel5->CCR |= DMA_CCR_EN;
	DMA2_Channel2->CCR |= DMA_CCR_EN;

	// ### Motors
	RCC->APB2ENR |= RCC_APB2ENR_TIM1EN;
	RCC->CFGR3 |= RCC_CFGR3_TIM1SW;	// TIM1 + TIM8 clocks now running at 144MHz

	// Timer config
	// TIM1
	TIM1->CR1 = TIM_CR1_CMS_0;	// center-aligned mode 1
	TIM1->PSC = 0;
	TIM1->ARR = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	TIM1->RCR = 1;				// update event only every second count (e.g. downcounting) = 20kHz

	// PWM mode 2, preload enable on CCR(1-3) registers
	TIM1->CCMR1 = TIM_CCMR1_OC1PE | (7 << 4) | TIM_CCMR1_OC2PE | (7 << 12);
	TIM1->CCMR2 = TIM_CCMR2_OC3PE | (7 << 4) | (7 << 12);
	TIM1->CCER = TIM_CCER_CC1E | TIM_CCER_CC1NE | TIM_CCER_CC2E | TIM_CCER_CC2NE | TIM_CCER_CC3E | TIM_CCER_CC3NE | TIM_CCER_CC4E;
	TIM1->CR2 = TIM_CR2_OIS1 | TIM_CR2_OIS2 | TIM_CR2_OIS3 | (7 << 20);	// output idle state = high, OC4 on TRGO2 (connected to ADC)
	TIM1->CCR1 = 0;
	TIM1->CCR2 = 0;
	TIM1->CCR3 = 0;
	TIM1->CCR4 = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM1->EGR = TIM_EGR_UG;

	TIM1->BDTR = TIM_BDTR_AOE | TIM_BDTR_BKP | TIM_BDTR_OSSR;
	TIM1->DIER = TIM_DIER_UIE;

	TIM1->CR1 |= TIM_CR1_CEN;

	ADC3->CR |= ADC_CR_ADSTART;
	ADC4->CR |= ADC_CR_ADSTART;

	TIM1->BDTR |= TIM_BDTR_MOE;

	// pin config - drive outputs
	GPIOA->ODR |= GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12;
	GPIOB->ODR |= GPIO_PIN_15;

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 6;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_50MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit);

	gpioInit.alternate = 4;
	GPIOInit(GPIOB, GPIO_PIN_15, &gpioInit);

	// ### Hall Sensors
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING_FALLING;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOC, GPIO_PIN_13 | GPIO_PIN_14 | GPIO_PIN_15, &gpioInit);

	// Control timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;
	TIM3->CR1 = 0;
	TIM3->CR2 = 0;
	TIM3->PSC = 719;	// = 10us
	TIM3->ARR = 999;	// = 10ms = 100Hz
	TIM3->RCR = 0;
	TIM3->DIER = TIM_DIER_UIE;
	TIM3->CR1 |= TIM_CR1_CEN;

	// Free running time measurement timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;
	TIM2->CR1 = 0;
	TIM2->CR2 = 0;
	TIM2->PSC = 0;
	TIM2->ARR = 0xFFFFFFFF;
	TIM2->RCR = 0;
	TIM2->CCMR2 = TIM_CCMR2_CC3S_0 | TIM_CCMR2_CC4S_0;
	TIM2->CCER = TIM_CCER_CC3E | TIM_CCER_CC4E;
	TIM2->CR1 |= TIM_CR1_CEN;

	NVICEnableIRQ(TIM1_UP_TIM16_IRQn, 3);
	NVICEnableIRQ(EXTI15_10_IRQn, 2);
	NVICEnableIRQ(TIM3_IRQn, IRQL_MOTOR_UPDATE+2);
}

/**
 * PRIVATE functions
 */
IN_CCM static float min3(float a, float b, float c)
{
	if(a < b && a < c)
		return a;

	if(b < a && b < c)
		return b;

	return c;
}

IN_CCM static float max3(float a, float b, float c)
{
	if(a > b && a > c)
		return a;

	if(b > a && b > c)
		return b;

	return c;
}

typedef union _u32u16
{
	uint32_t u32;
	uint16_t u16[2];
} u32u16_t;

IN_CCM static void updateADC(Motor* pDrive)
{
	float i0f;
	float i1f;

	uint16_t i016;
	uint16_t i116;

	u32u16_t i1u, i0u;

	i0u.u32 = pDrive->adc32[0][0];
	i1u.u32 = pDrive->adc32[1][0];

	for(uint8_t i = 1; i < DRIVE_ADC_SAMPLES/2; i++)
	{
		i0u.u32 = __UQADD16(i0u.u32, pDrive->adc32[0][i]);
		i1u.u32 = __UQADD16(i1u.u32, pDrive->adc32[1][i]);
	}

	i016 = i0u.u16[0]+i0u.u16[1];
	i116 = i1u.u16[0]+i1u.u16[1];

	i0f = i016*(1.0f/DRIVE_ADC_SAMPLES);
	i1f = i116*(1.0f/DRIVE_ADC_SAMPLES);

	i0f = (i0f-1706.6667f)*-MOT_ADC_TO_CURRENT; // convert to real current
	i1f = (i1f-1706.6667f)*-MOT_ADC_TO_CURRENT; // convert to real current

	float alpha = i0f;
	float beta = 0.57735026918963f*i0f + 1.154700538379252f*i1f;

	float abs = sqrtf(alpha*alpha + beta*beta); //*0.70710678118655f;

	EMAFilterUpdate(&motor.avgCurrent, abs);
}

IN_CCM static void setPWM(uint32_t pwmVal, uint8_t halPos)
{
	halPos ^= 0x07;

	TIM1->CCER = ccerLUT[halPos];
	TIM1->CCR1 = hallMod[halPos][0]*pwmVal;
	TIM1->CCR2 = hallMod[halPos][1]*pwmVal;
	TIM1->CCR3 = hallMod[halPos][2]*pwmVal;
}

/**
 * PUBLIC functions
 */
void MotorInit()
{
	EMAFilterInit(&motor.avgCurrent, 0.99995f);
	EMAFilterInit(&motor.biasCurrent, 0.995f);

	PIDParams speedPIDParams;
	speedPIDParams.kp = 0.2f;
	speedPIDParams.ki = 10.0f;
	speedPIDParams.kd = 0.0f;
	speedPIDParams.outputMin = -45000.0f;	// [RPM]
	speedPIDParams.outputMax =  45000.0f;	// [RPM]

	PIDInit(&motor.pidSpeed, &speedPIDParams, 0.01f, 1);

	LagElementPT1Init(&motor.tempPT1, 70, 240, 0.01f);

	initHw();

	MotorEnableAutoCommutation(1);

	// trigger initial hall IRQ to get a valid hall position
	EXTI->SWIER |= EXTI_SWIER_SWIER14;
}

void MotorEnablePhase(uint32_t ccer)
{
	uint32_t oldCCER = TIM1->CCER & (TIM_CCER_CC1P | TIM_CCER_CC1NP | TIM_CCER_CC2P | TIM_CCER_CC2NP | TIM_CCER_CC3P | TIM_CCER_CC3NP);
	TIM1->CCER = oldCCER | ccer;
}

IN_CCM void MotorSetVoltage(float vol)
{
	float Vd = supply.bat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float da = vol/Vd;

	if(da > 1.0f)
		da = 1.0f;
	if(da < -1.0f)
		da = -1.0f;

	motor.pwmVal = (int32_t)(((float)PWM_MAX_VAL)*da);
	setPWM(motor.pwmVal, M5Pos);
}

void MotorSetSpeed(float rpm)
{
	motor.pidSpeed.setpoint = rpm;
}

void MotorEnableSpeedPID(uint8_t enable)
{
	PIDEnable(&motor.pidSpeed, enable);
}

void MotorEnableAutoCommutation(uint8_t enable)
{
	motor.enableAutoCommutation = enable;
}

void MotorSetPhaseVoltagesSVPWM(float Va, float Vb, float Vc)
{
	float Vd = supply.bat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float Vcom = Vd/2-0.5f*(max3(Va, Vb, Vc)+min3(Va, Vb, Vc));

	float da = 1/Vd*(Va+Vcom);
	float db = 1/Vd*(Vb+Vcom);
	float dc = 1/Vd*(Vc+Vcom);

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	TIM1->CCR1 = (uint16_t)(da*PWM_MAX_VAL);
	TIM1->CCR2 = (uint16_t)(db*PWM_MAX_VAL);
	TIM1->CCR3 = (uint16_t)(dc*PWM_MAX_VAL);
}

void MotorSetAlphaBeta(float Valpha, float Vbeta)
{
	float Va = Valpha;
	float Vb = -0.5f*Valpha + 0.86602540378444f*Vbeta;
	float Vc = -0.5f*Valpha - 0.86602540378444f*Vbeta;

	MotorSetPhaseVoltagesSVPWM(Va, Vb, Vc);
}

void MotorSetVoltageAngle(float Vref, float angle)
{
	float sin, cos;
	while(angle > 180.0f)
		angle -= 360.0f;

	arm_sin_cos_f32(angle, &sin, &cos);

	MotorSetAlphaBeta(cos*Vref, sin*Vref);
}
