/*
 * io.c
 *
 *  Created on: 24.12.2014
 *      Author: AndreR
 */

#include "io.h"
#include "util/intercom.h"
#include "intercom_constants.h"
#include "hal/motor.h"
#include "supply.h"
#include "hal/kicker.h"
#include "kick.h"

#define KICKER_MODE_FORCE		0
#define KICKER_MODE_ARM			1
#define KICKER_MODE_DISARM		2
#define KICKER_MODE_DRIBBLER	3

static uint8_t matchCtrlSeq = 1;

static void ctrlCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	switch(pHeader->command)
	{
		case IC_KD_MATCH_COMMAND:
		{
			KdCommand* pCmd = (KdCommand*)pData;

			// Dribbler
			if((pCmd->dribblerSpeed > DRIVE_MOVE_THRESHOLD || pCmd->dribblerSpeed < -DRIVE_MOVE_THRESHOLD) && supply.batEmpty == 0)
			{
				MotorSetSpeed(pCmd->dribblerSpeed);
			}
			else
			{
				MotorSetSpeed(0);
			}

			// Kicker
			if(pCmd->seq != matchCtrlSeq)
			{
				matchCtrlSeq = pCmd->seq;

				if(pCmd->autoChg && supply.batEmpty == 0)
				{
					KickerEnableAutoRecharge(1);
					KickerEnableCharge(1);
				}
				else
				{
					KickDischarge();
					return;
				}

				switch(pCmd->mode)
				{
					case KICKER_MODE_FORCE: KickFire(pCmd->duration, pCmd->device); break;
					case KICKER_MODE_ARM: KickArmIr(pCmd->duration, pCmd->device); break;
					case KICKER_MODE_DRIBBLER: KickArmDribbler(pCmd->duration, pCmd->device); break;
					default: KickDisarm(); break;
				}
			}
		}
		break;
		case IC_KD_AUTOCHG:
		{
			KickerChargeAuto* pCharge = (KickerChargeAuto*)pData;

			if(pCharge->maxCapLevel == 0)
			{
				KickDischarge();
			}
			else if(supply.batEmpty == 0)
			{
				KickerSetConfig(pCharge->maxCapLevel*0.1f, kicker.config.cooldown);
				KickerEnableCharge(1);
				KickerEnableAutoRecharge(1);
			}
		}
		break;
		case IC_KD_CONFIG:
		{
			KickerConfig* pConfig = (KickerConfig*)pData;

			KickerSetConfig(pConfig->maxCapLevel, pConfig->cooldown);
			KickSetConfig(pConfig->irThreshold, pConfig->irFilterLimit);
		}
		break;
	}
}

void IoInit()
{
	IntercomClient iClient;
	iClient.pUserData = 0;
	iClient.rxCallback = &ctrlCallback;
	iClient.section = INTERCOM_SECTION_KD;

	IntercomAddClient(iClient);
}

msg_t IoTask(void* params)
{
	(void)params;

	uint32_t lastPowerGoodTime = chTimeNow();
	IntercomHeader iHeader;
	KdFeedback kdFeedback;

	iHeader.address = INTERCOM_ADDRESS_MAIN | INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_KD;
	iHeader.command = IC_KD_FEEDBACK;

	chRegSetThreadName("I/O");

	while(1)
	{
		chThdSleepMilliseconds(5);

		if(supply.bat > 10.5f)
			lastPowerGoodTime = chTimeNow();

		if(chTimeElapsedSince(lastPowerGoodTime) > S2ST(10) && supply.batEmpty == 0)
		{
			supply.batEmpty = 1;

			KickDischarge();
		}

		kdFeedback.kicker.cap = kicker.vCap;
		kdFeedback.kicker.chg = kicker.iChg;
		kdFeedback.kicker.state = kicker.state | (kicker.autoCharge << 8) | (kicker.autoDischarge << 12);
		kdFeedback.kicker.kickCounter = kicker.kickCounter;

		kdFeedback.dribbler.velocity = motor.curVelocity;
		kdFeedback.dribbler.temp = motor.temp;
		kdFeedback.dribbler.overheated = motor.overheated;

		kdFeedback.barrier.level[0] = supply.irResult[0];
		kdFeedback.barrier.level[1] = supply.irResult[1];
		kdFeedback.barrier.interrupted = kick.interrupted;

		kdFeedback.power.bat = supply.bat;
		kdFeedback.power.cur = supply.shunt;
		kdFeedback.power.batEmpty = supply.batEmpty;

		kdFeedback.config.maxCapLevel = kicker.config.maxVoltage;
		kdFeedback.config.cooldown = kicker.config.cooldown;
		kdFeedback.config.irThreshold = kick.config.irThreshold;
		kdFeedback.config.irFilterLimit = kick.config.irFilterThreshold;

		IntercomSend(&iHeader, sizeof(KdFeedback), &kdFeedback);
	}

	return 0;
}
