/*
 * kick.c
 *
 *  Created on: 21.05.2014
 *      Author: AndreR
 */

#include "kick.h"
#include "commands.h"
#include "util/fixed.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "hal/kicker.h"
#include "hal/motor.h"
#include "supply.h"
#include "util/network_print.h"
#include "util/sys_time.h"
#include "config_file_ids.h"

#include <math.h>

#define EVENT_BARRIER_INTERRUPTED 0
#define EVENT_DRIBBLER_SPEED_REACHED 1

#define KICKER_STATE_STANDBY		0
#define KICKER_STATE_IR_ARMED		1
#define KICKER_STATE_DRIBBLER_ARMED	2
#define KICKER_STATE_DISCHARGE		3

KickGlobal kick = { .config = { .irFilterThreshold = 1, .irThreshold = 20, .kickVoltageLimit = 0.9f } };

static const ConfigFileDesc kickFileDesc =
	{ CFG_ID_KICK, 3, "kick/cfg", 3, (ConfigElementDesc[]) {
		{ CET_UINT16, "IR Threshold" },
		{ CET_UINT32, "Filter Limit" },
		{ CET_FLOAT, "Ignore Kick Threshold" },
	} };

static void irCallback()
{
	if(supply.irResult[1] > supply.irResult[0])
		kick.irDiff = 0;
	else
		kick.irDiff = supply.irResult[0] - supply.irResult[1];

	if(kick.irDiff < kick.config.irThreshold)
	{
		if(kick.irFilterCounter < kick.config.irFilterThreshold)
			kick.irFilterCounter++;
	}
	else
	{
		kick.irFilterCounter = 0;
		kick.interrupted = 0;
	}

	if(kick.irFilterCounter >= kick.config.irFilterThreshold)
	{
		kick.interrupted = 1;
		chSysLockFromIsr();
		chMBPostI(&kick.eventQueue, EVENT_BARRIER_INTERRUPTED);
		chSysUnlockFromIsr();
	}
}

void KickInit()
{
	kick.pConfigFile = ConfigOpenOrCreate(&kickFileDesc, &kick.config, sizeof(KickInternalConfig), 0);

	kick.state = KICKER_STATE_STANDBY;
	kick.irFilterCounter = 0;

	supply.irDoneCb = &irCallback;

	chMBInit(&kick.eventQueue, kick.eventQueueData, KICK_EVENT_QUEUE_SIZE);
}

void KickSetConfig(uint16_t irThreshold, uint32_t irFilterLimit)
{
	kick.config.irThreshold = irThreshold;
	kick.config.irFilterThreshold = irFilterLimit;

	ConfigNotifyUpdate(kick.pConfigFile);
}

msg_t KickTask(void* params)
{
	(void)params;

	msg_t event;
	msg_t result;

	chRegSetThreadName("Kick");

	while(1)
	{
		result = chMBFetch(&kick.eventQueue, &event, MS2ST(10));
		if(result == RDY_OK)
		{
			switch(event)
			{
				case EVENT_BARRIER_INTERRUPTED:
				{
					if(kick.state == KICKER_STATE_IR_ARMED)
					{
						if(kicker.vCap > kick.config.kickVoltageLimit*kicker.config.maxVoltage)
						{
							KickFire(kick.armed.time, kick.armed.device);
						}
						else
							kick.irFilterCounter = 0;
					}
				}
				break;
				case EVENT_DRIBBLER_SPEED_REACHED:
				{
					// TODO: not completed yet
				}
				break;
			}
		}

		if(kick.state == KICKER_STATE_DRIBBLER_ARMED && fabsf(pMotor->curVelocity - kick.armed.rpm) < 250.0f)
		{
			chMBPost(&kick.eventQueue, EVENT_DRIBBLER_SPEED_REACHED, TIME_IMMEDIATE);
		}
	}

	return 0;
}

void KickFire(float duration, uint8_t device)
{
	if(kicker.vCap < kick.config.kickVoltageLimit*kicker.config.maxVoltage)
		return;

	if(device == KICKER_DEVICE_STRAIGHT)
		KickerFire1(duration);

	if(device == KICKER_DEVICE_CHIP)
		KickerFire2(duration);

	kick.state = KICKER_STATE_STANDBY;
}

void KickArmIr(float fireDuration, uint8_t device)
{
	kick.armed.device = device;
	kick.armed.time = fireDuration;
	kick.state = KICKER_STATE_IR_ARMED;
	kick.irFilterCounter = 0;
}

void KickArmDribbler(float fireDuration, uint8_t device)
{
	kick.armed.device = device;
	kick.armed.time = fireDuration;
	kick.state = KICKER_STATE_IR_ARMED;	// TODO: implement correctly
	kick.irFilterCounter = 0;
}

void KickDisarm()
{
	kick.state = KICKER_STATE_STANDBY;
}

void KickDischarge()
{
	KickerAutoDischarge();
	kick.state = KICKER_STATE_STANDBY;
}
