#include "ch.h"

#include "hal/led.h"
#include "hal/usart1.h"
#include "hal/kicker.h"

#include "util/boot.h"
#include "util/flash_fs.h"
#include "util/crc.h"
#include "util/init_hal.h"
#include "util/sys_time.h"
#include "util/network_print.h"
#include "util/intercom_reliable.h"
#include "util/prog_client.h"
#include "util/config.h"

#include "intercom1.h"
#include "cli.h"
#include "supply.h"
#include "hal/motor.h"
#include "kick.h"
#include "io.h"

#ifdef ENV_RUN
static WORKING_AREA(waKick, 512);
static WORKING_AREA(waKicker, 1024);
static WORKING_AREA(waCLI, 4096);
static WORKING_AREA(waIO, 512);
static WORKING_AREA(waConfig, 512);
#endif
static WORKING_AREA(waUSART1, 256);
static WORKING_AREA(waIntercom, 1024);

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	RCC->AHBENR |= RCC_AHBENR_DMA1EN | RCC_AHBENR_DMA2EN | RCC_AHBENR_GPIOAEN
			| RCC_AHBENR_GPIOBEN | RCC_AHBENR_GPIOCEN | RCC_AHBENR_GPIOFEN;
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;

	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	SystemClockInitData clkInit;
	clkInit.pllMul = 12;	// 6MHz*12 = 72MHz
	clkInit.HSEBypass = 1;
	clkInit.APB1Div = 2;	// = 36MHz
	clkInit.APB2Div = 1;	// = 72MHz
	clkInit.flashLatency = 2;
	clkInit.sysTickFreq = CH_FREQUENCY;
	SystemClockInit(&clkInit, 72000000UL);

	chSysInit();

	// set vector table offset
#ifdef ENV_BOOT
	SCB->VTOR = 0;
#else
	SCB->VTOR = BOOT_APP_OFFSET;
#endif

	CPULoadCalibrate(20);
	chThdSleep(30);

/*	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	// -- HAL
	LEDInit();
	CRCInit();
	FlashFSInit(0x0801F800, 0x800);	// 2k
	ConfigInit(&ConfigOutputIntercom);
	USART1Init();
	Intercom1Init();
	IntercomReliableInit();
	SysTimeInitCorrection(10);
	ProgClientInit();
	ConfigIntercomInit();
#ifdef ENV_RUN
	CLIInit();
	NetworkPrintInitIntercom();
	KickerInit();
	SupplyInit();
	MotorInit();
	KickInit();
	IoInit();
#endif

	// -- Threads
	chThdCreateStatic(waUSART1, sizeof(waUSART1), NORMALPRIO+2, USARTTask, &usart1);
	chThdCreateStatic(waIntercom, sizeof(waIntercom), NORMALPRIO-1, IntercomTask, 0);
#ifdef ENV_RUN
	chThdCreateStatic(waIO, sizeof(waIO), NORMALPRIO-1, IoTask, 0);
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-1, CLITask, 0);
	chThdCreateStatic(waKicker, sizeof(waKicker), NORMALPRIO+10, KickerTask, 0);
	chThdCreateStatic(waKick, sizeof(waKick), NORMALPRIO+3, KickTask, 0);
	chThdCreateStatic(waConfig, sizeof(waConfig), NORMALPRIO-8, ConfigTask, 0);
#endif

	while(1)
	{
		chThdSleepMilliseconds(10);

		if(chTimeElapsedSince(sysTime.lastCorrTime) > MS2ST(2000))
		{
			KickDischarge();
			MotorSetSpeed(0);

			while(1)
			{
				LEDToggle(LED_GREEN);
				chThdSleepMilliseconds(50);
			}
		}

#ifdef ENV_RUN
		uint32_t time = SysTimeUSec()%1000000;

		if(time > 400000 && time < 600000)
#else
		uint32_t time = SysTimeUSec()%500000;

		if(time > 200000 && time < 300000)
#endif
			LEDMustOn(LED_GREEN);
		else
			LEDMustOff(LED_GREEN);
	}

	return 0;
}
