/*
, * supply.c
 *
 *  Created on: 28.01.2014
 *      Author: AndreR
 */

#include "supply.h"

#include "util/intercom.h"
#include "util/init_hal.h"
#include "constants.h"

#include "intercom_constants.h"
#include "intercom1.h"
#include "kick.h"

#define EVENT_SAMPLING_DONE		0

#define ADC_BAT_TO_VOLTAGE 0.0048014322916667f
#define ADC_SHUNT_TO_CURRENT 0.0146484375f

Supply supply;

static float32_t lowpass500Hz[] = {
	0.006026, 0.0087914, 0.016392, 0.02869, 0.044663, 0.062521,
	0.079979, 0.094637, 0.10439, 0.10782, 0.10439, 0.094637,
	0.079979, 0.062521, 0.044663, 0.02869, 0.016392, 0.0087914, 0.006026 };

void DMA1_Channel1_IRQHandler()
{
	DMA1->IFCR = DMA_IFCR_CGIF1;

	float32_t batMeas[NUM_SUPPLY_SAMPLES];
	float32_t shuntMeas[NUM_SUPPLY_SAMPLES];

	for(uint8_t i = 0; i < NUM_SUPPLY_SAMPLES*2; i += 2)
	{
		batMeas[i/2] = supply.adcData[i]*ADC_BAT_TO_VOLTAGE;
		shuntMeas[i/2] = supply.adcData[i+1]*ADC_SHUNT_TO_CURRENT;
	}

	float32_t bat, shunt;

	arm_fir_decimate_f32(&supply.batFir, batMeas, &bat, NUM_SUPPLY_SAMPLES);
	arm_fir_decimate_f32(&supply.shuntFir, shuntMeas, &shunt, NUM_SUPPLY_SAMPLES);

	supply.bat = LagElementPT1Process(&supply.batLag, bat);
	supply.shunt = LagElementPT1Process(&supply.shuntLag, shunt);
}

void DMA2_Channel1_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	DMA2->IFCR = DMA_IFCR_CGIF1;

	uint16_t ir = 0;

	for(uint8_t i = 0; i < NUM_IR_SAMPLES; i++)
		ir += supply.irData[i];

	ir /= NUM_IR_SAMPLES;

	supply.irResult[supply.irCur++] = ir;
	supply.irCur %= 2;

	if(supply.irDoneCb)
		(*supply.irDoneCb)();

	CH_IRQ_EPILOGUE();
}

void SupplyInit()
{
	GPIOInitData gpioInit;

#ifdef DEBUG
	supply.pDMACh = DMA1_Channel1;
	supply.pADC = ADC1;
	supply.pDMA = DMA1;
#endif

	// Initialize FIR filter
	// Fs = 20kHz, Fc = 500Hz (-6dB), design approach: Window-Hamming
	// Takes 20 input samples and decimates by 20 = 1 output value
	arm_fir_decimate_init_f32(&supply.batFir, SUPPLY_LOWPASS_500_TAPS, NUM_SUPPLY_SAMPLES, lowpass500Hz, supply.batFirData, NUM_SUPPLY_SAMPLES);
	arm_fir_decimate_init_f32(&supply.shuntFir, SUPPLY_LOWPASS_500_TAPS, NUM_SUPPLY_SAMPLES, lowpass500Hz, supply.shuntFirData, NUM_SUPPLY_SAMPLES);

	LagElementPT1Init(&supply.batLag, 1.0f, 0.25f, 0.001f);
	LagElementPT1Init(&supply.shuntLag, 1.0f, 0.25f, 0.001f);

	supply.bat = 12.0f;

	// DMA configuration
	// Medium prority, half-words, memory increment, circular mode, periph to mem, transfer complete IRQ
	DMA1_Channel1->CCR = DMA_PL_MEDIUM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE
			| DMA_CCR_MINC | DMA_CCR_CIRC | DMA_DIR_PERIPH2MEM | DMA_CCR_TCIE;
	DMA1_Channel1->CNDTR = 2*NUM_SUPPLY_SAMPLES;
	DMA1_Channel1->CMAR = (uint32_t)supply.adcData;
	DMA1_Channel1->CPAR = (uint32_t)&ADC1->DR;

	// Medium prority, half-words, memory increment, circular mode, periph to mem, transfer complete IRQ
	DMA2_Channel1->CCR = DMA_PL_MEDIUM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE
			| DMA_CCR_MINC | DMA_CCR_CIRC | DMA_DIR_PERIPH2MEM | DMA_CCR_TCIE;
	DMA2_Channel1->CNDTR = NUM_IR_SAMPLES;
	DMA2_Channel1->CMAR = (uint32_t)supply.irData;
	DMA2_Channel1->CPAR = (uint32_t)&ADC2->DR;

	// ### ADC
	RCC->AHBENR |= RCC_AHBENR_ADC12EN;

	// independent, synchronous clock from AHB
	ADC1_2->CCR = ADC12_CCR_CKMODE_0;

	// enable voltage regulator for each ADC
	ADC1->CR = ADC_CR_ADVREGEN_0;
	ADC2->CR = ADC_CR_ADVREGEN_0;
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	// start single-ended calibration
	ADC1->CR |= ADC_CR_ADCAL;
	ADC2->CR |= ADC_CR_ADCAL;
	chThdSleep(MS2ST(4));	// calibration delay

	supply.adcCalSupply  = ADC1->CALFACT;
	supply.adcCalIr  = ADC2->CALFACT;

	// single conversion, 12Bit, trigger TIM15_TRGO
	ADC1->CFGR = ADC_CFGR_EXTEN_0 | (14 << 6) | ADC_CFGR_DMACFG;
	ADC1->SQR1 = 1;	// 2 conversions (shunt and battery)

	// single conversion, 12Bit, trigger TIM3_TRGO
	ADC2->CFGR = ADC_CFGR_EXTEN_0 | (4 << 6) | ADC_CFGR_DMACFG;
	ADC2->SQR1 = NUM_IR_SAMPLES-1;

	// ADC1 regular channel configuration
	ADCRegularChannelConfig(ADC1, 3, 1, ADC_SampleTime_181Cycles5);	// battery voltage
	ADCRegularChannelConfig(ADC1, 4, 2, ADC_SampleTime_181Cycles5);	// current consumption

	// ADC2 regular channel configuration
	for(uint8_t i = 1; i <= NUM_IR_SAMPLES; i++)
		ADCRegularChannelConfig(ADC2, 12, i, ADC_SampleTime_181Cycles5);	// IR LED

	// Enable the ADC DMA
	ADC1->CFGR |= ADC_CFGR_DMAEN;
	ADC2->CFGR |= ADC_CFGR_DMAEN;

	ADC1->CR |= ADC_CR_ADEN;
	ADC2->CR |= ADC_CR_ADEN;
	chThdSleep(MS2ST(4));	// ADC ready delay

	NVICEnableIRQ(DMA1_Channel1_IRQn, IRQL_SUPPLY_MEAS);
	NVICEnableIRQ(DMA2_Channel1_IRQn, IRQL_IR_BARRIER);

	// IR Timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;

	TIM3->CR1 = 0;
	TIM3->PSC = 71; // = 1MHz
	TIM3->ARR = 249;	// = 2kHz, rise/fall time: 200us, sample time: 50us
	// ADC 181.5 cycles, 16 samples ~= 44us
	TIM3->RCR = 0;

	// Toggle mode
	// OC3 in PWM mode 2, OC4 in toggle mode
	TIM3->CCMR2 = (3 << 12) | (7 << 4);
	TIM3->CCER = TIM_CCER_CC3E | TIM_CCER_CC4E;

	TIM3->CR2 = (6 << 4);	// OC3 connected to TRGO (ADC)

	TIM3->CCR3 = 199;
	TIM3->CCR4 = 1;

	TIM3->CR1 |= TIM_CR1_CEN;

	// Supply Timer
	RCC->APB2ENR |= RCC_APB2ENR_TIM15EN;

	// Normal Counter, Update event connected to TRGO (ADC)
	TIM15->CR1 = 0;
	TIM15->PSC = 71; // = 1MHz
	TIM15->ARR = 49; // = 20kHz
	TIM15->RCR = 0;
	TIM15->CR2 = (2 << 4);

	TIM15->CR1 |= TIM_CR1_CEN;

	ADC1->CR |= ADC_CR_ADSTART;
	ADC2->CR |= ADC_CR_ADSTART;

	DMA1_Channel1->CCR |= DMA_CCR_EN;
	DMA2_Channel1->CCR |= DMA_CCR_EN;

	// OPAMP1 pins
	gpioInit.mode = GPIO_MODE_ANALOG;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_1 | GPIO_PIN_2, &gpioInit);

	// IN+: PA1, IN-: Output (follower mode)
	OPAMP1->CSR = OPAMP_CSR_VPSEL | OPAMP_CSR_VMSEL;
	OPAMP1->CSR |= OPAMP1_CSR_OPAMP1EN;

	// Configure ADC Channel4 as analog input
	GPIOInit(GPIOA, GPIO_PIN_3, &gpioInit);

	// ADC2 for IR
	GPIOInit(GPIOB, GPIO_PIN_0 | GPIO_PIN_2, &gpioInit);

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 2;
	gpioInit.pupd = GPIO_PUPD_UP;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_2MHZ;
	GPIOInit(GPIOB, GPIO_PIN_1, &gpioInit);
}
