/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include "hal/buzzer.h"
#include "hal/led.h"
#include "hal/usart1.h"
#include "hal/usart6.h"

#include "util/console.h"
#include "util/flash_fs.h"
#include "util/flash.h"
#include "util/network_print.h"
#include "util/terminal.h"
#include "util/boot.h"
#include "util/sys_time.h"
#include "util/log.h"
#include "util/crc.h"
#include "util/trajectory.h"
#include "util/trajectory_orient.h"
#include "util/arm_mat_util_f32.h"
#include "util/intercom_reliable.h"
#include "util/hermite_spline_cubic.h"
#include "util/ukf.h"
#include "util/angle_math.h"

#include "ctrl/ctrl_test_step.h"
#include "ctrl/ctrl_motor.h"
#include "ctrl/ctrl_calibrate.h"
#include "ctrl/ctrl_tigga.h"

#include "skills.h"
#include "skills/skill_sine.h"
#include "skills/skill_enc_train.h"
#include "skills/skill_basics.h"

#include "commands.h"
#include "intercom_constants.h"
#include "spi3.h"
#include "spi2.h"
#include "cpu_load.h"
#include "ctrl.h"
#include "motors.h"
#include "version.h"
#include "hal/power.h"
#include "network.h"
#include "ui.h"

#include <stdio.h>
#include <string.h>

CLIGlobal cli;

static void processCommand();

#define COLOR_NONE	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_BLUE	3
#define COLOR_YELLOW 4

#define COLOR_MEDIA COLOR_BLUE
#define COLOR_MAIN COLOR_NONE
#define COLOR_KD COLOR_RED
#define COLOR_LEFT COLOR_GREEN
#define COLOR_RIGHT COLOR_YELLOW

static void echoColor(uint8_t color)
{
	const char* pCol;

	switch(color)
	{
	case COLOR_RED:
		pCol = "\e[0;31m";
		break;
	case COLOR_GREEN:
		pCol = "\e[0;32m";
		break;
	case COLOR_BLUE:
		pCol = "\e[0;36m";
		break;
	case COLOR_YELLOW:
		pCol = "\e[0;33m";
		break;
	default:
		pCol = "\e[0m";
		break;
	}

	TerminalWrite((uint8_t*)pCol, strlen(pCol));
}

static void echoTargetColor(uint8_t addr)
{
	switch(addr)
	{
	case INTERCOM_ADDRESS_MEDIA:
		echoColor(COLOR_MEDIA);
		break;
	case INTERCOM_ADDRESS_LEFT:
		echoColor(COLOR_LEFT);
		break;
	case INTERCOM_ADDRESS_RIGHT:
		echoColor(COLOR_RIGHT);
		break;
	case INTERCOM_ADDRESS_KD:
		echoColor(COLOR_KD);
		break;
	default:
		echoColor(COLOR_MAIN);
		break;
	}
}

static void consoleTxCallback(void* param)
{
	(void)param;

	while(console.txFifo.numPackets > 0)
	{
		echoTargetColor(INTERCOM_ADDRESS_MAIN);

		uint8_t* pData;
		uint32_t len32;
		int16_t result = FifoLinGet(&console.txFifo, &pData, &len32);
		if(result)
			break;

		TerminalWrite(pData, len32);

		FifoLinDelete(&console.txFifo);

		echoTargetColor(cli.currentTarget);
	}
}

// IRQ Level: Task
static void intercomCallback(const IntercomHeader* pHeader, uint16_t length, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	// called after new console intercom rx
	echoTargetColor(pHeader->command);
	TerminalWrite(pData, length);
	echoTargetColor(cli.currentTarget);
}

static void cmdCallback(char* pBuf, uint16_t length)
{
	ConsolePrint("\r\n");

	// target selection logic
	if(strcmp(pBuf, "target main") == 0)
	{
		cli.currentTarget = INTERCOM_ADDRESS_MAIN;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: MAIN\r\n");
		return;
	}

	if(strcmp(pBuf, "target media") == 0)
	{
		cli.currentTarget = INTERCOM_ADDRESS_MEDIA;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: MEDIA\r\n");
		return;
	}

	if(strcmp(pBuf, "target left") == 0)
	{
		cli.currentTarget = INTERCOM_ADDRESS_LEFT;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: LEFT\r\n");
		return;
	}

	if(strcmp(pBuf, "target right") == 0)
	{
		cli.currentTarget = INTERCOM_ADDRESS_RIGHT;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: RIGHT\r\n");
		return;
	}

	if(strcmp(pBuf, "target kd") == 0)
	{
		cli.currentTarget = INTERCOM_ADDRESS_KD;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: KD\r\n");
		return;
	}

	// check current target
	if(cli.currentTarget == INTERCOM_ADDRESS_MAIN)
	{
		// process onboard
		uint8_t* pData;
		int16_t result = FifoLinReserve(&console.rxFifo, length, &pData);
		if(result)
			return;

		memcpy(pData, pBuf, length);

		FifoLinCommit(&console.rxFifo, length);
	}
	else
	{
		// forward to intercom
		IntercomHeader header;
		header.address = cli.currentTarget;
		header.command = 0;
		header.section = INTERCOM_SECTION_CONSOLE;

		IntercomReliableSend(&header, length, (uint8_t*)pBuf);
	}
}

static void escCallback(VT100Result* pResult)
{
	if(pResult->code == FUNCTION_KEY)
	{
		switch(pResult->param1)
		{
			case FK_F1:
			{
				cli.currentTarget = INTERCOM_ADDRESS_MAIN;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: MAIN\r\n");
			}
			break;
			case FK_F2:
			{
				cli.currentTarget = INTERCOM_ADDRESS_LEFT;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: LEFT\r\n");
			}
			break;
			case FK_F3:
			{
				cli.currentTarget = INTERCOM_ADDRESS_RIGHT;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: RIGHT\r\n");
			}
			break;
			case FK_F4:
			{
				cli.currentTarget = INTERCOM_ADDRESS_KD;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: KD\r\n");
			}
			break;
			case FK_F5:
			{
				cli.currentTarget = INTERCOM_ADDRESS_MEDIA;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: MEDIA\r\n");
			}
			break;
		}
	}
}

void CLIInit()
{
	cli.currentTarget = INTERCOM_ADDRESS_MAIN;

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = &consoleTxCallback;
	conData.pUserData = 0;

	ConsoleInit(conData);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &intercomCallback;
	client.section = INTERCOM_SECTION_CONSOLE;

	IntercomAddClient(client);

	terminal.cmdCb = &cmdCallback;
	terminal.escCb = &escCallback;
}

msg_t CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

	NetworkPrint("Firmware: %s - %s - %.7s\r\n", __DATE__, __TIME__,GIT_REF);

	while(1)
	{
		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		if(cli.intercomFwd.inUse)
		{
			// forward to intercom
			IntercomHeader header;
			header.address = cli.intercomFwd.target;
			header.command = 0;
			header.section = INTERCOM_SECTION_CONSOLE;

			IntercomReliableSend(&header, cli.intercomFwd.bufSize, cli.intercomFwd.buf);

			cli.intercomFwd.inUse = 0;
		}

		chThdSleepMilliseconds(5);
	}

	return 0;
}

static void sensorsCommands()
{
	if(ConsoleCmpCmd("motors"))
	{
		ConsolePrint("%f %f %f %f %hu\r\n", motors.mot[MOT_FR].cur.velocity, motors.mot[MOT_RR].cur.velocity, motors.mot[MOT_FL].cur.velocity, motors.mot[MOT_RL].cur.velocity, (uint16_t) ctrl.input.enc.updated);
		ConsolePrint("lastMotRx %u %u\r\n", motors.lastMotRx[0], motors.lastMotRx[1]);
	}

	if(ConsoleCmpCmd("imu"))
	{
		ConsolePrint("GYR: %6hd  %6hd  %6hd\r\n", spi2.gyrRaw[0], spi2.gyrRaw[1], spi2.gyrRaw[2]);
		ConsolePrint("ACC: %6hd  %6hd  %6hd\r\n", spi2.accRaw[0], spi2.accRaw[1], spi2.accRaw[2]);
		ConsolePrint("AvgGyrTime: %uus\r\n", spi2.avgGyrPeriod);
		ConsolePrint("AvgAccTime: %uus\r\n", spi2.avgAccPeriod);
		ConsolePrint("GYR: % .5f % .5f % .5f\r\n", spi2.gyr[0], spi2.gyr[1], spi2.gyr[2]);
		ConsolePrint("ACC: % .5f % .5f % .5f\r\n", spi2.acc[0], spi2.acc[1], spi2.acc[2]);
	}
}

static void controllerCommands()
{
	float f32[6];
	uint16_t u16;

	if(ConsoleCmpCmd("ctrl off"))
	{
		CtrlSetController(0);

		ConsolePrint("Controller disabled\r\n");
	}

	if(ConsoleCmpCmd("ctrl motor"))
	{
		CtrlSetController(&ctrlMotorInstance);

		ConsolePrint("Controller: Motors\r\n");
	}

	if(ConsoleCmpCmd("ctrl step"))
	{
		CtrlSetController(&ctrlTestStepInstance);

		ConsolePrint("Controller: Test Step\r\n");
	}

	if(ConsoleCmpCmd("ctrl calibrate"))
	{
		CtrlSetController(&ctrlCalibrateInstance);

		ConsolePrint("Controller: Calibrate\r\n");
	}

	if(ConsoleCmpCmd("ctrl tigga"))
	{
		CtrlSetController(&ctrlTiggaInstance);

		ConsolePrint("Controller: TIGGA!!!\r\n");
	}

	if(ConsoleCmpCmd("tigga"))
	{
		ConsolePrint("A:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].A);
		chThdSleepMilliseconds(10);
		ConsolePrint("B:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].B);
		chThdSleepMilliseconds(10);
		ConsolePrint("C:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].C);
		chThdSleepMilliseconds(10);
		ConsolePrint("Ex:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].Ex);
		chThdSleepMilliseconds(10);
		ConsolePrint("Ez:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].Ez);
		chThdSleepMilliseconds(10);
		ConsolePrint("z:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].z);
		chThdSleepMilliseconds(10);
		ConsolePrint("KF states:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].x);
		chThdSleepMilliseconds(10);
		ConsolePrint("KF Sigma:\r\n");
		arm_mat_print(&ctrlTigga.kf[0].Sigma);
		chThdSleepMilliseconds(10);
		ConsolePrint("I: %f\r\n", ctrl.momentOfInertia);
		ConsolePrint("F (XYW): %f, %f, %f\r\n", ctrlTigga.outXYWForce[0], ctrlTigga.outXYWForce[1], ctrlTigga.outXYWForce[2]);
		ConsolePrint("M1-4: %f, %f, %f, %f\r\n", ctrlTigga.outMotorTorque[0],
				ctrlTigga.outMotorTorque[1], ctrlTigga.outMotorTorque[2], ctrlTigga.outMotorTorque[3]);
	}

	if(ConsoleCmpCmd("tigga perf"))
	{
		float total = ctrlTigga.avgTraj.value + ctrlTigga.avgSensorFusion.value
				+ ctrlTigga.avgControl.value;
		float traj = ctrlTigga.avgTraj.value/total*100.0f;
		float fusion = ctrlTigga.avgSensorFusion.value/total*100.0f;
		float control = ctrlTigga.avgControl.value/total*100.0f;

		ConsolePrint("TIGGA performance\r\n");
		ConsolePrint("Runtime:    %9.3fus\r\n", ctrl.avgCtrlRuntime.value);
		ConsolePrint("Fusion:     %9.1f\t%5.2f%%\r\n", ctrlTigga.avgSensorFusion.value, fusion);
		ConsolePrint("Trajectory: %9.1f\t%5.2f%%\r\n", ctrlTigga.avgTraj.value, traj);
		ConsolePrint("Controller: %9.1f\t%5.2f%%\r\n", ctrlTigga.avgControl.value, control);
	}

	if(ConsoleCmpCmd("ctrl perf"))
	{
		float maxVelXY;
		float maxVelW;

		CtrlUtilGetVMax(&ctrl.perfConstants, ctrl.lastKdFeedback.power.bat, &maxVelXY, &maxVelW);

		ConsolePrint("Bat: %.3fV\r\n", ctrl.lastKdFeedback.power.bat);
		ConsolePrint("Acc. (XY): %f, U: %f\r\n", ctrl.botLimits.accMax, ctrl.perfConstants.Uacc);
		ConsolePrint("Acc. (W): %f, U: %f\r\n", ctrl.botLimits.accMaxW, ctrl.perfConstants.UaccW);
		ConsolePrint("Vel. (XY): %f, (W): %f\r\n", maxVelXY, maxVelW);
	}

	if(ConsoleCmpCmd("encoders"))
	{
		ConsolePrint("w2w: %f\r\n", ctrl.encoders.w2w);
		chThdSleepMilliseconds(20);
		ConsolePrint("xy2w: %.6f, %.6f, %.6f, %.6f, %.6f\r\n", ctrl.encoders.xy2w[0],
				ctrl.encoders.xy2w[1], ctrl.encoders.xy2w[2], ctrl.encoders.xy2w[3], ctrl.encoders.xy2w[4]);
		ConsolePrint("      %.6f, %.6f, %.6f, %.6f, %.6f\r\n", ctrl.encoders.xy2w[5],
				ctrl.encoders.xy2w[6], ctrl.encoders.xy2w[7], ctrl.encoders.xy2w[8], ctrl.encoders.xy2w[9]);
		chThdSleepMilliseconds(20);
		ConsolePrint("xy2x: %.6f, %.6f, %.6f, %.6f, %.6f\r\n", ctrl.encoders.xy2x[0],
				ctrl.encoders.xy2x[1], ctrl.encoders.xy2x[2], ctrl.encoders.xy2x[3], ctrl.encoders.xy2x[4]);
		ConsolePrint("      %.6f, %.6f, %.6f, %.6f, %.6f\r\n", ctrl.encoders.xy2x[5],
				ctrl.encoders.xy2x[6], ctrl.encoders.xy2x[7], ctrl.encoders.xy2x[8], ctrl.encoders.xy2x[9]);
		chThdSleepMilliseconds(20);
		ConsolePrint("xy2y: %.6f, %.6f, %.6f, %.6f, %.6f\r\n", ctrl.encoders.xy2w[0],
				ctrl.encoders.xy2y[1], ctrl.encoders.xy2y[2], ctrl.encoders.xy2y[3], ctrl.encoders.xy2y[4]);
		ConsolePrint("      %.6f, %.6f, %.6f, %.6f, %.6f\r\n", ctrl.encoders.xy2y[5],
				ctrl.encoders.xy2y[6], ctrl.encoders.xy2y[7], ctrl.encoders.xy2y[8], ctrl.encoders.xy2y[9]);

		NetworkPrint("w2w: %f", ctrl.encoders.w2w);
		chThdSleepMilliseconds(20);
		NetworkPrint("xy2w: %.6f, %.6f, %.6f, %.6f, %.6f", ctrl.encoders.xy2w[0],
				ctrl.encoders.xy2w[1], ctrl.encoders.xy2w[2], ctrl.encoders.xy2w[3], ctrl.encoders.xy2w[4]);
		NetworkPrint("      %.6f, %.6f, %.6f, %.6f, %.6f", ctrl.encoders.xy2w[5],
				ctrl.encoders.xy2w[6], ctrl.encoders.xy2w[7], ctrl.encoders.xy2w[8], ctrl.encoders.xy2w[9]);
		chThdSleepMilliseconds(20);
		NetworkPrint("xy2x: %.6f, %.6f, %.6f, %.6f, %.6f", ctrl.encoders.xy2x[0],
				ctrl.encoders.xy2x[1], ctrl.encoders.xy2x[2], ctrl.encoders.xy2x[3], ctrl.encoders.xy2x[4]);
		NetworkPrint("      %.6f, %.6f, %.6f, %.6f, %.6f", ctrl.encoders.xy2x[5],
				ctrl.encoders.xy2x[6], ctrl.encoders.xy2x[7], ctrl.encoders.xy2x[8], ctrl.encoders.xy2x[9]);
		chThdSleepMilliseconds(20);
		NetworkPrint("xy2y: %.6f, %.6f, %.6f, %.6f, %.6f", ctrl.encoders.xy2w[0],
				ctrl.encoders.xy2y[1], ctrl.encoders.xy2y[2], ctrl.encoders.xy2y[3], ctrl.encoders.xy2y[4]);
		NetworkPrint("      %.6f, %.6f, %.6f, %.6f, %.6f", ctrl.encoders.xy2y[5],
				ctrl.encoders.xy2y[6], ctrl.encoders.xy2y[7], ctrl.encoders.xy2y[8], ctrl.encoders.xy2y[9]);
	}

	if(ConsoleCmpCmd("matQ"))
	{
		arm_mat_print(&ctrl.matForceXYZ2DownForce);

		arm_matrix_instance_f32 matDownForces = {4, 1, (float[4]){}};
		arm_matrix_instance_f32 matForcesXYZ = {3, 1, (float[3]){0, 0, ctrl.botParams.mass*9.81f}};

		arm_mat_mult_f32(&ctrl.matForceXYZ2DownForce, &matForcesXYZ, &matDownForces);

		arm_mat_print(&matDownForces);
	}

	if(ConsoleScanCmd("accfwd %hu", &u16) == 1)
	{
		NetworkPrint("TIGGA acc feedforward: %hu\n", u16);

		ctrlTigga.configPos.accFeedforward = u16;
		ConfigNotifyUpdate(ctrlTigga.pConfigFilePos);
	}

	if(ConsoleScanCmd("set pos %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("Setting internal state pos: % 7.5f % 7.5f % 7.5f\r\n", f32[0], f32[1], f32[2]);

		ctrl.state.pos[0] = f32[0];
		ctrl.state.pos[1] = f32[1];
		ctrl.state.pos[2] = f32[2];
	}

	if(ConsoleScanCmd("set target %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("Setting target pos: % 7.5f % 7.5f % 7.5f\r\n", f32[0], f32[1], f32[2]);

		CtrlTiggaSetPosTarget(f32[0], f32[1], f32[2], -1);
		CtrlMotorSetPosTarget(f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("rot %f", f32) == 1)
	{
		ConsolePrint("Rotating %fdeg\r\n", f32[0]);

		float rad = f32[0]*M_PI/180.0f;

		CtrlTiggaSetPosTarget(ctrl.state.pos[0], ctrl.state.pos[1],
				AngleNormalize(ctrl.state.pos[2]+rad), -1);
		CtrlMotorSetPosTarget(ctrl.state.pos[0], ctrl.state.pos[1],
				AngleNormalize(ctrl.state.pos[2]+rad));
	}

	if(ConsoleScanCmd("sine %f %f %f %f", f32, f32+1, f32+2, f32+3) == 4)
	{
		uint8_t data[SYSTEM_MATCH_CTRL_USER_DATA_SIZE];
		SkillSineInput* pSine = (SkillSineInput*)data;

		pSine->vel[0] = (int16_t)(f32[0]*1000.0f);
		pSine->vel[1] = (int16_t)(f32[1]*1000.0f);
		pSine->vel[2] = (int16_t)(f32[2]*1000.0f);
		pSine->freq = (uint16_t)(f32[3]*1000.0f);

		SkillsSetInput(4, data);
	}

	if(ConsoleCmpCmd("skill off"))
	{
		uint8_t data[SYSTEM_MATCH_CTRL_USER_DATA_SIZE];

		SkillsSetInput(0, data);
	}

	if(ConsoleScanCmd("skill train %f %f %f %f %f %f %hu", f32, f32+1, f32+2, f32+3, f32+4, f32+5, &u16) == 7)
	{
		NetworkPrint("Engaging encoder training skill");

		uint8_t data[SYSTEM_MATCH_CTRL_USER_DATA_SIZE];

		SkillEncTrainInput* pEnc = (SkillEncTrainInput*)data;
		pEnc->startPos[0] = (int16_t)(f32[0]*1000.0f);
		pEnc->startPos[1] = (int16_t)(f32[1]*1000.0f);
		pEnc->endPos[0] = (int16_t)(f32[2]*1000.0f);
		pEnc->endPos[1] = (int16_t)(f32[3]*1000.0f);
		pEnc->firstAngleSteps = (uint8_t)f32[4];
		pEnc->secondAngleSteps = (uint8_t)f32[5];
		pEnc->xyRepeat = u16;

		SkillsSetInput(5, data);
	}

	if(ConsoleScanCmd("skill override %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Skill override enabled\r\n");
			NetworkPrint("Skill override enabled\n");
			network.skillOverride = 1;
		}
		else
		{
			ConsolePrint("Skill override disabled\r\n");
			NetworkPrint("Skill override disabled\n");
			network.skillOverride = 0;
		}
	}

	if(ConsoleScanCmd("calib %f %f %f", f32+2, f32, f32+1) == 3)
	{
		ConsolePrint("Doing calibration on angle %.3f over %.3fm and acc. of %.3fm/s^2\r\n", f32[2], f32[0], f32[1]);

		CtrlCalibrateMove(f32[2]*(M_PI/180.0f), f32[0], f32[1]);
	}

	if(ConsoleScanCmd("calib print %hu", &u16) == 1)
	{
		CtrlCalibratePrintResult(u16);
	}

	if(ConsoleCmpCmd("ctrl input"))
	{
		ConsolePrint("--- Currents ---\r\n");
		ConsolePrint("FR: % .4fA\r\nFL: % .4fA\r\nRL: % .4fA\r\nRR: % .4fA\r\n",
				motors.mot[MOT_FR].cur.rmsCurrent, motors.mot[MOT_FL].cur.rmsCurrent,
				motors.mot[MOT_RL].cur.rmsCurrent, motors.mot[MOT_RR].cur.rmsCurrent);
	}

	if(ConsoleScanCmd("motors %hu", &u16) == 1)
	{
		ConsolePrint("Motors: %hu\r\n", u16);

		CtrlSetMotorsOn(u16);
	}

	if(ConsoleScanCmd("voltage %f", f32) == 1)
	{
		ConsolePrint("Setting all motor voltages to %.2fV\r\n", *f32);

		CtrlTestStepSet(*f32);
	}

	if(ConsoleScanCmd("move %f %f %f", f32, f32+1, f32+2) == 3)
	{
		CtrlMotorSetVelTarget(f32[0], f32[1], f32[2]);
		CtrlTiggaSetVelTarget(f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("structure %f %f %f %f %f", f32, f32+1, f32+2, f32+3, f32+4) == 5)
	{
		ConsolePrint("New structural parameters:\r\n");
		ConsolePrint("Front angle: %fdeg\r\n", f32[0]);
		ConsolePrint("Back angle: %fdeg\r\n", f32[1]);
		ConsolePrint("Bot radius: %fm\r\n", f32[2]);
		ConsolePrint("Wheel radius: %fm\r\n", f32[3]);
		ConsolePrint("Mass: %fkg\r\n", f32[4]);

		CtrlSetParams(f32[0], f32[1], f32[2], f32[3], f32[4]);

		// NetworkPrint("xyw2m: [ [%.2f %.2f %.2f] [%.2f %.2f %.2f] [%.2f %.2f %.2f] ]", ctrl.matXYW2Motor.pData[0+0], ctrl.matXYW2Motor.pData[0+1], ctrl.matXYW2Motor.pData[0+2], ctrl.matXYW2Motor.pData[3+0], ctrl.matXYW2Motor.pData[3+1], ctrl.matXYW2Motor.pData[3+2], ctrl.matXYW2Motor.pData[6+0], ctrl.matXYW2Motor.pData[6+1], ctrl.matXYW2Motor.pData[6+2]);
		NetworkPrint("structure: %f %f %f %f %f", f32[0], f32[1], f32[2], f32[3], f32[4]);
	}

	if(ConsoleCmpCmd("pos"))
	{
		ConsolePrint("X:%8hd Y:%8hd Z:%8hd\r\n", network.lastVisionPos[0], network.lastVisionPos[1], network.lastVisionPos[2]);
	}

	if(ConsoleCmpCmd("pos delay"))
	{
		ConsolePrint("Avg. position delay: %.5f\r\n", ctrl.avgPosDelay.value);
	}

	if(ConsoleCmpCmd("nullVec"))
	{
		ConsolePrint("motorNullDir: % .4f % .4f % .4f % .4f\r\n", ctrl.matMotorNullDir.pData[0], ctrl.matMotorNullDir.pData[1],
				ctrl.matMotorNullDir.pData[2], ctrl.matMotorNullDir.pData[3]);
	}
}

static void ukfStateTest(arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU)
{
	(void)pX;
	(void)pU;

	return;
}

static void ukfMeasTest(const arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU, arm_matrix_instance_f32* pY)
{
	(void)pX;
	(void)pU;
	arm_mat_zero_f32(pY);
}

static void miscCommands()
{
	if(ConsoleCmpCmd("reset"))
	{
		BootReset();
	}

	if(ConsoleCmpCmd("fault"))
	{
		volatile uint32_t* pFault = 0;
		*pFault = 42;
	}

	if(ConsoleCmpCmd("bootloader"))
	{
		ConsolePrint("Selecting bootloader\r\n");

		IntercomHeader iHeader;
		iHeader.section = INTERCOM_SECTION_BOOTLOADER;
		iHeader.address = INTERCOM_ADDRESS_MEDIA | INTERCOM_ADDRESS_KD | INTERCOM_ADDRESS_LEFT | INTERCOM_ADDRESS_RIGHT | INTERCOM_ADDRESS_MAIN;
		iHeader.command = BOOTLOADER_SELECT;

		IntercomReliableSend(&iHeader, 0, 0);
	}

	if(ConsoleCmpCmd("features"))
	{
		ConsolePrint("Features: 0x%04X\r\n", ui.cfgMisc.features);
	}

	if(ConsoleCmpCmd("hwid"))
	{
		ConsolePrint("Hardware ID: %hu\r\n", (uint16_t)ui.cfgMisc.hardwareId);
	}

	uint16_t u16;
	if(ConsoleScanCmd("set hwid %hu", &u16) == 1)
	{
		UISetHardwareID(u16);

		ConsolePrint("New hardware ID: %hu\r\n", u16);
	}

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	float f = 21.0f;
	if(ConsoleScanCmd("float %f", &f) == 1)
	{
		ConsolePrint("Float: %f\r\n", f);
	}

	if(ConsoleCmpCmd("float"))
	{
		ConsolePrint("Float: %f\r\n", 42.0f);
	}

	if(ConsoleCmpCmd("usart6 stats"))
	{
		ConsolePrint("--- USART6 ---\r\n");
		ConsolePrint("Rx invalid: %d\r\n", usart6.stats.rxInvalid);
	}

	if(ConsoleCmpCmd("led get"))
	{
		ConsolePrint("LED State: 0x%X\r\n", LEDGetState());
	}

	if(ConsoleCmpCmd("cpu"))
	{
		ConsolePrint("CPU Load: %3u.%1u%%\r\n", cpuLoad.load/10, cpuLoad.load%10);
		ConsolePrint("Max: %3u.%1u%%\r\n", cpuLoad.maxLoad/10, cpuLoad.maxLoad%10);
		ConsolePrint("incPerTick: %10u/%u\r\n", cpuLoad.incPerSysTick, cpuLoad.maxIncPerSysTick);
		cpuLoad.maxLoad = 0;
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("Name            rPrio\tTime\r\n");

		for(Thread* pThread = chRegFirstThread(); pThread != 0; pThread = chRegNextThread(pThread))
		{
			ConsolePrint("%-16s%u\t%u\r\n", pThread->p_name, pThread->p_realprio, pThread->p_time);

			chThdSleepMilliseconds(10);
		}
	}

	if(ConsoleCmpCmd("wifi stats"))
	{
		ConsolePrint("Avg: %u\r\n", network.avgRxPeriod);

		NRF24QueueStats* pStat = &network.queue.stats;

		ConsolePrint(" %2hu        toBS     fromBS     rxLoss     txLoss\r\n", (uint16_t)network.queue.id);
		ConsolePrint("LL P %10u %10u %10u %10u\r\n", pStat->nrfIO.txPackets, pStat->nrfIO.rxPackets,
				pStat->nrfFeedback.rxPacketsLost, pStat->nrfFeedback.txPacketsMaxRT);
		ConsolePrint("   B %10u %10u\r\n", pStat->nrfIO.txBytes, pStat->nrfIO.rxBytes);
		ConsolePrint("HL P %10u %10u %10u %10u\r\n", pStat->queueIO.txPackets, pStat->queueIO.rxPackets,
				pStat->queueFeedback.rxPacketsLost, pStat->queueFeedback.txPacketsLost);
		ConsolePrint("   B %10u %10u\r\n", pStat->queueIO.txBytes, pStat->queueIO.rxBytes);
	}

	if(ConsoleCmpCmd("float16"))
	{
		float x = 1.0f;
		float16_t f16 = floatToHalfFloat(x);

		float y = halfFloatToFloat(f16);

		ConsolePrint("In: %f, F16: 0x%04X, out: %f\r\n", x, f16, y);
	}

	if(ConsoleCmpCmd("net hello"))
	{
		ConsolePrint("Saying hello to network print\r\n");
		NetworkPrint("Hello World");
	}

	if(ConsoleScanCmd("set botid %hu", &u16) == 1)
	{
		NRF24ExChangeId(&network.nrf24, u16);
		network.config.botId = u16;

		NetworkPrint("Changed BotID to %hu", u16);
		ConsolePrint("Changed BotID to %hu\r\n", u16);
	}

	if(ConsoleScanCmd("write botid %hu", &u16) == 1)
	{
		NRF24ExChangeId(&network.nrf24, u16);
		network.config.botId = u16;

		NetworkSaveConfig();

		NetworkPrint("Changed BotID to %hu", u16);
		ConsolePrint("Changed BotID to %hu\r\n", u16);
	}

	if(ConsoleScanCmd("set channel %hu", &u16) == 1)
	{
		NRF24ExChangeChannel(&network.nrf24, u16);
		network.config.channel = u16;

		ConsolePrint("Changed channel to %hu\r\n", u16);
	}

	uint16_t speed;
	if(ConsoleScanCmd("set speed %hu", &speed) == 1)
	{
		NRF24ExChangeSpeed(&network.nrf24, speed);
		network.config.speed = speed;

		if(network.config.speed == 0)
			ConsolePrint("Speed: 250k\r\n");
		if(network.config.speed == 1)
			ConsolePrint("Speed: 1M\r\n");
		if(network.config.speed == 2)
			ConsolePrint("Speed: 2M\r\n");
	}

	if(ConsoleCmpCmd("config write"))
	{
		NetworkSaveConfig();
		ConsolePrint("Config saved to FLASH\r\n");
	}

	if(ConsoleCmpCmd("config"))
	{
		ConsolePrint("ID: %hu\r\n", network.config.botId);
		ConsolePrint("Channel: %hu\r\n", network.config.channel);
		if(network.config.speed == 0)
			ConsolePrint("Speed: 250k\r\n");
		if(network.config.speed == 1)
			ConsolePrint("Speed: 1M\r\n");
		if(network.config.speed == 2)
			ConsolePrint("Speed: 2M\r\n");
	}

	if(ConsoleCmpCmd("ukftest"))
	{
		UKF ukf;
		static float ukfData[UKF_DATA_SIZE(6, 7, 3, 7)];

		UKFInit(&ukf, 6, 7, 3, ukfData, &ukfStateTest, &ukfMeasTest);

		volatile uint32_t start = DWT->CYCCNT;
		UKFPredict(&ukf);
		UKFUpdate(&ukf);
		volatile uint32_t m1 = DWT->CYCCNT-start;

		start = DWT->CYCCNT;
		UKFPredictNoStateFunc(&ukf);
		UKFUpdate(&ukf);
		volatile uint32_t m2 = DWT->CYCCNT-start;

		ConsolePrint("m1: %u\r\nm2: %u\r\n", m1, m2);
	}

	if(ConsoleCmpCmd("trig"))
	{
		float s;
		float c;
		float x = SysTimePrec();

		volatile uint32_t start = DWT->CYCCNT;
		x = sinf(x);
		volatile uint32_t bi = DWT->CYCCNT-start;
		start = DWT->CYCCNT;
		x = atan2(x, x);
		volatile uint32_t at = DWT->CYCCNT-start;
		start = DWT->CYCCNT;
		arm_sin_f32(1);
		volatile uint32_t a1 = DWT->CYCCNT-start;
		start = DWT->CYCCNT;
		arm_sin_cos_f32(1, &s, &c);
		volatile uint32_t a2 = DWT->CYCCNT-start;

		ConsolePrint("Build-In: %u\r\narm1: %u\r\narm2: %u\r\nat: %u\r\n", bi, a1, a2, at);
		ConsolePrint("%f\r\n", x);
	}

	if(ConsoleCmpCmd("inv2"))
	{
		float a[] = {1, 2, 3, 4};
		float b[4];
		arm_matrix_instance_f32 in = (arm_matrix_instance_f32){2, 2, a};
		arm_matrix_instance_f32 out = (arm_matrix_instance_f32){2, 2, b};

		arm_mat_inv_2x2_f32(&in, &out);
		arm_mat_print(&out);

		arm_mat_inverse_f32(&in, &out);
		arm_mat_print(&out);
	}

	if(ConsoleCmpCmd("inv3"))
	{
		float a[] = {6, 2, 3, 42, 5, 6, 3, 8, 9};
		float b[9];
		arm_matrix_instance_f32 in = (arm_matrix_instance_f32){3, 3, a};
		arm_matrix_instance_f32 out = (arm_matrix_instance_f32){3, 3, b};

		arm_mat_inv_3x3_f32(&in, &out);
		arm_mat_print(&out);

		arm_mat_inverse_f32(&in, &out);
		arm_mat_print(&out);
	}

	if(ConsoleCmpCmd("skill input"))
	{
		int16_t* pVel = (int16_t*)skills.input.data;

		ConsolePrint("%hu, %hu, %hu\r\n", pVel[0], pVel[1], pVel[2]);
	}

	if(ConsoleCmpCmd("enc table"))
	{
		for(uint16_t i = 0; i < 128; i++)
		{
			if(i % 8 == 0)
				ConsolePrint("\r\n");

			ConsolePrint("%5.3f ", ctrlTigga.enc.xy2x[i]);
		}

		ConsolePrint("\r\n");
		chThdSleepMilliseconds(20);

		for(uint16_t i = 0; i < 128; i++)
		{
			if(i % 8 == 0)
				ConsolePrint("\r\n");

			ConsolePrint("%5.3f ", ctrlTigga.enc.xy2y[i]);
		}

		ConsolePrint("\r\n");
		chThdSleepMilliseconds(20);

		for(uint16_t i = 0; i < 128; i++)
		{
			if(i % 8 == 0)
				ConsolePrint("\r\n");

			ConsolePrint("%5.3f ", ctrlTigga.enc.xy2w[i]);
		}

		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("enc cover"))
	{
		float cover = CtrlTiggaGetEncoderCoverage();

		ConsolePrint("Encoder coverage: %7.3f%%\r\n", cover);
		NetworkPrint("Encoder coverage: %7.3f%%\r\n", cover);
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleCmpCmd("flash compact"))
	{
		FlashFSCompact();

		ConsolePrint("Flash compacted\r\n");
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	if(ConsoleScanCmd("log %hu", &u16) == 1)
	{
		LogPrintAllRecent(u16);
	}

	if(ConsoleCmpCmd("spi3time"))
	{
		ConsolePrint("Avg. SPI3 TX Period: %uus\r\n", spi3.avgTxPeriod);
	}

	if(ConsoleCmpCmd("crc32"))
	{
		uint8_t data[12];
		uint32_t crc;
		uint32_t check;

		for(uint8_t i = 0; i < 12; i++)
			data[i] = (i+1)*4;

		crc = CRC32CalcChecksum(data, 8);
		ConsolePrint("8: 0x%08X\r\n", crc);
		crc = CRC32CalcChecksumAlt(data, 8);
		ConsolePrint("8: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 8, crc);
		ConsolePrint("8: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 7);
		ConsolePrint("7: 0x%08X\r\n", crc);
		crc = CRC32CalcChecksumAlt(data, 7);
		ConsolePrint("7: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 7, crc);
		ConsolePrint("7: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 6);
		ConsolePrint("6: 0x%08X\r\n", crc);
		crc = CRC32CalcChecksumAlt(data, 6);
		ConsolePrint("6: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 6, crc);
		ConsolePrint("6: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 5);
		ConsolePrint("5: 0x%08X\r\n", crc);
		crc = CRC32CalcChecksumAlt(data, 5);
		ConsolePrint("5: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 5, crc);
		ConsolePrint("5: 0x%08X\r\n", check);
	}

	if(ConsoleCmpCmd("5v on"))
	{
		ConsolePrint("Enabling 5V supply\r\n");
		PowerEnable5VSupply(1);
	}

	if(ConsoleCmpCmd("5v off"))
	{
		ConsolePrint("Disabling 5V supply\r\n");
		PowerEnable5VSupply(0);
	}

	if(ConsoleCmpCmd("power"))
	{
		ConsolePrint("Bat: %2.3fV\r\nCur: %2.3fA\r\n", ctrl.lastKdFeedback.power.bat, ctrl.lastKdFeedback.power.cur);
	}

	if(ConsoleCmpCmd("structure"))
	{
		ConsolePrint("Front angle: %fdeg\r\n", ctrl.botParams.frontAngle);
		ConsolePrint("Back angle: %fdeg\r\n", ctrl.botParams.backAngle);
		ConsolePrint("Bot radius: %fm\r\n", ctrl.botParams.botRadius);
		ConsolePrint("Wheel radius: %fm\r\n", ctrl.botParams.wheelRadius);
		ConsolePrint("Mass: %fkg\r\n", ctrl.botParams.mass);
	}
}

static void buzzerCommands()
{
	uint16_t u16;

	if(ConsoleCmpCmd("buzz off"))
	{
		ConsolePrint("Disabling buzzer\r\n");

		BuzzerPlay(0);
	}

	if(ConsoleScanCmd("buzz %hu", &u16) == 1)
	{
		ConsolePrint("Buzzer freq: %hu\r\n", u16);

		BuzzerTone(u16);
	}

	if(ConsoleCmpCmd("buzz up20"))
	{
		BuzzerPlay(&buzzSeqUp20);
	}

	if(ConsoleCmpCmd("buzz up50"))
	{
		BuzzerPlay(&buzzSeqUp50);
	}

	if(ConsoleCmpCmd("buzz down50"))
	{
		BuzzerPlay(&buzzSeqDown50);
	}

	if(ConsoleCmpCmd("buzz up100"))
	{
		BuzzerPlay(&buzzSeqUp100);
	}

	if(ConsoleCmpCmd("buzz down100"))
	{
		BuzzerPlay(&buzzSeqDown100);
	}

	if(ConsoleCmpCmd("buzz updown20"))
	{
		BuzzerPlay(&buzzSeqUpDown20);
	}

	if(ConsoleCmpCmd("buzz beep"))
	{
		BuzzerPlay(&buzzSeqBeepFast);
	}

	if(ConsoleCmpCmd("buzz doublebeep"))
	{
		BuzzerPlay(&buzzSeqDoubleBeepSlow);
	}

	if(ConsoleCmpCmd("buzz tada"))
	{
		BuzzerPlay(&buzzSeqTada);
	}
}

static void processCommand()
{
	sensorsCommands();
	miscCommands();
	buzzerCommands();
	controllerCommands();

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Firmware Date: %s - %s\r\n", __DATE__, __TIME__);
	}

	if(ConsoleCmpCmd("gitref"))
	{
		ConsolePrint("GIT ref: %s\r\n", GIT_REF);
	}
}
