/*
 * ctrl.c
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#include "ctrl.h"

#include "constants.h"
#include "commands.h"
#include "network.h"
#include "util/console.h"
#include "util/sys_time.h"
#include "util/init_hal.h"
#include "util/intercom.h"
#include "config_file_ids.h"
#include "hal/power.h"
#include "hal/buzzer.h"
#include "spi2.h"
#include "spi3.h"
#include "motors.h"
#include "ui.h"
#include "util/arm_mat_util_f32.h"

#include <string.h>

Ctrl ctrl = {
	.botParams = {
		0.0245f, 30.0f, 45.0f, 0.076f, 3.0f,
	},
	.botLimits = {
		3.0f, 5.0f, 30.0f, 60.0f, 2.0f, 1.2f,
	},
	.botLogging = {
		0, 0, 0, 0,
	},
};

static const ConfigFileDesc configFileDescPhysical =
	{ CFG_ID_CTRL_PHYSICAL_PARAMS, 1, "ctrl/physical", 5, (ConfigElementDesc[]) {
		{ CET_FLOAT, "Wheel Radius [m]" },
		{ CET_FLOAT, "Front Angle [deg]" },
		{ CET_FLOAT, "Back  Angle [deg]" },
		{ CET_FLOAT, "Bot Radius [m]" },
		{ CET_FLOAT, "Mass [kg]" },
	} };

static const ConfigFileDesc configFileDescLimits =
	{ CFG_ID_CTRL_LIMITS, 2, "ctrl/limits", 6, (ConfigElementDesc[]) {
		{ CET_FLOAT, "Max. acceleration XY [m/s^2]" },
		{ CET_FLOAT, "Max. deceleration XY [m/s^2]" },
		{ CET_FLOAT, "Max. acceleration W  [rad/s^2]" },
		{ CET_FLOAT, "Max. deceleration W  [rad/s^2]" },
		{ CET_FLOAT, "Battery buffer [V]" },
		{ CET_FLOAT, "Limited Velocity [m/s]" },
	} };

static const ConfigFileDesc configFileDescLogging =
	{ CFG_ID_CTRL_LOGGING, 0, "ctrl/logs", 4, (ConfigElementDesc[]) {
		{ CET_UINT8, "Ctrl. Input" },
		{ CET_UINT8, "Ctrl. State" },
		{ CET_UINT8, "Ctrl. Output" },
		{ CET_UINT8, "Ctrl. Tigga PIDs" },
	} };

static float pXYW2Motor[12];
static float pMotor2XYW[12];
static float pForceXYW2Motor[12];
static float pForceMotor2XYW[12];
static float pMotorNullDir[4];
static float pForceXYZ2DownForce[12];

static void updateMotorInput(uint8_t m);
static void sendMatchFeedback();
static void updatePhysicalParameters();

#define CTRL_TIMER_RESOLUTION 0.0005f

void TIM1_BRK_TIM9_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM9->SR = 0;

	chSysLockFromIsr();
	chMBPostI(&ctrl.eventQueue, CTRL_EVENT_RUN);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

static void kdCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)dataLength;
	(void)pUserData;

	switch(pHeader->command)
	{
		case IC_KD_FEEDBACK:
		{
			KdFeedback* pFb = (KdFeedback*)pData;

			ctrl.lastKdFeedback = *pFb;

			chMBPost(&ctrl.eventQueue, CTRL_EVENT_UPDATE_POWER, TIME_IMMEDIATE);
		}
		break;
	}
}

static void motorsUpdated(uint8_t source)
{
	msg_t event;

	if(source == MOT_SOURCE_RIGHT)
		event = CTRL_EVENT_UPDATE_RIGHT;
	else
		event = CTRL_EVENT_UPDATE_LEFT;

	chMBPost(&ctrl.eventQueue, event, TIME_IMMEDIATE);
}

static void gyrUpdated()
{
	chMBPost(&ctrl.eventQueue, CTRL_EVENT_UPDATE_GYR, TIME_IMMEDIATE);
}

static void accUpdated()
{
	chMBPost(&ctrl.eventQueue, CTRL_EVENT_UPDATE_ACC, TIME_IMMEDIATE);
}

static void posUpdated()
{
	chMBPost(&ctrl.eventQueue, CTRL_EVENT_UPDATE_POS, TIME_IMMEDIATE);
}

static void configUpdateCallback(uint16_t cfgId)
{
	switch(cfgId)
	{
		case CFG_ID_CTRL_PHYSICAL_PARAMS:
		{
			updatePhysicalParameters();

			// reload controller so that its init function can take the new parameters into account
			CtrlSetController(ctrl.pInstance);
		}
		/* no break */
		case CFG_ID_CTRL_LIMITS:
		{
			CtrlUtilGetPerfConstants(ctrl.botLimits.accMax, ctrl.botLimits.accMaxW, &ctrl.perfConstants);
		}
		break;
	}
}

void CtrlInit()
{
	ctrl.logInterval = 100000;
	ctrl.buzzOnTimeViolation = 1;

	ctrl.input.header.type = CTYPE_CTRL_INPUT;
	ctrl.input.header.length = sizeof(LogCtrlInput);
	ctrl.state.header.type = CTYPE_CTRL_STATE;
	ctrl.state.header.length = sizeof(LogCtrlState);
	ctrl.output.header.type = CTYPE_CTRL_OUTPUT;
	ctrl.output.header.length = sizeof(LogCtrlOutput);

	EMAFilterInit(&ctrl.avgPosDelay, 0.96f);
	EMAFilterInit(&ctrl.avgCtrlRuntime, 0.96f);

	// initial bot position depends on botID and should be out of field
	uint8_t botNumber = network.config.botId;
	if(botNumber > 11)
	{
		botNumber -= 11;

		// that's a blue bot, put him on -Y
		ctrl.defaultPosition[1] = -5.0f;
	}
	else
	{
		// a yellow bot, put him on +Y
		ctrl.defaultPosition[1] = 5.0f;
	}
	ctrl.defaultPosition[0] = -2.4f+(botNumber*0.4);
	ctrl.defaultPosition[2] = 0;

	memcpy(ctrl.input.pos.pos, ctrl.defaultPosition, 3*sizeof(float_t));

	// construct matrices
	ctrl.matXYW2Motor = (arm_matrix_instance_f32){4, 3, pXYW2Motor};
	ctrl.matMotor2XYW = (arm_matrix_instance_f32){3, 4, pMotor2XYW};
	ctrl.matForceXYW2Motor = (arm_matrix_instance_f32){4, 3, pForceXYW2Motor};
	ctrl.matForceMotor2XYW = (arm_matrix_instance_f32){3, 4, pForceMotor2XYW};
	ctrl.matMotorNullDir = (arm_matrix_instance_f32){4, 1, pMotorNullDir};
	ctrl.matForceXYZ2DownForce = (arm_matrix_instance_f32){4, 3, pForceXYZ2DownForce};
	ctrl.matPos = (arm_matrix_instance_f32){3, 1, ctrl.state.pos};
	ctrl.matVel = (arm_matrix_instance_f32){3, 1, ctrl.state.vel};
	ctrl.matAcc = (arm_matrix_instance_f32){3, 1, ctrl.state.acc};

	ctrl.pBotParamsConfig = ConfigOpenOrCreate(&configFileDescPhysical, &ctrl.botParams, sizeof(BotParams), &configUpdateCallback);
	ctrl.pBotLimitsConfig = ConfigOpenOrCreate(&configFileDescLimits, &ctrl.botLimits, sizeof(BotLimits), &configUpdateCallback);
	ctrl.pBotLoggingConfig = ConfigOpenOrCreate(&configFileDescLogging, &ctrl.botLogging, sizeof(BotLogging), &configUpdateCallback);

	FlashFSOpenOrCreate("ctrl/encoders", 0, &ctrl.encoders, sizeof(BotEncoders), &ctrl.pEncodersFile);

	// init event queue
	chMBInit(&ctrl.eventQueue, ctrl.eventQueueData, CTRL_EVENT_QUEUE_SIZE);

	// register encoder update callback
	motors.updateCb = &motorsUpdated;

	// register acc and gyr callbacks
	SPI2SetCallbacks(&gyrUpdated, &accUpdated);

	// register pos callback
	NetworkSetPosCallback(&posUpdated);

	ConfigNotifyUpdate(ctrl.pBotParamsConfig);

	// init precision timer for accurate controller dT
	RCC->APB2ENR |= RCC_APB2ENR_TIM9EN;

	TIM9->CR1 = 0;
	TIM9->SMCR = 0;
	TIM9->SR = 0;
	TIM9->DIER = TIM_DIER_UIE;
	TIM9->PSC = 41999;	// = 0.5ms
	TIM9->ARR = 1;

	NVICEnableIRQ(TIM1_BRK_TIM9_IRQn, IRQL_CTRL_TIMER);

	TIM9->CR1 |= TIM_CR1_CEN;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kdCallback;
	client.section = INTERCOM_SECTION_KD;
	IntercomAddClient(client);
}

void CtrlSetParams(float frontAngle, float backAngle, float botRadius,
		float wheelRadius, float mass)
{
	ctrl.botParams.frontAngle = frontAngle;
	ctrl.botParams.backAngle = backAngle;
	ctrl.botParams.botRadius = botRadius;
	ctrl.botParams.wheelRadius = wheelRadius;
	ctrl.botParams.mass = mass;

	ConfigNotifyUpdate(ctrl.pBotParamsConfig);
}

static void updatePhysicalParameters()
{
	arm_matrix_instance_f32 matCM2Wheel = {4, 3, (float[12]){}};
	const float centerOfMass[3] = {0, 0, 0.056};

	float frontAngle = ctrl.botParams.frontAngle;
	float backAngle = ctrl.botParams.backAngle;
	float botRadius = ctrl.botParams.botRadius;
	float mass = ctrl.botParams.mass;

	ctrl.momentOfInertia = 0.75f*mass*botRadius*botRadius;

	float alpha = frontAngle*PI/180.0f;
	float beta = backAngle*PI/180.0f;

	ctrl.theta[0] = alpha;
	ctrl.theta[1] = PI-alpha;
	ctrl.theta[2] = PI+beta;
	ctrl.theta[3] = 2*PI-beta;

	for(uint8_t i = 0; i < 4; i++)
	{
		// construct matrix for XYW velocity to motor velocity conversion
		MAT_ELEMENT(ctrl.matXYW2Motor, i, 0) = -sinf(ctrl.theta[i])*1.42f;
		MAT_ELEMENT(ctrl.matXYW2Motor, i, 1) = cosf(ctrl.theta[i])*1.12f;
		MAT_ELEMENT(ctrl.matXYW2Motor, i, 2) = botRadius;

		// construct matrix for motor forces to XYW forces/moment conversion
		MAT_ELEMENT(ctrl.matForceMotor2XYW, 0, i) = -sinf(ctrl.theta[i]);
		MAT_ELEMENT(ctrl.matForceMotor2XYW, 1, i) = cosf(ctrl.theta[i]);
		MAT_ELEMENT(ctrl.matForceMotor2XYW, 2, i) = botRadius;

		// construct vectors from center of mass to wheels
		MAT_ELEMENT(matCM2Wheel, i, 0) = cosf(ctrl.theta[i])*ctrl.botParams.botRadius - centerOfMass[0];
		MAT_ELEMENT(matCM2Wheel, i, 1) = sinf(ctrl.theta[i])*ctrl.botParams.botRadius - centerOfMass[1];
		MAT_ELEMENT(matCM2Wheel, i, 2) = -centerOfMass[2];
	}

	arm_mat_pinv(&ctrl.matXYW2Motor, &ctrl.matMotor2XYW);
	arm_mat_pinv(&ctrl.matForceMotor2XYW, &ctrl.matForceXYW2Motor);

	// construct matrix to calculate wheel down forces from XY acceleration
	arm_matrix_instance_f32 matQ = {3, 4, (float[12]){}};
	arm_mat_pinv(&matCM2Wheel, &matQ);

	for(uint8_t i = 0; i < 4; i++)
	{
		MAT_ELEMENT(ctrl.matForceXYZ2DownForce, i, 0) = MAT_ELEMENT(matCM2Wheel, i, 2)*MAT_ELEMENT(matQ, 0, i);
		MAT_ELEMENT(ctrl.matForceXYZ2DownForce, i, 1) = MAT_ELEMENT(matCM2Wheel, i, 2)*MAT_ELEMENT(matQ, 1, i);
		MAT_ELEMENT(ctrl.matForceXYZ2DownForce, i, 2) = MAT_ELEMENT(matCM2Wheel, i, 2)*MAT_ELEMENT(matQ, 2, i);
	}

	// Construct motor null vector
	// this simple method is only valid for robots with an equal angle between the side wheels
	pMotorNullDir[0] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 2);
	pMotorNullDir[1] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 3);
	pMotorNullDir[2] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 1);
	pMotorNullDir[3] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 0);
	float length;
	arm_power_f32(pMotorNullDir, 4, &length);
	length = sqrtf(length);
	arm_scale_f32(pMotorNullDir, 1/length, pMotorNullDir, 4);
}

// takes 4 motor or encoder values
float CtrlGetSlippage(float* pMotorVel)
{
	float slip;
	arm_dot_prod_f32(pMotorVel, pMotorNullDir, 4, &slip);

	return slip;
}

// subtract the motor vector in the null direction from the wrong values to
// project them back on the energy-saving hyperplane
void CtrlCorrectSlippage(float* pMotorIn, float* pMotorOut)
{
	float slip = CtrlGetSlippage(pMotorIn);

	pMotorOut[0] = pMotorIn[0] - slip*pMotorNullDir[0];
	pMotorOut[1] = pMotorIn[1] - slip*pMotorNullDir[1];
	pMotorOut[2] = pMotorIn[2] - slip*pMotorNullDir[2];
	pMotorOut[3] = pMotorIn[3] - slip*pMotorNullDir[3];
}

static float fourier5th(float* pParams, float x)
{
	return pParams[0]*arm_cos_f32(x) + pParams[1]*arm_sin_f32(x) +
			pParams[2]*arm_cos_f32(2*x) + pParams[3]*arm_sin_f32(2*x) +
			pParams[4]*arm_cos_f32(3*x) + pParams[5]*arm_sin_f32(3*x) +
			pParams[6]*arm_cos_f32(4*x) + pParams[7]*arm_sin_f32(4*x) +
			pParams[8]*arm_cos_f32(5*x) + pParams[9]*arm_sin_f32(5*x);
}

void CtrlCorrectEncodersOut(float* pIn, float* pOut)
{
	float angle = arm_atan2_f32(pIn[1], pIn[0]);
	float abs = sqrtf(pIn[0]*pIn[0] + pIn[1]*pIn[1]);

	float corrW = fourier5th(ctrl.encoders.xy2w, angle)*abs + ctrl.encoders.w2w*abs;
	float corrX = fourier5th(ctrl.encoders.xy2x, angle)*abs;
	float corrY = fourier5th(ctrl.encoders.xy2y, angle)*abs;

	pOut[0] = pIn[0]+corrX;
	pOut[1] = pIn[1]+corrY;
	pOut[2] = pIn[2]+corrW;
}

void CtrlCorrectEncodersMeas(float* pState, float* pMeas, float* pOut)
{
	float angle = arm_atan2_f32(pState[1], pState[0]);
	float abs = sqrtf(pState[0]*pState[0] + pState[1]*pState[1]);

	float corrW = fourier5th(ctrl.encoders.xy2w, angle)*abs + ctrl.encoders.w2w*abs;
	float corrX = fourier5th(ctrl.encoders.xy2x, angle)*abs;
	float corrY = fourier5th(ctrl.encoders.xy2y, angle)*abs;

	pOut[0] = pMeas[0]-corrX;
	pOut[1] = pMeas[1]-corrY;
	pOut[2] = pMeas[2]-corrW;
}

msg_t CtrlTask(void* params)
{
	(void) params;

	msg_t event;
	float execTime;
	float lastExecTime = SysTime();

	chRegSetThreadName("CTRL");

	while(1)
	{
		if(chMBFetch(&ctrl.eventQueue, &event, TIME_INFINITE) != RDY_OK)
			continue;

		switch(event)
		{
			case CTRL_EVENT_RUN:
			{
				float now = SysTime();
				execTime = now - lastExecTime;
				lastExecTime = now;

				if(chTimeElapsedSince(ctrl.lastDefaultPosStartTime) < MS2ST(1000))
					ctrl.input.pos.updated = 1;

				if(ctrl.pInstance && ctrl.pInstance->pControllerFunc)
				{
					// check controller execution time violation
					if(execTime > 2*ctrl.pInstance->deltaT && ctrl.buzzOnTimeViolation)
						BuzzerPlay(&buzzSeqUp20);

					// input timestamp is used for execution time analysis
					ctrl.input.header.timestamp = SysTimeUSec();

					// execute state estimation
					(*ctrl.pInstance->pStateEstimationFunc)(&ctrl.input, &ctrl.output);

					// call skill
					SkillsExecute();
					if(ctrl.pInstance->pInput && skills.output.updated)
						(*ctrl.pInstance->pInput)(skills.output.mode, skills.output.xyw);

					// execute the controller
					(*ctrl.pInstance->pControllerFunc)(&ctrl.input, &ctrl.output);

					// Update average filter for runtime measurement
					EMAFilterUpdate(&ctrl.avgCtrlRuntime, SysTimeUSec()-ctrl.input.header.timestamp);

					// transfer controller output to motors
					if(ctrl.motorsOn && (ui.cfgMisc.features & SYSTEM_MATCH_FEEDBACK_FEATURE_MOVE))
						motors.ctrlMode = ctrl.output.ctrlMode;
					else
						motors.ctrlMode = MOT_CTRL_OFF;

					for(uint8_t i = 0; i < 4; i++)
					{
						motors.mot[i].set.velocity = ctrl.output.motorVel[i] * 1/(2*M_PI);
						motors.mot[i].set.voltage = ctrl.output.motorVoltage[i];
						motors.mot[i].set.torque = ctrl.output.motorTorque[i];
					}

					// this is the second part of execution time analysis
					ctrl.output.header.timestamp = SysTimeUSec();

					// State logging
					ctrl.state.header.timestamp = SysTimeUSec();

					if(ctrl.state.posUpdated)
						ctrl.feedbackStateUpdated[0] = 1;
					if(ctrl.state.velUpdated)
						ctrl.feedbackStateUpdated[1] = 1;
				}

				// send logs to media for USB logging
				IntercomHeader iHeader;
				iHeader.address = INTERCOM_ADDRESS_MEDIA;
				iHeader.section = INTERCOM_SECTION_LOGFILE;
				iHeader.command = IC_LOGFILE_MESSAGE;

				uint16_t dataLen = 0;
				void* pSend = &ctrl.input;
				if(ctrl.botLogging.ctrlInput)
					dataLen = sizeof(CtrlInput);
				else
					pSend = &ctrl.state;

				if(ctrl.botLogging.ctrlState)
					dataLen += sizeof(CtrlState);
				else
					pSend = &ctrl.output;

				if(ctrl.botLogging.ctrlOutput)
				{
					if(ctrl.botLogging.ctrlState)
						dataLen += sizeof(CtrlOutput);
					else
						dataLen = sizeof(CtrlOutput);
				}

				if(dataLen > 0)
					IntercomSend(&iHeader, dataLen, pSend);

				// send Feedback to Sumatra
				sendMatchFeedback();

				ctrl.input.enc.updated = 0;
				ctrl.input.motorCur.updated = 0;
				ctrl.input.gyr.updated = 0;
				ctrl.input.acc.updated = 0;
				ctrl.input.pos.updated = 0;
				ctrl.input.power.updated = 0;
				ctrl.state.posUpdated = 0;
				ctrl.state.velUpdated = 0;
				ctrl.state.accUpdated = 0;
			}
			break;
			case CTRL_EVENT_UPDATE_LEFT:
			{
				updateMotorInput(MOT_FL);
				updateMotorInput(MOT_RL);
			}
			break;
			case CTRL_EVENT_UPDATE_RIGHT:
			{
				updateMotorInput(MOT_FR);
				updateMotorInput(MOT_RR);
			}
			break;
			case CTRL_EVENT_UPDATE_GYR:
			{
				memcpy(ctrl.input.gyr.rotVel, spi2.gyr, 3*sizeof(float));
				ctrl.input.gyr.time = spi2.lastGyrTime;
				ctrl.input.gyr.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_ACC:
			{
				memcpy(ctrl.input.acc.linAcc, spi2.acc, 3*sizeof(float));
				ctrl.input.acc.time = spi2.lastAccTime;
				ctrl.input.acc.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_POS:
			{
				ctrl.input.pos.pos[0] = network.lastVisionPos[0]*0.001f;
				ctrl.input.pos.pos[1] = network.lastVisionPos[1]*0.001f;
				ctrl.input.pos.pos[2] = network.lastVisionPos[2]*0.001f;
//				ctrl.input.pos.time = network.lastVisionPosTimestamp;
				ctrl.input.pos.time = SysTimeUSec();
				ctrl.input.pos.updated = 1;

				EMAFilterUpdate(&ctrl.avgPosDelay, network.lastVisionDelay*0.001f);
			}
			break;
			case CTRL_EVENT_UPDATE_POWER:
			{
				if(ctrl.lastKdFeedback.power.batEmpty)
				{
					ui.cfgMisc.features = 0;
					if(ctrl.pInstance)
						CtrlSetController(0);

					BuzzerPlay(&buzzSeqBeepFast);
				}

				ctrl.input.power.voltage = ctrl.lastKdFeedback.power.bat;
				ctrl.input.power.updated = 1;
			}
			break;
			default:
				break;
		}

		if(ctrl.pInstance != ctrl.pNewInstance)
		{
			// controller change requested

			if(ctrl.pInstance)
			{
				if(ctrl.pInstance->pExit)
					(*ctrl.pInstance->pExit)();
			}

			for(uint8_t i = 0; i < 4; i++)
			{
				motors.mot[i].set.velocity = 0;
				motors.mot[i].set.voltage = 0;
				motors.mot[i].set.torque = 0;
			}

			ctrl.pInstance = ctrl.pNewInstance;

			if(ctrl.pInstance)
			{
				TIM9->ARR = ((uint32_t)((ctrl.pInstance->deltaT+0.0001f)/CTRL_TIMER_RESOLUTION))-1;
				TIM9->EGR = TIM_EGR_UG;

				ctrl.lastDefaultPosStartTime = chTimeNow();

				if(ctrl.pInstance->pInit)
					(*ctrl.pInstance->pInit)();
			}
			else
			{
				motors.ctrlMode = MOT_CTRL_OFF;
			}
		}
	}

	return 0;
}

void CtrlSetController(CtrlInstance* pInstance)
{
	ctrl.pNewInstance = pInstance;
}

void CtrlSetMotorsOn(uint8_t enable)
{
	ctrl.motorsOn = enable;
}

void CtrlSetLogRate(uint32_t rate)
{
	uint32_t newInterval = 1000000/rate;
	if(newInterval != ctrl.logInterval)
		ctrl.logInterval = newInterval;
}

void CtrlTurnLocal2Global(float globalOrientation, float localX, float localY, float* pGlobalX, float* pGlobalY)
{
	float localToGlobalAngle = -PI / 2.0f + globalOrientation;
	float sinA = arm_sin_f32(localToGlobalAngle);
	float cosA = arm_cos_f32(localToGlobalAngle);

	// turn to global system
	*pGlobalX = localX * cosA - localY * sinA;
	*pGlobalY = localY * cosA + localX * sinA;
}

void CtrlTurnGlobal2Local(float globalOrientation, float globalX, float globalY, float* pLocalX, float* pLocalY)
{
	float globalToLocalAngle = PI / 2.0f - globalOrientation;
	float sinA = arm_sin_f32(globalToLocalAngle);
	float cosA = arm_cos_f32(globalToLocalAngle);

	// turn to local system
	*pLocalX = globalX * cosA - globalY * sinA;
	*pLocalY = globalY * cosA + globalX * sinA;
}

static void updateMotorInput(uint8_t m)
{
	uint8_t updateMask;
	if(m == MOT_FL || m == MOT_RL)
		updateMask = 0x01;
	else
		updateMask = 0x02;

	ctrl.input.enc.vel[m] = motors.mot[m].cur.velocity * (2*PI);
	ctrl.input.enc.time[m] = motors.mot[m].cur.velocityTimestamp;
	ctrl.input.enc.updated |= updateMask;

	ctrl.input.motorVol.time[m] = motors.mot[m].cur.setVoltageTimestamp;
	ctrl.input.motorVol.vol[m] = motors.mot[m].cur.setVoltage;

	ctrl.input.motorTorque.time[m] = motors.mot[m].cur.torqueTimestamp;
	ctrl.input.motorTorque.M[m] = motors.mot[m].cur.torque;

	ctrl.input.motorCur.time[m] = motors.mot[m].cur.rmsCurrentTimestamp;
	ctrl.input.motorCur.cur[m] = motors.mot[m].cur.rmsCurrent;
	ctrl.input.motorCur.updated |= updateMask;
}

static void sendMatchFeedback()
{
	// regular performance feedback for path planning
	if(SysTimeUSec() - ctrl.lastPerfTime > 500000)
	{
		ctrl.lastPerfTime = SysTimeUSec();

		float velMax;
		float velMaxW;
		CtrlUtilGetVMax(&ctrl.perfConstants, ctrl.lastKdFeedback.power.bat-ctrl.botLimits.batBuffer, &velMax, &velMaxW);

		SystemPerformance performance;
		performance.accMax = (uint16_t)(ctrl.botLimits.accMax*1000.0f);
		performance.accMaxW = (uint16_t)(ctrl.botLimits.accMaxW*100.0f);
		performance.brkMax = (uint16_t)(ctrl.botLimits.brkMax*1000.0f);
		performance.brkMaxW = (uint16_t)(ctrl.botLimits.brkMaxW*100.0f);
		performance.velMax = (uint16_t)(velMax*1000.0f);
		performance.velMaxW = (uint16_t)(velMaxW*100.0f);

		PacketHeader header;
		header.section = SECTION_SYSTEM;
		header.cmd = CMD_SYSTEM_PERFORMANCE;
		NetworkSendPacket(&header, (uint8_t*) &performance, sizeof(SystemPerformance));
	}

	// send status
	if(SysTimeUSec() - ctrl.lastLogTime < ctrl.logInterval)
		return;

	ctrl.lastLogTime = SysTimeUSec();

	SystemMatchFeedback feedback;
	memset(&feedback, 0, sizeof(SystemMatchFeedback));

	feedback.batteryLevel = (uint16_t)(ctrl.lastKdFeedback.power.bat*1000.0f);
	feedback.kickerLevel = ctrl.lastKdFeedback.kicker.cap;
	feedback.barrierKickCounter = (ctrl.lastKdFeedback.barrier.interrupted << 7) | (ctrl.lastKdFeedback.kicker.kickCounter & 0x7F);
	feedback.dribblerSpeed = ctrl.lastKdFeedback.dribbler.velocity;
	feedback.dribblerTemp = (uint8_t)(ctrl.lastKdFeedback.dribbler.temp*2.0f);
	feedback.hardwareId = ui.cfgMisc.hardwareId;

	feedback.features = ui.cfgMisc.features;
	if(ctrl.lastKdFeedback.dribbler.overheated)
	{
		feedback.features &= ~SYSTEM_MATCH_FEEDBACK_FEATURE_DRIBBLE;
	}

	if(ctrl.feedbackStateUpdated[0])
	{
		ctrl.feedbackStateUpdated[0] = 0;

		feedback.curPosition[0] = (int16_t)(ctrl.state.pos[0]*1000.0f);
		feedback.curPosition[1] = (int16_t)(ctrl.state.pos[1]*1000.0f);
		feedback.curPosition[2] = (int16_t)(ctrl.state.pos[2]*1000.0f);
	}
	else
	{
		feedback.curPosition[0] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curPosition[1] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curPosition[2] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
	}

	if(ctrl.feedbackStateUpdated[1])
	{
		ctrl.feedbackStateUpdated[1] = 0;
		feedback.curVelocity[0] = (int16_t)(ctrl.state.vel[0]*1000.0f);
		feedback.curVelocity[1] = (int16_t)(ctrl.state.vel[1]*1000.0f);
		feedback.curVelocity[2] = (int16_t)(ctrl.state.vel[2]*1000.0f);
	}
	else
	{
		feedback.curVelocity[0] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curVelocity[1] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curVelocity[2] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
	}

	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_MATCH_FEEDBACK;
	NetworkSendPacket(&header, (uint8_t*) &feedback, sizeof(SystemMatchFeedback));
}
