/*
 * ctrl.h
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#ifndef CTRL_H_
#define CTRL_H_

#include "arm_math.h"
#include "util/vector.h"
#include "intercom_constants.h"
#include "util/config.h"
#include "util/ema_filter.h"
#include "log_msgs.h"
#include "ctrl_util.h"
#include "ch.h"
#include "skills.h"

#define CTRL_MOTOR_TO_WHEEL_RATIO (15.0f/50.0f)
#define CTRL_WHEEL_TO_MOTOR_RATIO (50.0f/15.0f)
#define CTRL_MOTOR_KM (0.0255f) 				// motor torque constant [Nm/A]
#define CTRL_MOTOR_KN (374.0f*2.0f*PI/60.0f)	// motor speed constant [rad/(s*V)]
#define CTRL_MOTOR_KE (1.0f/CTRL_MOTOR_KN)		// motor back-EMF constant [V/rad]
#define CTRL_MOTOR_KF 0.94e-4f					// motor friction constant [Nm*s]
#define CTRL_MOTOR_R 1.2f						// motor resistance [Ohm]
#define CTRL_MOTOR_L (560e-6f)					// motor inductance [H]

#define CTRL_EVENT_RUN				1
#define CTRL_EVENT_UPDATE_LEFT		2
#define CTRL_EVENT_UPDATE_RIGHT		3
#define CTRL_EVENT_UPDATE_GYR		4
#define CTRL_EVENT_UPDATE_ACC		5
#define CTRL_EVENT_UPDATE_POS		6
#define CTRL_EVENT_UPDATE_POWER		7

#define CTRL_EVENT_QUEUE_SIZE 10

typedef LogCtrlInput CtrlInput;
typedef LogCtrlOutput CtrlOutput;
typedef LogCtrlState CtrlState;

typedef void(*CtrlInitFunction)();
typedef void(*CtrlStateFunction)(CtrlInput*, const CtrlOutput*);
typedef void(*CtrlFunction)(CtrlInput*, CtrlOutput*);
typedef void(*CtrlExitFunction)();
typedef void(*CtrlInputFunction)(uint8_t mode, float* pCmd);

typedef struct _CtrlInstance
{
	CtrlInitFunction pInit;
	CtrlStateFunction pStateEstimationFunc;
	CtrlFunction pControllerFunc;
	CtrlExitFunction pExit;
	CtrlInputFunction pInput;

	float deltaT; // in [s], min. accuracy is 0.5ms
} CtrlInstance;

// physical bot params
typedef struct PACKED _BotParams
{
	float wheelRadius;
	float frontAngle;
	float backAngle;
	float botRadius;
	float mass; // [kg]
} BotParams;

// bot limits
typedef struct PACKED _BotLimits
{
	float accMax;
	float brkMax;
	float accMaxW;
	float brkMaxW;
	float batBuffer;
	float limitedVel;
} BotLimits;

typedef struct PACKED _BotLogging
{
	uint8_t ctrlInput;
	uint8_t ctrlState;
	uint8_t ctrlOutput;
	uint8_t ctrlTiggaPIDs;
} BotLogging;

typedef struct PACKED _BotEncoders
{
	float w2w;		// constant
	float xy2w[10];	// Fourier Series of 5th Order
	float xy2x[10];	// Fourier Series of 5th Order
	float xy2y[10];	// Fourier Series of 5th Order
} BotEncoders;

typedef struct _Ctrl
{
	struct PACKED
	{
		CtrlInput input;
		CtrlState state;
		CtrlOutput output;
	};

	uint8_t feedbackStateUpdated[2];

	CtrlInstance* pInstance;
	CtrlInstance* pNewInstance;

	Mailbox eventQueue;
	msg_t eventQueueData[CTRL_EVENT_QUEUE_SIZE];

	BotParams botParams;
	ConfigFile* pBotParamsConfig;
	BotLimits botLimits;
	ConfigFile* pBotLimitsConfig;
	BotLogging botLogging;
	ConfigFile* pBotLoggingConfig;
	float momentOfInertia;
	float theta[4];	// wheel angles in [rad]
	CtrlUtilPerfConstants perfConstants;

	BotEncoders encoders;
	FlashFile* pEncodersFile;

	arm_matrix_instance_f32 matXYW2Motor; // 4x3
	arm_matrix_instance_f32 matMotor2XYW; // 3x4

	arm_matrix_instance_f32 matForceXYW2Motor; // 4x3
	arm_matrix_instance_f32 matForceMotor2XYW; // 3x4

	arm_matrix_instance_f32	matMotorNullDir; // 4x1, linear combinations of this motor vector do not produce any movement = null space of matMotor2XYW

	arm_matrix_instance_f32 matForceXYZ2DownForce; // 4x3

	arm_matrix_instance_f32 matPos;
	arm_matrix_instance_f32 matVel;
	arm_matrix_instance_f32 matAcc;

	KdFeedback lastKdFeedback;

	uint32_t lastLogTime;
	uint32_t lastPerfTime;
	uint32_t logInterval;

	float defaultPosition[3];
	uint32_t lastDefaultPosStartTime;

	EMAFilter avgPosDelay;	// in [ms]

	EMAFilter avgCtrlRuntime;

	uint8_t motorsOn;
	uint8_t buzzOnTimeViolation;
	uint8_t limitedVel;
} Ctrl;

extern Ctrl ctrl;

void CtrlInit();
void CtrlSetParams(float frontAngle, float backAngle, float botRadius, float wheelRadius, float mass);
msg_t CtrlTask(void* params);
void CtrlSetController(CtrlInstance* pInstance);
void CtrlSetLogRate(uint32_t rate);
void CtrlSetMotorsOn(uint8_t enable);

float CtrlGetSlippage(float* pMotorVel);
void CtrlCorrectSlippage(float* pMotorIn, float* pMotorOut);
void CtrlTurnLocal2Global(float globalOrientation, float localX, float localY, float* pGlobalX, float* pGlobalY);
void CtrlTurnGlobal2Local(float globalOrientation, float globalX, float globalY, float* pLocalX, float* pLocalY);

void CtrlCorrectEncodersOut(float* pIn, float* pOut);
void CtrlCorrectEncodersMeas(float* pState, float* pMeas, float* pOut);

#endif /* CTRL_H_ */
