/*
 * ctrl_calibrate.c
 *
 *  Created on: 20.01.2015
 *      Author: AndreR
 */

#include "ctrl_calibrate.h"
#include "util/sys_time.h"
#include "util/console.h"
#include "util/network_print.h"
#include "util/log.h"

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);
static void stateEstimation(CtrlInput* pInput, const CtrlOutput* pOutput);

CtrlCalibrate ctrlCalibrate;
CtrlInstance ctrlCalibrateInstance = {
	.pInit = &init,
	.pExit = 0,
	.pStateEstimationFunc = &stateEstimation,
	.pControllerFunc = &controller,
	.pInput = 0,
	.deltaT = 0.001f };

static void init()
{
	memset(&ctrlCalibrate, 0, sizeof(CtrlCalibrate));

	// set 100ms lag to smooth test function "knicks"
	LagElementPT1Init(&ctrlCalibrate.trajectoryLag, 1.0f, 0.1f, 0.001f);

	ctrlCalibrate.matCurMotor.pData = ctrlCalibrate.curWheel;
	ctrlCalibrate.matCurMotor.numCols = 1;
	ctrlCalibrate.matCurMotor.numRows = 4;

	ctrlCalibrate.matSetpoint.pData = ctrlCalibrate.setpoint;
	ctrlCalibrate.matSetpoint.numCols = 1;
	ctrlCalibrate.matSetpoint.numRows = 4;

	ctrlCalibrate.matEncPosXYW.pData = ctrlCalibrate.encPosXYW;
	ctrlCalibrate.matEncPosXYW.numCols = 1;
	ctrlCalibrate.matEncPosXYW.numRows = 3;

	ctrlCalibrate.matEncPos.pData = ctrlCalibrate.encPos;
	ctrlCalibrate.matEncPos.numCols = 1;
	ctrlCalibrate.matEncPos.numRows = 4;
}

static void stateEstimation(CtrlInput* pInput, const CtrlOutput* pOutput)
{
	(void)pOutput;

	if(ctrlCalibrate.active == 0)
		return;

	// process encoder update
	if(pInput->enc.updated)
	{
		memcpy(ctrlCalibrate.curWheel, pInput->enc.vel, sizeof(float)*4);

		for(uint8_t i = 0; i < 4; i++)
			ctrlCalibrate.curWheel[i] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now speed over ground [m/s]

		// velocity state update
		arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlCalibrate.matCurMotor, &ctrl.matVel);
		ctrl.state.velUpdated = 1;

		// integrate motor velocity to obtain position later on
		for(uint8_t i = 0; i < 4; i++)
		{
			if(ctrlCalibrate.lastEncTimes[i] != 0)
			{
				float dt = (pInput->enc.time[i]-ctrlCalibrate.lastEncTimes[i])*1e-6f;
				float integral = dt*(ctrlCalibrate.lastEncVel[i]+pInput->enc.vel[i])/2.0f;
				ctrlCalibrate.encPos[i] += integral;
			}

			ctrlCalibrate.lastEncTimes[i] = pInput->enc.time[i];
			ctrlCalibrate.lastEncVel[i] = pInput->enc.vel[i];
		}
	}

	if(pInput->acc.updated)
	{
		pInput->acc.updated = 0;

		float acc = sqrtf(pInput->acc.linAcc[0]*pInput->acc.linAcc[0] + pInput->acc.linAcc[1]*pInput->acc.linAcc[1]);
		if(ctrlCalibrate.lastAccTime != 0)
		{
			float dt = (pInput->acc.time-ctrlCalibrate.lastAccTime)*1e-6f;
			float integral = dt*(ctrlCalibrate.lastAccVal+acc)/2.0f;
			ctrlCalibrate.accVel += integral;
		}

		ctrlCalibrate.lastAccTime = pInput->acc.time;
		ctrlCalibrate.lastAccVal = acc;
	}

	if(pInput->gyr.updated)
	{
		pInput->gyr.updated = 0;

		if(ctrlCalibrate.lastGyrTime != 0)
		{
			float dt = (pInput->gyr.time-ctrlCalibrate.lastGyrTime)*1e-6f;
			float integral = dt*(ctrlCalibrate.lastGyrVal+pInput->gyr.rotVel[2])/2.0f;
			ctrlCalibrate.gyrPos += integral;
		}

		ctrlCalibrate.lastGyrTime = pInput->gyr.time;
		ctrlCalibrate.lastGyrVal = pInput->gyr.rotVel[2];
	}

	if(pInput->pos.updated)
	{
		pInput->pos.updated = 0;

		float orientDiff = ctrlCalibrate.lastOrientation-pInput->pos.pos[2];
		if(orientDiff > PI)
			++ctrlCalibrate.turns;
		if(orientDiff < -PI)
			--ctrlCalibrate.turns;

		ctrlCalibrate.lastOrientation = pInput->pos.pos[2];
	}
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	uint32_t trajTime = SysTimeUSec()-ctrlCalibrate.trajectory.tStart;
//	uint8_t axis = ctrlCalibrate.trajectoryAxis;
	const float plateauAlpha = 0.98;

	pOutput->ctrlMode = MOT_CTRL_VELOCITY;

	if(ctrlCalibrate.active == 0)
	{
		memset(pOutput->motorVel, 0, sizeof(float)*4);

		return;
	}

	if(trajTime > ctrlCalibrate.tRuntime)
	{
		ctrlCalibrate.active = 0;

		// trajectory + wait time complete
		memcpy(ctrlCalibrate.endPos, pInput->pos.pos, sizeof(float)*3);

		for(uint8_t i = 0; i < 4; i++)
			ctrlCalibrate.encPos[i] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now distance over ground [m]

		// position transform to XYW
		arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlCalibrate.matEncPos, &ctrlCalibrate.matEncPosXYW);

		// bring vision data to local frame
		float pos[3];
		pos[0] = ctrlCalibrate.endPos[0]-ctrlCalibrate.startPos[0];
		pos[1] = ctrlCalibrate.endPos[1]-ctrlCalibrate.startPos[1];
		pos[2] = ctrlCalibrate.endPos[2]-ctrlCalibrate.startPos[2];
		CtrlTurnGlobal2Local(ctrlCalibrate.startPos[2], pos[0], pos[1], &ctrlCalibrate.localPos[0], &ctrlCalibrate.localPos[1]);
		ctrlCalibrate.localPos[2] = pos[2]+ctrlCalibrate.turns*2*PI;

		CtrlCalibratePrintResult(0);
		CtrlCalibratePrintResult(1);

		return;
	}

	if(ctrlCalibrate.startPosTaken == 0)
	{
		ctrlCalibrate.startPosTaken = 1;
		memcpy(ctrlCalibrate.startPos, pInput->pos.pos, sizeof(float)*3);
		ctrlCalibrate.lastOrientation = pInput->pos.pos[2];
	}

	// Measurements during the constant-velocity phase of the trajectory
	float velAbs = sqrtf(ctrl.state.vel[0]*ctrl.state.vel[0] + ctrl.state.vel[1]*ctrl.state.vel[1]);
	if(trajTime > ctrlCalibrate.tPlateauStart && trajTime < ctrlCalibrate.tPlateauEnd)
	{
		if(ctrlCalibrate.onPlateau == 0)
		{
			ctrlCalibrate.onPlateau = 1;

			// just entered plateau phase, initialize average measurements
			ctrlCalibrate.encAvg = velAbs;
			ctrlCalibrate.accVelAvg = ctrlCalibrate.accVel;
			ctrlCalibrate.gyrAvg = ctrlCalibrate.lastGyrVal;
		}
	}

	if(ctrlCalibrate.onPlateau)
	{
		// on plateau phase, do measurements here
		ctrlCalibrate.encAvg = (plateauAlpha*ctrlCalibrate.encAvg + (1.0f-plateauAlpha)*velAbs);
		ctrlCalibrate.accVelAvg = (plateauAlpha*ctrlCalibrate.accVelAvg + (1.0f-plateauAlpha)*ctrlCalibrate.accVel);
		ctrlCalibrate.gyrAvg = (plateauAlpha*ctrlCalibrate.gyrAvg + (1.0f-plateauAlpha)*ctrlCalibrate.lastGyrVal);
	}

	if(trajTime > ctrlCalibrate.tPlateauEnd && ctrlCalibrate.onPlateau == 1)
	{
		ctrlCalibrate.onPlateau = 0;

		// just left plateau, complete measurements here
	}

	// process output
	float xyw[3] = {0, 0, 0};
	arm_matrix_instance_f32 matXYW = {.numCols = 1, .numRows = 3, .pData = xyw};

	// get current trajectory speed
	float speed = PiecewiseFunctionValue(&ctrlCalibrate.trajectory);

	// feed through PT1 lag element for smoothing
	speed = LagElementPT1Process(&ctrlCalibrate.trajectoryLag, speed);
	xyw[0] = cosf(ctrlCalibrate.trajectoryAngle)*speed;
	xyw[1] = sinf(ctrlCalibrate.trajectoryAngle)*speed;
//	xyw[ctrlCalibrate.trajectoryAxis] = LagElementPT1Process(&ctrlCalibrate.trajectoryLag, speed);

	// convert to motor speeds over ground
	arm_mat_mult_f32(&ctrl.matXYW2Motor, &matXYW, &ctrlCalibrate.matSetpoint);

	// convert to motor RPS and assign to output
	for(uint8_t i = 0; i < 4; i++)
		pOutput->motorVel[i] = ctrlCalibrate.setpoint[i]*CTRL_WHEEL_TO_MOTOR_RATIO * 1/ctrl.botParams.wheelRadius;
}

void CtrlCalibrateMove(float angle, float distance, float acceleration)
{
	ctrlCalibrate.distance = distance;
	ctrlCalibrate.acceleration = acceleration;
	ctrlCalibrate.trajectoryAngle = angle;

	PiecewiseFunctionTestTrajectory(&ctrlCalibrate.trajectory, distance, acceleration);
	ctrlCalibrate.tRuntime = ctrlCalibrate.trajectory.pieces[2].tEnd + 500000;

	ctrlCalibrate.tPlateauStart = ctrlCalibrate.trajectory.pieces[0].tEnd + 100000;
	ctrlCalibrate.tPlateauEnd = ctrlCalibrate.trajectory.pieces[1].tEnd;
	ctrlCalibrate.onPlateau = 0;
	ctrlCalibrate.plateauVelocity = ctrlCalibrate.trajectory.pieces[1].a[0];

	memset(ctrlCalibrate.encPos, 0, sizeof(float)*4);
	memset(ctrlCalibrate.lastEncTimes, 0, sizeof(float)*4);

	ctrlCalibrate.lastAccTime = 0;
	ctrlCalibrate.accVel = 0;

	ctrlCalibrate.lastGyrTime = 0;
	ctrlCalibrate.gyrPos = 0;

	ctrlCalibrate.startPosTaken = 0;
	ctrlCalibrate.turns = 0;

	PiecewiseFunctionStart(&ctrlCalibrate.trajectory);
	ctrlCalibrate.active = 1;
}

void CtrlCalibratePrintResult(uint8_t network)
{
	if(ctrlCalibrate.active)
	{
		ConsolePrint("Calibration active\r\n");
		return;
	}

	float distVision = sqrtf(ctrlCalibrate.localPos[0]*ctrlCalibrate.localPos[0] + ctrlCalibrate.localPos[1]*ctrlCalibrate.localPos[1]);
	float distEnc = sqrtf(ctrlCalibrate.encPosXYW[0]*ctrlCalibrate.encPosXYW[0] + ctrlCalibrate.encPosXYW[1]*ctrlCalibrate.encPosXYW[1]);


	if(network)
	{
		NetworkPrint("%.3f,%.3f,%.3f,%.3f,%.3f,%.3f", ctrlCalibrate.trajectoryAngle*(180.0f/M_PI), ctrlCalibrate.distance, distVision, distEnc, ctrlCalibrate.plateauVelocity, ctrlCalibrate.encAvg);
//		NetworkPrint("Angle: %.3f", ctrlCalibrate.trajectoryAngle);
//		NetworkPrint("Distance: %.3fm",	ctrlCalibrate.distance);
//		NetworkPrint("Velocity: %.3fm/s", ctrlCalibrate.plateauVelocity);
//		NetworkPrint("Acceleration: %.3fm/s^2", ctrlCalibrate.acceleration);
//		NetworkPrint("Runtime: %.3fs", ctrlCalibrate.tRuntime*1e-6f);
//		NetworkPrint("--- Position ---");
//		NetworkPrint("Vision: % .3f", distVision);
//		NetworkPrint("Encoder: %.4f [m or rad]", distEnc);
//		NetworkPrint("Gyro (Z): %.4frad", ctrlCalibrate.gyrPos);
//		NetworkPrint("--- Velocity ---");
//		NetworkPrint("Encoder: %.4f [m/s or rad/s]", ctrlCalibrate.encAvg);
//		NetworkPrint("Gyro (Z): %.4frad/s", ctrlCalibrate.gyrAvg);
//		NetworkPrint("Acc: %.4fm/s", ctrlCalibrate.accVelAvg);
	}
	else
	{
		ConsolePrint("Angle: %.3f\r\nDistance: %.3fm\r\nVelocity: %.3fm/s\r\nAcceleration: %.3fm/s^2\r\n", ctrlCalibrate.trajectoryAngle,
				ctrlCalibrate.distance, ctrlCalibrate.plateauVelocity, ctrlCalibrate.acceleration);
		ConsolePrint("Runtime: %.3fs\r\n", ctrlCalibrate.tRuntime*1e-6f);
		ConsolePrint("--- Position ---\r\n");
		ConsolePrint("Vision: % .3f\r\n", distVision);
		ConsolePrint("Encoder: %.4f [m or rad]\r\n", distEnc);
		ConsolePrint("Gyro (Z): %.4frad\r\n", ctrlCalibrate.gyrPos);
		ConsolePrint("--- Velocity ---\r\n");
		ConsolePrint("Encoder: %.4fm/s\r\n", ctrlCalibrate.encAvg);
		ConsolePrint("Gyro (Z): %.4frad/s\r\n", ctrlCalibrate.gyrAvg);
		ConsolePrint("Acc: %.4fm/s\r\n", ctrlCalibrate.accVelAvg);
	}
}
