/*
 * ctrl_motor.c
 *
 *  Created on: 08.03.2013
 *      Author: AndreR
 */

#include "ctrl_motor.h"
#include "config_file_ids.h"
#include "util/angle_math.h"
#include "util/arm_mat_util_f32.h"

#include <string.h>

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);
static void stateEstimation(CtrlInput* pInput, const CtrlOutput* pOutput);
static void input(uint8_t mode, float* pCmd);

static const ConfigFileDesc configFileDescVel =
	{ CFG_ID_CTRL_MOTOR_VEL, 0, "ctrl_motor/vel", 3, (ConfigElementDesc[]) {
		{ CET_FLOAT, "maxVel/XY" },
		{ CET_FLOAT, "maxVel/W" },
		{ CET_UINT8, "Override Bat." },
	} };


CtrlMotor ctrlMotor = {
	.configVel = {
		.maxXY = 1.0f,
		.maxW = 3.0f,
		.overrideBat = 1.0f,
	},
};
CtrlInstance ctrlMotorInstance = {
	.pInit = &init,
	.pExit = 0,
	.pStateEstimationFunc = &stateEstimation,
	.pControllerFunc = &controller,
	.pInput = &input,
	.deltaT = 0.001f };

static void init()
{
	ctrlMotor.pConfigFileVel = ConfigOpenOrCreate(&configFileDescVel, &ctrlMotor.configVel, sizeof(CtrlMotorConfigVel), 0);

	memset(ctrlMotor.velSetpoint, 0, sizeof(float)*4);

	ctrlMotor.matCurMotor.pData = ctrlMotor.curWheel;
	ctrlMotor.matCurMotor.numCols = 1;
	ctrlMotor.matCurMotor.numRows = 4;

	ctrlMotor.matVelSetpoint.pData = ctrlMotor.velSetpoint;
	ctrlMotor.matVelSetpoint.numCols = 1;
	ctrlMotor.matVelSetpoint.numRows = 4;

	ctrlMotor.posMode = 0;
	ctrlMotor.trajTime = INFINITY;
}

static void stateEstimation(CtrlInput* pInput, const CtrlOutput* pOutput)
{
	(void)pOutput;

	if(pInput->enc.updated)
	{
		memcpy(ctrlMotor.curWheel, pInput->enc.vel, sizeof(float)*4);

		for(uint8_t i = 0; i < 4; i++)
			ctrlMotor.curWheel[i] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now speed over ground [m/s]

		// velocity state update
		arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlMotor.matCurMotor, &ctrl.matVel);

		float globalVel[2];
		CtrlTurnLocal2Global(ctrl.state.pos[2], ctrl.state.vel[0], ctrl.state.vel[1], globalVel, globalVel+1);
		ctrl.state.vel[0] = globalVel[0];
		ctrl.state.vel[1] = globalVel[1];
		ctrl.state.velUpdated = 1;
	}

	if(pInput->pos.updated)
	{
		// position update (feed-through)
		memcpy(ctrl.state.pos, pInput->pos.pos, sizeof(float)*3);
		ctrl.state.posUpdated = 1;
	}
	else
	{
		// integrate encoders if no pos update is available
		ctrl.state.pos[0] += ctrl.state.vel[0]*ctrlMotorInstance.deltaT;
		ctrl.state.pos[1] += ctrl.state.vel[1]*ctrlMotorInstance.deltaT;
		float newOrient = AngleNormalize(ctrl.state.pos[2] + ctrl.state.vel[2]*ctrlMotorInstance.deltaT);
		ctrl.state.pos[2] = newOrient;
		ctrl.state.posUpdated = 1;
	}
}

static uint8_t isTrajActive()
{
	static const float postRun = 0.0f;
	if(	ctrlMotor.trajTime > Trajectory2DGetTotalTime(&ctrlMotor.trajXY)+postRun &&
		ctrlMotor.trajTime > TrajectoryOrientGetTotalTime(&ctrlMotor.trajW)+postRun)
		return 0;

	return 1;
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	(void)pOutput;
	(void)pInput;

	if(ctrlMotor.posMode)
	{
		const float trajDt = ctrlMotorInstance.deltaT;

		float vel[3] = {0, 0, 0};
		float acc[3];
		if(ctrlMotor.newSetpointPos)
		{
			ctrlMotor.newSetpointPos = 0;

			float pos[3] = {0, 0, 0};
			float Umax = ctrl.input.power.voltage;
			float velMax = ctrlMotor.configVel.maxXY;
			float velMaxW = ctrlMotor.configVel.maxW;

			if(ctrlMotor.configVel.overrideBat == 0)
				CtrlUtilGetVMax(&ctrl.perfConstants, Umax, &velMax, &velMaxW);

			if(ctrl.limitedVel && velMax > ctrl.botLimits.limitedVel)
				velMax = ctrl.botLimits.limitedVel;

			// copy state to local variable to prevent unaligned access faults
			memcpy(pos, ctrl.state.pos, sizeof(float)*3);
			memcpy(vel, ctrl.state.vel, sizeof(float)*3);

			Trajectory2DCreate(&ctrlMotor.trajXY, pos, ctrlMotor.setpointPos, vel,
					ctrl.botLimits.accMax, ctrl.botLimits.brkMax, velMax);

			TrajectoryOrientCreateAngles(&ctrlMotor.trajW, pos[2], ctrlMotor.setpointPos[2], vel[2],
					ctrl.botLimits.accMaxW, ctrl.botLimits.brkMaxW, velMaxW, &ctrlMotor.trajXY, velMax);

			ctrlMotor.trajTime = 0;
		}

		if(isTrajActive())
		{
			Trajectory2DValuesAtTime(&ctrlMotor.trajXY, ctrlMotor.trajTime, acc, vel, 0);
			TrajectoryOrientValuesAtTime(&ctrlMotor.trajW, ctrlMotor.trajTime, acc+2, vel+2, 0);

			memcpy(ctrl.state.acc, acc, sizeof(float)*3);
			ctrl.state.accUpdated = 1;

			ctrlMotor.trajTime += trajDt;

			CtrlTurnGlobal2Local(ctrl.state.pos[2], vel[0], vel[1], vel, vel+1);
			arm_matrix_instance_f32 matXYW = {3, 1, vel};
			arm_mat_mult_f32(&ctrl.matXYW2Motor, &matXYW, &ctrlMotor.matVelSetpoint);
		}
		else
		{
			arm_mat_zero_f32(&ctrlMotor.matVelSetpoint);

			memset(ctrl.state.acc, 0, sizeof(float)*3);
			ctrl.state.accUpdated = 1;
		}
	}

	// convert speed over ground to motor output
	for(uint8_t i = 0; i < 4; i++)
		pOutput->motorVel[i] = ctrlMotor.velSetpoint[i]*CTRL_WHEEL_TO_MOTOR_RATIO * 1/ctrl.botParams.wheelRadius;

	pOutput->ctrlMode = MOT_CTRL_VELOCITY;
}

static void input(uint8_t mode, float* pCmd)
{
	switch(mode)
	{
		case SKILL_OUT_MODE_LOCAL_VEL:
		{
			CtrlMotorSetVelTarget(pCmd[0], pCmd[1], pCmd[2]);
		}
		break;
		case SKILL_OUT_MODE_GLOBAL_POS:
		{
			CtrlMotorSetPosTarget(pCmd[0], pCmd[1], pCmd[2]);
		}
		break;
	}
}

void CtrlMotorSetVelTarget(float x, float y, float w)
{
	float xyw[3] = {x, y , w};
	arm_matrix_instance_f32 matXYW = {3, 1, xyw};

	float absVel = sqrtf(x*x + y*y);
	if(ctrl.limitedVel && absVel > ctrl.botLimits.limitedVel)
	{
		float f = ctrl.botLimits.limitedVel/absVel;
		xyw[0] *= f;
		xyw[1] *= f;
	}

	arm_mat_mult_f32(&ctrl.matXYW2Motor, &matXYW, &ctrlMotor.matVelSetpoint);

	ctrlMotor.posMode = 0;
}

void CtrlMotorSetPosTarget(float x, float y, float w)
{
	ctrlMotor.posMode = 1;

	float newPos[3] = {x, y, w};

	if(!isTrajActive())
	{
		memcpy(ctrlMotor.setpointPos, newPos, sizeof(float)*3);
		ctrlMotor.newSetpointPos = 1;
	}
}
