/*
 * ctrl_tigga.c
 *
 *  Created on: 26.02.2015
 *      Author: AndreR
 */

#include "ctrl_tigga.h"
#include "config_file_ids.h"
#include "util/arm_mat_util_f32.h"
#include "util/angle_math.h"
#include "util/sys_time.h"
#include "util/console.h"
#include "util/network_print.h"
#include "util/log.h"
#include "../ctrl.h"
#include "../hal/buzzer.h"

#include <math.h>

#include "stm32f407xx.h"

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);
static void sensorFusion(CtrlInput* pInput, const CtrlOutput* pOutput);
static void input(uint8_t mode, float* pCmd);

CtrlTigga ctrlTigga = {
		.configPos = {
			.pidParamsXY = {10, 0, 0.08, 1.5, -1.5},
			.pidParamsW = {30, 0, 1, 8, -8},
			.posCtrlInterval = 1,
			.posPostRuntime = 60.0f,
			.accFeedforward = 1,
			.trajTimeOffset = 0,
			.lagVelPT1TimeConstant = 0.05f,
		},
		.configFilter = {
			.statePos = 0.001f,
			.stateVel = 0.01f,
			.sensorVision = 0.01f,
			.sensorGyro = 0.05f,
			.sensorEncodersXYOffset = 0.01f,
			.sensorEncodersXYGain = 100.0f,
			.sensorEncodersW = 0.05f,
			.sensorAccelerometerOffset = 0.01f,
			.sensorAccelerometerGain = 1.0f,
			.visionDelay = 25.0f,
			.encMeasTrust = 0.3f,
		}, };

static const ConfigFileDesc configFileDescPos =
	{ CFG_ID_CTRL_TIGGA_POS, 6, "ctrl_tigga/pos", 15, (ConfigElementDesc[]) {
		{ CET_FLOAT, "pidPosXY/p" },
		{ CET_FLOAT, "pidPosXY/i" },
		{ CET_FLOAT, "pidPosXY/d" },
		{ CET_FLOAT, "pidPosXY/outMax" },
		{ CET_FLOAT, "pidPosXY/outMin" },
		{ CET_FLOAT, "pidPosW/p" },
		{ CET_FLOAT, "pidPosW/i" },
		{ CET_FLOAT, "pidPosW/d" },
		{ CET_FLOAT, "pidPosW/outMax" },
		{ CET_FLOAT, "pidPosW/outMin" },
		{ CET_UINT8, "Pos. Ctrl. Interval" },
		{ CET_FLOAT, "Keep pos. time [s]" },
		{ CET_UINT8, "Acceleration Feedforward [bool]" },
		{ CET_FLOAT, "Trajectory time offset [s]" },
		{ CET_FLOAT, "Lag Element Time const. [s]" },
	} };

static const ConfigFileDesc configFileDescFilter =
	{ CFG_ID_CTRL_TIGGA_FILTER, 11, "ctrl_tigga/filter", 11, (ConfigElementDesc[]) {
		{ CET_FLOAT, "state/pos" },
		{ CET_FLOAT, "state/vel" },
		{ CET_FLOAT, "sensor/vision" },
		{ CET_FLOAT, "sensor/gyr" },
		{ CET_FLOAT, "sensor/encXYOff" },
		{ CET_FLOAT, "sensor/encXYGain" },
		{ CET_FLOAT, "sensor/encW" },
		{ CET_FLOAT, "sensor/accOff" },
		{ CET_FLOAT, "sensor/accGain" },
		{ CET_FLOAT, "vision/delay" },
		{ CET_FLOAT, "enc/measTrust" },
	} };

CtrlInstance ctrlTiggaInstance = {
	.pInit = &init,
	.pExit = 0,
	.pStateEstimationFunc = &sensorFusion,
	.pControllerFunc = &controller,
	.pInput = &input,
	.deltaT = 0.002f };

static void trajectoryUpdate();
static void extrapolatePos(CtrlInput* pInput, float* newPos);
static void control(CtrlOutput* pOutput);

static void configUpdateCallback(uint16_t cfgId)
{
	switch(cfgId)
	{
		case CFG_ID_CTRL_TIGGA_POS:
		{
			PIDInit(&ctrlTigga.pidPos[0], &ctrlTigga.configPos.pidParamsXY, ctrlTiggaInstance.deltaT*ctrlTigga.configPos.posCtrlInterval, 1);
			PIDInit(&ctrlTigga.pidPos[1], &ctrlTigga.configPos.pidParamsXY, ctrlTiggaInstance.deltaT*ctrlTigga.configPos.posCtrlInterval, 1);
			PIDInit(&ctrlTigga.pidPos[2], &ctrlTigga.configPos.pidParamsW, ctrlTiggaInstance.deltaT*ctrlTigga.configPos.posCtrlInterval, 1);
			for(uint8_t i = 0; i < 3; i++)
			{
				LagElementPT1Init(&ctrlTigga.velLag[i], 1, ctrlTigga.configPos.lagVelPT1TimeConstant, ctrlTiggaInstance.deltaT);
			}
		}
		break;
		case CFG_ID_CTRL_TIGGA_FILTER:
		{
			// KF
			for(uint8_t i = 0; i < 3; i++)
			{
				MAT_ELEMENT(ctrlTigga.kf[i].Ex, 0, 0) = ctrlTigga.configFilter.statePos;
				MAT_ELEMENT(ctrlTigga.kf[i].Ex, 1, 1) = ctrlTigga.configFilter.stateVel;
				MAT_ELEMENT(ctrlTigga.kf[i].Sigma, 0, 0) = ctrlTigga.configFilter.statePos;
				MAT_ELEMENT(ctrlTigga.kf[i].Sigma, 1, 1) = ctrlTigga.configFilter.stateVel;

				MAT_ELEMENT(ctrlTigga.kf[i].Ez, 0, 0) = ctrlTigga.configFilter.sensorVision;
				MAT_ELEMENT(ctrlTigga.kf[i].Ez, 1, 1) = ctrlTigga.configFilter.sensorEncodersXYOffset;
			}
			MAT_ELEMENT(ctrlTigga.kf[2].Ez, 1, 1) = ctrlTigga.configFilter.sensorEncodersW;

			MAT_ELEMENT(ctrlTigga.kf[0].Ez, 2, 2) = ctrlTigga.configFilter.sensorAccelerometerOffset;
			MAT_ELEMENT(ctrlTigga.kf[1].Ez, 2, 2) = ctrlTigga.configFilter.sensorAccelerometerOffset;
			MAT_ELEMENT(ctrlTigga.kf[2].Ez, 2, 2) = ctrlTigga.configFilter.sensorGyro;
		}
		break;
	}
}

static void init()
{
	FlashFSOpenOrCreate("ctrl_tigga/enc", 0, &ctrlTigga.enc, sizeof(ctrlTigga.enc), &ctrlTigga.pEncFile);

	ctrlTigga.matOutXYWForce = (arm_matrix_instance_f32){3, 1, ctrlTigga.outXYWForce};
	ctrlTigga.matOutMotorTorque = (arm_matrix_instance_f32){4, 1, ctrlTigga.outMotorTorque};

	ctrlTigga.log.header.type = CTYPE_CTRL_TIGGA_PIDS;
	ctrlTigga.log.header.length = sizeof(LogCtrlTiggaPIDs);

	// state vector: [p, v]
	// sensor vector: [vision, encoder, gyro/acc]
	for(uint8_t i = 0; i < 3; i++)
	{
		KFInit(&ctrlTigga.kf[i], 2, 1, 3, ctrlTigga.kfData[i]);

		KF* pKF = &ctrlTigga.kf[i];

		arm_mat_identity_f32(&pKF->A);
		MAT_ELEMENT(pKF->A, 0, 1) = ctrlTiggaInstance.deltaT;

		MAT_ELEMENT(pKF->B, 0, 0) = 0.5f*ctrlTiggaInstance.deltaT*ctrlTiggaInstance.deltaT;
		MAT_ELEMENT(pKF->B, 1, 0) = ctrlTiggaInstance.deltaT;

		MAT_ELEMENT(pKF->C, 0, 0) = 0;
		MAT_ELEMENT(pKF->C, 1, 1) = 1;
		MAT_ELEMENT(pKF->C, 2, 1) = 1;
	}

	KFSetOrientationComponent(&ctrlTigga.kf[2], 0);

	LagElementPT1Init(&ctrlTigga.accPT1[0], 1, 0.5f, ctrlTiggaInstance.deltaT);
	LagElementPT1Init(&ctrlTigga.accPT1[1], 1, 0.5f, ctrlTiggaInstance.deltaT);

	ConfigNotifyUpdate(ctrlTigga.pConfigFilePos);
	ConfigNotifyUpdate(ctrlTigga.pConfigFileFilter);

	const float measAlpha = 0.98f;
	EMAFilterInit(&ctrlTigga.avgTraj, measAlpha);
	EMAFilterInit(&ctrlTigga.avgSensorFusion, measAlpha);
	EMAFilterInit(&ctrlTigga.avgControl, measAlpha);
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	(void)pInput;

	uint32_t start;

//	start = DWT->CYCCNT;
//	sensorFusion(pInput, pOutput);
//	EMAFilterUpdate(&ctrlTigga.avgSensorFusion, DWT->CYCCNT-start);

	start = DWT->CYCCNT;
	trajectoryUpdate();
	EMAFilterUpdate(&ctrlTigga.avgTraj, DWT->CYCCNT-start);

	start = DWT->CYCCNT;
	control(pOutput);
	EMAFilterUpdate(&ctrlTigga.avgControl, DWT->CYCCNT-start);

	// send logs to media for USB logging
	ctrlTigga.log.header.timestamp = SysTimeUSec();

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_LOGFILE;
	iHeader.command = IC_LOGFILE_MESSAGE;

	if(ctrl.botLogging.ctrlTiggaPIDs)
		IntercomSend(&iHeader, sizeof(LogCtrlTiggaPIDs), &ctrlTigga.log);
}

static void trajectoryUpdate()
{
	const float trajDt = ctrlTiggaInstance.deltaT;

	ctrlTigga.log.trajActive = ctrlTigga.enableTraj;

	if(!ctrlTigga.enableTraj)
		return;

	float pos[3] = {0, 0, 0};
	float vel[3] = {0, 0, 0};
	float accNext[3] = {0, 0, 0};
//	float Umax = 10.5f;
	float Umax = ctrl.input.power.voltage-ctrl.botLimits.batBuffer;

	if(ctrlTigga.newSetpointPos)
	{
		float velMax;
		float velMaxW;

		CtrlUtilGetVMax(&ctrl.perfConstants, Umax, &velMax, &velMaxW);

		if(ctrl.limitedVel && velMax > ctrl.botLimits.limitedVel)
			velMax = ctrl.botLimits.limitedVel;

		// copy state to local variable to prevent unaligned access faults
		memcpy(pos, ctrl.state.pos, sizeof(float)*3);
		memcpy(vel, ctrl.state.vel, sizeof(float)*3);

		if(ctrlTigga.transitionTime >= 0 && ctrlTigga.trajTime < Trajectory2DGetTotalTime(&ctrlTigga.trajXY))
		{
			// reuse XY-Velocity
			Trajectory2DValuesAtTime(&ctrlTigga.trajXY, ctrlTigga.transitionTime, 0, vel, pos);
		}

		Trajectory2DCreate(&ctrlTigga.trajXY, pos, ctrlTigga.setpointPos, vel,
				ctrl.botLimits.accMax, ctrl.botLimits.brkMax, velMax);

		TrajectoryOrientCreateAngles(&ctrlTigga.trajW, pos[2], ctrlTigga.setpointPos[2], vel[2],
				ctrl.botLimits.accMaxW, ctrl.botLimits.brkMaxW, velMaxW, &ctrlTigga.trajXY, velMax);

		ctrlTigga.trajTime = trajDt + ctrlTigga.configPos.trajTimeOffset;

		ctrlTigga.log.newTrajTime = SysTimeUSec();

		// target position changed
		ctrlTigga.newSetpointPos = 0;
	}

	if(ctrlTigga.trajTime > Trajectory2DGetTotalTime(&ctrlTigga.trajXY)+ctrlTigga.configPos.posPostRuntime &&
			ctrlTigga.trajTime > TrajectoryOrientGetTotalTime(&ctrlTigga.trajW)+ctrlTigga.configPos.posPostRuntime)
	{
		// trajectory finished
		ctrlTigga.enableTraj = 0;

		accNext[0] = 0;
		accNext[1] = 0;
		accNext[2] = 0;
		vel[0] = 0;
		vel[1] = 0;
		vel[2] = 0;
	}
	else
	{
		float futureTraj = ctrlTigga.trajTime;
		Trajectory2DValuesAtTime(&ctrlTigga.trajXY, futureTraj, accNext, 0, 0);
		TrajectoryOrientValuesAtTime(&ctrlTigga.trajW, futureTraj, accNext+2, 0, 0);

		// and another future point for position control setpoint
		futureTraj = ctrlTigga.trajTime+(ctrlTigga.configPos.posCtrlInterval/2)*trajDt;

		Trajectory2DValuesAtTime(&ctrlTigga.trajXY, futureTraj, 0, 0, pos);
		TrajectoryOrientValuesAtTime(&ctrlTigga.trajW, futureTraj, 0, 0, pos+2);

		// and the current one for velocity
		Trajectory2DValuesAtTime(&ctrlTigga.trajXY, ctrlTigga.trajTime, 0, vel, 0);
		TrajectoryOrientValuesAtTime(&ctrlTigga.trajW, ctrlTigga.trajTime, 0, vel+2, 0);

		ctrlTigga.trajTime += trajDt;

		memcpy(ctrlTigga.log.trajPos, pos, sizeof(float)*3);
	}

	memcpy(ctrlTigga.log.trajVel, vel, sizeof(float)*3);
	memcpy(ctrlTigga.log.trajAcc, accNext, sizeof(float)*3);
}

float CtrlTiggaGetEncoderCoverage()
{
	float cover = 0;
	for(uint16_t i = 0; i < 127; i++)
		cover += ctrlTigga.encCover[i];

	cover /= 128;

	return cover;
}

static void sensorFusion(CtrlInput* pInput, const CtrlOutput* pOutput)
{
	// control data is not used yet
	(void)pOutput;

	uint32_t start = DWT->CYCCNT;

	float extPos[3];
	extrapolatePos(pInput, extPos);

	if(pInput->pos.updated)
	{
		MAT_ELEMENT(ctrlTigga.kf[0].C, 0, 0) = 1.0f;
		MAT_ELEMENT(ctrlTigga.kf[1].C, 0, 0) = 1.0f;
		MAT_ELEMENT(ctrlTigga.kf[2].C, 0, 0) = 1.0f;

//		MAT_ELEMENT(ctrlTigga.kf[0].z, 0, 0) = extPos[0];
//		MAT_ELEMENT(ctrlTigga.kf[1].z, 0, 0) = extPos[1];
//		MAT_ELEMENT(ctrlTigga.kf[2].z, 0, 0) = extPos[2];
		MAT_ELEMENT(ctrlTigga.kf[0].z, 0, 0) = pInput->pos.pos[0];
		MAT_ELEMENT(ctrlTigga.kf[1].z, 0, 0) = pInput->pos.pos[1];
		MAT_ELEMENT(ctrlTigga.kf[2].z, 0, 0) = pInput->pos.pos[2];
	}
	else
	{
		MAT_ELEMENT(ctrlTigga.kf[0].C, 0, 0) = 0.0f;
		MAT_ELEMENT(ctrlTigga.kf[1].C, 0, 0) = 0.0f;
		MAT_ELEMENT(ctrlTigga.kf[2].C, 0, 0) = 0.0f;
	}

	// predict now to get the best orientation estimate
	KFPredict(&ctrlTigga.kf[0]);
	KFPredict(&ctrlTigga.kf[1]);
	KFPredict(&ctrlTigga.kf[2]);
	AngleNormalizePtr(&ctrlTigga.kf[2].x.pData[0]);

	// current global orientation
	float orient = MAT_ELEMENT(ctrlTigga.kf[2].x, 0, 0);

	// now process encoders
	arm_matrix_instance_f32 encMat = (arm_matrix_instance_f32){4, 1, (float[4]){}};
	arm_matrix_instance_f32 encVelLocalMat = (arm_matrix_instance_f32){3, 1, (float[3]){}};

	memcpy(encMat.pData, pInput->enc.vel, sizeof(float)*4);

	for(uint8_t i = 0; i < 4; i++)
		encMat.pData[i] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now speed over ground [m/s]

	arm_mat_mult_f32(&ctrl.matMotor2XYW, &encMat, &encVelLocalMat);
	float* pEncVel = encVelLocalMat.pData;

	float encGlobal[2];
	CtrlTurnLocal2Global(orient, pEncVel[0], pEncVel[1], encGlobal, encGlobal+1);

	MAT_ELEMENT(ctrlTigga.kf[0].z, 1, 0) = encGlobal[0];
	MAT_ELEMENT(ctrlTigga.kf[1].z, 1, 0) = encGlobal[1];
	MAT_ELEMENT(ctrlTigga.kf[2].z, 1, 0) = pEncVel[2];

	// Adjust measurement error linear to velocity
	MAT_ELEMENT(ctrlTigga.kf[0].Ez, 1, 1) = ctrlTigga.configFilter.sensorEncodersXYOffset
			+ ctrlTigga.configFilter.sensorEncodersXYGain*fabsf(encGlobal[0]);
	MAT_ELEMENT(ctrlTigga.kf[1].Ez, 1, 1) = ctrlTigga.configFilter.sensorEncodersXYOffset
			+ ctrlTigga.configFilter.sensorEncodersXYGain*fabsf(encGlobal[1]);

	// Accelerometers
	float encSpeedXY = sqrtf(pEncVel[0]*pEncVel[0] + pEncVel[1]*pEncVel[1]);
	if(encSpeedXY < 0.002f)
	{
		ctrlTigga.accBias[0] = LagElementPT1Process(&ctrlTigga.accPT1[0], pInput->acc.linAcc[0]);
		ctrlTigga.accBias[1] = LagElementPT1Process(&ctrlTigga.accPT1[1], pInput->acc.linAcc[1]);
		ctrlTigga.accVel[0] = 0;
		ctrlTigga.accVel[1] = 0;

		MAT_ELEMENT(ctrlTigga.kf[0].Ez, 2, 2) -= ctrlTigga.configFilter.sensorAccelerometerGain*ctrlTiggaInstance.deltaT*10.0f;
		MAT_ELEMENT(ctrlTigga.kf[1].Ez, 2, 2) -= ctrlTigga.configFilter.sensorAccelerometerGain*ctrlTiggaInstance.deltaT*10.0f;

		if(MAT_ELEMENT(ctrlTigga.kf[0].Ez, 2, 2) < ctrlTigga.configFilter.sensorAccelerometerOffset)
			MAT_ELEMENT(ctrlTigga.kf[0].Ez, 2, 2) = ctrlTigga.configFilter.sensorAccelerometerOffset;

		if(MAT_ELEMENT(ctrlTigga.kf[1].Ez, 2, 2) < ctrlTigga.configFilter.sensorAccelerometerOffset)
			MAT_ELEMENT(ctrlTigga.kf[1].Ez, 2, 2) = ctrlTigga.configFilter.sensorAccelerometerOffset;
	}
	else
	{
		ctrlTigga.accVel[0] += (pInput->acc.linAcc[0]-ctrlTigga.accBias[0])*ctrlTiggaInstance.deltaT;
		ctrlTigga.accVel[1] += (pInput->acc.linAcc[1]-ctrlTigga.accBias[1])*ctrlTiggaInstance.deltaT;
		MAT_ELEMENT(ctrlTigga.kf[0].Ez, 2, 2) += ctrlTigga.configFilter.sensorAccelerometerGain*ctrlTiggaInstance.deltaT;
		MAT_ELEMENT(ctrlTigga.kf[1].Ez, 2, 2) += ctrlTigga.configFilter.sensorAccelerometerGain*ctrlTiggaInstance.deltaT;
	}

	float globalAcc[2];
	CtrlTurnLocal2Global(orient, ctrlTigga.accVel[0], ctrlTigga.accVel[1], globalAcc, globalAcc+1);
	MAT_ELEMENT(ctrlTigga.kf[0].z, 2, 0) = globalAcc[0];
	MAT_ELEMENT(ctrlTigga.kf[1].z, 2, 0) = globalAcc[1];

	// process gyro
	MAT_ELEMENT(ctrlTigga.kf[2].z, 2, 0) = pInput->gyr.rotVel[2];

	// KF measurements are now all set, perform update
	KFUpdate(&ctrlTigga.kf[0]);
	KFUpdate(&ctrlTigga.kf[1]);
	KFUpdate(&ctrlTigga.kf[2]);
	AngleNormalizePtr(&ctrlTigga.kf[2].x.pData[0]);

	arm_mat_clip_f32(&ctrlTigga.kf[0].Sigma, -100, 100);
	arm_mat_clip_f32(&ctrlTigga.kf[1].Sigma, -100, 100);
	arm_mat_clip_f32(&ctrlTigga.kf[2].Sigma, -100, 100);

	// now update states
	ctrl.state.pos[0] = ctrlTigga.kf[0].x.pData[0];
	ctrl.state.pos[1] = ctrlTigga.kf[1].x.pData[0];
	ctrl.state.pos[2] = ctrlTigga.kf[2].x.pData[0];
	ctrl.state.vel[0] = ctrlTigga.kf[0].x.pData[1];
	ctrl.state.vel[1] = ctrlTigga.kf[1].x.pData[1];
	ctrl.state.vel[2] = ctrlTigga.kf[2].x.pData[1];
	ctrl.state.acc[0] = ctrlTigga.kf[0].u.pData[0];
	ctrl.state.acc[1] = ctrlTigga.kf[1].u.pData[0];
	ctrl.state.acc[2] = ctrlTigga.kf[2].u.pData[0];

	ctrl.state.posUpdated = 1;
	ctrl.state.velUpdated = 1;
	ctrl.state.accUpdated = 1;

	// TODO: testing
	float localVel[2];
	CtrlTurnGlobal2Local(orient, ctrl.state.vel[0], ctrl.state.vel[1], localVel, localVel+1);
	ctrlTigga.log.pidVel[0].in = localVel[0];
	ctrlTigga.log.pidVel[0].out = ctrlTigga.accVel[0];
	ctrlTigga.log.pidVel[1].in = localVel[1];
	ctrlTigga.log.pidVel[1].out = ctrlTigga.accVel[1];

	EMAFilterUpdate(&ctrlTigga.avgSensorFusion, DWT->CYCCNT-start);
}

static void extrapolatePos(CtrlInput* pInput, float* newPos)
{
	uint32_t timeBackup = pInput->pos.time;

	pInput->pos.time -= (uint32_t) (ctrlTigga.configFilter.visionDelay * 1000.0f);

	// store global velocity in state buffer
	ctrlTigga.velStateBuffer[ctrlTigga.velStateBufIndex].timestamp = SysTimeUSec();
	ctrlTigga.velStateBuffer[ctrlTigga.velStateBufIndex].vel[0] = ctrlTigga.kf[0].x.pData[1];
	ctrlTigga.velStateBuffer[ctrlTigga.velStateBufIndex].vel[1] = ctrlTigga.kf[1].x.pData[1];
	ctrlTigga.velStateBuffer[ctrlTigga.velStateBufIndex++].vel[2] = ctrlTigga.kf[2].x.pData[1];
	ctrlTigga.velStateBufIndex %= CTRL_TIGGA_VEL_STATE_BUFFER_SIZE;

	// ### state updates!
	// position
	if(pInput->pos.updated && pInput->pos.time > ctrlTigga.velStateBuffer[ctrlTigga.velStateBufIndex].timestamp)
	{
		memcpy(newPos, pInput->pos.pos, sizeof(float)*3);

		// go through velocity state buffer, oldest entry first
		for(uint16_t i = 0; i < CTRL_TIGGA_VEL_STATE_BUFFER_SIZE; i++)
		{
			CtrlTiggaVelState* pEntry = &ctrlTigga.velStateBuffer[(ctrlTigga.velStateBufIndex+i)%CTRL_TIGGA_VEL_STATE_BUFFER_SIZE];
			if(pEntry->timestamp > pInput->pos.time)
			{
				// this entry occurred between "vision time" and "now" => apply it to input pos
				newPos[0] += pEntry->vel[0]*ctrlTiggaInstance.deltaT;
				newPos[1] += pEntry->vel[1]*ctrlTiggaInstance.deltaT;
				newPos[2] += pEntry->vel[2]*ctrlTiggaInstance.deltaT;
				AngleNormalizePtr(newPos+2);
			}
		}
	}
	else
	{
		// TODO: only extrapolating once does not seem right
		// if there is no position update use linear velocity model to predict position
		for(uint8_t i = 0; i < 3; i++)
			ctrl.state.pos[i] += ctrl.state.vel[i] * ctrlTiggaInstance.deltaT;

		ctrl.state.pos[2] = AngleNormalize(ctrl.state.pos[2]);
	}

	pInput->pos.time = timeBackup;
}

static void control(CtrlOutput* pOutput)
{
	static uint8_t posCounter = 0;
	static float posOutput[3];

	// gather setpoints
	float* setPos = ctrlTigga.log.trajPos;
	float setVel[3] = {0, 0, ctrlTigga.log.trajVel[2]};

	// turn to local frame
	CtrlTurnGlobal2Local(ctrl.state.pos[2], ctrlTigga.log.trajVel[0], ctrlTigga.log.trajVel[1], setVel, setVel+1);

	// ### CONTROL

	if(!ctrlTigga.posActive)
	{
		ctrlTigga.kf[0].u.pData[0] = 0;
		ctrlTigga.kf[1].u.pData[0] = 0;
		ctrlTigga.kf[2].u.pData[0] = 0;

		for(uint8_t i = 0; i < 3; i++)
		{
			if(ctrl.motorsOn)
				setVel[i] = LagElementPT1Process(&ctrlTigga.velLag[i], ctrlTigga.setpointVel[i]);
			else
				setVel[i] = LagElementPT1Process(&ctrlTigga.velLag[i], 0);
		}
	}
	else
	{
		// position controller
		if(posCounter < ctrlTigga.configPos.posCtrlInterval)
		{
			posCounter++;
		}
		else
		{
			posCounter = 0;

			ctrlTigga.pidPos[0].setpoint = setPos[0];
			ctrlTigga.pidPos[1].setpoint = setPos[1];
			ctrlTigga.pidPos[2].setpoint = setPos[2];

			PIDUpdate(&ctrlTigga.pidPos[0], ctrl.state.pos[0]);
			PIDUpdate(&ctrlTigga.pidPos[1], ctrl.state.pos[1]);
			PIDUpdateAngular(&ctrlTigga.pidPos[2], ctrl.state.pos[2]);

			memcpy(&ctrlTigga.log.pidPos[0], &ctrlTigga.pidPos[0].log, sizeof(PIDLog));
			memcpy(&ctrlTigga.log.pidPos[1], &ctrlTigga.pidPos[1].log, sizeof(PIDLog));
			memcpy(&ctrlTigga.log.pidPos[2], &ctrlTigga.pidPos[2].log, sizeof(PIDLog));

			posOutput[2] = ctrlTigga.pidPos[2].output;

			CtrlTurnGlobal2Local(ctrl.state.pos[2], ctrlTigga.pidPos[0].output,
					ctrlTigga.pidPos[1].output, posOutput, posOutput+1);
		}

		setVel[0] += posOutput[0];
		setVel[1] += posOutput[1];
		setVel[2] += posOutput[2];
	}

	// put acc in KF input
//	ctrlTigga.kf[0].u.pData[0] = ctrlTigga.log.trajAcc[0];
//	ctrlTigga.kf[1].u.pData[0] = ctrlTigga.log.trajAcc[1];
//	ctrlTigga.kf[2].u.pData[0] = ctrlTigga.log.trajAcc[2];

	arm_matrix_instance_f32 matVelOut = {3, 1, setVel};
	arm_matrix_instance_f32 matMotVel = {4, 1, (float[4]){}};

	arm_mat_mult_f32(&ctrl.matXYW2Motor, &matVelOut, &matMotVel);

	// convert speed over ground to motor output
	for(uint8_t i = 0; i < 4; i++)
		pOutput->motorVel[i] = matMotVel.pData[i]*CTRL_WHEEL_TO_MOTOR_RATIO * 1/ctrl.botParams.wheelRadius;

	pOutput->ctrlMode = MOT_CTRL_VELOCITY;
}

static void input(uint8_t mode, float* pCmd)
{
	switch(mode)
	{
		case SKILL_OUT_MODE_LOCAL_VEL:
		{
			CtrlTiggaSetVelTarget(pCmd[0], pCmd[1], pCmd[2]);
		}
		break;
		case SKILL_OUT_MODE_GLOBAL_VEL:
		{
			// TODO: implement
		}
		break;
		case SKILL_OUT_MODE_GLOBAL_POS:
		{
			CtrlTiggaSetPosTarget(pCmd[0], pCmd[1], pCmd[2], pCmd[3]);
		}
		break;
	}
}

void CtrlTiggaInitCfg()
{
	ctrlTigga.pConfigFilePos = ConfigOpenOrCreate(&configFileDescPos, &ctrlTigga.configPos, sizeof(CtrlTiggaConfigPos), &configUpdateCallback);
	ctrlTigga.pConfigFileFilter = ConfigOpenOrCreate(&configFileDescFilter, &ctrlTigga.configFilter, sizeof(CtrlTiggaConfigFilter), &configUpdateCallback);
}

void CtrlTiggaSetPIDPosParamsXY(float p, float i, float d, float max)
{
	PIDParams params = {p, i, d, max, -max};

	for(uint8_t i = 0; i < 2; i++)
		PIDSetParamsStruct(&ctrlTigga.pidPos[i], &params);
}

void CtrlTiggaSetPIDPosParamsW(float p, float i, float d, float max)
{
	PIDParams params = {p, i, d, max, -max};

	PIDSetParamsStruct(&ctrlTigga.pidPos[2], &params);
}

void CtrlTiggaSetVelTarget(float x, float y, float w)
{
	ctrlTigga.enableTraj = 0;
	ctrlTigga.posActive = 0;

	float absVel = sqrtf(x*x + y*y);
	if(ctrl.limitedVel && absVel > ctrl.botLimits.limitedVel)
	{
		float f = ctrl.botLimits.limitedVel/absVel;
		x *= f;
		y *= f;
	}

	ctrlTigga.setpointVel[0] = x;
	ctrlTigga.setpointVel[1] = y;
	ctrlTigga.setpointVel[2] = w;
}

void CtrlTiggaSetPosTarget(float x, float y, float w, float t)
{
	if(!ctrlTigga.posActive)
	{
		ctrlTigga.newSetpointPos = 1;
	}
	ctrlTigga.enableTraj = 1;
	ctrlTigga.posActive = 1;

	uint8_t reset = t < 0 && ctrlTigga.transitionTime != t;
	uint8_t posChanged = ctrlTigga.setpointPos[0] != x || ctrlTigga.setpointPos[1] != y || ctrlTigga.setpointPos[2] != w;
	if(reset || (posChanged && ctrlTigga.trajTime >= t))
	{
		ctrlTigga.setpointPos[0] = x;
		ctrlTigga.setpointPos[1] = y;
		ctrlTigga.setpointPos[2] = w;
		ctrlTigga.setpointVel[0] = 0;
		ctrlTigga.setpointVel[1] = 0;
		ctrlTigga.setpointVel[2] = 0;
		ctrlTigga.newSetpointPos = 1;
		ConsolePrint("New setpoints %f %f %f - %f\r\n", x, y, w, t);
	}
	ctrlTigga.transitionTime = t;
}

void CtrlTiggaSetPosVelTarget(float* data)
{
	ctrlTigga.enableTraj = 0;
	ctrlTigga.posActive = 1;

	memcpy(ctrlTigga.log.trajPos, data+3, sizeof(float)*3);
	memcpy(ctrlTigga.log.trajVel, data, sizeof(float)*3);
	memset(ctrlTigga.log.trajAcc, 0, sizeof(float)*3);
}

/* Code Graveyard
static void trainEncoders(const float* trueVel, const float* encVel)
{
	const float minVelAbs = 0.5f;
	const float angle2Index = 128.0f/(2*M_PI);
	const float angleStep = 1.0f/angle2Index;
	const float measTrust = ctrlTigga.configFilter.encMeasTrust;

	float trueVelAbs = sqrtf(trueVel[0]*trueVel[0]+trueVel[1]*trueVel[1]);
	if(trueVelAbs < minVelAbs)
	{
		// this is too slow for being a reliable dataset
		return;
	}

	float angle = arm_atan2_f32(trueVel[1], trueVel[0]) + M_PI;

	float error[3];
	error[0] = (encVel[0]-trueVel[0])/trueVelAbs;
	error[1] = (encVel[1]-trueVel[1])/trueVelAbs;
	error[2] = (encVel[2]-trueVel[2])/trueVelAbs;

	// find indices which are closest to the given angle
	uint16_t left = (uint16_t)(angle*angle2Index);
	uint16_t right = left+1;

	float leftAngle = angleStep*((float)left);
	float rightAngle = angleStep*((float)right);

	float rightFactor = (angle-leftAngle)*angle2Index;
	float leftFactor = (rightAngle-angle)*angle2Index;

	left &= 0x7F;
	right &= 0x7F;

	if(ctrlTigga.encCover[left] < 100)
		++ctrlTigga.encCover[left];

	if(ctrlTigga.encCover[right] < 100)
		++ctrlTigga.encCover[right];

//	ConsolePrint("Angle: %f\r\n", angle);
//	ConsolePrint("Left, %f, %f\r\n", leftAngle, leftFactor);
//	ConsolePrint("Right, %f, %f\r\n", rightAngle, rightFactor);

	// use exponential moving average filter to update support points
	ctrlTigga.enc.xy2x[left] = (1.0f - measTrust*leftFactor)*ctrlTigga.enc.xy2x[left] + (measTrust*leftFactor)*error[0];
	ctrlTigga.enc.xy2x[right] = (1.0f - measTrust*rightFactor)*ctrlTigga.enc.xy2x[right] + (measTrust*rightFactor)*error[0];

	ctrlTigga.enc.xy2y[left] = (1.0f - measTrust*leftFactor)*ctrlTigga.enc.xy2y[left] + (measTrust*leftFactor)*error[1];
	ctrlTigga.enc.xy2y[right] = (1.0f - measTrust*rightFactor)*ctrlTigga.enc.xy2y[right] + (measTrust*rightFactor)*error[1];

	ctrlTigga.enc.xy2w[left] = (1.0f - measTrust*leftFactor)*ctrlTigga.enc.xy2w[left] + (measTrust*leftFactor)*error[2];
	ctrlTigga.enc.xy2w[right] = (1.0f - measTrust*rightFactor)*ctrlTigga.enc.xy2w[right] + (measTrust*rightFactor)*error[2];

	if(ctrlTigga.enc.saved == 0 && CtrlTiggaGetEncoderCoverage() > 95.0f)
	{
		ctrlTigga.enc.saved = 1;
		FlashFSWrite(&ctrlTigga.pEncFile, &ctrlTigga.enc, sizeof(ctrlTigga.enc));
		BuzzerPlay(&buzzSeqTada);
	}
}

static void correctEncOut(float* pState, float* pOut)
{
	const float angle2Index = 128.0f/(2*M_PI);
	const float angleStep = 1.0f/angle2Index;

	float angle = arm_atan2_f32(pState[1], pState[0]) + M_PI;
	float velAbs = sqrtf(pState[0]*pState[0]+pState[1]*pState[1]);

	// find indices which are closest to the given angle
	uint16_t left = (uint16_t)(angle*angle2Index);
	uint16_t right = left+1;

	float leftAngle = angleStep*((float)left);
	float rightAngle = angleStep*((float)right);

	float rightFactor = (angle-leftAngle)*angle2Index;
	float leftFactor = (rightAngle-angle)*angle2Index;

	left &= 0x7F;
	right &= 0x7F;

	float cor[3];

	cor[0] = ctrlTigga.enc.xy2x[left]*leftFactor + ctrlTigga.enc.xy2x[right]*rightFactor;
	cor[1] = ctrlTigga.enc.xy2y[left]*leftFactor + ctrlTigga.enc.xy2y[right]*rightFactor;
	cor[2] = ctrlTigga.enc.xy2w[left]*leftFactor + ctrlTigga.enc.xy2w[right]*rightFactor;

	pOut[0] = pState[0] + cor[0]*velAbs;
	pOut[1] = pState[1] + cor[1]*velAbs;
	pOut[2] = pState[2] + cor[2]*velAbs;
}

static void correctEncMeas(float* pState, float* pMeas, float* pOut)
{
	const float angle2Index = 128.0f/(2*M_PI);
	const float angleStep = 1.0f/angle2Index;

	float angle = arm_atan2_f32(pState[1], pState[0]) + M_PI;
	float velAbs = sqrtf(pState[0]*pState[0]+pState[1]*pState[1]);

	// find indices which are closest to the given angle
	uint16_t left = (uint16_t)(angle*angle2Index);
	uint16_t right = left+1;

	float leftAngle = angleStep*((float)left);
	float rightAngle = angleStep*((float)right);

	float rightFactor = (angle-leftAngle)*angle2Index;
	float leftFactor = (rightAngle-angle)*angle2Index;

	left &= 0x7F;
	right &= 0x7F;

	float cor[3];

	cor[0] = ctrlTigga.enc.xy2x[left]*leftFactor + ctrlTigga.enc.xy2x[right]*rightFactor;
	cor[1] = ctrlTigga.enc.xy2y[left]*leftFactor + ctrlTigga.enc.xy2y[right]*rightFactor;
	cor[2] = ctrlTigga.enc.xy2w[left]*leftFactor + ctrlTigga.enc.xy2w[right]*rightFactor;

	pOut[0] = pMeas[0] - cor[0]*velAbs;
	pOut[1] = pMeas[1] - cor[1]*velAbs;
	pOut[2] = pMeas[2] - cor[2]*velAbs;
}

static void updateState()
{
	KFStartLookAhead(&ctrlTigga.botKF);

	uint16_t steps = ctrl.logInterval/1000+1;
	if(steps > 50)
		steps = 50;

	for(uint16_t i = 0; i < steps; i++)
	{
		KFDoLookAhead(&ctrlTigga.botKF);

		float trajTime = ctrlTigga.trajTime+i*ctrlTiggaInstance.deltaT;
		Trajectory2DValuesAtTime(&ctrlTigga.trajXY, trajTime, ctrlTigga.botKF.u.pData, 0, 0);
		TrajectoryOrientValuesAtTime(&ctrlTigga.trajW, trajTime, ctrlTigga.botKF.u.pData+2, 0, 0);
	}

	memcpy(ctrl.state.pos, ctrlTigga.botKF.x.pData, sizeof(float)*3);
	memcpy(ctrl.state.vel, ctrlTigga.botKF.x.pData+3, sizeof(float)*3);

	KFFinishLookAhead(&ctrlTigga.botKF);
}
 */
