/*
 * buzzer.c
 *
 *  Created on: 23.12.2010
 *      Author: AndreR
 */

#include "buzzer.h"

#include "stm32f407xx.h"
#include "util/init_hal.h"

Buzzer buzzer;

const BuzzerSequence buzzSeqBeepFast = {
		(uint16_t[]){1, 100, 0, 100},
		4, 0,
};

const BuzzerSequence buzzSeqDoubleBeepSlow = {
		(uint16_t[]){1, 100, 0, 100, 1, 100, 0, 2000},
		8, 0,
};

const BuzzerSequence buzzSeqUp100 = {
		(uint16_t[]){2500, 100, 3000, 100, 3500, 100},
		6, 1,
};

const BuzzerSequence buzzSeqTada = {
		(uint16_t[]){3200, 200, 3000, 200, 2500, 200, 3200, 300, 2500, 100, 3200, 500},
		12, 1,
};

const BuzzerSequence buzzSeqDown100 = {
		(uint16_t[]){3500, 100, 3000, 100, 2500, 100},
		6, 1,
};

static uint16_t buzzSeqUp50Data[32];

const BuzzerSequence buzzSeqUp50 = {
		buzzSeqUp50Data, 32, 1,
};

static uint16_t buzzSeqDown50Data[32];

const BuzzerSequence buzzSeqDown50 = {
		buzzSeqDown50Data, 32, 1,
};

static uint16_t buzzSeqUp20Data[32];

const BuzzerSequence buzzSeqUp20 = {
		buzzSeqUp20Data, 32, 1,
};

static uint16_t buzzSeqUpDown20Data[64];

const BuzzerSequence buzzSeqUpDown20 = {
		buzzSeqUpDown20Data, 64, 0,
};

void BuzzerInit()
{
	RCC->APB2ENR |= RCC_APB2ENR_TIM1EN;	// enable TIM1 clock

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 1;
	gpioInit.ospeed = GPIO_OSPEED_F4_2MHZ;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_9, &gpioInit);

	TIM1->CR1 = TIM_CR1_ARPE;	// enable ARR preload
	TIM1->CR2 = 0;
	TIM1->CCMR1 = TIM_CCMR1_OC2M;	// PWM mode 2
	TIM1->CCER = TIM_CCER_CC2E | TIM_CCER_CC2P;	// enable CC2 output and set polarity to active low
	TIM1->PSC = 83;		// prescaler
	TIM1->ARR = 320;	// period
	TIM1->CCR2 = 103;	// pulse
	TIM1->BDTR = 0;	// disable outputs
	TIM1->EGR = TIM_EGR_UG;	// generate update event to take new ARR value
	TIM1->CR1 |= TIM_CR1_CEN;	// enable timer

	// initalize buzzer sequences
	for(uint16_t i = 0; i < 32; i += 2)
	{
		buzzSeqUp50Data[i] = 2300+i*50;
		buzzSeqUp50Data[i+1] = 50;

		buzzSeqUp20Data[i] = 2300+i*50;
		buzzSeqUp20Data[i+1] = 20;

		buzzSeqUpDown20Data[i] = 2300+i*50;
		buzzSeqUpDown20Data[i+1] = 20;
	}

	for(uint16_t i = 32; i < 64; i += 2)
	{
		buzzSeqUpDown20Data[i] = 3900-(i-32)*50;
		buzzSeqUpDown20Data[i+1] = 20;

		buzzSeqDown50Data[i-32] = 3500-(i-32)*50;
		buzzSeqDown50Data[i-31] = 50;
	}
}

void BuzzerTone(uint16_t freq)
{
	if(freq == 1)
		freq = 3350;

	if(freq < 16 || freq > 20000)
	{
		TIM1->BDTR = 0;
		return;
	}

	TIM1->BDTR = TIM_BDTR_MOE;

	uint16_t arr = 1000000/freq;

	TIM1->ARR = arr;
	TIM1->CCR2 = arr/2;

	TIM1->EGR = TIM_EGR_UG;
}

static void buzzTimer(void* param)
{
	(void)param;

	if(!buzzer.pSeq)
		return;

	if(buzzer.curTone == 0 && buzzer.pSeq->once)
	{
		buzzer.pSeq = 0;
		BuzzerTone(0);
		return;
	}

	BuzzerTone(buzzer.pSeq->pSong[buzzer.curTone]);

	chVTSetI(&buzzer.timer, MS2ST(buzzer.pSeq->pSong[buzzer.curTone+1]), &buzzTimer, 0);

	buzzer.curTone += 2;
	buzzer.curTone %= buzzer.pSeq->numTones;
}

void BuzzerPlay(const BuzzerSequence* pSeq)
{
	if(pSeq == buzzer.pSeq)
		return;

	chVTReset(&buzzer.timer);
	buzzer.pSeq = 0;
	BuzzerTone(0);

	if(pSeq == 0)
		return;

	if(pSeq->numTones % 2 == 1 || pSeq->numTones == 0)
		return;

	buzzer.curTone = 2;
	buzzer.pSeq = pSeq;

	BuzzerTone(buzzer.pSeq->pSong[0]);
	chVTSet(&buzzer.timer, MS2ST(buzzer.pSeq->pSong[1]), &buzzTimer, 0);
}
