#include "ch.h"

#include "hal/mco.h"
#include "hal/led.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "hal/usart1.h"
#include "hal/usart2.h"
#include "hal/usart6.h"
#include "hal/hw_sync1.h"

#include "util/boot.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/sys_time.h"
#include "util/network_print.h"
#include "util/terminal.h"
#include "util/intercom_reliable.h"
#include "util/prog_client.h"

#include "cli.h"
#include "network.h"
#include "spi2.h"
#include "spi3.h"
#include "intercom1.h"
#include "hal/power.h"
#include "motors.h"
#include "ctrl.h"
#include "ctrl/ctrl_tigga.h"
#include "ui.h"
#include "config_main.h"
#include "skills.h"

static WORKING_AREA(waUSART2, 512);
static WORKING_AREA(waSPI3, 512);
static WORKING_AREA(waSPI2, 512);
static WORKING_AREA(waUSART6, 512);
static WORKING_AREA(waIntercom, 512);
static WORKING_AREA(waTerminal, 4096);
#ifdef ENV_RUN
static WORKING_AREA(waCLI, 8192);
static WORKING_AREA(waNetwork, 4096);
static WORKING_AREA(waNetworkRel, 4096);
static WORKING_AREA(waUI, 512);
static WORKING_AREA(waCTRL, 8192);
static WORKING_AREA(waMotors, 4096);
static WORKING_AREA(waConfig, 512);
#endif

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	// Enable important clocks
	RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN | RCC_AHB1ENR_DMA2EN | RCC_AHB1ENR_GPIOAEN
			| RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_GPIODEN;
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;

	// Enable compensation cell to reduce noise on power lines with 50MHz/100MHz I/Os
	SYSCFG->CMPCR = SYSCFG_CMPCR_CMP_PD;

	// Enable usage fault, bus fault, and mem fault
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	MPU->CTRL = MPU_CTRL_ENABLE_Msk | MPU_CTRL_PRIVDEFENA_Msk;
	// Region number: 0
	MPU->RNR = 0;
	// Base address: flash
	MPU->RBAR = 0x08000000;
	// Size: 1MB, disable for 384k-512k (FlashFS, sub-region 3), Read-Only
	MPU->RASR = MPU_RASR_ENABLE_Msk | (19 << 1) | (0b1000 << 8) | MPU_RASR_C_Msk | (0x7 << 24);

//	SCB->CCR |= SCB_CCR_DIV_0_TRP_Msk;

	// initialize Data Watch Trace support (instruction cycle counting)
	DWT->CTRL = DWT_CTRL_CYCCNTENA_Msk | DWT_CTRL_NUMCOMP_Msk;

	SystemClockInitData clkInit;
	clkInit.pll.M = 12;
	clkInit.pll.N = 336;
	clkInit.pll.P = 2;
	clkInit.pll.Q = 7;
	clkInit.HSEBypass = 1;
	clkInit.APB1Div = 4;	// = 42MHz
	clkInit.APB2Div = 4;	// = 42MHz
	clkInit.RTCDiv = 12;	// = 1MHz
	clkInit.flashLatency = 5;
	clkInit.sysTickFreq = CH_FREQUENCY;
	SystemClockInit(&clkInit, 168000000UL);

	chSysInit();

	// set vector table offset
#ifdef ENV_BOOT
	SCB->VTOR = 0;
#else
	SCB->VTOR = BOOT_APP_OFFSET;
#endif

	CPULoadCalibrate(10);
	chThdSleep(20);

	// --- HAL ---
	LEDInit();		// init LEDs to signal RTOS startup failures
	HWSync1Init();	// Init sync mechanism for timed distributed startup
	PowerInit();
	PowerEnable5VSupply(1);
	FlashFSInit(0x08060000, 0x20000);	// 128k
	ConfigMainInit();
	BuzzerInit();
	ExtiInit();
	USART1Init();
	USART2Init();
	USART6Init();
	SPI2Init();
	SPI3Init();
	CRCInit();

	// --- High-Level Systems ---
	TerminalInit(&usart1.inQueue, &usart1.outQueue);
	Intercom1Init();
	IntercomReliableInit();
	ProgClientInit();

#ifdef ENV_BOOT
	BuzzerPlay(&buzzSeqUp20);
	chThdSleepMilliseconds(100);
//	ConsolePrint("\f--- Bootloader (MAIN) ---\r\n");
#else
	CLIInit();
	ConsolePrint("\f--- Main ---\r\n");
#endif

#ifdef ENV_RUN
	SkillsInit();
	NetworkInit();
	NetworkPrintInit(&NetworkSendPacketRaw, INTERCOM_ADDRESS_MAIN);
	MotorsInit();
	UIInit();
	CtrlInit();
	CtrlTiggaInitCfg();
	CtrlSetController(&ctrlTiggaInstance);
#endif

	// now start motor left/right and kicker processor
	MCOInit();

#ifdef ENV_RUN
	chThdCreateStatic(waCTRL, sizeof(waCTRL), NORMALPRIO+5, CtrlTask, 0);
	chThdCreateStatic(waMotors, sizeof(waMotors), NORMALPRIO-5, MotorsTask, 0);
	chThdCreateStatic(waNetwork, sizeof(waNetwork), NORMALPRIO+1, NetworkTask, 0);
	chThdCreateStatic(waNetworkRel, sizeof(waNetworkRel), NORMALPRIO-4, NetworkReliableTask, 0);
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-2, CLITask, 0);
	chThdCreateStatic(waConfig, sizeof(waConfig), NORMALPRIO-8, ConfigTask, 0);
#endif
	chThdCreateStatic(waUSART2, sizeof(waUSART2), NORMALPRIO, USARTTask, &usart2);
	chThdCreateStatic(waUSART6, sizeof(waUSART6), NORMALPRIO, USARTTask, &usart6);
	chThdCreateStatic(waSPI3, sizeof(waSPI3), NORMALPRIO+10, SPI3Task, 0);
	chThdCreateStatic(waIntercom, sizeof(waIntercom), NORMALPRIO-1, IntercomTask, 0);
	chThdCreateStatic(waTerminal, sizeof(waTerminal), NORMALPRIO-2, TerminalTask, 0);
	chThdCreateStatic(waSPI2, sizeof(waSPI2), NORMALPRIO-3, SPI2Task, 0);

	HWSyncBarrier(&hwSync1);

#ifdef ENV_RUN
	chThdCreateStatic(waUI, sizeof(waUI), NORMALPRIO-10, UITask, 0);
#endif

/*
	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	LEDMustOff(LED_GREEN | LED_RED);

	while(1)
	{
		chThdSleepMilliseconds(10);

#ifdef ENV_RUN
		uint32_t time = SysTimeUSec()%1000000;

		if(time < 200000)
			LEDMustOn(LED_GREEN);
		else
			LEDMustOff(LED_GREEN);
#else
		uint32_t time = SysTimeUSec()%500000;

		if(time < 100000)
			LEDMustOn(LED_RED);
		else
			LEDMustOff(LED_RED);
#endif
	}

	return 0;
}
