/*
 * motors.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

#include "motors.h"
#include "spi3.h"
#include "hal/exti.h"
#include "constants.h"
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/boot.h"
#include "util/init_hal.h"
#include "hal/buzzer.h"
#include "ctrl.h"
#include "log_msgs.h"

#include "ch.h"

Motors motors;

#define EVENT_MASK_MOTOR_FAULT 0x0F

static void sendFaultEvent(uint8_t event)
{
	msg_t msg = event;

	chSysLockFromIsr();
	chMBPostI(&motors.eventQueue, msg);
	chSysUnlockFromIsr();
}

// Motor 2
void EXTI1_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR1;

	sendFaultEvent(1);

	CH_IRQ_EPILOGUE();
}

// Motor 3
void EXTI2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR2;

	sendFaultEvent(2);

	CH_IRQ_EPILOGUE();
}

static void faultM1()
{
	sendFaultEvent(0);
}

static void faultM4()
{
	sendFaultEvent(3);
}

static void faultM5()
{
	sendFaultEvent(4);
}

static void spi3AboutToSend(SPI3Slave* pSlave, uint8_t slaveID)
{
	(void)pSlave;

	IntercomHeader iHeader;
	iHeader.command = IC_MOTORS_MAIN2MOT;
	iHeader.section = INTERCOM_SECTION_MOTORS;

	MotorsMain2Mot main2Mot;
	main2Mot.bat = ctrl.lastKdFeedback.power.bat;
	main2Mot.ctrlMode = motors.ctrlMode;

	if(slaveID == SPI3_SLAVE_RIGHT)
	{
		iHeader.address = INTERCOM_ADDRESS_RIGHT;

		memcpy(&main2Mot.mot[MOT_FRONT], &motors.mot[MOT_FR].set, sizeof(MotorSetpoint));
		memcpy(&main2Mot.mot[MOT_REAR], &motors.mot[MOT_RR].set, sizeof(MotorSetpoint));
	}
	else
	{
		iHeader.address = INTERCOM_ADDRESS_LEFT;

		memcpy(&main2Mot.mot[MOT_FRONT], &motors.mot[MOT_FL].set, sizeof(MotorSetpoint));
		memcpy(&main2Mot.mot[MOT_REAR], &motors.mot[MOT_RL].set, sizeof(MotorSetpoint));
	}

	IntercomSend(&iHeader, sizeof(main2Mot), (uint8_t*)&main2Mot);
}

static void motorsCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_MOTORS_MOT2MAIN && dataLength >= sizeof(MotorsMot2Main))
	{
		MotorsMot2Main* pMot2Main = (MotorsMot2Main*)pData;

		if(pMot2Main->source == MOT_SOURCE_RIGHT)
		{
			memcpy(&motors.mot[MOT_FR].cur, &pMot2Main->mot[MOT_FRONT], sizeof(MotorFeedback));
			memcpy(&motors.mot[MOT_RR].cur, &pMot2Main->mot[MOT_REAR], sizeof(MotorFeedback));

			motors.lastMotRx[0] = chTimeNow();
		}
		else
		{
			memcpy(&motors.mot[MOT_FL].cur, &pMot2Main->mot[MOT_FRONT], sizeof(MotorFeedback));
			memcpy(&motors.mot[MOT_RL].cur, &pMot2Main->mot[MOT_REAR], sizeof(MotorFeedback));

			motors.lastMotRx[1] = chTimeNow();
		}

		if(motors.updateCb)
			(*motors.updateCb)(pMot2Main->source);
	}
}

void MotorsInit()
{
	motors.ctrlMode = MOT_CTRL_VELOCITY;
	chMBInit(&motors.eventQueue, motors.eventQueueBuf, MOTORS_EVENT_QUEUE_SIZE);

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.pupd = GPIO_PUPD_UP;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;
	GPIOInit(GPIOA, GPIO_PIN_10 | GPIO_PIN_12, &gpioInit);
	GPIOInit(GPIOC, GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_8, &gpioInit);

	NVICEnableIRQ(EXTI1_IRQn, IRQL_MOTOR_FAULT);
	NVICEnableIRQ(EXTI2_IRQn, IRQL_MOTOR_FAULT);

	ExtiSetCallback(&faultM1, 12);
	ExtiSetCallback(&faultM4, 8);
	ExtiSetCallback(&faultM5, 10);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &motorsCallback;
	client.section = INTERCOM_SECTION_MOTORS;

	IntercomAddClient(client);

	spi3.aboutToSendCb = &spi3AboutToSend;
}

msg_t MotorsTask(void* params)
{
	(void)params;

	chRegSetThreadName("Motors");

	msg_t event;

	chThdSleep(MS2ST(2000));

	motors.lastMotRx[0] = chTimeNow();
	motors.lastMotRx[1] = chTimeNow();

	while(1)
	{
		if(chMBFetch(&motors.eventQueue, &event, MS2ST(20)) == RDY_OK)
		{
			if(event & EVENT_MASK_MOTOR_FAULT)
				ConsolePrint("Motor %u fault!\r\n", event);
		}

		if(chTimeElapsedSince(motors.lastMotRx[0]) > MS2ST(500) && chTimeElapsedSince(motors.lastMotRx[0]) < MS2ST(1500))
		{
//			BuzzerPlay(&buzzSeqUp100);
//			BootReset();
		}

		if(chTimeElapsedSince(motors.lastMotRx[1]) > MS2ST(500) && chTimeElapsedSince(motors.lastMotRx[1]) < MS2ST(1500))
		{
//			BuzzerPlay(&buzzSeqDown100);
		}
	}

	return 0;
}
