/*
 * dect.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include "network.h"

#include "stm32f407xx.h"
#include "constants.h"

#include "util/bit_banding.h"
#include "util/console.h"
#include "errors.h"
#include "commands.h"
#include "util/float16_t.h"
#include "intercom_constants.h"
#include "intercom1.h"
#include "util/boot.h"
#include "util/sys_time.h"
#include "ctrl/ctrl_motor.h"
#include "ctrl/ctrl_calibrate.h"
#include "ctrl/ctrl_tigga.h"
#include "hal/power.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/intercom_reliable.h"
#include "intercom1.h"
#include "util/log.h"
#include "hal/buzzer.h"
#include "motors.h"
#include "ui.h"
#include "cli.h"
#include "config_main.h"
#include "skills.h"

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#define EVENT_REL_ACK 0x10000
#define EVENT_REL_RUN 0x20000

Network network = {
	.config = { 0xFF, 100, 2 },
};

static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength);

void DMA2_Stream0_IRQHandler()
{
	SPISyncDmaRxInterrupt(&network.nrf24.spi);
}

void EXTI4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = (1 << 4);

	NRF24ExIRQ(&network.nrf24);

	CH_IRQ_EPILOGUE();
}

void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;
	TIM7->CR1 = 0;

	chSysLockFromIsr();
	chBSemSignalI(&network.timerSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

static void startNRF24Timeout(uint32_t microSeconds)
{
	TIM7->ARR = microSeconds;
	TIM7->CR1 = TIM_CR1_CEN;
}

static void networkIntercomCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_NETWORK_NORMAL)
		NetworkSendPacket((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);

	if(pHeader->command == IC_NETWORK_RELIABLE)
		NetworkSendPacketReliable((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);
}

void NetworkInit()
{
	// data setup
	chBSemInit(&network.timerSem, TRUE);
	chMtxInit(&network.printMutex);
	chMtxInit(&network.sendPacketMutex);

	FifoLinInit(&network.reliableCmds, network.reliableBuf, NETWORK_RELIABLE_BUF_SIZE);
	chMBInit(&network.relEvents, network.relEventsData, NETWORK_REL_EVENT_SIZE);

	network.botDisabled = 0;

	FlashFSOpenOrCreate("net/config", 2, &network.config, sizeof(ConfigNetwork), &network.pConfigFile);

	network.posCb = 0;
	network.avgRxPeriod = 0;
	network.lastUSec = 0;

	// NRF24 IRQ line setup
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.pupd = GPIO_PUPD_UP;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;
	GPIOInit(GPIOC, GPIO_PIN_4, &gpioInit);

	NVICEnableIRQ(EXTI4_IRQn, IRQL_SPI1);

	// TIM7 setup
	RCC->APB1ENR |= RCC_APB1ENR_TIM7EN;

	TIM7->CR1 = 0;
	TIM7->CR2 = 0;
	TIM7->SR = 0;
	TIM7->DIER = TIM_DIER_UIE;
	TIM7->PSC = 83; 	// clock is now 1MHz
	TIM7->ARR = 1000;	// = 1ms

	NVICEnableIRQ(TIM7_IRQn, IRQL_NRF24_TIMER);

	// SPI1 init
	RCC->APB2ENR |= RCC_APB2ENR_SPI1EN;

	NVICEnableIRQ(DMA2_Stream0_IRQn, IRQL_SPI1);

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_25MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7, &gpioInit);

	// NRF24 setup
	NRF24ExData data;

	static uint8_t rxData[NRF24_RX_SIZE];
	static uint8_t txData[NRF24_TX_SIZE];

	NRF24QueueInit(&network.queue, txData, NRF24_TX_SIZE, rxData, NRF24_RX_SIZE);
	network.queue.id = network.config.botId;

	data.spiData.csPin = GPIO_PIN_2;
	data.spiData.pCSPort = GPIOB;
	data.spiData.pRegister = SPI1;
	data.spiData.pDma = DMA2;
	data.spiData.dmaChRx = 0;
	data.spiData.dmaChTx = 3;
	data.spiData.timeoutTicks = MS2ST(20);
	data.rx = 1;
	data.channel = network.config.channel;
	data.speed = network.config.speed;
	data.pCEPort = GPIOA;
	data.cePin = GPIO_PIN_4;

	NRF24ExInit(&network.nrf24, &data);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &networkIntercomCallback;
	client.section = INTERCOM_SECTION_NETWORK;

	IntercomAddClient(client);
}

void NetworkSendAnnouncement(uint32_t id)
{
	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_ANNOUNCEMENT;

	SystemAnnouncement an;
	an.cpuId[0] = id;
	an.cpuId[1] = 0;
	an.cpuId[2] = 0;

	NetworkSendPacketReliable(&header, (uint8_t*)&an, sizeof(SystemAnnouncement));
}

void NetworkSaveConfig()
{
	FlashFSWrite(&network.pConfigFile, &network.config, sizeof(ConfigNetwork));
}

msg_t NetworkTask(void* params)
{
	(void)params;

	chRegSetThreadName("Network");

	uint32_t len;
	int16_t result;
	uint32_t numDelimiters = 0;
	uint32_t i;
	network.avgRxPeriod = 99999;

	NRF24ExSetup(&network.nrf24);
	NRF24ExUseQueue(&network.nrf24, &network.queue);

	network.lastMatchCtrlTime = chTimeNow()-MS2ST(2000);

	while(1)
	{
		if(network.botDisabled)
		{
			chThdSleep(MS2ST(100));
			continue;
		}

		result = NRF24ExWaitForIRQ(&network.nrf24, MS2ST(100));
		if(result)
		{
			LogErrorC("NRF24ExWaitForIRQ", result);

			NRF24ExSetup(&network.nrf24);
			NRF24ExUseQueue(&network.nrf24, &network.queue);
		}

		const uint32_t alpha = 96;
		uint32_t tNow = SysTimeUSec();
		uint32_t tPeriod = tNow - network.lastUSec;
		network.lastUSec = tNow;
		network.avgRxPeriod = (alpha*network.avgRxPeriod + (100-alpha)*tPeriod)/100;

		if(network.avgRxPeriod > 80000 && network.bsOnline)
		{
			ConsolePrint("BS offline\r\n");
			network.bsOnline = 0;

			CtrlSetMotorsOn(0);
		}

		if(network.avgRxPeriod < 80000 && !network.bsOnline)
		{
			ConsolePrint("BS online\r\n");
			network.bsOnline = 1;

			NetworkSendAnnouncement(42);
		}

		uint32_t readDelay;
		uint32_t writeDelay;

		NRF24ExGetReadWriteDelay(&network.nrf24, &readDelay, &writeDelay);

		chBSemReset(&network.timerSem, TRUE);
		startNRF24Timeout(readDelay);
		if(chBSemWaitTimeout(&network.timerSem, MS2ST(100)) != RDY_OK)
		{
			ConsolePrint("RxDownDelay error\r\n");
			LogError("Timeout");
		}

		result = NRF24ExRxDownload(&network.nrf24);
		if(result)
			ConsolePrint("NRF24ExRxDownload result: 0x%02X\r\n", result);

		chBSemReset(&network.timerSem, TRUE);
		startNRF24Timeout(writeDelay);

		numDelimiters = network.queue.rxFifo.numPackets;
		for(i = 0; i < numDelimiters; i++)
		{
			result = NRF24QueueGetPacket(&network.queue, network.rxProcessingBuf, NETWORK_RX_PROCESSING_SIZE, &len);
			if(result)
			{
				ConsolePrint("NRF24QueueGetPacket failed: 0x%0X\r\n", result);
				LogErrorC("GetPacket", result);
				break;
			}

			if(len > PACKET_HEADER_SIZE)
			{
				PacketHeader* pHeader = (PacketHeader*)network.rxProcessingBuf;
				uint8_t* pData = network.rxProcessingBuf+PACKET_HEADER_SIZE;

				processPacket(pHeader, pData, len-PACKET_HEADER_SIZE);
			}
		}

		if(chTimeElapsedSince(network.lastUIWifiStatTime) > MS2ST(50))
		{
			network.lastUIWifiStatTime = chTimeNow();

			IntercomHeader iHeader;
			iHeader.address = INTERCOM_ADDRESS_MEDIA;
			iHeader.section = INTERCOM_SECTION_PRESENTER;
			iHeader.command = IC_PRESENTER_WIFI_STAT;

			PresenterWifiStat stat;
			stat.bsOnline = network.bsOnline;
			stat.sumatraOnline = network.sumatraOnline;
			stat.updateFreq = 1000000/network.avgRxPeriod;
			stat.visionDelay = network.lastVisionDelay;

			IntercomSend(&iHeader, sizeof(PresenterWifiStat), &stat);
		}

		if(chTimeElapsedSince(network.lastMatchCtrlTime) > MS2ST(1000))
		{
			if(network.sumatraOnline)
			{
				CtrlSetMotorsOn(0);
				SkillsSetInput(0, 0);
				network.sumatraOnline = 0;

				BuzzerPlay(&buzzSeqDown50);
			}
		}
		else
		{
			if(network.sumatraOnline == 0)
			{
				network.sumatraOnline = 1;

				BuzzerPlay(&buzzSeqUp50);
			}
		}

		if(chBSemWaitTimeout(&network.timerSem, MS2ST(100)) != RDY_OK)
			ConsolePrint("RxUpDelay error\r\n");

		result = NRF24ExRxUpload(&network.nrf24);
		if(result)
			ConsolePrint("NRF24ExRxUpload result: 0x%02X\r\n", result);
	}

	return 0;
}

msg_t NetworkReliableTask(void* params)
{
	(void)params;
	msg_t event;
	msg_t chResult;

	uint8_t* pBuf = 0;
	PacketHeaderEx* pHeaderEx = 0;
	uint32_t cmdSize;
	uint32_t lastSendTime = chTimeNow()-MS2ST(200);

	chRegSetThreadName("NetRel");

	while(1)
	{
		chResult = chMBFetch(&network.relEvents, &event, MS2ST(10));

		if(FifoLinGet(&network.reliableCmds, &pBuf, &cmdSize) == 0)
			pHeaderEx = (PacketHeaderEx*)pBuf;
		else
			pHeaderEx = 0;

		if(chResult == RDY_OK && (event & EVENT_REL_ACK) && pHeaderEx)
		{
			if((event & 0xFFFF) == pHeaderEx->seq)
			{
				FifoLinDelete(&network.reliableCmds);

//				ConsolePrint("Pkt %hu ack'ed\r\n", pHeaderEx->seq);

				if(FifoLinGet(&network.reliableCmds, &pBuf, &cmdSize) == 0)
					pHeaderEx = (PacketHeaderEx*)pBuf;
				else
					pHeaderEx = 0;

				lastSendTime = chTimeNow()-MS2ST(200);
			}
		}

		if(pHeaderEx && chTimeElapsedSince(lastSendTime) > MS2ST(100))
		{
			chMtxLock(&network.sendPacketMutex);	// prevent re-entering

			// data to send is now in network.txProcessingBuf
			int16_t result = NRF24QueueSend(&network.queue, pBuf, cmdSize);
			if(result)
			{
				FifoLinClear(&network.queue.txFifo);
			}

			chMtxUnlock();

			lastSendTime = chTimeNow();

//			ConsolePrint("Pkt %hu sent\r\n", pHeaderEx->seq);
		}
	}

	return 0;
}

void NetworkSetPosCallback(PosCallback cb)
{
	network.posCb = cb;
}

int16_t NetworkSendPacket(PacketHeader* pHeader, void* _pData, uint16_t dataLength)
{
	uint8_t* pData = (uint8_t*)_pData;

	if(dataLength + PACKET_HEADER_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	memcpy(network.txProcessingBuf, pHeader->_data, PACKET_HEADER_SIZE);
	memcpy(network.txProcessingBuf + PACKET_HEADER_SIZE, pData, dataLength);

	// data to send is now in network.txProcessingBuf
	int16_t result = NRF24QueueSend(&network.queue, network.txProcessingBuf, dataLength+PACKET_HEADER_SIZE);
	if(result)
	{
		LogErrorC("NRF24QueueSend error", result);
	}
	if(result)
	{
		FifoLinClear(&network.queue.txFifo);
		LogWarn("txFifo full => cleared\r\n");
	}

	chMtxUnlock();

	return result;
}

void NetworkSendPacketRaw(void* _pData, uint16_t dataLength)
{
	uint8_t* pData = (uint8_t*)_pData;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	// data to send is now in network.txProcessingBuf
	int16_t result = NRF24QueueSend(&network.queue, pData, dataLength);
	if(result)
	{
		LogErrorC("NRF24QueueSend error", result);
	}
	if(result)
	{
		FifoLinClear(&network.queue.txFifo);
		LogWarn("txFifo full => cleared\r\n");
	}

	chMtxUnlock();
}

int16_t NetworkSendPacketReliable(PacketHeader* pHeader, void* _pData, uint16_t dataLength)
{
	int16_t result;
	uint8_t* pBuf;
	uint8_t* pData = (uint8_t*)_pData;

	result = FifoLinReserve(&network.reliableCmds, dataLength+PACKET_HEADER_EX_SIZE, &pBuf);
	if(result)
		return result;

	PacketHeaderEx* pHeaderEx = (PacketHeaderEx*)pBuf;
	pHeaderEx->cmd = pHeader->cmd;
	pHeaderEx->section = pHeader->section | SECTION_EXTENDED_PACKET;
	pHeaderEx->seq = network.seq++;

	memcpy(pBuf+PACKET_HEADER_EX_SIZE, pData, dataLength);

	result = FifoLinCommit(&network.reliableCmds, dataLength+PACKET_HEADER_EX_SIZE);
	if(result)
		return result;

	chMBPost(&network.relEvents, EVENT_REL_RUN, TIME_IMMEDIATE);

	return 0;
}

// this is the main network processing function
static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength)
{
//	ConsolePrint("Sec/Cmd: %hhu, %hhu\r\n", pHeader->section, pHeader->cmd);

	if(pHeader->section & SECTION_EXTENDED_PACKET)
	{
		PacketHeaderEx* pHeaderEx = (PacketHeaderEx*)pHeader;

		PacketHeader ack;
		ack.cmd = CMD_SYSTEM_ACK;
		ack.section = SECTION_SYSTEM;

		NetworkSendPacket(&ack, &pHeaderEx->seq, sizeof(uint16_t));

//		ConsolePrint("Recv reliable cmd, ack'ed: %hu", pHeaderEx->seq);

		uint8_t* pBufOrig = pBuf;
		pHeader->section &= ~SECTION_EXTENDED_PACKET;
		pBuf += PACKET_HEADER_EX_SIZE-PACKET_HEADER_SIZE;
		dataLength -= PACKET_HEADER_EX_SIZE-PACKET_HEADER_SIZE;

		memcpy(pBufOrig, pHeader, sizeof(PacketHeader));
		pHeader = (PacketHeader*)pBufOrig;
	}

	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_PING:
				{
					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;

					NetworkSendPacket(&header, pBuf, dataLength);

					return 1;
				}
				break;
				case CMD_SYSTEM_CONSOLE_COMMAND:
				{
					SystemConsoleCommand* pCmd = (SystemConsoleCommand*)pBuf;

					if(pCmd->target == INTERCOM_ADDRESS_MAIN)
					{
						// process onboard
						uint8_t* pData;
						uint32_t length = dataLength-sizeof(SystemConsoleCommand);
						int16_t result = FifoLinReserve(&console.rxFifo, length, &pData);
						if(result)
							return 0;

						memcpy(pData, pBuf+sizeof(SystemConsoleCommand), length);

						FifoLinCommit(&console.rxFifo, length);
					}
					else
					{
						if(cli.intercomFwd.inUse == 0)
						{
							cli.intercomFwd.bufSize = dataLength-sizeof(SystemConsoleCommand);

							if(cli.intercomFwd.bufSize <= CLI_INTERCOM_BUF_SIZE)
							{
								cli.intercomFwd.inUse = 1;
								cli.intercomFwd.target = pCmd->target;

								memcpy(cli.intercomFwd.buf, pBuf+sizeof(SystemConsoleCommand), cli.intercomFwd.bufSize);
							}
						}
					}
				}
				break;
				case CMD_SYSTEM_ACK:
				{
					uint16_t* pSeq = (uint16_t*)pBuf;

					chMBPost(&network.relEvents, EVENT_REL_ACK + *pSeq, TIME_IMMEDIATE);
				}
				break;
				case CMD_SYSTEM_MATCH_CTRL:
				{
					SystemMatchCtrl* pCtrl = (SystemMatchCtrl*)pBuf;

					network.lastMatchCtrlTime = chTimeNow();

					CtrlSetLogRate(pCtrl->feedbackFreq);

					if(pCtrl->curPosition[0] != SYSTEM_MATCH_CTRL_UNUSED_FIELD)
					{
						uint32_t delay = pCtrl->posDelay;
						network.lastVisionDelay = (delay*25000)/100;	// delay now in [us]
						network.lastVisionPosTimestamp = SysTimeUSec()-network.lastVisionDelay;
						memcpy(network.lastVisionPos, pCtrl->curPosition, sizeof(int16_t)*3);

						if(network.posCb)
							(*network.posCb)();
					}

					if(pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_CHEER)
					{
						BuzzerPlay(&buzzSeqUpDown20);
					}
					else
					{
						if(buzzer.pSeq == &buzzSeqUpDown20)
						{
							BuzzerPlay(0);
						}
					}

					if(pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_LIMITED_VEL)
					{
						ctrl.limitedVel = 1;
					}
					else
					{
						ctrl.limitedVel = 0;
					}

					if(network.skillOverride == 0)
						SkillsSetInput(pCtrl->skillId, pCtrl->skillData);

					// forward to KD to process kicker and dribbler
					IntercomHeader iHeader;
					iHeader.address = INTERCOM_ADDRESS_KD;
					iHeader.section = INTERCOM_SECTION_KD;
					iHeader.command = IC_KD_MATCH_COMMAND;

					KdCommand kd;
					kd.seq = pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_SEQ_MASK;
					kd.autoChg = pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_AUTOCHG_MASK;
					kd.device = (pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_DEVICE_MASK);
					kd.mode = (pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_MODE_MASK) >> 1;
					kd.duration = pCtrl->kickDuration*1e-6f;
					kd.dribblerSpeed = pCtrl->dribblerSpeed;

					IntercomSend(&iHeader, sizeof(KdCommand), &kd);
				}
				break;
				case CMD_SYSTEM_QUERY:
				{
					SystemQuery* pQry = (SystemQuery*)pBuf;

					switch(pQry->parameterId)
					{
						case SYSTEM_QUERY_PID_PARAMS:
						{
						}
						break;
						case SYSTEM_QUERY_EX_STATE:
						case SYSTEM_QUERY_EZ_SENSOR:
						{
						}
						break;
						case SYSTEM_QUERY_KICKER_CONFIG:
						{
							PacketHeader header;
							header.section = SECTION_KICKER;
							header.cmd = CMD_KICKER_CONFIG;

							NetworkSendPacketReliable(&header, &ctrl.lastKdFeedback.config, sizeof(KickerConfig));
						}
						break;
						default: break;
					}
				}
				break;
			}
		}
		break;
		case SECTION_BOOTLOADER:
		{
			if(pHeader->cmd == CMD_BOOTLOADER_CHECK)
			{
				IntercomHeader iHeader;
				iHeader.address = INTERCOM_ADDRESS_MEDIA;
				iHeader.section = INTERCOM_SECTION_NETWORK;
				iHeader.command = 1;

				IntercomReliableSend(&iHeader, 0, 0);
			}
			else
			{
				IntercomHeader iHeader;
				iHeader.address = INTERCOM_ADDRESS_MEDIA;
				iHeader.section = INTERCOM_SECTION_NETWORK;
				iHeader.command = 0;

				IntercomSend(&iHeader, dataLength+PACKET_HEADER_SIZE, pHeader);
			}
		}
		break;

#ifdef ENV_RUN
		case SECTION_KICKER:
		{
			switch(pHeader->cmd)
			{
				case CMD_KICKER_CONFIG:
				{
					IntercomHeader iHeader;
					iHeader.address = INTERCOM_ADDRESS_KD;
					iHeader.section = INTERCOM_SECTION_KD;
					iHeader.command = IC_KD_CONFIG;

					IntercomSend(&iHeader, sizeof(KickerConfig), pBuf);
				}
				break;
			}
		}
		break;
		case SECTION_MOTOR:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOTOR_SET_PID_PARAMS:
				{
				}
				break;
			}
		}
		break;
		case SECTION_CTRL:
		{
			switch(pHeader->cmd)
			{
				case CMD_CTRL_SET_FILTER_PARAMS:
				{
//					CtrlSetFilterParams* pParams = (CtrlSetFilterParams*)pBuf;

//					CtrlFusionSetFilterParams(pParams);
				}
				break;
				case CMD_CTRL_SET_PID_PARAMS:
				{
//					CtrlSetPIDParams* pParams = (CtrlSetPIDParams*)pBuf;

//					CtrlFusionSetPIDParams(pParams);
				}
				break;
				case CMD_CTRL_SET_CONTROLLER_TYPE:
				{
					CtrlSetControllerType* pCtrl = (CtrlSetControllerType*)pBuf;

//					SkillReset();
					switch(pCtrl->controllerId)
					{
						case CTRL_TYPE_CALIBRATE:
						{
							CtrlSetController(&ctrlCalibrateInstance);
						}
						break;
						case CTRL_TYPE_MOTOR:
						{
							CtrlSetController(&ctrlMotorInstance);
						}
						break;
						case CTRL_TYPE_FUSION:
						{
//							CtrlSetController(&ctrlFusionInstance);
						}
						break;
						case CTRL_TYPE_TIGGA:
						{
							CtrlSetController(&ctrlTiggaInstance);
						}
						break;
						default:
						{
							CtrlSetController(0);
						}
						break;
					}
					ConsolePrint("Controller changed to %u\r\n", pCtrl->controllerId);
				}
				break;
				case CMD_CTRL_RESET:
				{
//					SkillReset();
//					CtrlFusionReset();
				}
				break;
			}
		}
		break;
		case SECTION_CONFIG:
		{
			ConfigNetworkInput(pHeader, pBuf, dataLength);
		}
		break;
#endif
		default: break;
	}

	return 0;
}
