/*
 * skill_enc_train.c
 *
 *  Created on: 25.06.2015
 *      Author: AndreR
 */

#include "skill_enc_train.h"
#include "util/angle_math.h"
#include "arm_math.h"
#include "../ctrl.h"
#include "util/sys_time.h"
#include "util/ukf.h"
#include "util/arm_mat_util_f32.h"
#include "util/ema_filter.h"
#include "../ctrl/ctrl_motor.h"

#include "util/network_print.h"
#include "util/console.h"

static void encTrainInit();
static void encTrainRun(void* pData, SkillOutput* pOutput);
static void encTrainExit();
static uint8_t angleMeas(void* pData, SkillOutput* pOutput);
static uint8_t distMeas(void* pData, SkillOutput* pOutput, uint8_t rotMode);
static void ukfMeasFunc(const arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU, arm_matrix_instance_f32* pY);

SkillInstance skillEncTrain = { &encTrainInit, &encTrainRun, &encTrainExit};

static struct _SkillEncTrain
{
	float angle;
	float orientDist;

	uint32_t lastMoveStart;

	uint32_t waitStart;

	float encSum[4];
	float initPos[3];
	float finalPos[3];

	uint8_t phase;
	uint8_t state;

	uint16_t xyRun;

	UKF ukf1;
	float ukf1Data[UKF_DATA_SIZE(10, 1, 1, 10)];
	UKF ukf2;
	float ukf2Data[UKF_DATA_SIZE(10, 1, 1, 10)];

	EMAFilter rotError;

	CtrlInstance* pOrigInstance;
} encTrain;

#define MOVE_THRESHOLD 0.001f
#define ANGLE_DIFF_THRESHOLD 0.05f
#define DIST_DIFF_THRESHOLD 0.01f

#define PHASE_ROT_ONLY	0
#define PHASE_XY_TO_ROT	1
#define PHASE_XY_TO_XY	2
#define PHASE_DONE		3
#define PHASE_WAIT		4

#define STATE_GOTO_START 0
#define STATE_GOTO_ANGLE 1
#define STATE_DIST_MEAS 2
#define STATE_WAIT_DIST_MEAS 3
#define STATE_WAIT_GOTO_START 4
#define STATE_ANGLE_PRE_MEAS 5
#define STATE_ANGLE_MEAS 6
#define STATE_ANGLE_POST_MEAS 7

static uint8_t isMoveComplete()
{
	float vel = sqrtf(ctrl.state.vel[0]*ctrl.state.vel[0] + ctrl.state.vel[1]*ctrl.state.vel[1]);

	if(vel < MOVE_THRESHOLD && fabsf(ctrl.state.vel[2]) < MOVE_THRESHOLD && (SysTimeUSec()-encTrain.lastMoveStart) > 500000)
		return 1;

	return 0;
}

static void resetUKF()
{
	const float stateUnc = 0.0001f;
	const float measUnc = 0.01f;

	arm_mat_zero_f32(&encTrain.ukf1.x);
	arm_mat_identity_f32(&encTrain.ukf1.Q);
	arm_mat_identity_f32(&encTrain.ukf1.R);
	arm_mat_scale_f32(&encTrain.ukf1.Q, stateUnc, &encTrain.ukf1.Q);
	arm_mat_scale_f32(&encTrain.ukf1.R, measUnc, &encTrain.ukf1.R);
	arm_mat_copy_f32(&encTrain.ukf1.Q, &encTrain.ukf1.Px);
	arm_mat_zero_f32(&encTrain.ukf1.Py);
	arm_mat_zero_f32(&encTrain.ukf1.Pxy);

	arm_mat_zero_f32(&encTrain.ukf2.x);
	arm_mat_identity_f32(&encTrain.ukf2.Q);
	arm_mat_identity_f32(&encTrain.ukf2.R);
	arm_mat_scale_f32(&encTrain.ukf2.Q, stateUnc, &encTrain.ukf2.Q);
	arm_mat_scale_f32(&encTrain.ukf2.R, measUnc, &encTrain.ukf2.R);
	arm_mat_copy_f32(&encTrain.ukf2.Q, &encTrain.ukf2.Px);
	arm_mat_zero_f32(&encTrain.ukf2.Py);
	arm_mat_zero_f32(&encTrain.ukf2.Pxy);
}

static void encTrainInit()
{
	ctrl.buzzOnTimeViolation = 0;

	UKFInit(&encTrain.ukf1, 10, 1, 1, encTrain.ukf1Data, 0, &ukfMeasFunc);
	UKFInit(&encTrain.ukf2, 10, 1, 1, encTrain.ukf2Data, 0, &ukfMeasFunc);

	EMAFilterInit(&encTrain.rotError, 0.98f);
	encTrain.rotError.value = 0.0f;

	encTrain.phase = PHASE_WAIT;
	encTrain.waitStart = SysTimeUSec();

	encTrain.pOrigInstance = ctrl.pInstance;

	memset(&ctrl.encoders, 0, sizeof(BotEncoders));

	CtrlSetController(&ctrlMotorInstance);
}

static void encTrainExit()
{
	CtrlSetController(encTrain.pOrigInstance);

	ctrl.buzzOnTimeViolation = 1;
}

static void encTrainRun(void* pData, SkillOutput* pOutput)
{
	SkillEncTrainInput* pInput = (SkillEncTrainInput*)pData;

	switch(encTrain.phase)
	{
		case PHASE_WAIT:
		{
			if(SysTimeUSec()-encTrain.waitStart > 1000000)
			{
				encTrain.phase = PHASE_ROT_ONLY;
				encTrain.orientDist = -2.5f;
				encTrain.state = STATE_ANGLE_PRE_MEAS;
				encTrain.waitStart = SysTimeUSec();
			}
		}
		break;
		case PHASE_ROT_ONLY:
		{
			if(angleMeas(pData, pOutput))
			{
				encTrain.phase = PHASE_XY_TO_ROT;
				encTrain.state = STATE_GOTO_START;
				encTrain.angle = 0.0f;
				encTrain.xyRun = 0;

				ctrl.encoders.w2w = encTrain.rotError.value;

				resetUKF();
			}
		}
		break;
		case PHASE_XY_TO_ROT:
		{
			if(distMeas(pData, pOutput, 1))
			{
				++encTrain.xyRun;
				if(encTrain.xyRun >= pInput->xyRepeat)
				{
					encTrain.phase = PHASE_XY_TO_XY;
					encTrain.state = STATE_GOTO_START;
					encTrain.angle = 0.0f;
					encTrain.xyRun = 0;

					memcpy(ctrl.encoders.xy2w, encTrain.ukf1.x.pData, sizeof(float)*10);

					resetUKF();
				}
				else
				{
					encTrain.state = STATE_GOTO_START;
					encTrain.angle = 0.0f;
				}
			}
		}
		break;
		case PHASE_XY_TO_XY:
		{
			if(distMeas(pData, pOutput, 0))
			{
				++encTrain.xyRun;
				if(encTrain.xyRun >= pInput->xyRepeat)
				{
					encTrain.phase = PHASE_DONE;

					memcpy(ctrl.encoders.xy2x, encTrain.ukf1.x.pData, sizeof(float)*10);
					memcpy(ctrl.encoders.xy2y, encTrain.ukf2.x.pData, sizeof(float)*10);

					FlashFSWrite(&ctrl.pEncodersFile, &ctrl.encoders, sizeof(BotEncoders));

					NetworkPrint("Done");
				}
				else
				{
					encTrain.state = STATE_GOTO_START;
					encTrain.angle = 0.0f;
				}
			}
		}
		break;
		case PHASE_DONE:
		{
		}
		break;
	}

	pOutput->mode = SKILL_OUT_MODE_GLOBAL_POS;
}

static uint8_t angleMeas(void* pData, SkillOutput* pOutput)
{
	SkillEncTrainInput* pInput = (SkillEncTrainInput*)pData;
	float orientAngleStep = 5.0f/((float)pInput->firstAngleSteps);

	switch(encTrain.state)
	{
		case STATE_ANGLE_PRE_MEAS:
		{
			if(SysTimeUSec()-encTrain.waitStart > 200000)
			{
				memcpy(encTrain.initPos, ctrl.state.pos, sizeof(float)*3);
				memset(encTrain.encSum, 0, sizeof(float)*4);

				pOutput->xyw[0] = ctrl.state.pos[0];
				pOutput->xyw[1] = ctrl.state.pos[1];
				pOutput->xyw[2] = AngleNormalize(ctrl.state.pos[2]+encTrain.orientDist);
				pOutput->updated = 1;
				encTrain.lastMoveStart = SysTimeUSec();

				encTrain.state = STATE_ANGLE_MEAS;
			}
		}
		break;
		case STATE_ANGLE_MEAS:
		{
			for(uint8_t i = 0; i < 4; i++)
			{
				encTrain.encSum[i] += ctrl.input.enc.vel[i]*ctrl.pInstance->deltaT;
			}

			if(isMoveComplete())
			{
				encTrain.waitStart = SysTimeUSec();
				encTrain.state = STATE_ANGLE_POST_MEAS;
			}

		}
		break;
		case STATE_ANGLE_POST_MEAS:
		{
			for(uint8_t i = 0; i < 4; i++)
			{
				encTrain.encSum[i] += ctrl.input.enc.vel[i]*ctrl.pInstance->deltaT;
			}

			if(SysTimeUSec()-encTrain.waitStart > 200000)
			{
				// save final pos
				memcpy(encTrain.finalPos, ctrl.state.pos, sizeof(float)*3);

				// calc enc pos
				arm_matrix_instance_f32 matEnc = {4, 1, encTrain.encSum};
				for(uint8_t i = 0; i < 4; i++)
					encTrain.encSum[i] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now dist over ground [m]

				arm_matrix_instance_f32 matEncPos = {3, 1, (float[3]){}};

				// calc XYW pos
				arm_mat_mult_f32(&ctrl.matMotor2XYW, &matEnc, &matEncPos);

				float setRot = encTrain.orientDist;
				float visRot = AngleNormalize(encTrain.finalPos[2]-encTrain.initPos[2]);
				float encRot = matEncPos.pData[2];

				if(fabsf(setRot) < 0.05f)
					break;

				EMAFilterUpdate(&encTrain.rotError, (encRot-visRot)/setRot);

//				ConsolePrint("rotErr: %f\r\n", encTrain.rotError.value);

				NetworkPrint("a %hd,%hd,%hd",
						(int16_t)(setRot*1000.0f),
						(int16_t)(visRot*1000.0f),
						(int16_t)(encRot*1000.0f));

				encTrain.orientDist += orientAngleStep;
				if(encTrain.orientDist > 2.5f)
					return 1;
				encTrain.waitStart = SysTimeUSec();
				encTrain.state = STATE_ANGLE_PRE_MEAS;
			}
		}
		break;
	}

	return 0;
}

static uint8_t distMeas(void* pData, SkillOutput* pOutput, uint8_t rotMode)
{
	SkillEncTrainInput* pInput = (SkillEncTrainInput*)pData;
	float startPos[2] = { pInput->startPos[0]*0.001f, pInput->startPos[1]*0.001f };
	float endPos[2] = { pInput->endPos[0]*0.001f, pInput->endPos[1]*0.001f };
	float distAngleStep = (2*M_PI)/((float)pInput->secondAngleSteps);

	float distXY[2] = {endPos[0]-startPos[0], endPos[1]-startPos[1]};
	float angleToX = atan2f(distXY[1], distXY[0]);
	float setAngle = encTrain.angle+angleToX;
	float dist = sqrtf(distXY[0]*distXY[0] + distXY[1]*distXY[1]);

	switch(encTrain.state)
	{
		case STATE_GOTO_START:
		{
			float angleDiff = AngleNormalize(setAngle-ctrl.state.pos[2]);
			float xDiff = startPos[0]-ctrl.state.pos[0];
			float yDiff = startPos[1]-ctrl.state.pos[1];
			float xyDiff = sqrtf(xDiff*xDiff + yDiff*yDiff);

			if(fabsf(angleDiff) < ANGLE_DIFF_THRESHOLD && xyDiff < DIST_DIFF_THRESHOLD)
			{
				if(isMoveComplete())
				{
					encTrain.waitStart = SysTimeUSec();
					encTrain.state = STATE_WAIT_DIST_MEAS;
				}
			}
			else
			{
				if(isMoveComplete())
				{
					pOutput->xyw[0] = startPos[0];
					pOutput->xyw[1] = startPos[1];
					pOutput->xyw[2] = setAngle;
					pOutput->updated = 1;
					encTrain.lastMoveStart = SysTimeUSec();
				}
			}
		}
		break;
		case STATE_WAIT_DIST_MEAS:
		{
			if(SysTimeUSec()-encTrain.waitStart > 200000)
			{
				memcpy(encTrain.initPos, ctrl.state.pos, sizeof(float)*3);
				memset(encTrain.encSum, 0, sizeof(float)*4);

				float toGo[3] = { endPos[0]-encTrain.initPos[0], endPos[1]-encTrain.initPos[1], 0 };
				CtrlTurnGlobal2Local(ctrl.state.pos[2], toGo[0], toGo[1], toGo, toGo+1);
				CtrlCorrectEncodersOut(toGo, toGo);
				CtrlTurnLocal2Global(ctrl.state.pos[2], toGo[0], toGo[1], toGo, toGo+1);

				pOutput->xyw[0] = ctrl.state.pos[0]+toGo[0];
				pOutput->xyw[1] = ctrl.state.pos[1]+toGo[1];
				pOutput->xyw[2] = ctrl.state.pos[2]+toGo[2];
				pOutput->updated = 1;
				encTrain.lastMoveStart = SysTimeUSec();

				encTrain.state = STATE_DIST_MEAS;
			}
		}
		break;
		case STATE_DIST_MEAS:
		{
			for(uint8_t i = 0; i < 4; i++)
			{
				encTrain.encSum[i] += ctrl.input.enc.vel[i]*ctrl.pInstance->deltaT;
			}

			if(isMoveComplete())
			{
				encTrain.waitStart = SysTimeUSec();
				encTrain.state = STATE_WAIT_GOTO_START;
			}
		}
		break;
		case STATE_WAIT_GOTO_START:
		{
			for(uint8_t i = 0; i < 4; i++)
			{
				encTrain.encSum[i] += ctrl.input.enc.vel[i]*ctrl.pInstance->deltaT;
			}

			if(SysTimeUSec()-encTrain.waitStart > 200000)
			{
				// save final pos
				memcpy(encTrain.finalPos, ctrl.state.pos, sizeof(float)*3);

				// calc enc pos
				arm_matrix_instance_f32 matEnc = {4, 1, encTrain.encSum};
				for(uint8_t i = 0; i < 4; i++)
					encTrain.encSum[i] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now dist over ground [m]

				arm_matrix_instance_f32 matEncPos = {3, 1, (float[3]){}};

				// calc XYW pos
				arm_mat_mult_f32(&ctrl.matMotor2XYW, &matEnc, &matEncPos);

				float visRot = AngleNormalize(encTrain.finalPos[2]-encTrain.initPos[2]);
				float encRot = matEncPos.pData[2];
				float angle = AngleNormalize(encTrain.initPos[2]-angleToX-M_PI/2.0f);

				if(rotMode)
				{
					MAT_ELEMENT(encTrain.ukf1.u, 0, 0) = angle;
//					MAT_ELEMENT(encTrain.ukf1.y, 0, 0) = (-visRot)/dist;
					MAT_ELEMENT(encTrain.ukf1.y, 0, 0) = (encRot-visRot)/dist;

					UKFPredictNoStateFunc(&encTrain.ukf1);
					UKFUpdate(&encTrain.ukf1);

//					ConsolePrint("UKF1\r\n");
//					arm_mat_print(&encTrain.ukf1.x);

					NetworkPrint("b %hd,%hd,%hd,%hd",
							(int16_t)(angle*1000.0f),
							(int16_t)(visRot*1000.0f),
							(int16_t)(encRot*1000.0f),
							(int16_t)(dist*1000.0f));
				}
				else
				{
					float setDist[2] = {endPos[0]-encTrain.initPos[0], endPos[1]-encTrain.initPos[1]};
					CtrlTurnGlobal2Local(encTrain.initPos[2], setDist[0], setDist[1], setDist, setDist+1);
					float driven[2] = {encTrain.finalPos[0]-encTrain.initPos[0], encTrain.finalPos[1]-encTrain.initPos[1]};
					CtrlTurnGlobal2Local(encTrain.initPos[2], driven[0], driven[1], driven, driven+1);

					MAT_ELEMENT(encTrain.ukf1.u, 0, 0) = angle;
//					MAT_ELEMENT(encTrain.ukf1.y, 0, 0) = (setDist[0]-driven[0])/dist;
					MAT_ELEMENT(encTrain.ukf1.y, 0, 0) = (matEncPos.pData[0]-driven[0])/dist;

					UKFPredictNoStateFunc(&encTrain.ukf1);
					UKFUpdate(&encTrain.ukf1);

					MAT_ELEMENT(encTrain.ukf2.u, 0, 0) = angle;
//					MAT_ELEMENT(encTrain.ukf2.y, 0, 0) = (setDist[1]-driven[1])/dist;
					MAT_ELEMENT(encTrain.ukf2.y, 0, 0) = (matEncPos.pData[1]-driven[1])/dist;

					UKFPredictNoStateFunc(&encTrain.ukf2);
					UKFUpdate(&encTrain.ukf2);

//					ConsolePrint("UKF1\r\n");
//					arm_mat_print(&encTrain.ukf1.x);
//					ConsolePrint("UKF2\r\n");
//					arm_mat_print(&encTrain.ukf2.x);

					NetworkPrint("c %hd,%hd,%hd,%hd,%hd,%hd",
							(int16_t)(angle*1000.0f),
							(int16_t)(driven[0]*1000.0f),
							(int16_t)(driven[1]*1000.0f),
							(int16_t)(matEncPos.pData[0]*1000.0f),
							(int16_t)(matEncPos.pData[1]*1000.0f),
							(int16_t)(dist*1000.0f));
				}

				encTrain.angle += distAngleStep;

				if(encTrain.angle >= 2*M_PI)
					return 1;

				encTrain.state = STATE_GOTO_START;
			}
		}
		break;
	}

	return 0;
}

// pU: angle
// pX: fourier parameters
// pY: error
static void ukfMeasFunc(const arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU, arm_matrix_instance_f32* pY)
{
	float* f = pX->pData;
	float x = MAT_ELEMENT(*pU, 0, 0);

	MAT_ELEMENT(*pY, 0, 0) = f[0]*cosf(x) + f[1]*sinf(x) +
		f[2]*cosf(2*x) + f[3]*sinf(2*x) +
		f[4]*cosf(3*x) + f[5]*sinf(3*x) +
		f[6]*cosf(4*x) + f[7]*sinf(4*x) +
		f[8]*cosf(5*x) + f[9]*sinf(5*x);
}
