/*
 * spi2.c
 *
 *  Created on: 19.01.2014
 *      Author: AndreR
 */

#include "spi2.h"

#include "constants.h"

#include "util/console.h"
#include "util/init_hal.h"
#include "util/intercom.h"
#include "util/sys_time.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "network.h"

SPI2Global spi2;

#define EVENT_GYR_RDY		0x01
#define EVENT_ACC_RDY		0x02

#define STATE_IDLE			0
#define STATE_ACTIVE		1

#define DMA_Stream_RX		DMA1_Stream3
#define DMA_Stream_TX		DMA1_Stream4

static uint8_t tx[10];
static uint8_t rx[10];

void DMA1_Stream3_IRQHandler()
{
	SPISyncDmaRxInterrupt(&spi2.slave[spi2.activeSlave]);
}

static void gyrDRDYIRQ()
{
	chSysLockFromIsr();
	chMBPostI(&spi2.eventQueue, EVENT_GYR_RDY);
	chSysUnlockFromIsr();
}

void EXTI0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = EXTI_PR_PR0;

	chSysLockFromIsr();
	chMBPostI(&spi2.eventQueue, EVENT_ACC_RDY);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void SPI2Init()
{
	// init variables
	chMBInit(&spi2.eventQueue, spi2.eventQueueData, SPI2_EVENT_QUEUE_SIZE);
	spi2.activeSlave = ACC;
	spi2.logAccUsed = 0;
	spi2.logGyrUsed = 0;

	for(uint8_t i = 0; i < 3; i++)
	{
		LagElementPT1Init(&spi2.gyrPT1[i], 1.0f, 1.0f, 1.0f/800.0f);
		LagElementPT1Init(&spi2.accPT1[i], 1.0f, 1.0f, 1.0f/800.0f);
	}

	RCC->APB1ENR |= RCC_APB1ENR_SPI2EN;

	SPISyncData spiData;
	spiData.cpha = 0;
	spiData.cpol = 0;
	spiData.csPin = GPIO_PIN_10;
	spiData.pCSPort = GPIOB;
	spiData.dmaChRx = 3;
	spiData.dmaChTx = 4;
	spiData.dmaPrio = 1;
	spiData.pDma = DMA1;
	spiData.pRegister = SPI2;
	spiData.prescaler = SPI_CR1_BRDIV16;
	spiData.timeoutTicks = MS2ST(20);
	SPISyncInit(&spi2.slave[GYR], &spiData);

	spiData.csPin = GPIO_PIN_12;
	SPISyncInit(&spi2.slave[ACC], &spiData);


#ifdef DEBUG
	spi2.pRxDma = DMA_Stream_RX;
	spi2.pTxDma = DMA_Stream_TX;
	spi2.pSPI = SPI2;
	spi2.pDMA = DMA1;
#endif

	// init GPIO
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_25MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOB, GPIO_PIN_13 | GPIO_PIN_14 | GPIO_PIN_15, &gpioInit); // SCK, MISO, MOSI

	// configure NVIC
	NVICEnableIRQ(DMA1_Stream3_IRQn, IRQL_SPI2);

	// DRDY GYR IRQ
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOC, GPIO_PIN_5, &gpioInit);

	// INT1 ACC IRQ
	GPIOInit(GPIOB, GPIO_PIN_0, &gpioInit);

	ExtiSetCallback(&gyrDRDYIRQ, 5);

	NVICEnableIRQ(EXTI0_IRQn, IRQL_SPI2);
}

static void gyrUpdate()
{
	spi2.activeSlave = GYR;

	tx[0] = 0x80 | 0x40 | 0x28;
	SPISyncTransfer(&spi2.slave[GYR], tx, rx, 7);

	spi2.gyrRaw[0] = *((int16_t*)(rx+1));
	spi2.gyrRaw[1] = *((int16_t*)(rx+3));
	spi2.gyrRaw[2] = *((int16_t*)(rx+5));

	// Timing
	const uint32_t alpha = 96;
	uint32_t tNow = SysTimeUSec();
	uint32_t tPeriod = tNow - spi2.lastGyrTime;
	spi2.lastGyrTime = tNow;
	spi2.avgGyrPeriod = (alpha*spi2.avgGyrPeriod + (100-alpha)*tPeriod)/100;

	float temp;
	for(uint8_t i = 0; i < 3; i++)
	{
//		spi2.gyr[i] = spi2.gyrRaw[i]*SPI2_GYR_UNIT_TO_RADS;

		temp = spi2.gyrRaw[i]*SPI2_GYR_UNIT_TO_RADS;
		spi2.gyrDelayed[i] = LagElementPT1Process(&spi2.gyrPT1[i], temp);

		spi2.gyr[i] = temp-spi2.gyrOffset[i];

		temp = fabsf(spi2.gyrDelayed[i]-spi2.gyrCompare[i]);
		if(temp > spi2.gyrDiff[i])
			spi2.gyrDiff[i] = temp;
	}

	// Callback
	if(spi2.gyrCb)
		(*spi2.gyrCb)();

	// Logging
	LogRawI16XYZ* pLog = &spi2.logGyr[spi2.logGyrUsed];
	pLog->header.type = CTYPE_GYRO_RAW;
	pLog->header.timestamp = SysTimeUSec();
	pLog->header.length = sizeof(LogRawI16XYZ);
	pLog->xyz[0] = spi2.gyrRaw[0];
	pLog->xyz[1] = spi2.gyrRaw[1];
	pLog->xyz[2] = spi2.gyrRaw[2];

	++spi2.logGyrUsed;

	if(spi2.logGyrUsed >= SPI2_IMU_BUF_SIZE)
	{
		IntercomHeader iHeader;
		iHeader.address = INTERCOM_ADDRESS_MEDIA;
		iHeader.section = INTERCOM_SECTION_LOGFILE;
		iHeader.command = IC_LOGFILE_MESSAGE;

		IntercomSend(&iHeader, SPI2_IMU_BUF_SIZE*sizeof(LogRawI16XYZ), (uint8_t*)spi2.logGyr);
		spi2.logGyrUsed = 0;
	}
}

static void accUpdate()
{
	spi2.activeSlave = ACC;

	tx[0] = 0x80 | 0x40 | 0x28;
	SPISyncTransfer(&spi2.slave[ACC], tx, rx, 7);

	spi2.accRaw[0] = *((int16_t*)(rx+1));
	spi2.accRaw[1] = *((int16_t*)(rx+3));
	spi2.accRaw[2] = *((int16_t*)(rx+5));

	// Timing
	const uint32_t alpha = 96;
	uint32_t tNow = SysTimeUSec();
	uint32_t tPeriod = tNow - spi2.lastAccTime;
	spi2.lastAccTime = tNow;
	spi2.avgAccPeriod = (alpha*spi2.avgAccPeriod + (100-alpha)*tPeriod)/100;

	float temp;
	for(uint8_t i = 0; i < 3; i++)
	{
		spi2.acc[i] = spi2.accRaw[i]*SPI2_ACC_UNIT_TO_MS2;

		temp = spi2.accRaw[i]*SPI2_ACC_UNIT_TO_MS2;
		spi2.accDelayed[i] = LagElementPT1Process(&spi2.accPT1[i], temp);
//
//		spi2.acc[i] = temp-spi2.accOffset[i];
//
//		temp = fabsf(spi2.accDelayed[i]-spi2.accCompare[i]);
//		if(temp > spi2.accDiff[i])
//			spi2.accDiff[i] = temp;
	}

	// Callback
	if(spi2.accCb)
		(*spi2.accCb)();

	// Logging
	LogRawI16XYZ* pLog = &spi2.logAcc[spi2.logAccUsed];
	pLog->header.type = CTYPE_ACC_RAW;
	pLog->header.timestamp = SysTimeUSec();
	pLog->header.length = sizeof(LogRawI16XYZ);
	pLog->xyz[0] = spi2.accRaw[0];
	pLog->xyz[1] = spi2.accRaw[1];
	pLog->xyz[2] = spi2.accRaw[2];

	++spi2.logAccUsed;

	if(spi2.logAccUsed >= SPI2_IMU_BUF_SIZE)
	{
		IntercomHeader iHeader;
		iHeader.address = INTERCOM_ADDRESS_MEDIA;
		iHeader.section = INTERCOM_SECTION_LOGFILE;
		iHeader.command = IC_LOGFILE_MESSAGE;

		IntercomSend(&iHeader, SPI2_IMU_BUF_SIZE*sizeof(LogRawI16XYZ), (uint8_t*)spi2.logAcc);
		spi2.logAccUsed = 0;
	}
}

void SPI2SetCallbacks(SPI2Callback gyr, SPI2Callback acc)
{
	spi2.gyrCb = gyr;
	spi2.accCb = acc;
}

msg_t SPI2Task(void* params)
{
	(void)params;

	chRegSetThreadName("SPI2");

	chThdSleepMilliseconds(100);

	spi2.activeSlave = ACC;

	// LSM303D
	tx[0] = 0x1F | 0x40;	// CTRL0 Register + multiple write
	tx[1] = 0x80;	// CTRL0: reboot memory content
	SPISyncTransfer(&spi2.slave[ACC], tx, rx, 2);

	chThdSleepMilliseconds(50);

	tx[0] = 0x20 | 0x40;	// CTRL1 Register + multiple write
	tx[1] = 0x97;	// CTRL1: ACC rate 800Hz, X/Y/Z on
	tx[2] = 0x00;	// CTRL2: ACC anti-alias bandwidth 773Hz, +-2g full-scale
//	tx[2] = 0x80;	// CTRL2: ACC anti-alias bandwidth 362Hz, +-2g full-scale
	tx[3] = 0x04;	// CTRL3: ACC data-ready on INT1
	tx[4] = 0x04;	// CTRL4: MAG data-ready on INT2
	tx[5] = 0x74;	// CTRL5: MAG high-resolution, rate 100Hz
	tx[6] = 0x00;	// CTRL6: MAG 2 gauss full-scale
	tx[7] = 0x00;	// CTRL7: ACC high-pass filter disabled, MAG continous conversion mode
//	tx[7] = 0xA0;	// CTRL7: ACC high-pass filter enabled, MAG continous conversion mode
	SPISyncTransfer(&spi2.slave[ACC], tx, rx, 8);

	spi2.activeSlave = GYR;

	// L3G4200D
	tx[0] = 0x20 | 0x40; // CTRL1 Register + multiple write bit
	tx[1] = 0xFF; // CTRL1: 800Hz, LPF1 Cut-off 110Hz, X,Y,Z on
	tx[2] = 0x06; // CTRL2: High Pass 1Hz
	tx[3] = 0x08; // CTRL3: Data Ready interrupt on DRDY/INT2
	tx[4] = 0x20; // CTRL4: FSS@2000dps
	tx[5] = 0x00; // CTRL5: FIFO disabled, HPF disabled, LPF2 disabled
	SPISyncTransfer(&spi2.slave[GYR], tx, rx, 6);

	gyrUpdate();
	accUpdate();

	msg_t event;
	uint32_t last1sCall = chTimeNow();
	while(1)
	{
		msg_t chResult = chMBFetch(&spi2.eventQueue, &event, MS2ST(10));
		if(chResult == RDY_OK)
		{
			if(event & EVENT_GYR_RDY)
				gyrUpdate();

			if(event & EVENT_ACC_RDY)
				accUpdate();
		}
		else
		{
			gyrUpdate();
			accUpdate();
		}

		if(chTimeElapsedSince(last1sCall) > S2ST(1))
		{
			last1sCall = chTimeNow();

			for(uint8_t i = 0; i < 3; i++)
			{
				if(spi2.gyrDiff[i] < 0.001f)
				{
					spi2.gyrOffset[i] = spi2.gyrCompare[i];
//					ConsolePrint("GYR %hu offset: % -.5f\r\n", (uint16_t)i, spi2.gyrOffset[i]);
				}

				if(spi2.accDiff[i] < 0.01f)
				{
					spi2.accOffset[i] = spi2.accCompare[i];
//					ConsolePrint("ACC %hu offset: % -.5f\r\n", (uint16_t)i, spi2.accOffset[i]);
				}

				spi2.gyrCompare[i] = spi2.gyrDelayed[i];
				spi2.accCompare[i] = spi2.accDelayed[i];
				spi2.gyrDiff[i] = 0;
				spi2.accDiff[i] = 0;
			}
		}
	}

	return 0;
}
