/*
 * spi3.c
 *
 *  Created on: 19.01.2014
 *      Author: AndreR
 */

#include "spi3.h"

#include "util/log.h"
#include "util/console.h"
#include "util/bit_banding.h"
#include "util/init_hal.h"
#include "util/sys_time.h"
#include "util/cobs.h"
#include "util/intercom.h"
#include "intercom_constants.h"
#include <string.h>

#define IRQL_SPI3 12

SPI3Global spi3;

#define EVENT_TRANSFER_MASK	0x01FF

#define DMA_TX_TE			0x0001
#define DMA_TX_DME			0x0002
#define DMA_TX_FE			0x0004
#define DMA_TX_TC			0x0008
#define DMA_RX_TE			0x0010
#define DMA_RX_DME			0x0020
#define DMA_RX_FE			0x0040
#define DMA_RX_TC			0x0080

#define EVENT_PREPARE_TX	0x0010 // + slave
#define EVENT_START_TX		0x0020 // + slave

#define DMA_Stream_RX		DMA1_Stream0
#define DMA_Stream_TX		DMA1_Stream7

#define DMA_IFCR_RX			DMA1->LIFCR
#define DMA_IFCR_TX			DMA1->HIFCR

#define DMA_ISR_RX			DMA1->LISR
#define DMA_ISR_TX			DMA1->HISR

#define DMA_IFCR_RX_CLEAR	(DMA_LIFCR_CDMEIF0 | DMA_LIFCR_CTCIF0 | DMA_LIFCR_CHTIF0 | DMA_LIFCR_CFEIF0 | DMA_LIFCR_CFEIF0)
#define DMA_IFCR_TX_CLEAR	(DMA_HIFCR_CDMEIF7 | DMA_HIFCR_CTCIF7 | DMA_HIFCR_CHTIF7 | DMA_HIFCR_CFEIF7 | DMA_HIFCR_CFEIF7)

#define DMAClearTCIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 4)
#define DMASetTCIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CR, 4)
#define DMAGetTCIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CR, 4)

#define DMAClearTEIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 2)
#define DMASetTEIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CR, 2)
#define DMAGetTEIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CR, 2)

static uint16_t stopDMA()
{
	uint16_t result = 0;

    if(DMA1->LISR & DMA_LISR_TEIF0)
    	result |= DMA_RX_TE;
    if(DMA1->LISR & DMA_LISR_DMEIF0)
    	result |= DMA_RX_DME;
    if(DMA1->LISR & DMA_LISR_FEIF0)
    	result |= DMA_RX_FE;
    if(DMA1->LISR & DMA_LISR_TCIF0)
    	result |= DMA_RX_TC;

    if(DMA1->HISR & DMA_HISR_TEIF7)
    	result |= DMA_TX_TE;
    if(DMA1->HISR & DMA_HISR_DMEIF7)
    	result |= DMA_TX_DME;
    if(DMA1->HISR & DMA_HISR_FEIF7)
    	result |= DMA_TX_FE;
    if(DMA1->HISR & DMA_HISR_TCIF7)
    	result |= DMA_TX_TC;

    if(DMA_Stream_TX->CR & DMA_SxCR_EN)
    	result |= 0x0200;
    if(DMA_Stream_RX->CR & DMA_SxCR_EN)
    	result |= 0x0400;

    // disable DMA
    DMA_Stream_TX->CR &= ~DMA_SxCR_EN;
    DMA_Stream_RX->CR &= ~DMA_SxCR_EN;

    // wait for it...
    while(DMA_Stream_TX->CR & DMA_SxCR_EN);
    while(DMA_Stream_RX->CR & DMA_SxCR_EN);

    return result;
}

void DMA1_Stream0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	DMAClearTCIE(DMA_Stream_RX);
	DMAClearTEIE(DMA_Stream_RX);

	chSysLockFromIsr();
	chBSemSignalI(&spi3.transferSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void TIM8_UP_TIM13_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM13->SR = 0;
	TIM13->CR1 = 0;

	chSysLockFromIsr();
	chBSemSignalI(&spi3.timerSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

static void startTimeout(uint32_t microSeconds)
{
	chBSemReset(&spi3.timerSem, TRUE);

	TIM13->ARR = microSeconds;
	TIM13->CR1 = TIM_CR1_CEN;
}


void SPI3Init()
{
	// init event queue/variables
	chMBInit(&spi3.eventQueue, spi3.eventQueueData, 4);
	chBSemInit(&spi3.transferSem, 0);
	chBSemInit(&spi3.timerSem, 0);
	spi3.slave[SPI3_SLAVE_LEFT].pCSPort = GPIOC;
	spi3.slave[SPI3_SLAVE_LEFT].csPin = GPIO_PIN_3;
	spi3.slave[SPI3_SLAVE_LEFT].id = SPI3_SLAVE_LEFT;
	spi3.slave[SPI3_SLAVE_RIGHT].pCSPort = GPIOB;
	spi3.slave[SPI3_SLAVE_RIGHT].csPin = GPIO_PIN_5;
	spi3.slave[SPI3_SLAVE_RIGHT].id = SPI3_SLAVE_RIGHT;

	for(uint8_t i = 0; i < 2; i++)
	{
		SPI3Slave* pSlave = &spi3.slave[i];

		FifoLinInit(&pSlave->txFifo, pSlave->fifoDataTx, SPI3_FIFO_SIZE_TX);
		FifoLinInit(&pSlave->rxFifo, pSlave->fifoDataRx, SPI3_FIFO_SIZE_RX);
	}

#ifdef DEBUG
	spi3.pRxDma = DMA_Stream_RX;
	spi3.pTxDma = DMA_Stream_TX;
	spi3.pSPI = SPI3;
	spi3.pDMA = DMA1;
#endif

	RCC->APB1ENR |= RCC_APB1ENR_SPI3EN;

	// init GPIO
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 6;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_25MHZ;
	gpioInit.pupd = GPIO_PUPD_DOWN;
	GPIOInit(GPIOC, GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit); // SCK, MISO, MOSI

	gpioInit.mode = GPIO_MODE_OUTPUT;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_25MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOB, GPIO_PIN_5, &gpioInit);	// MOTR-NSS
	GPIOInit(GPIOC, GPIO_PIN_3, &gpioInit);	// MOTL-NSS

	GPIOB->BSRRL = GPIO_PIN_5;	// deselect slaves
	GPIOC->BSRRL = GPIO_PIN_3;

	// init SPI3
	SPI3->CR1 = SPI_CR1_SSI | SPI_CR1_SSM | SPI_CR1_DFF | SPI_CR1_BRDIV4 | SPI_CR1_MSTR;
	SPI3->CR2 = 0;

	// init DMA
	// DMA TX setup
	DMA_Stream_TX->CR = DMA_CHSEL(0) | DMA_DIR_MEM2PERIPH | DMA_SxCR_MINC |
			DMA_PSIZE_2BYTE | DMA_MSIZE_2BYTE | DMA_PL_HIGH | DMA_MBURST_SINGLE | DMA_PBURST_SINGLE;
	DMA_Stream_TX->FCR = DMA_SxFCR_DMDIS | DMA_FTH_FULL;
	DMA_Stream_TX->PAR = (uint32_t)&SPI3->DR;
	DMA_Stream_TX->NDTR = 0;
	DMA_Stream_TX->M0AR = 0;

	// DMA RX setup
	DMA_Stream_RX->CR = DMA_CHSEL(0) | DMA_DIR_PERIPH2MEM | DMA_SxCR_MINC |
			DMA_PSIZE_2BYTE | DMA_MSIZE_2BYTE | DMA_PL_HIGH | DMA_MBURST_SINGLE | DMA_PBURST_SINGLE;
	DMA_Stream_RX->FCR = DMA_SxFCR_DMDIS | DMA_FTH_FULL;
	DMA_Stream_RX->PAR = (uint32_t)&SPI3->DR;
	DMA_Stream_RX->NDTR = 0;
	DMA_Stream_RX->M0AR = 0;

	DMA_Stream_RX->CR |= DMA_SxCR_TCIE | DMA_SxCR_TEIE | DMA_SxCR_DMEIE;
	DMA_Stream_RX->FCR |= DMA_SxFCR_FEIE;

	DMA_IFCR_RX = DMA_IFCR_RX_CLEAR;
	DMA_IFCR_TX = DMA_IFCR_TX_CLEAR;

	// configure NVIC
	NVICEnableIRQ(DMA1_Stream0_IRQn, IRQL_SPI3);

	SPI3->CR2 |= SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;

    // init timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM13EN;

	TIM13->CR1 = 0;
	TIM13->PSC = 83;	// clock is now 1MHz
	TIM13->ARR = 999;	// = 500us
	TIM13->SR = 0;
	TIM13->DIER = TIM_DIER_UIE;

	NVICEnableIRQ(TIM8_UP_TIM13_IRQn, IRQL_SPI3);
}

static void prepareTx(SPI3Slave* pSlave)
{
	if(spi3.aboutToSendCb)
		(*spi3.aboutToSendCb)(pSlave, pSlave->id);

	IntercomHeader iHeader;
	iHeader.section = INTERCOM_SECTION_LOGFILE;
	iHeader.command = IC_LOGFILE_TIMESYNC;

	if(spi3.lastSyncTime == 49 || spi3.lastSyncTime == 50)
	{
		if(spi3.lastSyncTime == 50)
			spi3.lastSyncTime = 0;

		if(pSlave->id == SPI3_SLAVE_LEFT)
			iHeader.address = INTERCOM_ADDRESS_MEDIA | INTERCOM_ADDRESS_LEFT | INTERCOM_ADDRESS_KD;
		else
			iHeader.address = INTERCOM_ADDRESS_RIGHT;

		LogFileTimeSync sync;
		sync.time = SysTimeUSec();

		IntercomSend(&iHeader, sizeof(LogFileTimeSync), (uint8_t*)&sync);
	}

	++spi3.lastSyncTime;

	// upload data
	uint8_t* pTxData;
	uint32_t pktSize;
	uint32_t bytesWritten;

	pSlave->txData.bytesWritten = 0;

	while(FifoLinGet(&pSlave->txFifo, &pTxData, &pktSize) == 0)
	{
		uint8_t* pOutData = pSlave->txData.blockData+pSlave->txData.bytesWritten;
		uint16_t sizeOut = sizeof(pSlave->txData.blockData)-pSlave->txData.bytesWritten;

		if(COBSEncode(pTxData, pktSize, pOutData, sizeOut, &bytesWritten) != 0)
			break;

		if(pSlave->txData.bytesWritten+bytesWritten < sizeof(pSlave->txData.blockData))
		{
			pOutData += bytesWritten;
			++bytesWritten;
			*pOutData = 0;
		}

		pSlave->txData.bytesWritten += bytesWritten;

		FifoLinDelete(&pSlave->txFifo);
	}

	// setup data
	DMA_Stream_TX->M0AR = (uint32_t)&pSlave->txData;
	DMA_Stream_RX->M0AR = (uint32_t)&pSlave->rxData;
	DMA_Stream_TX->NDTR = sizeof(pSlave->txData)/2;
	DMA_Stream_RX->NDTR = sizeof(pSlave->rxData)/2;

	// clear flags
    DMA_IFCR_RX = DMA_IFCR_RX_CLEAR;
    DMA_IFCR_TX = DMA_IFCR_TX_CLEAR;

	// Enable the DMA channels
	DMA_Stream_RX->CR |= DMA_SxCR_EN;
	DMA_Stream_TX->CR |= DMA_SxCR_EN;
	DMASetTCIE(DMA_Stream_RX);
    DMASetTEIE(DMA_Stream_RX);
}

static void doTransfer()
{
    chBSemReset(&spi3.transferSem, 1);	// reset semaphore and take it

    SPI3->SR = 0;
    SPI3->CR1 |= SPI_CR1_SPE;	// enable SPI => start transfer

	chBSemWait(&spi3.transferSem);	// wait for DMA done signal through semaphore

	stopDMA();	// get DMA flags and disable DMA
}

msg_t SPI3Task(void* params)
{
	(void)params;

	chRegSetThreadName("SPI3");

	// TODO: left and right need some time, sync method?
	chThdSleep(MS2ST(1000));

	chMBReset(&spi3.eventQueue);

	SPI3Slave* pSlave = &spi3.slave[SPI3_SLAVE_LEFT];

	uint32_t periodAdjustTime = chTimeNow();
	int32_t prepareDelay = 20;

	while(1)
	{
		const uint32_t alpha = 96;
		uint32_t tNow = SysTimeUSec();
		uint32_t tPeriod = tNow - spi3.lastUSec;
		spi3.lastUSec = tNow;
		spi3.avgTxPeriod = (alpha*spi3.avgTxPeriod + (100-alpha)*tPeriod)/100;

		if(chTimeNow()-periodAdjustTime > MS2ST(500))
		{
			periodAdjustTime = chTimeNow();

			int32_t adjust = 500-spi3.avgTxPeriod;
			prepareDelay += adjust;
			if(prepareDelay < 0)
				prepareDelay = 0;
			if(prepareDelay > 500)
				prepareDelay = 500;
		}

		pSlave->pCSPort->BSRRH = pSlave->csPin;	// select slave

		startTimeout(50+prepareDelay);

		prepareTx(pSlave);

		chBSemWait(&spi3.timerSem);

		doTransfer();

	    pSlave->pCSPort->BSRRL = pSlave->csPin;	// deselect slave

	    SPI3->SR = 0;
	    SPI3->CR1 &= ~SPI_CR1_SPE;	// disable SPI

	    // clear OVR (just in case)
	    SPI3->DR;
	    SPI3->SR;

		// Decode data
		uint16_t bytesToRead = pSlave->rxData.bytesWritten;
		uint8_t* pFifoData;
		uint32_t bytesWritten;
		uint32_t pktSize = 0;

		if(bytesToRead > MOT_EXCHANGE_DATA_MSG_SIZE)
		{
			LogWarnC("bytesToRead > blockSize", bytesToRead);
			bytesToRead = MOT_EXCHANGE_DATA_MSG_SIZE;
		}

		for(uint16_t bytesRead = 0; bytesRead < bytesToRead; bytesRead += pktSize+1)
		{
			uint8_t* pData = pSlave->rxData.blockData+bytesRead;
			uint8_t* pPktEnd = memchr(pData, 0, bytesToRead-bytesRead);

			if(pPktEnd == 0)
				break;

			pktSize = pPktEnd-pData;

			if(FifoLinReserve(&pSlave->rxFifo, SPI3_RESERVE_SIZE, &pFifoData) != 0)
			{
				LogWarnC("RX Fifo full", pktSize);
				break;
			}

			if(COBSDecode(pData, pktSize, pFifoData, SPI3_RESERVE_SIZE, &bytesWritten) != 0)
			{
				LogWarn("COBSDecode error");
				break;
			}

			FifoLinCommit(&pSlave->rxFifo, bytesWritten);
		}

		if(pSlave->rxFifo.numPackets && spi3.callback)
			(*spi3.callback)(spi3.pUserData);

		if(pSlave == spi3.slave)
			pSlave = &spi3.slave[1];
		else
			pSlave = spi3.slave;
	}

	return 0;
}
