/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"
#include "util/log.h"
#include "intercom_constants.h"
#include "hal/led.h"
#include "hal/usart2.h"
#include "hal/sram.h"
#include "hal/tft.h"
#include "util/intercom.h"
#include "GUI.h"
#include "util/sys_time.h"
#include "usb/usb_msc.h"
#include "fatfs/ff.h"
#include "log_file.h"
#include "util/network_print.h"
#include "prog_store.h"

CLI cli;

#define SRAM_TEST_SIZE 256*1024
static volatile EXT_SRAM uint16_t sram[SRAM_TEST_SIZE];

void processCommand();

void CLIInit()
{
	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

	ConsoleIntercomAdapterInit(&cli.adapter, INTERCOM_ADDRESS_MEDIA, INTERCOM_ADDRESS_MAIN);
}

msg_t CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

	LEDMustOff(LED_GREEN | LED_RED);

	while(1)
	{
		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		chThdSleepMilliseconds(5);
	}

	return 0;
}

void processCommand()
{
	static const uint16_t bufLen = 128;
	static uint8_t buf8[128];
	static uint32_t buf32[10];

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("Name            rPrio\tTime\r\n");

		for(Thread* pThread = chRegFirstThread(); pThread != 0; pThread = chRegNextThread(pThread))
		{
			ConsolePrint("%-16s%u\t%u\r\n", pThread->p_name, pThread->p_realprio, pThread->p_time);
		}
	}

	if(ConsoleCmpCmd("usart2 state"))
	{
		ConsolePrint("TX: %d\r\n", usart2.txState);
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	if(ConsoleScanCmd("tft brightness %d", buf32) == 1)
	{
		ConsolePrint("TFT brightness set to: %d\r\n");
		TFTSetBrightness((uint8_t)buf32[0]);
	}

	if(ConsoleCmpCmd("prog usb"))
	{
		ProgStoreStart(PROG_STORE_SOURCE_USB);
	}

	if(ConsoleCmpCmd("prog wifi"))
	{
		ProgStoreStart(PROG_STORE_SOURCE_WIFI);
	}

	FRESULT fresult;

	if(ConsoleCmpCmd("pwd"))
	{
		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("%s\r\n", buf8);
	}

	if(ConsoleCmpCmd("ls"))
	{
		DIR dir;
		FILINFO finfo;
		finfo.lfname = (TCHAR*)buf8;
		finfo.lfsize = bufLen-1;

		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("Content of: %s\r\n", buf8);

		// show folders
		fresult = f_opendir(&dir, "");
		if(fresult != FR_OK)
			return;

		do
		{
			fresult = f_readdir(&dir, &finfo);

			if(finfo.fname[0] == 0)
				break;

			if(finfo.fattrib & AM_DIR)
			{
				ConsolePrint("<DIR>      %s\r\n", *finfo.lfname ? finfo.lfname : finfo.fname);
			}
		}
		while(fresult == FR_OK);

		f_closedir(&dir);

		// show files
		fresult = f_opendir(&dir, "");
		if(fresult != FR_OK)
			return;

		do
		{
			fresult = f_readdir(&dir, &finfo);

			if(finfo.fname[0] == 0)
				break;

			if((finfo.fattrib & AM_DIR) == 0)
			{
				if(finfo.fsize > 1024*1024*16)	// larger than 16M?
					ConsolePrint("%7uM   %s\r\n", finfo.fsize/(1024*1024), *finfo.lfname ? finfo.lfname : finfo.fname);
				else if(finfo.fsize > 1024*16)	// larget than 16k?
					ConsolePrint("%7uk   %s\r\n", finfo.fsize/1024, *finfo.lfname ? finfo.lfname : finfo.fname);
				else
					ConsolePrint("%7u    %s\r\n", finfo.fsize, *finfo.lfname ? finfo.lfname : finfo.fname);
			}
		}
		while(fresult == FR_OK);

		f_closedir(&dir);
	}

	if(ConsoleScanCmd("cd %127[^\n]", buf8) == 1)
	{
		f_chdir((TCHAR*)buf8);

		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("%s\r\n", buf8);
	}

	if(ConsoleScanCmd("touch %127[^\n]", buf8) == 1)
	{
		FIL file;
		fresult = f_open(&file, (TCHAR*)buf8, FA_CREATE_NEW);
		if(fresult == FR_EXIST)
			ConsolePrint("File already exists\r\n");
		f_close(&file);
	}

	if(ConsoleScanCmd("mkdir %127[^\n]", buf8) == 1)
	{
		fresult = f_mkdir((TCHAR*)buf8);
		if(fresult == FR_OK)
			ConsolePrint("Created dir: %s\r\n", buf8);
	}

	if(ConsoleScanCmd("rm %127[^\n]", buf8) == 1)
	{
		fresult = f_unlink((TCHAR*)buf8);
		if(fresult != FR_OK)
			ConsolePrint("Could not remove object\r\n");
	}

	if(ConsoleScanCmd("read %u %127[^\n]", buf32, buf8) == 2)
	{
		ConsolePrint("Reading %uB from file: %s\r\n", *buf32, buf8);

		FIL file;
		fresult = f_open(&file, (TCHAR*)buf8, FA_READ);
		if(fresult != FR_OK)
		{
			ConsolePrint("Open file failed\r\n");
			return;
		}

		uint32_t bytesRead;
		uint32_t timeStart = SysTimeUSec();
		fresult = f_read(&file, (void*)sram, *buf32, (UINT*)&bytesRead);
		uint32_t timeEnd = SysTimeUSec();
		ConsolePrint("Result: %d\r\n", fresult);
		ConsolePrint("Bytes read: %u\r\n", bytesRead);
		uint32_t timeUs = timeEnd-timeStart;
		ConsolePrint("Time: %uus\r\n", timeUs);
		f_close(&file);
	}

	if(ConsoleScanCmd("write %u %127[^\n]", buf32, buf8) == 2)
	{
		ConsolePrint("Writing %uB to file: %s\r\n", *buf32, buf8);

		FIL file;
		fresult = f_open(&file, (TCHAR*)buf8, FA_CREATE_ALWAYS | FA_WRITE);
		if(fresult != FR_OK)
		{
			ConsolePrint("Open file failed\r\n");
			return;
		}

		uint32_t bytesRead;
		uint32_t timeStart = SysTimeUSec();
		fresult = f_write(&file, (void*)sram, *buf32, (UINT*)&bytesRead);
		fresult = f_sync(&file);
		uint32_t timeEnd = SysTimeUSec();
		ConsolePrint("Result: %d\r\n", fresult);
		ConsolePrint("Bytes written: %u\r\n", bytesRead);
		uint32_t timeUs = timeEnd-timeStart;
		ConsolePrint("Time: %uus\r\n", timeUs);
		f_close(&file);
	}

	if(ConsoleScanCmd("writec %u %127[^\n]", buf32, buf8) == 2)
	{
		ConsolePrint("Writing %uB chunks to file: %s\r\n", *buf32, buf8);

		uint32_t chunks = (2*1024*1024)/(*buf32);
		uint32_t bytesWrittenSum = 0;

		FIL file;
		fresult = f_open(&file, (TCHAR*)buf8, FA_CREATE_ALWAYS | FA_WRITE);
		if(fresult != FR_OK)
		{
			ConsolePrint("Open file failed\r\n");
			return;
		}

		uint32_t bytesWritten;
		uint32_t timeStart = SysTimeUSec();

		for(uint32_t i = 0; i < chunks; i++)
		{
			fresult = f_write(&file, (void*)0x08000000, *buf32, (UINT*)&bytesWritten);
			if(fresult != FR_OK)
			{
				ConsolePrint("write failed\r\n");
				break;
			}
			bytesWrittenSum += bytesWritten;
		}

		fresult = f_sync(&file);
		uint32_t timeEnd = SysTimeUSec();
		ConsolePrint("Result: %d\r\n", fresult);
		ConsolePrint("Bytes written: %u\r\n", bytesWrittenSum);
		uint32_t timeUs = timeEnd-timeStart;
		ConsolePrint("Time: %uus\r\n", timeUs);
		f_close(&file);
	}

	if(ConsoleScanCmd("logfile %127[^\n]", buf8) == 1)
	{
		LogFileClose();

		uint32_t freeCluster;
		FATFS* fs;
		fresult = f_getfree(usbMsc.fatFsDrivePath, &freeCluster, &fs);
		if(fresult)
		{
			NetworkPrint("No media attached\r\n");
			ConsolePrint("No media attached\r\n");
			return;
		}

		uint32_t freeSectors = freeCluster * fs->csize;
		uint32_t freeSizeMB = freeSectors/((1024*1024)/usbMsc.blockSize);

		int16_t result = LogFileOpen((char*)buf8);
		if(!result)
		{
			NetworkPrint("Logfile %s open\r\nFree Space: %uMB\r\n", logFile.logFilename, freeSizeMB);
			ConsolePrint("Logfile %s open\r\nFree Space: %uMB\r\n", logFile.logFilename, freeSizeMB);
		}
	}

	if(ConsoleCmpCmd("startlog"))
	{
		ConsolePrint("Writing log header\r\n");

		int16_t result = LogFileOpen(0);
		ConsolePrint("Open result: %hu\r\n", result);
	}

	if(ConsoleCmpCmd("stoplog"))
	{
		LogFileClose();

		NetworkPrint("Log file closed\r\n");
		ConsolePrint("Log file closed\r\n");
	}

	if(ConsoleScanCmd("log %u", buf32) == 1)
	{
		LogPrintAllRecent(*buf32);
	}

	if(ConsoleCmpCmd("netprint"))
	{
		NetworkPrint("Hello Network - MEDIA\n");
	}
}
