/*-----------------------------------------------------------------------*/
/* Low level disk I/O module skeleton for FatFs     (C)ChaN, 2013        */
/*-----------------------------------------------------------------------*/
/* If a working storage control module is available, it should be        */
/* attached to the FatFs via a glue function rather than modifying it.   */
/* This is an example of glue functions to attach various exsisting      */
/* storage control module to the FatFs module with a defined API.        */
/*-----------------------------------------------------------------------*/

#include "diskio.h"		/* FatFs lower layer API */
#include "../usb/usb_msc.h"
#include <stdint.h>

/* Definitions of physical drive number for each media */
#define USB0	0
#define USB1	1
#define SD0		2

/*-----------------------------------------------------------------------*/
/* Inidialize a Drive                                                    */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize(BYTE pdrv /* Physical drive nmuber (0..) */)
{
	switch(pdrv)
	{
		case USB0:
			return 0;
		case USB1:
		case SD0:
			return STA_NODISK;
	}

	return STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status(BYTE pdrv /* Physical drive nmuber (0..) */)
{
	DRESULT res;
	SCSIRequestSenseResult reqSense;

	switch(pdrv)
	{
		case USB0:
			if(UsbMscScsiRequestSense(&reqSense) != 0)
			{
				return STA_NOINIT;
			}
			else
			{
				if(reqSense.sense != SCSI_SENSE_KEY_NO_SENSE)
					return STA_NOINIT;
			}

			if(UsbMscScsiTestUnitReady() == 0)
			{
				res = 0;
			}
			else
			{
				res = STA_NOINIT;
			}

			return res;
		case USB1:
			return STA_NOINIT;
		case SD0:
			return STA_NOINIT;
	}
	return STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read(BYTE pdrv, /* Physical drive nmuber (0..) */
BYTE *buff, /* Data buffer to store read data */
DWORD sector, /* Sector address (LBA) */
UINT count /* Number of sectors to read (1..128) */
)
{
	switch(pdrv)
	{
		case USB0:
		{
			int16_t result = UsbMscScsiRead(sector, count, buff);
			if(result)
				return RES_ERROR;

			return RES_OK;
		}
		case USB1:
		case SD0:
			return RES_NOTRDY;
	}
	return RES_PARERR;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if _USE_WRITE
DRESULT disk_write(BYTE pdrv, /* Physical drive nmuber (0..) */
const BYTE *buff, /* Data to be written */
DWORD sector, /* Sector address (LBA) */
UINT count /* Number of sectors to write (1..128) */
)
{
	switch(pdrv)
	{
		case USB0:
		{
			int16_t result = UsbMscScsiWrite(sector, count, (uint8_t*) buff);
			if(result)
				return RES_ERROR;

			return RES_OK;
		}
		case USB1:
		case SD0:
			return RES_NOTRDY;
	}
	return RES_PARERR;
}
#endif

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL
DRESULT disk_ioctl(BYTE pdrv, /* Physical drive nmuber (0..) */
BYTE cmd, /* Control code */
void *buff /* Buffer to send/receive control data */
)
{
	switch(pdrv)
	{
		case USB0:
			switch(cmd)
			{
				case CTRL_SYNC:
					return RES_OK;
				case GET_SECTOR_COUNT:
					*((uint32_t*) buff) = usbMsc.numBlocks;
					return RES_OK;
				case GET_SECTOR_SIZE:
					*((uint32_t*) buff) = usbMsc.blockSize;
					return RES_OK;
				case GET_BLOCK_SIZE:
					*((uint32_t*) buff) = usbMsc.blockSize;
					return RES_OK;
				default:
					return RES_PARERR;
			}
			break;
		case USB1:
		case SD0:
			return RES_NOTRDY;
	}
	return RES_PARERR;
}
#endif
