/*
 * rtc.c
 *
 *  Created on: 30.11.2012
 *      Author: AndreR
 */

#include "rtc.h"

#include "stm32f407xx.h"

#include <time.h>

static uint8_t byteToBcd2(uint8_t Byte)
{
  uint8_t bcdhigh = 0;

  while (Byte >= 10)
  {
    bcdhigh++;
    Byte -= 10;
  }

  return  ((uint8_t)(bcdhigh << 4) | Byte);
}

static uint8_t bcd2ToByte(uint8_t bcd)
{
  uint8_t tmp = 0;
  tmp = ((uint8_t)(bcd & (uint8_t)0xF0) >> (uint8_t)0x4) * 10;
  return (tmp + (bcd & (uint8_t)0x0F));
}

void RTCInit()
{
	/* Enable the PWR clock */
	RCC->APB1ENR |= RCC_APB1ENR_PWREN;

	/* Allow access to RTC */
	PWR->CR |= PWR_CR_DBP;

	// Select high-speed external oscillator (divided by 12) as clk source => 1MHz
	RCC->BDCR |= RCC_BDCR_RTCSEL;

	// Enable RTC Clock
	RCC->BDCR |= RCC_BDCR_RTCEN;

	// enable write access
	RTC->WPR = 0xCA;
	RTC->WPR = 0x53;

	// enable INIT mode
	RTC->ISR = RTC_ISR_INIT;
	while((RTC->ISR & RTC_ISR_INITF) == 0)
		asm volatile("nop");

	RTC->PRER = 8000;			// SyncDic
	RTC->PRER |= (125 << 16);	//AsyncDiv

//	RTC->TR = 0;	// time is 00:00:00
	if((RTC->DR & RTC_DR_YT) == 0)
		RTC->DR = (1 << 20) | (4 << 16) | (3 << 13) | (1 << 8) | 1;	// 1.1.2014

	// exit INIT mode
	RTC->ISR &= ~RTC_ISR_INIT;

	// enable write-protection
	RTC->WPR = 0xFF;
}

// Return unix timestamp
uint32_t RTCGetTime()
{
	uint32_t dr = RTC->DR;
	uint32_t tr = RTC->TR;

	uint32_t day = bcd2ToByte(dr & 0x3F);
	uint32_t month = bcd2ToByte((dr & 0x1F00) >> 8);
	uint32_t year = bcd2ToByte((dr & 0xFF0000) >> 16);

	uint32_t sec = bcd2ToByte(tr & 0x7F);
	uint32_t min = bcd2ToByte((tr & 0x7F00) >> 8);
	uint32_t hour = bcd2ToByte((tr & 0x3F0000) >> 16);

	struct tm t;
	t.tm_year = year + 100;
	t.tm_mon = month-1; // Month, 0 - jan
	t.tm_mday = day; // Day of the month
	t.tm_hour = hour;
	t.tm_min = min;
	t.tm_sec = sec;
	t.tm_isdst = -1; // Is DST on? 1 = yes, 0 = no, -1 = unknown

	time_t time;
	time = mktime(&t);

	return (uint32_t)time;
}

void RTCSetTime(uint32_t unixTime)
{
	time_t time = (time_t)unixTime;

	struct tm* pT = localtime(&time);

	uint32_t dr = 0;
	uint32_t tr = 0;

	dr |= byteToBcd2(pT->tm_mday);
	dr |= byteToBcd2(pT->tm_mon+1) << 8;
	dr |= (((pT->tm_wday+6)%7)+1) << 13;
	dr |= byteToBcd2(pT->tm_year-100) << 16;

	tr |= byteToBcd2(pT->tm_sec);
	tr |= byteToBcd2(pT->tm_min) << 8;
	tr |= byteToBcd2(pT->tm_hour) << 16;

	// enable write access
	RTC->WPR = 0xCA;
	RTC->WPR = 0x53;

	// enable INIT mode
	RTC->ISR = RTC_ISR_INIT;
	while((RTC->ISR & RTC_ISR_INITF) == 0)
		asm volatile("nop");

	RTC->DR = dr;
	RTC->TR = tr;

	// exit INIT mode
	RTC->ISR &= ~RTC_ISR_INIT;

	// enable write-protection
	RTC->WPR = 0xFF;
}

DateTime RTCGetDateTime()
{
	DateTime result;

	uint32_t dr = RTC->DR;
	uint32_t tr = RTC->TR;

	uint32_t day = bcd2ToByte(dr & 0x3F);
	uint32_t month = bcd2ToByte((dr & 0x1F00) >> 8);
	uint32_t year = bcd2ToByte((dr & 0xFF0000) >> 16);

	uint32_t sec = bcd2ToByte(tr & 0x7F);
	uint32_t min = bcd2ToByte((tr & 0x7F00) >> 8);
	uint32_t hour = bcd2ToByte((tr & 0x3F0000) >> 16);

	result.year = year+2000;
	result.month = month;
	result.day = day;
	result.hour = hour;
	result.min = min;
	result.sec = sec;

	return result;
}
