/*
 * tft.c
 *
 *  Created on: 04.12.2012
 *      Author: AndreR
 */
#include "tft.h"

#include "util/init_hal.h"
#include "../constants.h"
#include "ch.h"

#define LCD_HPS		(LCD_WIDTH-1)  //Set the horizontal panel size
#define LCD_VPS		(LCD_HEIGHT-1)  //Set the vertical panel size

#define LCD_Panel_Data_Width				0//0:18bit; 1:24bit;
#define LCD_Pixel_Data_Interface_Format		3 // 16-bit (565 format)

#define LCD_DotClk_Polarity			1 //0:Data latch in rising edge; 1:Data latch in falling edge
#define LCD_Hsync_Polarity			0 //0:Active low; 1:Active high
#define LCD_Vsync_Polarity			0 //0:Active low; 1:Active high

#define LCD_Panel_Mode				0 //0:Hsync+Vsync +DE mode; 1:TTL mode
#define LCD_TFT_Type				0 //0 or 1:TFT mode; 2:Serial RGB mode; 3:Serial RGB+dummy mode

#define LCD_Panel_Mode1	(LCD_Panel_Data_Width<<5)|(0<<4)|(0<<3)|(LCD_DotClk_Polarity<<2)|(LCD_Hsync_Polarity<<1)|(LCD_Vsync_Polarity)
#define LCD_Panel_Mode2	(LCD_Panel_Mode<<7)|(LCD_TFT_Type<<5)

//Horizontal Cycle = Horizontal Pulse Width + Horizontal Back Porch + Horizontal Display Period + Horizontal Front Porch
//Horizontal Start Position = Horizontal Pulse Width + Horizontal Back Porch
#define LCD_HORI_FRONT_PORCH		70//0x90//0x30//2
#define LCD_HORI_BACK_PORCH			8//0x05//0x0F//2
#define LCD_HORI_PULSE_WIDTH		2//0x1e//0x60//41

//Vertical Cycle = Vertical Pulse Width + Vertical Back Porch + Vertical Display Period + Vertical Front Porch
//Vertical Start Position = Vertical Pulse Width + Vertical Back Porch
#define LCD_VERT_FRONT_PORCH		13//0x20//3//2
#define LCD_VERT_BACK_PORCH			8//2//0//2
#define LCD_VERT_PULSE_WIDTH		2//2//10

#define LCD_HORI_CYCLE				(LCD_HORI_PULSE_WIDTH + LCD_HORI_BACK_PORCH + LCD_WIDTH + LCD_HORI_FRONT_PORCH)
#define LCD_VERT_CYCLE				(LCD_VERT_PULSE_WIDTH + LCD_VERT_BACK_PORCH + LCD_HEIGHT + LCD_VERT_FRONT_PORCH)

#define LCD_HORI_START				(LCD_HORI_PULSE_WIDTH + LCD_HORI_BACK_PORCH)
#define LCD_VERT_START				(LCD_VERT_PULSE_WIDTH + LCD_VERT_BACK_PORCH)

static void regInit()
{
	*TFT_CMD_REG = TFT_REG_SOFT_RESET;	// soft reset
	chThdSleep(MS2ST(10));

	// set LSHIFT frequency (DotClk)
	*TFT_CMD_REG = TFT_REG_SET_LSHIFT_FREQ;
	*TFT_DATA_REG = 0x01;
	*TFT_DATA_REG = 0x90;
	*TFT_DATA_REG = 0xFF;

	// Set LCD mode
	*TFT_CMD_REG = TFT_REG_SET_LCD_MODE;
	*TFT_DATA_REG = LCD_Panel_Mode1; //  24Bit data width, dot clk polatrity 1, vsync active low, hsync active low
	*TFT_DATA_REG = LCD_Panel_Mode2; // Mode: hsync+vsync +DE, TFT mode
	*TFT_DATA_REG = LCD_HPS >> 8;
	*TFT_DATA_REG = LCD_HPS;
	*TFT_DATA_REG = LCD_VPS >> 8;
	*TFT_DATA_REG = LCD_VPS;
	*TFT_DATA_REG = 0x00; // RGB sequence

	// Set Horizontal Period
	*TFT_CMD_REG = TFT_REG_SET_HORI_PERIOD;
	*TFT_DATA_REG = LCD_HORI_CYCLE >> 8;
	*TFT_DATA_REG = LCD_HORI_CYCLE;
	*TFT_DATA_REG = LCD_HORI_FRONT_PORCH >> 8;
	*TFT_DATA_REG = LCD_HORI_FRONT_PORCH;
	*TFT_DATA_REG = LCD_HORI_PULSE_WIDTH;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;

	// Set Vertical Period
	*TFT_CMD_REG = TFT_REG_SET_VERT_PERIOD;
	*TFT_DATA_REG = LCD_VERT_CYCLE >> 8;
	*TFT_DATA_REG = LCD_VERT_CYCLE;
	*TFT_DATA_REG = LCD_VERT_FRONT_PORCH >> 8;
	*TFT_DATA_REG = LCD_VERT_FRONT_PORCH;
	*TFT_DATA_REG = LCD_VERT_PULSE_WIDTH;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;

	// Set GPIO value
	*TFT_CMD_REG = TFT_REG_SET_GPIO_VALUE;
	*TFT_DATA_REG = 0x0F;

	chThdSleep(MS2ST(100));

	// Display on
	*TFT_CMD_REG = TFT_REG_SET_DISPLAY_ON;

	chThdSleep(MS2ST(100));

	// Set column address
	*TFT_CMD_REG = TFT_REG_SET_COLUMN_ADDRESS;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = LCD_HPS >>8;
	*TFT_DATA_REG = LCD_HPS;

	// Set Page address
	*TFT_CMD_REG = TFT_REG_SET_PAGE_ADDRESS;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = LCD_VPS >> 8;
	*TFT_DATA_REG = LCD_VPS;

	// Set address mode
	*TFT_CMD_REG = TFT_REG_SET_ADDRESS_MODE;
	*TFT_DATA_REG = 0x00;

	// Set pixel data interface
	*TFT_CMD_REG = TFT_REG_SET_PIXEL_DATA_INTERFACE;
	*TFT_DATA_REG = LCD_Pixel_Data_Interface_Format;

	// Set GPIO config
	*TFT_CMD_REG = TFT_REG_SET_GPIO_CONF;
	*TFT_DATA_REG = 0x0F; // GPIO 0-3 as output
	*TFT_DATA_REG = 0x01; // normal GPIOs

	// Set GPIO value
	*TFT_CMD_REG = TFT_REG_SET_GPIO_VALUE;
	*TFT_DATA_REG = 0x0F;

	// Set PWM config
	*TFT_CMD_REG = TFT_REG_SET_PWM_CONF;
	*TFT_DATA_REG = 0x0E;
	*TFT_DATA_REG = 0xFF;
	*TFT_DATA_REG = 0x01;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x0F;
}

void TFTWritePixel(uint32_t color)
{
	*TFT_DATA_REG = TFT_COLOR(color);
}

void TFTInit()
{
	GPIOInitData gpioInit;

	// Configure RESET Pin
	GPIOG->BSRRL = GPIO_PIN_15;

	gpioInit.mode = GPIO_MODE_OUTPUT;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_2MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOG, GPIO_PIN_15, &gpioInit);

	chThdSleep(MS2ST(1000));	// power up delay for display

	// Generate RESET pulse
	GPIOG->BSRRH = GPIO_PIN_15;
	chThdSleep(MS2ST(10));
	GPIOG->BSRRL = GPIO_PIN_15;
	chThdSleep(MS2ST(10));

	FSMC_Bank1->BTCR[0] = FSMC_BCR1_MWID_0 | FSMC_BCR1_WREN;
	FSMC_Bank1->BTCR[1] = (200 << 8);	// DataSetupTime = 200, display is very slow as long as PLL is not set
	FSMC_Bank1->BTCR[0] |= FSMC_BCR1_MBKEN;

	*TFT_CMD_REG = 0x01; // soft reset
	chThdSleep(MS2ST(10));

	// init display PLL to 110Mhz
	*TFT_CMD_REG = TFT_REG_SET_PLL_MN;
	*TFT_DATA_REG = 0x20;
	*TFT_DATA_REG = 0x02;
	*TFT_DATA_REG = 0x04;

	*TFT_CMD_REG = TFT_REG_SET_PLL;
	*TFT_DATA_REG = 0x01;
	chThdSleep(MS2ST(100)); // wait for PLL lock
	*TFT_CMD_REG = TFT_REG_SET_PLL;
	*TFT_DATA_REG = 0x03;

	FSMC_Bank1->BTCR[0] &= ~FSMC_BCR1_MBKEN;
	FSMC_Bank1->BTCR[0] |= FSMC_BCR1_EXTMOD;	// extended mode
	FSMC_Bank1->BTCR[1] = (2 << 8);	// DataSetupTime = 2, write can be very fast :)
	FSMC_Bank1E->BWTR[0] = (3 << 8);// DataSetupTime = 3, reading is not so fast :/

	FSMC_Bank1->BTCR[0] |= FSMC_BCR1_MBKEN;

	chThdSleep(MS2ST(2000));

	regInit();
}

void TFTSetBrightness(uint8_t level)
{
	// Set PWM config
	*TFT_CMD_REG = TFT_REG_SET_PWM_CONF;
	*TFT_DATA_REG = 0x0E;  // frequency
	*TFT_DATA_REG = level;  //  PWM duty cycle
	*TFT_DATA_REG = 0x01;  //  PWM enabled & PWM controlled by host
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x0F;
}
