/*
 * touch.c
 *
 *  Created on: 05.07.2014
 *      Author: AndreR
 */

#include "touch.h"
#include "../constants.h"
#include "util/bit_banding.h"
#include "util/init_hal.h"
#include "util/console.h"
#include "GUI.h"
#include "tft.h"
#include <math.h>

Touch touch;

#define TOUCH_X_MIN		2048	// 204.8*10
#define TOUCH_X_MAX		63488	// 204.8*310
#define TOUCH_Y_MIN		2730	// 273.066*10
#define TOUCH_Y_MAX		62805	// 273.066*230
#define TOUCH_X_RANGE	(TOUCH_X_MAX-TOUCH_X_MIN)
#define TOUCH_Y_RANGE	(TOUCH_Y_MAX-TOUCH_Y_MIN)

typedef struct _TSMeas
{
	uint32_t x;
	uint32_t y;
	uint32_t z1;
	uint32_t z2;
} TSMeas;

// touch event active
void EXTI2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	BBPeriphSet((uint32_t)&EXTI->PR, 2);

	chSysLockFromIsr();
	chBSemSignalI(&touch.touchedSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void DMA1_Stream2_IRQHandler()
{
	SPISyncDmaRxInterrupt(&touch.spi3);
}

void TouchInit()
{
	chBSemInit(&touch.touchedSem, TRUE);

	RCC->APB1ENR |= RCC_APB1ENR_SPI3EN;

	SPISyncData data3;
	data3.cpha = 0;
	data3.cpol = 0;
	data3.csPin = GPIO_PIN_6;
	data3.pCSPort = GPIOD;
	data3.dmaChRx = 2;
	data3.dmaChTx = 7;
	data3.dmaPrio = 1;
	data3.pDma = DMA1;
	data3.pRegister = SPI3;
	data3.prescaler = SPI_CR1_BRDIV2;
	data3.timeoutTicks = MS2ST(20);
	SPISyncInit(&touch.spi3, &data3);

	NVICEnableIRQ(DMA1_Stream2_IRQn, IRQL_SPI3);

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 6;	// SPI3
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_25MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOC, GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit);	// SCK, MISO, MOSI

	// TP-IRQ: E2
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING_FALLING;
	GPIOInit(GPIOE, GPIO_PIN_2, &gpioInit);

	NVICEnableIRQ(EXTI2_IRQn, IRQL_EXTI2);

	touch.buf[0] = 0x03 << 1;
	touch.buf[1] = 0xFF; // 16 samples, median filter
	touch.buf[2] = 0x07 << 1;
	touch.buf[3] = 0x60; // set PUR to 1ms
	touch.buf[4] = 0x0B << 1; // write Mode Config Register
	touch.buf[5] = 0x06; // avg filter on x,y,z1,z2
	touch.buf[6] = 0x04 << 1;
	touch.buf[7] = 0x55; // ADC sample time 128us on all channels
	touch.buf[8] = 0x05 << 1;
	touch.buf[9] = 0x44; // Panel setup time 1ms
	touch.buf[10]= 0x06 << 1;
	touch.buf[11]= 0x00; // Delayed conversion 1ms
	touch.buf[12]= 0x01 << 1;
	touch.buf[13]= 0xC0;	//

	SPISyncTransfer(&touch.spi3, touch.buf, 0, 12);
}

msg_t TouchTask(void* params)
{
	(void)params;
	TSMeas result;

	chRegSetThreadName("Touch");

	uint32_t touchTime = chTimeNow();

	while(1)
	{
		chBSemWaitTimeout(&touch.touchedSem, MS2ST(10));

		// CMD_MEAS_XYZ
		touch.buf[0] = 0x74 << 1;
		SPISyncTransfer(&touch.spi3, touch.buf, 0, 1);

		chThdSleep(MS2ST(10));

		// CMD_READ_XYZ
		touch.buf[0] = 0x52 << 1 | 0x01;
		SPISyncTransfer(&touch.spi3, touch.buf, touch.buf, 9);

		result.x =  touch.buf[1] << 8;
		result.x |= touch.buf[2];
		result.y =  touch.buf[3] << 8;
		result.y |= touch.buf[4];
		result.z1 =  touch.buf[5] << 8;
		result.z1 |= touch.buf[6];
		result.z2 =  touch.buf[7] << 8;
		result.z2 |= touch.buf[8];

		result.y = 65536-result.y;	// Y starts at 0 in lower left corner, so we invert it to match pixel positions

		float x = result.x;
		float z1 = result.z1;
		float z2 = result.z2;

		touch.touchZ1 = result.z1;
		touch.touchZ2 = result.z2;

		touch.touchZ = (x/65536.0f)*((z2/z1)-1.0f);
		if(touch.touchZ < 2.0f && z1 > 1000.0f && z2 < 65000.0f
			&& result.x > TOUCH_X_MIN && result.x < TOUCH_X_MAX && result.y > TOUCH_Y_MIN && result.y < TOUCH_Y_MAX)
		{
			touch.touchY = LCD_WIDTH-((result.x-TOUCH_X_MIN)*LCD_WIDTH)/TOUCH_X_RANGE;
			touch.touchX = ((result.y-TOUCH_Y_MIN)*LCD_HEIGHT)/TOUCH_Y_RANGE;

#ifdef TFT_FLIP
			touch.touchX = LCD_HEIGHT-touch.touchX;
			touch.touchY = LCD_WIDTH-touch.touchY;
#endif

			if(touch.touchDown == 0)
			{
				touch.touchPressed = 1;

				touch.swipe.downPos[0] = touch.touchX;
				touch.swipe.downPos[1] = touch.touchY;
				touch.swipe.downTime = chTimeNow();
			}

			touch.touchDown = 1;
		}
		else
		{
			if(touch.touchDown)
			{
				touch.touchReleased = 1;

				touch.swipe.upPos[0] = touch.touchX;
				touch.swipe.upPos[1] = touch.touchY;
				touch.swipe.upTime = chTimeNow();

				touch.swipe.move[0] = touch.swipe.upPos[0] - touch.swipe.downPos[0];
				touch.swipe.move[1] = touch.swipe.upPos[1] - touch.swipe.downPos[1];
				touch.swipe.deltaT = touch.swipe.upTime - touch.swipe.downTime;

				int32_t x = touch.swipe.move[0];
				int32_t y = touch.swipe.move[1];
				touch.swipe.length = (int16_t)sqrtf(x*x + y*y);

				if(touch.swipe.length > 50 && touch.swipe.deltaT > MS2ST(50))
				{
//					ConsolePrint("Swipe from %hu,%hu to %hu,%hu\r\n", touch.swipe.downPos[0], touch.swipe.downPos[1],
//							touch.swipe.upPos[0], touch.swipe.upPos[1]);
//					ConsolePrint("Relative movement: %hd,%hd\r\n", touch.swipe.move[0], touch.swipe.move[1]);
//					ConsolePrint("Time: %hums, Length: %hu\r\n", touch.swipe.deltaT, touch.swipe.length);

					if(touch.swipeCb)
						(*touch.swipeCb)(&touch.swipe);
				}
			}

			touch.touchDown = 0;
		}

		// forward to STemWin
		if(touch.touchDown)
			GUI_TOUCH_StoreState(touch.touchX, touch.touchY);
		else
			GUI_TOUCH_StoreState(-1, -1);

		if(touch.touchPressed)
		{
			touch.touchPressed = 0;
			touchTime = chTimeNow();
			TFTSetBrightness(255);
		}

		if(chTimeElapsedSince(touchTime) > S2ST(5))
		{
			TFTSetBrightness(20);
		}
	}

	return 0;
}
