/*
 * log_file.c
 *
 *  Created on: 24.08.2014
 *      Author: AndreR
 */

#include "log_file.h"

#include "ch.h"
#include "util/intercom.h"
#include "intercom_constants.h"
#include "util/console.h"
#include "util/sys_time.h"
#include <string.h>
#include <stdio.h>
#include "log_msgs.h"

enum LogElementType
{
	UINT8 = 0,
	INT8,
	UINT16,
	INT16,
	UINT32,
	INT32,
	FLOAT,
	DOUBLE,
//	HALF_FLOAT,	// TODO: support will be added later, if reasonable
//	STRING,
};

//#define ARRAY_OF(t, n) (t + (n << 4))	// TODO: support will be added later

// describes a single structure element
typedef struct PACKED _LogElementDesc
{
	uint8_t type;
	const char* pName;
	const char* pUnit;
	const char* pDesc;
} LogElementDesc;

// describes a message
typedef struct PACKED _LogMessageDesc
{
	uint16_t msgId;
	const char* pName;
	uint16_t numElements;
	LogElementDesc* elements;
} LogMessageDesc;

static __attribute__ ((section (".extram"))) uint8_t writeBuf[FIFO_BLOCK_DATA(BUFFER_CHUNKS, BUFFER_SIZE)];

LogFile logFile;

static const LogMessageDesc msgs[] = {
	{ CTYPE_GYRO_RAW, "gyro_raw", 3, (LogElementDesc[]) {
		{ INT16, "X", "raw", "X axis" },
		{ INT16, "Y", "raw", "Y axis" },
		{ INT16, "Z", "raw", "Z axis" },
	}, },
	{ CTYPE_ACC_RAW, "acc_raw", 3, (LogElementDesc[]) {
		{ INT16, "X", "raw", "X axis" },
		{ INT16, "Y", "raw", "Y axis" },
		{ INT16, "Z", "raw", "Z axis" },
	}, },
	{ CTYPE_CTRL_INPUT, "ctrl_input", 51, (LogElementDesc[]) {
		{ UINT8, 	"cur_updated", "bool", "Motor current updated" },
		{ UINT32,	"cur_m1_t", "us", "Motor 1 current timestamp" },
		{ UINT32,	"cur_m2_t", "us", "Motor 2 current timestamp" },
		{ UINT32,	"cur_m3_t", "us", "Motor 3 current timestamp" },
		{ UINT32,	"cur_m4_t", "us", "Motor 4 current timestamp" },
		{ FLOAT,	"cur_m1_i", "A",  "Motor 1 current" },
		{ FLOAT,	"cur_m2_i", "A",  "Motor 2 current" },
		{ FLOAT,	"cur_m3_i", "A",  "Motor 3 current" },
		{ FLOAT,	"cur_m5_i", "A",  "Motor 4 current" },

		{ UINT8,	"vel_updated", "bool", "Motor velocities updated" },
		{ UINT32,	"vel_m1_t", "us",	"Motor 1 velocity timestamp" },
		{ UINT32,	"vel_m2_t", "us",	"Motor 2 velocity timestamp" },
		{ UINT32,	"vel_m3_t", "us",	"Motor 3 velocity timestamp" },
		{ UINT32,	"vel_m4_t", "us",	"Motor 4 velocity timestamp" },
		{ FLOAT,	"vel_m1_v", "rad/s","Motor 1 angular velocity" },
		{ FLOAT,	"vel_m2_v", "rad/s","Motor 2 angular velocity" },
		{ FLOAT,	"vel_m3_v", "rad/s","Motor 3 angular velocity" },
		{ FLOAT,	"vel_m4_v", "rad/s","Motor 4 angular velocity" },

		{ UINT8,	"gyr_updated", "bool", "Gyroscope data updated" },
		{ UINT32,	"gyr_t", "us", "Timestamp" },
		{ FLOAT,	"gyr_x", "rad/s", "Angular velocity around X axis" },
		{ FLOAT,	"gyr_y", "rad/s", "Angular velocity around Y axis" },
		{ FLOAT,	"gyr_z", "rad/s", "Angular velocity around Z axis" },

		{ UINT8,	"acc_updated", "bool", "Accelerometer data updated" },
		{ UINT32,	"acc_t", "us", "Timestamp" },
		{ FLOAT,	"acc_x", "m/s^2", "Acceleration along X axis" },
		{ FLOAT,	"acc_y", "m/s^2", "Acceleration along Y axis" },
		{ FLOAT,	"acc_z", "m/s^2", "Acceleration along Z axis" },

		{ UINT8,	"pos_updated", "bool", "Position data updated" },
		{ UINT32,	"pos_t", "us", "Timestamp" },
		{ FLOAT,	"pos_x", "m", "X Position" },
		{ FLOAT,	"pos_y", "m", "Y Position" },
		{ FLOAT,	"pos_z", "m", "Z Position" },

		{ UINT8,	"supply_updated", "bool", "Supply data updated" },
		{ FLOAT,	"supply_bat", "V", "Battery voltage" },

		{ UINT32,	"vol_m1_t", "us",	"Motor 1 voltage timestamp" },
		{ UINT32,	"vol_m2_t", "us",	"Motor 2 voltage timestamp" },
		{ UINT32,	"vol_m3_t", "us",	"Motor 3 voltage timestamp" },
		{ UINT32,	"vol_m4_t", "us",	"Motor 4 voltage timestamp" },
		{ FLOAT,	"vol_m1_u", "V","Motor 1 voltage" },
		{ FLOAT,	"vol_m2_u", "V","Motor 2 voltage" },
		{ FLOAT,	"vol_m3_u", "V","Motor 3 voltage" },
		{ FLOAT,	"vol_m4_u", "V","Motor 4 voltage" },

		{ UINT32,	"M_m1_t", "us",	"Motor 1 torque timestamp" },
		{ UINT32,	"M_m2_t", "us",	"Motor 2 torque timestamp" },
		{ UINT32,	"M_m3_t", "us",	"Motor 3 torque timestamp" },
		{ UINT32,	"M_m4_t", "us",	"Motor 4 torque timestamp" },
		{ FLOAT,	"M_m1_u", "Nm","Motor 1 torque" },
		{ FLOAT,	"M_m2_u", "Nm","Motor 2 torque" },
		{ FLOAT,	"M_m3_u", "Nm","Motor 3 torque" },
		{ FLOAT,	"M_m4_u", "Nm","Motor 4 torque" },
	}, },
	{ CTYPE_CTRL_STATE, "ctrl_state", 12, (LogElementDesc[]) {
		{ UINT8,	"pos_updated", "bool", "Position estimation updated" },
		{ FLOAT,	"pos_x", "m", "X Position" },
		{ FLOAT,	"pos_y", "m", "Y Position" },
		{ FLOAT,	"pos_z", "m", "Z Position" },

		{ UINT8,	"vel_updated", "bool", "Velocity estimation updated" },
		{ FLOAT,	"vel_x", "m/s", "X Velocity" },
		{ FLOAT,	"vel_y", "m/s", "Y Velocity" },
		{ FLOAT,	"vel_z", "m/s", "Z Velocity" },

		{ UINT8,	"acc_updated", "bool", "Acceleration estimation updated" },
		{ FLOAT,	"acc_x", "m/s^2", "X Acceleration" },
		{ FLOAT,	"acc_y", "m/s^2", "Y Acceleration" },
		{ FLOAT,	"acc_z", "m/s^2", "Z Acceleration" },
	}, },
	{ CTYPE_CTRL_OUTPUT, "ctrl_output", 13, (LogElementDesc[]) {
		{ FLOAT,	"vel_m1", "rad/s", "Motor 1 velocity" },
		{ FLOAT,	"vel_m2", "rad/s", "Motor 2 velocity" },
		{ FLOAT,	"vel_m3", "rad/s", "Motor 3 velocity" },
		{ FLOAT,	"vel_m4", "rad/s", "Motor 4 velocity" },
		{ UINT8,	"ctrlMode", "-", "Motor controller mode"},
		{ FLOAT,	"vol_m1", "V", "Motor 1 voltage" },
		{ FLOAT,	"vol_m2", "V", "Motor 2 voltage" },
		{ FLOAT,	"vol_m3", "V", "Motor 3 voltage" },
		{ FLOAT,	"vol_m4", "V", "Motor 4 voltage" },
		{ FLOAT,	"torque_m1", "Nm", "Motor 1 torque" },
		{ FLOAT,	"torque_m2", "Nm", "Motor 2 torque" },
		{ FLOAT,	"torque_m3", "Nm", "Motor 3 torque" },
		{ FLOAT,	"torque_m4", "Nm", "Motor 4 torque" },
	}, },
	{ CTYPE_CTRL_TIGGA_PIDS, "ctrl_tigga_pids", 45, (LogElementDesc[]) {
		{ UINT32,	"newTrajTime", "us", "Trajectory Generation Time" },

		{ FLOAT,	"traj_pos_x", "m", "Trajectory Position X" },
		{ FLOAT,	"traj_pos_y", "m", "Trajectory Position Y" },
		{ FLOAT,	"traj_pos_z", "m", "Trajectory Orientation" },

		{ FLOAT,	"traj_vel_x", "m/s", "Trajectory Velocity X" },
		{ FLOAT,	"traj_vel_y", "m/s", "Trajectory Velocity Y"},
		{ FLOAT,	"traj_vel_z", "m/s", "Trajectory Angular Velocity" },

		{ FLOAT,	"traj_acc_x", "m/s^2", "Trajectory Acceleration X" },
		{ FLOAT,	"traj_acc_y", "m/s^2", "Trajectory Acceleration Y" },
		{ FLOAT,	"traj_acc_z", "m/s^2", "Trajectory Angular Acceleration" },

		{ FLOAT,	"acc_fwd_x", "m/s^2", "Trajectory Acceleration Feedforward X" },
		{ FLOAT,	"acc_fwd_y", "m/s^2", "Trajectory Acceleration Feedforward Y" },
		{ FLOAT,	"acc_fwd_z", "m/s^2", "Trajectory Angular Acceleration Feedforward " },

		{ FLOAT,	"pid_pos_x_in", "m", "PID Position X, In" },
		{ FLOAT,	"pid_pos_x_out", "m", "PID Position X, Out" },
		{ FLOAT,	"pid_pos_x_err", "m", "PID Position X, Error" },
		{ FLOAT,	"pid_pos_x_iTerm", "m", "PID Position X, Integral Term" },
		{ FLOAT,	"pid_pos_x_dTerm", "m", "PID Position X, Differential Term" },

		{ FLOAT,	"pid_pos_y_in", "m", "PID Position Y, In" },
		{ FLOAT,	"pid_pos_y_out", "m", "PID Position Y, Out" },
		{ FLOAT,	"pid_pos_y_err", "m", "PID Position Y, Error" },
		{ FLOAT,	"pid_pos_y_iTerm", "m", "PID Position Y, Integral Term" },
		{ FLOAT,	"pid_pos_y_dTerm", "m", "PID Position Y, Differential Term" },

		{ FLOAT,	"pid_pos_z_in", "m", "PID Position Z, In" },
		{ FLOAT,	"pid_pos_z_out", "m", "PID Position Z, Out" },
		{ FLOAT,	"pid_pos_z_err", "m", "PID Position Z, Error" },
		{ FLOAT,	"pid_pos_z_iTerm", "m", "PID Position Z, Integral Term" },
		{ FLOAT,	"pid_pos_z_dTerm", "m", "PID Position Z, Differential Term" },

		{ FLOAT,	"pid_vel_x_in", "m/s", "PID Velocity X, In" },
		{ FLOAT,	"pid_vel_x_out", "m/s", "PID Velocity X, Out" },
		{ FLOAT,	"pid_vel_x_err", "m/s", "PID Velocity X, Error" },
		{ FLOAT,	"pid_vel_x_iTerm", "m/s", "PID Velocity X, Integral Term" },
		{ FLOAT,	"pid_vel_x_dTerm", "m/s", "PID Velocity X, Differential Term" },

		{ FLOAT,	"pid_vel_y_in", "m/s", "PID Velocity Y, In" },
		{ FLOAT,	"pid_vel_y_out", "m/s", "PID Velocity Y, Out" },
		{ FLOAT,	"pid_vel_y_err", "m/s", "PID Velocity Y, Error" },
		{ FLOAT,	"pid_vel_y_iTerm", "m/s", "PID Velocity Y, Integral Term" },
		{ FLOAT,	"pid_vel_y_dTerm", "m/s", "PID Velocity Y, Differential Term" },

		{ FLOAT,	"pid_vel_z_in", "m/s", "PID Velocity Z, In" },
		{ FLOAT,	"pid_vel_z_out", "m/s", "PID Velocity Z, Out" },
		{ FLOAT,	"pid_vel_z_err", "m/s", "PID Velocity Z, Error" },
		{ FLOAT,	"pid_vel_z_iTerm", "m/s", "PID Velocity Z, Integral Term" },
		{ FLOAT,	"pid_vel_z_dTerm", "m/s", "PID Velocity Z, Differential Term" },

		{ UINT8,	"trajActive", "-", "Trajectory active"},
		{ UINT8,	"accFwd", "-", "Acceleration Feedforward enabled"},
	}, },
};


static void writeFile(uint32_t bufIndex)
{
	chMBPost(&logFile.eventQueue, bufIndex, TIME_INFINITE);
}

static void logfileCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	uint16_t bytesWritten = 0;

	if(pHeader->command == IC_LOGFILE_MESSAGE)
	{
		while(bytesWritten + sizeof(LogChunkHeader) < dataLength)
		{
			LogFileWrite(pData);

			LogChunkHeader* pHeader = (LogChunkHeader*)pData;
			pData += pHeader->length;
			bytesWritten += pHeader->length;
		}
	}
}

void LogFileInit()
{
	logFile.ready = 0;
	chMBInit(&logFile.eventQueue, logFile.eventQueueData, LOG_FILE_QUEUE_SIZE);
	chMtxInit(&logFile.fileMutex);

	FifoBlockInit(&logFile.writeFifo, BUFFER_CHUNKS, BUFFER_SIZE, writeBuf);
	logFile.writeFifo.fullCb = &writeFile;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &logfileCallback;
	client.section = INTERCOM_SECTION_LOGFILE;
	IntercomAddClient(client);
}

int16_t LogFileOpen(const char* pName)
{
	FRESULT fResult;
	LogChunkHeader cHeader;

	chMtxLock(&logFile.fileMutex);
	if(pName)
	{
		snprintf(logFile.logFilename, 64, "%s.dat", pName);
		fResult = f_open(&logFile.file, logFile.logFilename, FA_WRITE | FA_CREATE_ALWAYS);
	}
	else
	{
		unsigned int id = 0;

		do
		{
			snprintf(logFile.logFilename, 64, "log%u.dat", id++);
			fResult = f_open(&logFile.file, logFile.logFilename, FA_WRITE | FA_CREATE_NEW);
		} while(fResult == FR_EXIST);
	}
	chMtxUnlock();

	if(fResult != FR_OK)
		return 1;

	// write message descs
	for(uint32_t msg = 0; msg < sizeof(msgs)/sizeof(msgs[0]); msg++)
	{
		const LogMessageDesc* pMsgDesc = &msgs[msg];

		// now write the message
		cHeader.type = CTYPE_MESSAGE_DESC;
		cHeader.length = sizeof(uint16_t)+sizeof(uint16_t)+strlen(pMsgDesc->pName)+1;	// msg ID, numElements, name[]
		FifoBlockPush(&logFile.writeFifo, &cHeader, sizeof(LogChunkHeader));
		FifoBlockPush(&logFile.writeFifo, &pMsgDesc->msgId, sizeof(uint16_t));
		FifoBlockPush(&logFile.writeFifo, &pMsgDesc->numElements, sizeof(uint16_t));
		FifoBlockPush(&logFile.writeFifo, pMsgDesc->pName, strlen(pMsgDesc->pName)+1);

		for(uint16_t el = 0; el < pMsgDesc->numElements; el++)
		{
			const LogElementDesc* pElDesc = &pMsgDesc->elements[el];

			cHeader.type = CTYPE_ELEMENT_DESC;
			cHeader.length = sizeof(uint8_t) + strlen(pElDesc->pUnit)		// type and unit
					+ strlen(pElDesc->pName) + strlen(pElDesc->pDesc) + 3;	// string lengths

			FifoBlockPush(&logFile.writeFifo, &cHeader, sizeof(LogChunkHeader));
			FifoBlockPush(&logFile.writeFifo, &pElDesc->type, sizeof(uint8_t));
			FifoBlockPush(&logFile.writeFifo, pElDesc->pName, strlen(pElDesc->pName)+1);
			FifoBlockPush(&logFile.writeFifo, pElDesc->pUnit, strlen(pElDesc->pUnit)+1);
			FifoBlockPush(&logFile.writeFifo, pElDesc->pDesc, strlen(pElDesc->pDesc)+1);
		}
	}

	logFile.event.pFilename = logFile.logFilename;
	logFile.event.fileSize = 0;
	logFile.event.eventType = LOGFILE_EVENT_START;
	logFile.lastEventSize = 0;

	logFile.ready = 1;

	if(logFile.eventCb)
		(*logFile.eventCb)(&logFile.event);

	return 0;
}

void LogFileClose()
{
	if(logFile.ready == 0)
		return;

	logFile.ready = 0;

	FifoBlockFlush(&logFile.writeFifo);

	chMtxLock(&logFile.fileMutex);
	f_close(&logFile.file);
	logFile.file.fs = 0;
	chMtxUnlock();
}

void LogFileWrite(void* pLogMsg)
{
	LogChunkHeader* pHeader = (LogChunkHeader*)pLogMsg;

	logFile.bytesPushed += pHeader->length;

	if(chTimeElapsedSince(logFile.lastPushedTime) > MS2ST(1000))
	{
		logFile.lastPushedTime = chTimeNow();

		logFile.event.bytesPerSecIn = logFile.bytesPushed;
		logFile.bytesPushed = 0;

		if(logFile.eventCb && logFile.ready)
			(*logFile.eventCb)(&logFile.event);
	}

	if(logFile.ready)
		FifoBlockPush(&logFile.writeFifo, pLogMsg, pHeader->length);
}

msg_t LogFileTask(void* params)
{
	(void)params;

	msg_t event;
	uint8_t* pData;
	uint32_t dataSize;
	UINT bytesWritten;
	FRESULT fresult;

	chRegSetThreadName("LogFile");

	while(1)
	{
		if(chMBFetch(&logFile.eventQueue, &event, MS2ST(500)) != RDY_OK)
		{
			if(logFile.ready == 0)
			{
				logFile.event.eventType = LOGFILE_EVENT_CLOSE;
				logFile.event.fileSize = 0;
				logFile.event.bytesPerSecIn = 0;
				logFile.event.writeSpeed = 0;
				snprintf(logFile.logFilename, 64, "N/A");

				if(logFile.eventCb)
					(*logFile.eventCb)(&logFile.event);
			}

			continue;
		}

		if(FifoBlockGetData(&logFile.writeFifo, event, &pData, &dataSize) != 0)
			continue;

		uint32_t start = SysTimeUSec();
		chMtxLock(&logFile.fileMutex);
		fresult = f_write(&logFile.file, pData, dataSize, &bytesWritten);
		chMtxUnlock();
		uint32_t writeTime = SysTimeUSec()-start;

		logFile.event.writeSpeed = ((1e9/writeTime)*bytesWritten)/1000;

		if(fresult == FR_OK)
		{
			logFile.event.fileSize += bytesWritten;
			logFile.event.eventType = LOGFILE_EVENT_UPDATE;

			if(logFile.event.fileSize - logFile.lastEventSize > 64000 && logFile.eventCb)
			{
				logFile.lastEventSize = logFile.event.fileSize;
				(*logFile.eventCb)(&logFile.event);
			}
		}

		FifoBlockFreeBuffer(&logFile.writeFifo, event);
	}

	return 0;
}

/**
General chunk header:
uint16_t chunkType	// MSB reserved for control chunks
uint16_t chunkLength

0x8000	message description
0x8001	element description
0x8002	unit description

File Order:
unit descriptions
message descriptions
- element desc
- ...
...
 */

