#include "hal/led.h"
#include "hal/hw_sync1.h"
#include "hal/usart2.h"
#include "hal/exti.h"
#include "hal/fsmc.h"
#include "hal/sram.h"
#include "hal/audio.h"
#include "hal/touch.h"
#include "hal/ov7725.h"

#include "cli.h"
#include "presenter.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "usb/usb.h"
#include "usb/usb_hcd.h"
#include "usb/usb_msc.h"
#include "log_file.h"
#include "prog_store.h"
#include "prog_source_usb.h"
#include "prog_source_wifi.h"

#include "util/boot.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/crc.h"
#include "util/network_print.h"
#include "util/sys_time.h"
#include "util/intercom.h"
#include "util/intercom_reliable.h"
#include "util/prog_client.h"
#include "util/config.h"

#include "ch.h"

#include "GUI.h"

#include <stdlib.h>

static WORKING_AREA(waUSART2, 1024);
static WORKING_AREA(waIntercom, 512);
#ifdef ENV_RUN
static WORKING_AREA(waCLI, 8192);
static WORKING_AREA(waLogFile, 1024);
static WORKING_AREA(waTouch, 1024);
static WORKING_AREA(waUSB, 4096);
static WORKING_AREA(waUSBHCD, 4096);
static WORKING_AREA(waPresenter, 4096);
static WORKING_AREA(waConfig, 512);
#endif

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	// Enable important clocks
	RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN | RCC_AHB1ENR_DMA2EN | RCC_AHB1ENR_GPIOAEN
			| RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_GPIODEN
			| RCC_AHB1ENR_GPIOEEN | RCC_AHB1ENR_GPIOFEN | RCC_AHB1ENR_GPIOGEN
			| RCC_AHB1ENR_CRCEN;
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;

	// Enable compensation cell to reduce noise on power lines with 50MHz/100MHz I/Os
	SYSCFG->CMPCR = SYSCFG_CMPCR_CMP_PD;

	// Enable usage fault, bus fault, and mem fault
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	SystemClockInitData clkInit;
	clkInit.pll.M = 12;
	clkInit.pll.N = 336;
	clkInit.pll.P = 2;
	clkInit.pll.Q = 7;
	clkInit.HSEBypass = 0;
	clkInit.APB1Div = 4;	// = 42MHz
	clkInit.APB2Div = 4;	// = 42MHz
	clkInit.RTCDiv = 12;	// = 1MHz
	clkInit.flashLatency = 5;
	clkInit.sysTickFreq = CH_FREQUENCY;
	SystemClockInit(&clkInit, 168000000UL);

	chSysInit();

	// set vector table offset
#ifdef ENV_BOOT
	SCB->VTOR = 0;
#else
	SCB->VTOR = BOOT_APP_OFFSET;
#endif

	CPULoadCalibrate(10);
	chThdSleep(14);

	// Init sync mechanism for timed distributed startup
	HWSync1Init();

	// Configure and start Independent watchdog
//	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
//	IWDG_SetPrescaler(IWDG_Prescaler_8);
//	IWDG_SetReload(2500);	// approx. 500ms
//	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);

	// -- HAL
	LEDInit();
	CRCInit();
	FlashFSInit(0x08060000, 0x20000);	// 128k
	ConfigInit(&ConfigOutputIntercom);
	ExtiInit();
	FSMCInit();
	SRAMInit();
	USART2Init();
	AudioInit();
	TouchInit();
	Intercom1Init();
	IntercomReliableInit();
	SysTimeInitCorrection(10);
	ProgStoreInit();
	ProgClientInit();
	ConfigIntercomInit();

#ifdef ENV_RUN
	CLIInit();
	ProgSourceUSBInit();
	ProgSourceWifiInit();
	LogFileInit();
	NetworkPrintInitIntercom();
	PresenterInit();
	OV7725Init();
	USBInit();
	USBHCDInit();
	USBMSCInit();
#endif

	chThdCreateStatic(waUSART2, sizeof(waUSART2), NORMALPRIO, USARTTask, &usart2);
	chThdCreateStatic(waIntercom, sizeof(waIntercom), NORMALPRIO-1, IntercomTask, 0);

#ifdef ENV_RUN
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-2, CLITask, 0);
	chThdCreateStatic(waLogFile, sizeof(waLogFile), NORMALPRIO-3, LogFileTask, 0);
	chThdCreateStatic(waPresenter, sizeof(waPresenter), NORMALPRIO-1, PresenterTask, 0);
	chThdCreateStatic(waTouch, sizeof(waTouch), NORMALPRIO, TouchTask, 0);
	chThdCreateStatic(waUSB, sizeof(waUSB), NORMALPRIO-5, USBTask, 0);
	chThdCreateStatic(waUSBHCD, sizeof(waUSBHCD), NORMALPRIO-6, USBHCDTask, 0);
	chThdCreateStatic(waConfig, sizeof(waConfig), NORMALPRIO-8, ConfigTask, 0);
#endif

	HWSyncBarrier(&hwSync1);

#ifdef ENV_BOOT
//	ConsolePrint("--- Bootloader (MEDIA) ---\r\n");
#else
	ConsolePrint("--- Media ---\r\n");
#endif

#ifndef ENV_RUN
	ProgStoreFlash();
#endif

	while(1)
	{
		chThdSleepMilliseconds(10);

#ifdef ENV_RUN
		uint32_t time = SysTimeUSec()%1000000;

		if(time > 600000 && time < 800000)
			LEDMustOn(LED_GREEN);
		else
			LEDMustOff(LED_GREEN);
#else
		uint32_t time = SysTimeUSec()%500000;

		if(time > 300000 && time < 400000)
			LEDMustOn(LED_RED);
		else
			LEDMustOff(LED_RED);
#endif
	}

	return 0;
}
