/*
 * presenter.c
 *
 *  Created on: 25.03.2014
 *      Author: rYan
 */

#include "presenter.h"

#include "GUI.h"
#include "PROGBAR.h"
#include "BUTTON.h"
#include "DIALOG.h"
#include "EDIT.h"
#include "MENU.h"
#include "FRAMEWIN.h"
#include "ICONVIEW.h"
#include "util/console.h"
#include "util/intercom.h"
#include "hal/touch.h"
#include "usb/usb_hcd.h"
#include "usb/usb_msc.h"
#include "log_msgs.h"
#include "prog_store.h"

#include "gui/numpad.h"
#include "gui/desktop.h"
#include "gui/select_botid.h"
#include "gui/ids.h"
#include "gui/features.h"
#include "gui/wifi_settings.h"
#include "gui/bg_com.h"
#include "gui/navbar.h"
#include "icons/icons.h"

#include <string.h>

#define EVENT_KD_UPDATE			0
#define EVENT_CFG_MISC_UPDATE	1
#define EVENT_CONFIG_UPDATE		2
#define EVENT_WIFI_STAT_UPDATE	3

#define ID_MENU_WIFI_BOTID 		GUI_ID_PRESENTER + 0
#define ID_MENU_WIFI_SETTINGS	GUI_ID_PRESENTER + 2
#define ID_MENU_FEATURES		GUI_ID_PRESENTER + 3

Presenter presenter;

static struct _PresenterHandler
{
	WM_HWIN hSelectBotID;
	WM_HWIN hFeatures;
	WM_HWIN hWifiSettings;

	PROGBAR_Handle hBatBar;

	WM_HWIN mainDesktop;
	WM_HWIN bgComDesktop;

	WM_HWIN hNav;
} presenterHandles;

static void presenterCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	switch(pHeader->command)
	{
		case IC_PRESENTER_CONFIG:
		{
			memcpy(&presenter.lastConfig, pData, sizeof(ConfigNetwork));

			chMBPost(&presenter.eventQueue, EVENT_CONFIG_UPDATE, TIME_IMMEDIATE);
		}
		break;
		case IC_PRESENTER_CFG_MISC:
		{
			memcpy(&presenter.lastCfgMisc, pData, sizeof(ConfigMisc));

			chMBPost(&presenter.eventQueue, EVENT_CFG_MISC_UPDATE, TIME_IMMEDIATE);
		}
		break;
		case IC_PRESENTER_WIFI_STAT:
		{
			memcpy(&presenter.lastWifiStat, pData, sizeof(PresenterWifiStat));

			chMBPost(&presenter.eventQueue, EVENT_WIFI_STAT_UPDATE, TIME_IMMEDIATE);
		}
		break;
	}
}

static void kdCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	if(pHeader->command == IC_KD_FEEDBACK)
	{
		KdFeedback* pFb = (KdFeedback*)pData;
		presenter.lastKdFeedback = *pFb;

		if(chTimeElapsedSince(presenter.lastKdTime) > MS2ST(100))
		{
			chMBPost(&presenter.eventQueue, EVENT_KD_UPDATE, TIME_IMMEDIATE);
			presenter.lastKdTime = chTimeNow();
		}
	}
}

static void swipeCallback(TouchSwipeEvent* pSwipe)
{
	if(pSwipe->move[0] > 0)
	{
		NavbarChangeTab(0);
	}
	else
	{
		NavbarChangeTab(1);
	}
}

static void usbHcdCallback(USBHCDConnectionEvent* pConEvent)
{
	BGComHCDUpdate(pConEvent);
}

static void usbMscCallback(USBMSCEvent* pMscEvent)
{
	BGComMSCUpdate(pMscEvent);

	if(pMscEvent->connect)
	{
		if(f_stat("autorun_bootloader", 0) == FR_OK)
			ProgStoreStart(PROG_STORE_SOURCE_USB);
		else
			LogFileOpen(0);
	}
	else
		LogFileClose();
}

static void logFileCallback(LogFileEvent* pLogEvent)
{
	BGComLogUpdate(pLogEvent);
}

void cbDesktop(WM_MESSAGE* pMsg)
{
	switch(pMsg->MsgId)
	{
		case WM_MENU:
		{
			MENU_MSG_DATA* pMenuData = (MENU_MSG_DATA*)pMsg->Data.p;
			if(pMenuData->MsgType == MENU_ON_ITEMSELECT)
			{
				switch(pMenuData->ItemId)
				{
					case ID_MENU_WIFI_BOTID:
					{
						WM_ShowWindow(presenterHandles.hSelectBotID);
						WM_SelectWindow(presenterHandles.hSelectBotID);
					}
					break;
					case ID_MENU_WIFI_SETTINGS:
					{
						WM_ShowWindow(presenterHandles.hWifiSettings);
						WM_SelectWindow(presenterHandles.hWifiSettings);

						NumpadSetVisible(1);
					}
					break;
					case ID_MENU_FEATURES:
					{
						WM_ShowWindow(presenterHandles.hFeatures);
						WM_SelectWindow(presenterHandles.hFeatures);
					}
					break;
				}
				break;
			}

			WM_DefaultProc(pMsg);
		}
		break;

		case WM_PAINT:
		{
			GUI_SetBkColor(GUI_BLACK);
			GUI_Clear();
		}
		break;

		case GUI_ID_DESKTOP_BOTID:
		{
			WM_ShowWindow(presenterHandles.hSelectBotID);
			WM_SelectWindow(presenterHandles.hSelectBotID);
		}
		break;

		default:
			WM_DefaultProc(pMsg);
	}
}

void PresenterInit()
{
	chMBInit(&presenter.eventQueue, presenter.eventQueueData, PRESENTER_EVENT_QUEUE_SIZE);

	GUI_Init();

	FRAMEWIN_SetDefaultTitleHeight(30);
	FRAMEWIN_SetDefaultFont(&GUI_Font16B_ASCII);

	WM_SetCallback(WM_HBKWIN, &cbDesktop);

	NumpadInit();
	NumpadSetVisible(0);

	presenterHandles.mainDesktop = DesktopInit();
	presenterHandles.bgComDesktop = BGComInit();

	NavbarInit();
	NavbarAddTab(&bmHome, presenterHandles.mainDesktop);
	NavbarAddTab(&bmLog, presenterHandles.bgComDesktop);

	presenterHandles.hSelectBotID = SelectBotIDInit();
	presenterHandles.hFeatures = FeaturesInit();
	presenterHandles.hWifiSettings = WifiSettingsInit();

	MENU_SetDefaultFont(&GUI_Font16B_ASCII);
	MENU_SetDefaultBorderSize(MENU_BI_LEFT, 5);
	MENU_SetDefaultBorderSize(MENU_BI_RIGHT, 5);
	MENU_SetDefaultBorderSize(MENU_BI_TOP, 8);
	MENU_SetDefaultBorderSize(MENU_BI_BOTTOM, 8);

	MENU_Handle hWifiMenu = MENU_CreateEx(0, 0, 0, 0, 0, 0, MENU_CF_VERTICAL, 0);

	static MENU_ITEM_DATA wifiBotId;
	wifiBotId.pText = "Bot ID";
	wifiBotId.Id = ID_MENU_WIFI_BOTID;
	MENU_AddItem(hWifiMenu, &wifiBotId);

	static MENU_ITEM_DATA wifiSettings;
	wifiSettings.pText = "Settings";
	wifiSettings.Id = ID_MENU_WIFI_SETTINGS;
	MENU_AddItem(hWifiMenu, &wifiSettings);

	MENU_Handle hMainMenu = MENU_CreateEx(0, 0, 240, 30, 0, WM_CF_SHOW, MENU_CF_HORIZONTAL,  0);

	static MENU_ITEM_DATA wifiMenu;
	wifiMenu.pText = " Wifi ";
	wifiMenu.Id = 0;
	wifiMenu.Flags = 0;
	wifiMenu.hSubmenu = hWifiMenu;
	MENU_AddItem(hMainMenu, &wifiMenu);

	static MENU_ITEM_DATA featuresMenu;
	featuresMenu.pText = "Features";
	featuresMenu.Id = ID_MENU_FEATURES;
	featuresMenu.Flags = 0;
	featuresMenu.hSubmenu = 0;
	MENU_AddItem(hMainMenu, &featuresMenu);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kdCallback;
	client.section = INTERCOM_SECTION_KD;
	IntercomAddClient(client);

	client.rxCallback = &presenterCallback;
	client.section = INTERCOM_SECTION_PRESENTER;
	IntercomAddClient(client);

	touch.swipeCb = &swipeCallback;
	usbHcd.pConCb = &usbHcdCallback;
	usbMsc.eventCb = &usbMscCallback;
	logFile.eventCb = &logFileCallback;
}

msg_t PresenterTask(void* params)
{
	(void)params;

	int32_t event;
	int32_t result;

	chRegSetThreadName("Presenter");

	while(1)
	{
		result = chMBFetch(&presenter.eventQueue, &event, MS2ST(20));

		GUI_Exec();
		GUI_X_ExecIdle();

		if(result != RDY_OK)
			continue;

		GUI_SetFont(&GUI_Font8x12_ASCII);

		switch(event)
		{
			case EVENT_KD_UPDATE:
			{
				DesktopKdUpdate(&presenter.lastKdFeedback);
			}
			break;
			case EVENT_CONFIG_UPDATE:
			{
				DesktopWifiUpdate(&presenter.lastConfig);
				WifiUpdateSettings(&presenter.lastConfig);
			}
			break;
			case EVENT_CFG_MISC_UPDATE:
			{
				FeaturesSet(presenter.lastCfgMisc.features);
				DesktopHwIdUpdate(presenter.lastCfgMisc.hardwareId);
			}
			break;
			case EVENT_WIFI_STAT_UPDATE:
			{
				// TODO: display interesting data
			}
			break;
			default: break;
		}
	}

	return 0;
}
