/*
 * prog_store.c
 *
 *  Created on: 01.04.2015
 *      Author: AndreR
 */

#include "prog_store.h"
#include "intercom_constants.h"
#include "util/intercom_reliable.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/boot.h"
#include "ch.h"

#include <string.h>

#define PROG_SIZE_MAIN (256*1024)
#define PROG_SIZE_MEDIA (256*1024)
#define PROG_SIZE_KD (92*1024)
#define PROG_SIZE_MOT (92*1024)

#define PROG_STORE_EVENT_TYPE_ERROR	0
#define PROG_STORE_EVENT_TYPE_DATA	1
#define PROG_STORE_EVENT_TYPE_SIZE	2
#define PROG_STORE_EVENT_TYPE_CRC32	3

static WORKING_AREA(waProgStore, 8192);

static const uint32_t progSizes[PROG_STORE_NUM_PROCESSORS] = {PROG_SIZE_MAIN, PROG_SIZE_MEDIA, PROG_SIZE_MOT, PROG_SIZE_MOT, PROG_SIZE_KD};

static struct _progs
{
	uint8_t main[PROG_SIZE_MAIN];
	uint8_t media[PROG_SIZE_MEDIA];
	uint8_t kd[PROG_SIZE_KD];
	uint8_t left[PROG_SIZE_MOT];
	uint8_t right[PROG_SIZE_MOT];

	struct _info
	{
		uint8_t update;
		uint32_t crc;
		uint32_t size;
	} info[PROG_STORE_NUM_PROCESSORS];
} progs __attribute__ ((section (".flashram")));

ProgStore progStore = {
	.pProgs = {progs.main, progs.media, progs.left, progs.right, progs.kd}
};

static void bootloaderCb(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)dataLength;
	(void)pUserData;

	switch(pHeader->command)
	{
		case BOOTLOADER_NACK:
		{
			BootloaderSeq* pSeq = (BootloaderSeq*)pData;

			if(pSeq->seq == progStore.expectedSeq)
			{
				progStore.progAckd = 0;
				chBSemSignal(&progStore.progSem);
			}
		}
		break;
		case BOOTLOADER_ACK:
		{
			BootloaderSeq* pSeq = (BootloaderSeq*)pData;

			if(pSeq->seq == progStore.expectedSeq)
			{
				progStore.progAckd = 1;
				chBSemSignal(&progStore.progSem);
			}
		}
		break;
	}
}

static uint8_t progClientErase(uint8_t procId)
{
	progStore.expectedSeq = progStore.progSeq;
	BootloaderSeq seq = {progStore.progSeq++};

	IntercomHeader header;
	header.address = (1 << procId);
	header.command = BOOTLOADER_ERASE;
	header.section = INTERCOM_SECTION_BOOTLOADER;

	progStore.progAckd = 0;
	chBSemReset(&progStore.progSem, 1);

	do
	{
		IntercomSend(&header, sizeof(BootloaderSeq), &seq);
	}
	while(chBSemWaitTimeout(&progStore.progSem, MS2ST(1000)) != RDY_OK);

	return progStore.progAckd;
}

static uint8_t progClientCrc(uint8_t procId, uint32_t start, uint32_t end, uint32_t expectedCRC)
{
	progStore.expectedSeq = progStore.progSeq;
	BootloaderChecksum crc;
	crc.seq = progStore.progSeq++;
	crc.start = start;
	crc.end = end;
	crc.crc = expectedCRC;

	IntercomHeader header;
	header.address = (1 << procId);
	header.command = BOOTLOADER_CHECKSUM;
	header.section = INTERCOM_SECTION_BOOTLOADER;

	progStore.progAckd = 0;
	chBSemReset(&progStore.progSem, 1);

	do
	{
		IntercomSend(&header, sizeof(BootloaderChecksum), &crc);
	}
	while(chBSemWaitTimeout(&progStore.progSem, MS2ST(100)) != RDY_OK);

	return progStore.progAckd;
}

#define MAX_PROG_DATA_SIZE 256

static uint8_t progClientData(uint8_t procId, uint32_t offset, uint8_t* pData, uint32_t dataLength)
{
	static uint8_t progData[MAX_PROG_DATA_SIZE+sizeof(BootloaderProgramData)];

	progStore.expectedSeq = progStore.progSeq;

	BootloaderProgramData* pProgData = (BootloaderProgramData*)progData;
	pProgData->seq = progStore.progSeq++;
	pProgData->offset = offset;

	memcpy(progData+sizeof(BootloaderProgramData), pData, dataLength);

	IntercomHeader header;
	header.address = (1 << procId);
	header.command = BOOTLOADER_DATA;
	header.section = INTERCOM_SECTION_BOOTLOADER;

	progStore.progAckd = 0;
	chBSemReset(&progStore.progSem, 1);

	do
	{
		IntercomSend(&header, dataLength+sizeof(BootloaderProgramData), progData);
	}
	while(chBSemWaitTimeout(&progStore.progSem, MS2ST(200)) != RDY_OK);

	return progStore.progAckd;
}

static void progClientSelectBootloader()
{
	IntercomHeader iHeader;
	iHeader.section = INTERCOM_SECTION_BOOTLOADER;
	iHeader.address = INTERCOM_ADDRESS_MEDIA | INTERCOM_ADDRESS_KD | INTERCOM_ADDRESS_LEFT | INTERCOM_ADDRESS_RIGHT | INTERCOM_ADDRESS_MAIN;
	iHeader.command = BOOTLOADER_SELECT;

	IntercomReliableSend(&iHeader, 0, 0);
}

void ProgStoreInit()
{
	chBSemInit(&progStore.progSem, 0);
	chMBInit(&progStore.eventQueue, progStore.eventQueueData, PROG_STORE_NUM_EVENTS);

	progStore.expectedSeq = 1;

	for(uint16_t i = 0; i < PROG_STORE_NUM_EVENTS; i++)
	{
		chBSemInit(&progStore.events[i].lockSem, 0);
	}

	IntercomClient iClient;
	iClient.pUserData = 0;
	iClient.section = INTERCOM_SECTION_BOOTLOADER;
	iClient.rxCallback = &bootloaderCb;

	IntercomAddClient(iClient);
}

static ProgStoreEvent* getProgStoreEvent()
{
	ProgStoreEvent* result = 0;

	for(uint16_t i = 0; i < PROG_STORE_NUM_EVENTS; i++)
	{
		result = &progStore.events[i];

		if(chBSemWaitTimeout(&result->lockSem, TIME_IMMEDIATE) == RDY_OK)
			return result;
	}

	return 0;
}

void ProgStoreProvideError(uint8_t procId, int16_t error)
{
	ConsolePrint("Error: %hd\r\n", error);

	// e.g. media detached, link lost
	if(procId >= PROG_STORE_NUM_PROCESSORS)
		return;

	ProgStoreEvent* pEvent = getProgStoreEvent();
	if(!pEvent)
		return;

	pEvent->type = PROG_STORE_EVENT_TYPE_ERROR;
	pEvent->procId = procId;
	pEvent->errorId = error;

	chMBPostAhead(&progStore.eventQueue, (msg_t)pEvent, TIME_IMMEDIATE);
}

void ProgStoreProvideData(uint8_t procId, uint32_t offset, uint8_t* pData, uint32_t size)
{
	if(procId >= PROG_STORE_NUM_PROCESSORS)
		return;

	ProgStoreEvent* pEvent = getProgStoreEvent();
	if(!pEvent)
		return;

	if(offset+size > progSizes[procId])
		return;

	memcpy(progStore.pProgs[procId]+offset, pData, size);

	pEvent->type = PROG_STORE_EVENT_TYPE_DATA;
	pEvent->procId = procId;
	pEvent->data.offset = offset;
	pEvent->data.size = size;

	chMBPostAhead(&progStore.eventQueue, (msg_t)pEvent, TIME_IMMEDIATE);
}

void ProgStoreProvideSize(uint8_t procId, uint32_t size)
{
	if(procId >= PROG_STORE_NUM_PROCESSORS)
		return;

	ProgStoreEvent* pEvent = getProgStoreEvent();
	if(!pEvent)
		return;

	pEvent->type = PROG_STORE_EVENT_TYPE_SIZE;
	pEvent->procId = procId;
	pEvent->size = size;

	chMBPostAhead(&progStore.eventQueue, (msg_t)pEvent, TIME_IMMEDIATE);
}

void ProgStoreProvideCRC32(uint8_t procId, uint32_t startAddr, uint32_t endAddr, uint32_t crc32)
{
	if(procId >= PROG_STORE_NUM_PROCESSORS)
		return;

	ProgStoreEvent* pEvent = getProgStoreEvent();
	if(!pEvent)
		return;

	pEvent->type = PROG_STORE_EVENT_TYPE_CRC32;
	pEvent->procId = procId;
	pEvent->crc32.startAddr = startAddr;
	pEvent->crc32.endAddr = endAddr;
	pEvent->crc32.crc = crc32;

	chMBPostAhead(&progStore.eventQueue, (msg_t)pEvent, TIME_IMMEDIATE);
}

static void timeoutFunc(void* param)
{
	(void)param;

	chSysLockFromIsr();

	chMBPostI(&progStore.eventQueue, 0);

	chSysUnlockFromIsr();
}

static int16_t waitForEvent(ProgStoreEvent* pEventIn)
{
	msg_t event;
	int16_t result = 0;
	ProgStoreEvent* pCurEvent;

	while(1)
	{
		if(chMBFetch(&progStore.eventQueue, &event, TIME_INFINITE) != RDY_OK)
			return PROG_STORE_ERROR_TIMEOUT;

		if(event == 0)
			return PROG_STORE_ERROR_TIMEOUT;

		pCurEvent = (ProgStoreEvent*)event;

		if(pCurEvent->type == PROG_STORE_EVENT_TYPE_ERROR)
		{
			result = pCurEvent->errorId;
			chBSemSignal(&pCurEvent->lockSem);
			break;
		}

		if(pCurEvent->type == pEventIn->type && pCurEvent->procId == pEventIn->procId)
		{
			if(pCurEvent->type == PROG_STORE_EVENT_TYPE_SIZE)
			{
				pEventIn->size = pCurEvent->size;
				chBSemSignal(&pCurEvent->lockSem);
				break;
			}

			if(	pCurEvent->type == PROG_STORE_EVENT_TYPE_DATA &&
				pEventIn->data.offset == pCurEvent->data.offset)
			{
				pEventIn->data.size = pCurEvent->data.size;
				chBSemSignal(&pCurEvent->lockSem);
				break;
			}

			if(	pCurEvent->type == PROG_STORE_EVENT_TYPE_CRC32 &&
				pEventIn->crc32.startAddr == pCurEvent->crc32.startAddr &&
				pEventIn->crc32.endAddr == pCurEvent->crc32.endAddr)
			{
				pEventIn->crc32.crc = pCurEvent->crc32.crc;
				chBSemSignal(&pCurEvent->lockSem);
				break;
			}
		}

		chBSemSignal(&pCurEvent->lockSem);
	}

	return result;
}

static int16_t waitForData(ProgSource* pSource, uint8_t procId, uint32_t offset, uint32_t size)
{
	int16_t result;
	ProgStoreEvent ev;

	ev.procId = procId;
	ev.type = PROG_STORE_EVENT_TYPE_DATA;
	ev.data.offset = offset;

	while(1)
	{
		(*pSource->pDataFunc)(procId, offset, size);

		chVTSet(&progStore.timeoutTimer, MS2ST(pSource->timeout), &timeoutFunc, 0);

		result = waitForEvent(&ev);
		if(result == PROG_STORE_ERROR_TIMEOUT)
			continue;

		if(ev.data.size != size)
			continue;

		chVTReset(&progStore.timeoutTimer);

		chMBReset(&progStore.eventQueue);

		if(result)
			return result;

		break;
	}

	return 0;
}

static int16_t waitForSize(ProgSource* pSource, int8_t procId, uint32_t* pSize)
{
	int16_t result;
	ProgStoreEvent ev;

	ev.procId = procId;
	ev.type = PROG_STORE_EVENT_TYPE_SIZE;

	while(1)
	{
		(*pSource->pSizeFunc)(procId);

		chVTSet(&progStore.timeoutTimer, MS2ST(pSource->timeout), &timeoutFunc, 0);

		result = waitForEvent(&ev);
		if(result == PROG_STORE_ERROR_TIMEOUT)
			continue;

		chVTReset(&progStore.timeoutTimer);

		chMBReset(&progStore.eventQueue);

		if(result)
			return result;

		break;
	}

	if(pSize)
		*pSize = ev.size;

	return 0;
}

static int16_t waitForCrc(ProgSource* pSource, int8_t procId, uint32_t startAddr, uint32_t endAddr, uint32_t* pCRC)
{
	int16_t result;
	ProgStoreEvent ev;

	ev.procId = procId;
	ev.type = PROG_STORE_EVENT_TYPE_CRC32;
	ev.crc32.startAddr = startAddr;
	ev.crc32.endAddr = endAddr;

	while(1)
	{
		(*pSource->pCRC32Func)(procId, startAddr, endAddr);

		chVTSet(&progStore.timeoutTimer, MS2ST(pSource->timeout), &timeoutFunc, 0);

		result = waitForEvent(&ev);
		if(result == PROG_STORE_ERROR_TIMEOUT)
			continue;

		chVTReset(&progStore.timeoutTimer);

		chMBReset(&progStore.eventQueue);

		if(result)
			return result;

		break;
	}

	if(pCRC)
		*pCRC = ev.crc32.crc;

	return 0;
}

msg_t ProgFlashTask(void* params)
{
	(void)params;

	for(uint16_t proc = 0; proc < PROG_STORE_NUM_PROCESSORS; proc++)
	{
		if(progs.info[proc].update == 0)
			continue;

		progClientErase(proc);

		for(uint32_t offset = 0; offset < progs.info[proc].size; offset += 128)
		{
			progClientData(proc, offset, progStore.pProgs[proc]+offset, 128);

			uint32_t crc = CRC32CalcChecksumAlt(progStore.pProgs[proc]+offset, 128);

			if(progClientCrc(proc, offset, offset+128, crc) == 0)
			{
				offset -= 128;
			}
		}

		uint32_t crc = CRC32CalcChecksumAlt(progStore.pProgs[proc], progSizes[proc]);

		if(progClientCrc(proc, 0, progSizes[proc], crc) == 0)
		{
			proc--;
		}
	}

	BootReset();

	return 0;
}

msg_t ProgStoreTask(void* params)
{
	(void)params;
	int16_t result;
	uint8_t sourceAvail;

	chRegSetThreadName("ProgStore");

	// 1. get program size for main to check source availability
	(*progStore.sources[progStore.useSource].pOpenFunc)();

	result = waitForSize(&progStore.sources[progStore.useSource], 0, 0);
	if(result)
		sourceAvail = 0;
	else
		sourceAvail = 1;

	(*progStore.sources[progStore.useSource].pCloseFunc)();

	// 2. select prog source
	if(!sourceAvail)
	{
		ConsolePrint("No program source available\r\n");
		return 1;
	}

	// 3. get CRCs and sizes from selected prog source
	ProgSource* pSource = &progStore.sources[progStore.useSource];

	(*pSource->pOpenFunc)();

	for(uint16_t proc = 0; proc < PROG_STORE_NUM_PROCESSORS; proc++)
	{
		ConsolePrint("Calculating proc %hu CRC\r\n", proc);

		memset(progStore.pProgs[proc], 0xFF, progSizes[proc]);

		result = waitForCrc(pSource, proc, 0, progSizes[proc], &progs.info[proc].crc);
		if(result)
		{
			ConsolePrint("waitForCrc: %hd\r\n", result);
			return result;
		}

		ConsolePrint("Proc %hu CRC: %08X\r\n", proc, progs.info[proc].crc);

		progs.info[proc].update = 1;

		// skip processors with that flash content
		if(progClientCrc(proc, 0, progSizes[proc], progs.info[proc].crc))
		{
			ConsolePrint("CRC matched\r\n");

			progs.info[proc].update = 0;
			continue;
		}

		result = waitForSize(pSource, proc, &progs.info[proc].size);
		if(result)
			return result;

		ConsolePrint("Proc %hu size: %u\r\n", proc, progs.info[proc].size);
	}


	// 4. load prog source binaries to ext. RAM
	for(uint16_t proc = 0; proc < PROG_STORE_NUM_PROCESSORS; proc++)
	{
		if(progs.info[proc].update == 0)
			continue;

		ConsolePrint("Load %hu\r\n", proc);

		for(uint32_t offset = 0; offset < progs.info[proc].size; offset += pSource->maxTransferSize)
		{
//			ConsolePrint("Req: %u\r\n", offset);

			if(offset > 0 && offset%(10*pSource->maxTransferSize) == 0)
			{
				uint32_t tmpCrc;
				uint32_t startAddr = offset-10*pSource->maxTransferSize;
				result = waitForCrc(pSource, proc, startAddr, offset, &tmpCrc);

				if(CRC32VerifyChecksum(progStore.pProgs[proc]+startAddr, offset-startAddr, tmpCrc) != CRC32_MAGIC_NUMBER)
				{
					ConsolePrint("Intermediate CRC check failure at %u of %hu\r\n", offset, proc);
					offset -= 11*pSource->maxTransferSize;
					continue;
				}
			}

			result = waitForData(pSource, proc, offset, pSource->maxTransferSize);
			if(result)
				return result;
		}

		// verify checksum
		if(CRC32VerifyChecksum(progStore.pProgs[proc], progSizes[proc], progs.info[proc].crc) != CRC32_MAGIC_NUMBER)
		{
			ConsolePrint("Final CRC check failure for proc %hu\r\n", proc);
			--proc;
		}

		ConsolePrint("Proc %hu loaded\r\n", proc);
	}

	(*pSource->pCloseFunc)();

	uint16_t proc;
	for(proc = 0; proc < PROG_STORE_NUM_PROCESSORS; proc++)
	{
		if(progs.info[proc].update)
			break;
	}

	if(proc == PROG_STORE_NUM_PROCESSORS)
	{
		ConsolePrint("up to date\r\n");
			return 0;
	}

	// 6. reboot & flash
	progClientSelectBootloader();
	BootReset();

	return 0;
}

void ProgStoreFlash()
{
	progStore.pTask = chThdCreateStatic(waProgStore, sizeof(waProgStore), NORMALPRIO-20, &ProgFlashTask, 0);
}

void ProgStoreStart(uint8_t source)
{
	progStore.useSource = source;
	progStore.pTask = chThdCreateStatic(waProgStore, sizeof(waProgStore), NORMALPRIO-20, &ProgStoreTask, 0);
}

int16_t ProgStoreAddSource(const ProgSource* pSource)
{
	if(progStore.usedSources >= PROG_STORE_MAX_SOURCES)
		return -1;

	int16_t id = progStore.usedSources;

	memcpy(&progStore.sources[progStore.usedSources++], pSource, sizeof(ProgSource));

	return id;
}

// TODO:
// ProgSourceWireless
// ProgSourceUSB
