/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "arm_math.h"
#include "hal/led.h"
#include "util/console.h"
#include "util/network_print.h"
#include "util/sys_time.h"
#include "util/flash.h"
#include "util/flash_fs.h"
#include "util/log.h"
#include "util/arm_mat_util_f32.h"
#include "util/intercom.h"
#include "intercom_constants.h"
#include "spi1.h"
#include "drive.h"
#include "cpu_load.h"
#include "motors.h"

CLI cli;

void processCommand();

void CLIInit()
{
	chMtxInit(&cli.procBufMutex);

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

#ifdef MOT_LEFT
	ConsoleIntercomAdapterInit(&cli.adapter, INTERCOM_ADDRESS_LEFT, INTERCOM_ADDRESS_MAIN);
#else
	ConsoleIntercomAdapterInit(&cli.adapter, INTERCOM_ADDRESS_RIGHT, INTERCOM_ADDRESS_MAIN);
#endif
}

msg_t CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

#ifdef ENV_BOOT
#ifdef MOT_LEFT
	ConsolePrint("--- Bootloader (Left) ---\r\n");
#else
	ConsolePrint("--- Bootloader (Right) ---\r\n");
#endif
#else
#ifdef MOT_LEFT
	ConsolePrint("--- Left ---\r\n");
#else
	ConsolePrint("--- Right ---\r\n");
#endif
#endif

	while(1)
	{
		chThdSleepMilliseconds(20);

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

	}

	return 0;
}

void processCommand()
{
	uint16_t u16;

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("test2"))
	{
		ConsolePrint("0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\r\n");
	}

	if(ConsoleCmpCmd("cpu"))
	{
//		DWT->CYCCNT = 0;

		CPULoad load = cpuLoad;

		ConsolePrint("CPU Load: %3u.%1u%%\r\n", load.load/10, load.load%10);
		ConsolePrint("Max: %3u.%1u%%\r\n", load.maxLoad/10, load.maxLoad%10);
		ConsolePrint("incPerTick: %10u/%u\r\n", load.incPerSysTick, load.maxIncPerSysTick);

//		uint32_t cycs = DWT->CYCCNT;
//
//		ConsolePrint("cyc: %u\r\n", cycs);
		chThdSleep(MS2ST(10));
		cpuLoad.maxLoad = 0;
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("Name            rPrio\tTime\r\n");

		for(Thread* pThread = chRegFirstThread(); pThread != 0; pThread = chRegNextThread(pThread))
		{
			ConsolePrint("%-16s%u\t%u\r\n", pThread->p_name, pThread->p_realprio, pThread->p_time);
		}
	}

	if(ConsoleCmpCmd("timsr"))
	{
		ConsolePrint("Resetting SR register\r\n");

		TIM1->SR = 0;
		TIM8->SR = 0;
	}

	if(ConsoleCmpCmd("tim"))
	{
		ConsolePrint("TIM1:\r\n");
		ConsolePrint("CR1: 0x%hX, SR: 0x%X, BDTR: 0x%X\r\n", TIM1->CR1, TIM1->SR, TIM1->BDTR);
		ConsolePrint("CCR: %hu, %hu, %hu\r\n", TIM1->CCR1, TIM1->CCR2, TIM1->CCR3);
		ConsolePrint("TIM8:\r\n");
		ConsolePrint("CR1: 0x%hX, SR: 0x%X, BDTR: 0x%X\r\n", TIM8->CR1, TIM8->SR, TIM8->BDTR);
		ConsolePrint("CCR: %hu, %hu, %hu\r\n", TIM8->CCR1, TIM8->CCR2, TIM8->CCR3);
		ConsolePrint("outLevel: %d, %d\r\n", pFront->outLevel, pRear->outLevel);
	}

	if(ConsoleCmpCmd("stubs"))
	{
		EXTI->SWIER = (1 << 3);
	}

	uint16_t mot;
	uint16_t vref, angle;
	if(ConsoleScanCmd("turn %hu %hu", &mot, &vref) == 2)
	{
		Drive* pDrive = pFront;
		if(mot)
			pDrive = pRear;

		motors.ignoreMainCmds = 1;

		chThdSleepMilliseconds(5);

		DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);
		DriveSetCommutationMode(pDrive, DRIVE_COM_MODE_NONE);

		float power = vref/1000.0f;

		DriveSetVoltageAngle(pDrive, power, 0);

		chThdSleep(MS2ST(50));

		ConsolePrint("enc: %hu\r\n", pDrive->pTIMEncoder->CNT);

		for(float a = 0; a <= 2880; a += 5)
		{
			DriveSetVoltageAngle(pDrive, power, a);

//			angle = ((float)a)*PI/180.0;
//
//			DriveSetSpaceVector(&rear, power, angle);

			chThdSleep(MS2ST(20));
		}

		ConsolePrint("enc: %hu\r\n", pDrive->pTIMEncoder->CNT);

		DriveSetVoltageAngle(pDrive, 0, 0);
	}

	if(ConsoleScanCmd("va %hu %hu", &vref, &angle) == 2)
	{
//		float a = ((float)angle)*PI/180.0;
		float r = ((float)vref)/1000.0f;

		DriveSetVoltageAngle(pRear, r, angle);
	}

	if(ConsoleScanCmd("mtest %hu %hu", &mot, &vref) == 2)
	{
		ConsolePrint("Motor hall test\r\n");

		float hallAngles[7][2];

		Drive* pDrive = pFront;
		if(mot)
			pDrive = pRear;

		motors.ignoreMainCmds = 1;

		chThdSleepMilliseconds(5);

		DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);
		DriveSetCommutationMode(pDrive, DRIVE_COM_MODE_NONE);

		uint16_t hall = pDrive->halPos;
		float v = ((float)vref)/1000.0f;

		for(float a = 0; a < 360; a += 1.0f)
		{
			DriveSetVoltageAngle(pDrive, v, a);
			chThdSleep(MS2ST(20));

			if(hall != pDrive->halPos)
			{
				hall = pDrive->halPos;
				ConsolePrint(".");

				hallAngles[hall][0] = a;
			}
		}

		for(float a = 360; a > 0; a -= 1.0f)
		{
			DriveSetVoltageAngle(pDrive, v, a);
			chThdSleep(MS2ST(20));

			if(hall != pDrive->halPos)
			{
				hall = pDrive->halPos;
				ConsolePrint(".");

				hallAngles[hall][1] = a;
			}
		}

		DriveSetVoltage(pDrive, 0);

		const uint16_t index2Angle[] = {0, 300, 60, 0, 180, 240, 120};

		ConsolePrint("\r\nResults:\r\n");

		for(uint16_t i = 1; i < 7; i++)
		{
			ConsolePrint("%3hu - %3hu - %3hu\r\n", (uint16_t)hallAngles[i][0], index2Angle[i], (uint16_t)hallAngles[i][1]);
		}

		motors.ignoreMainCmds = 0;
	}

	if(ConsoleCmpCmd("adc"))
	{
		for(uint8_t a = 0; a < 2; a++)
		{
			ConsolePrint("ADC%hu: ", (uint16_t)a);
			for(uint8_t i = 0; i < DRIVE_ADC_SAMPLES; i++)
			{
				ConsolePrint("% 5hd ", pRear->adc[a][i]);
			}
			ConsolePrint("\r\n");
		}
	}

//	if(ConsoleCmpCmd("current"))
//	{
//		ConsolePrint("iA: %d, iB: %d, iC: %d\r\n", (int32_t)(pRear->i[0]*1000), (int32_t)(pRear->i[1]*1000), (int32_t)(pRear->i[2]*1000));
//		ConsolePrint("iAlpha: %d, iBeta: %d\r\n", (int32_t)(pRear->iAlpha*1000), (int32_t)(pRear->iBeta*1000));
//		ConsolePrint("iD: %d, iQ: %d\r\n", (int32_t)(rear.pOut->iD*1000), (int32_t)(rear.pOut->iQ*1000));
//	}

	if(ConsoleCmpCmd("enc"))
	{
		float angle = (pRear->pTIMEncoder->CNT%256)*1.40625f;
		ConsolePrint("ENC pos: %u\r\n", (uint32_t)(angle*1000));
	}

	if(ConsoleCmpCmd("cyc"))
	{
//		ConsolePrint("updateADC: % 8u\r\n", pFront->updateADCCycles);
	}

	if(ConsoleCmpCmd("speed"))
	{
		ConsolePrint("Front vel: %d\r\n", (int32_t)(pFront->curVelocity*1000));
	}

	if(ConsoleCmpCmd("kf"))
	{
		ConsolePrint("Cur - KF: % 8dmA, Meas: % 8dmA\r\n", (int32_t)(pFront->estCurrent*1000), (int32_t)(pFront->rmsCurrent*1000));
		ConsolePrint("Speed-KF: % 8d, Meas: % 8d\r\n", (int32_t)(pFront->estVelocity*1000), (int32_t)(pFront->curVelocity*1000));
		ConsolePrint("M: % 8d, ML: % 8d\r\n", (int32_t)(pFront->estTorque*1000000), (int32_t)(pFront->estLoadTorque*1000000));
		ConsolePrint("Cur - KF: % 8dmA, Meas: % 8dmA\r\n", (int32_t)(pRear->estCurrent*1000), (int32_t)(pRear->rmsCurrent*1000));
		ConsolePrint("Speed-KF: % 8d, Meas: % 8d\r\n", (int32_t)(pRear->estVelocity*1000), (int32_t)(pRear->curVelocity*1000));
		ConsolePrint("M: % 8d, ML: % 8d\r\n", (int32_t)(pRear->estTorque*1000000), (int32_t)(pRear->estLoadTorque*1000000));
	}

/*	if(ConsoleCmpCmd("input"))
	{
		DriveMain2Mot* pFront = &spi1.pRxData->periodic.mot[0];
		DriveMain2Mot* pRear = &spi1.pRxData->periodic.mot[1];

		ConsolePrint("Front:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(pFront->velocity*1000), (int32_t)(pFront->iD*1000), (int32_t)(pFront->iQ*1000));
		ConsolePrint("Rear:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(pRear->velocity*1000), (int32_t)(pRear->iD*1000), (int32_t)(pRear->iQ*1000));
	}

	if(ConsoleCmpCmd("output"))
	{
		ConsolePrint("Front:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(front.pOut->velocity*1000), (int32_t)(front.pOut->iD*1000), (int32_t)(front.pOut->iQ*1000));
		ConsolePrint("Rear:\r\n");
		ConsolePrint("Vel: %d\r\niD:  %d\r\niQ:  %d\r\n",
				(int32_t)(rear.pOut->velocity*1000), (int32_t)(rear.pOut->iD*1000), (int32_t)(rear.pOut->iQ*1000));
	}
*/

	uint32_t p, i, d;
	if(ConsoleScanCmd("speedpid %u %u %u", &p, &i, &d) == 3)
	{
		ConsolePrint("Setting Speed PID: %u, %u, %u\r\n", p, i, d);

		DriveSpeedSetPIDParams(&drive[0], p*0.001f, i*0.001f, d*0.001f);
		DriveSpeedSetPIDParams(&drive[1], p*0.001f, i*0.001f, d*0.001f);
	}

	if(ConsoleScanCmd("torquepid %u %u %u", &p, &i, &d) == 3)
	{
		ConsolePrint("Setting Torque PID: %u, %u, %u\r\n", p, i, d);

		DriveTorqueSetPIDParams(&drive[0], p*0.001f, i*0.001f, d*0.001f);
		DriveTorqueSetPIDParams(&drive[1], p*0.001f, i*0.001f, d*0.001f);
	}

	if(ConsoleCmpCmd("bat"))
	{
		ConsolePrint("Front: %hu\tRear: %hu\r\n", (uint16_t)(pFront->curBat*1000), (uint16_t)(pRear->curBat*1000));
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	if(ConsoleCmpCmd("hall"))
	{
		ConsolePrint("Front: %hu\tRear: %hu\r\n", pFront->halPos, pRear->halPos);
	}

	if(ConsoleScanCmd("ignore %hu", &u16) == 1)
	{
		if(u16)
		{
			motors.ignoreMainCmds = 1;
			ConsolePrint("Ingoring commands from main\r\n");
		}
		else
		{
			motors.ignoreMainCmds = 0;
			ConsolePrint("Main cmds re-enabled\r\n");
		}
	}

	if(ConsoleCmpCmd("ctrl"))
	{
		ConsolePrint("Ctrl: %hu\r\nCom: %hu\r\n", (uint16_t)pFront->ctrlMode, (uint16_t)pFront->commutationMode);
	}

 	if(ConsoleCmpCmd("ctrl none"))
	{
 		DriveSetCtrlMode(&drive[0], DRIVE_CTRL_MODE_NONE);
 		DriveSetCtrlMode(&drive[1], DRIVE_CTRL_MODE_NONE);

		ConsolePrint("Controller disabled\r\n");
	}

	if(ConsoleCmpCmd("ctrl speed"))
	{
 		DriveSetCtrlMode(&drive[0], DRIVE_CTRL_MODE_PID_SPEED);
 		DriveSetCtrlMode(&drive[1], DRIVE_CTRL_MODE_PID_SPEED);


		ConsolePrint("Speed PID controller active\r\n");
	}

	if(ConsoleCmpCmd("ctrl torque"))
	{
 		DriveSetCtrlMode(&drive[0], DRIVE_CTRL_MODE_PID_TORQUE);
 		DriveSetCtrlMode(&drive[1], DRIVE_CTRL_MODE_PID_TORQUE);


		ConsolePrint("Torque PID controller active\r\n");
	}

	if(ConsoleCmpCmd("com none"))
	{
		DriveSetCommutationMode(&drive[0], DRIVE_COM_MODE_NONE);
		DriveSetCommutationMode(&drive[1], DRIVE_COM_MODE_NONE);

		ConsolePrint("Commutation mode: NONE\r\n");
	}

	if(ConsoleCmpCmd("com block"))
	{
		DriveSetCommutationMode(&drive[0], DRIVE_COM_MODE_BLOCK);
		DriveSetCommutationMode(&drive[1], DRIVE_COM_MODE_BLOCK);

		ConsolePrint("Commutation mode: BLOCK\r\n");
	}

	if(ConsoleCmpCmd("com sinus"))
	{
		DriveSetCommutationMode(&drive[0], DRIVE_COM_MODE_SINUSOIDAL);
		DriveSetCommutationMode(&drive[1], DRIVE_COM_MODE_SINUSOIDAL);

		ConsolePrint("Commutation mode: SINUSOIDAL\r\n");
	}

	int16_t i16;
	if(ConsoleScanCmd("vel %hd", &i16) == 1)
	{
		ConsolePrint("Set velocity to %hdRPS\r\n", i16);

		DriveSpeedSetVelocity(&drive[0], (float)i16);
	}

	if(ConsoleScanCmd("vol %hd", &i16) == 1)
	{
		ConsolePrint("Set voltage to %hdmV\r\n", i16);

		float volt = i16*0.001f;

		DriveSetVoltage(&drive[0], volt);
	}

	if(ConsoleScanCmd("current %hd", &i16) == 1)
	{
		ConsolePrint("Set current to %hdmA\r\n", i16);

		float current = i16*0.001f;

		pFront->torque.setCurrent = current;
	}

//	if(ConsoleCmpCmd("input"))
//	{
//		ConsolePrint("speedOn: %hu\r\n", (uint16_t)step[0].speedOn);
//		ConsolePrint("Vel: %hd, Vol: %hd\r\n", (int16_t)(step[0].setVelocity*1000), (int16_t)(step[0].setVoltage*1000));
//	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleScanCmd("log %hu", &u16) == 1)
	{
		LogPrintAllRecent(u16);
	}

	if(ConsoleCmpCmd("netprint"))
	{
#ifdef MOT_LEFT
		NetworkPrint("Hello Network - LEFT\n");
#else
		NetworkPrint("Hello Network - RIGHT\n");
#endif
	}
}
