/*
 * drive.c
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

/**	Timer/ADC/Motor assignments
 *
 * 			Right:	Left:
 * 	TIM1:	M1		M3
 * 	TIM2:	ENC4	ENC2
 * 	TIM4:	ENC1	ENC3
 * 	TIM8:	M4		M2
 * 	ADC1:	M4B		M2B
 * 	ADC2:	M4A		M2A
 * 	ADC3:	M1C		M3B
 * 	ADC4:	M1B		M3C
 */

#pragma GCC optimize ("Ofast")

#include "drive.h"

#include "spi1.h"
#include "intercom1.h"
#include "constants.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/arm_mat_util_f32.h"
#include <math.h>

#define PWM_MAX_VAL 3599
#define PI_3RD 1.047197551196598f
#define MOT_ADC_TO_CURRENT 0.0029296875f
#define SINUSOIDAL_ANGLE 15 // full circle = 256

Drive drive[2] = {
#ifdef MOT_RIGHT
		{ .currentMapping = {{ 1,  1}, { 0, -1}, {-1,  0}} },	// M1
		{ .currentMapping = {{ 0,  1}, { 1,  0}, {-1, -1}} },	// M4
#else
		{ .currentMapping = {{ 0,  1}, { 1,  0}, {-1, -1}} },	// M2
		{ .currentMapping = {{ 1,  1}, {-1,  0}, { 0, -1}} },	// M3
#endif
};

Drive* pFront = &drive[0];
Drive* pRear = &drive[1];

// Motor Kalman filter
// state vector: [speed, current, load Torque]
// Units: [rad/s, A, Nm]
// A & B created by MATLAB from continous-time state-space model
//static const float A[9] = { 0.99050027f, 0.30360765f, -28.95448228f, -0.01918461f, 0.11256390f, 0.37253838f, 0.0f, 0.0f, 1.0f };
//static const float B[3] = { 0.36227637f, 0.73161364f, 0.0f };

static const float A[4] = { 0.94815814f, 0.29493162f, -0.01863638f, 0.11265899f };
static const float B[2] = { 0.35629318f, 0.73166514f };

// Luenberger Observer matrices for motor
// x[0] = current [A]
// x[1] = motor speed [rad/s]
static const float lbA[4] = { 0.11265899f, -0.01863638f, 0.29493162f, 0.94815814f };
static const float lbB[2] = { 0.73166514f, 0.35629318f };
static const float lbC[2] = { 0.0f, 1.0f };
static const float lbL[2] = { -0.01363333f, 0.95081713f };

static const int32_t sinTableFwd[256][3];
static const int32_t sinTableRev[256][3];

// W3  W2  W1
//  5   5   5
static const uint32_t ccerLUT[8] = {0x555, 0x505, 0x055, 0x550, 0x550, 0x055, 0x505, 0x000 };

// W1, W2, W3 modifier
static const int32_t hallMod[8][3] = {
	{0, 0, 0},
	{36000, 0, 0},
	{0, 36000, 0},
	{0, 36000, 0},
	{0, 0, 36000},
	{36000, 0, 0},
	{0, 0, 36000},
	{0, 0, 0}
};

static float min3(float a, float b, float c);
static void updateADC(Drive* pDrive);
static void updateSpeed(Drive* pDrive);
static void updateHall(Drive* pDrive, uint16_t hallPos);
static void updateOutput(Drive* pDrive);
static void enablePhase(Drive* pDrive, uint32_t ccer);

/**
 * IRQs
 */
// Encoder time measurement
IN_CCM void TIM2_IRQHandler()
{
	static IN_CCM_VAR uint16_t last = 0;

	register uint16_t now = TIM3->CNT;
	TIM3->CCR3 = now+64000;

	TIM2->SR = 0;

#ifdef MOT_RIGHT
	++pRear->encIndex;		// M4
	pRear->encIndex %= DRIVE_ENC_SAMPLES;
	pRear->encDtA[pRear->encIndex] = now-last;

	if(pRear->commutationMode == DRIVE_COM_MODE_SINUSOIDAL)
		updateOutput(pRear);
#else
	++pFront->encIndex;		// M2
	pFront->encIndex %= DRIVE_ENC_SAMPLES;
	pFront->encDtA[pFront->encIndex] = now-last;

	if(pFront->commutationMode == DRIVE_COM_MODE_SINUSOIDAL)
		updateOutput(pFront);
#endif

	last = now;
}

IN_CCM void TIM4_IRQHandler()
{
	static IN_CCM_VAR uint16_t last = 0;

	register uint16_t now = TIM3->CNT;
	TIM3->CCR4 = now+64000;

	TIM4->SR = 0;

#ifdef MOT_RIGHT
	++pFront->encIndex;		// M1
	pFront->encIndex %= DRIVE_ENC_SAMPLES;
	pFront->encDtA[pFront->encIndex] = now-last;

	if(pFront->commutationMode == DRIVE_COM_MODE_SINUSOIDAL)
		updateOutput(pFront);
#else
	++pRear->encIndex;		// M3
	pRear->encIndex %= DRIVE_ENC_SAMPLES;
	pRear->encDtA[pRear->encIndex] = now-last;

	if(pRear->commutationMode == DRIVE_COM_MODE_SINUSOIDAL)
		updateOutput(pRear);
#endif

	last = now;
}

IN_CCM void TIM3_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	register uint16_t sr = TIM3->SR;
	TIM3->SR = 0;
	uint16_t now = TIM3->CNT;

	if(sr & TIM_SR_CC1IF)
	{
		TIM3->CCR1 = now+3999;
		updateSpeed(pFront);
		updateSpeed(pRear);
	}

	if(sr & TIM_SR_CC3IF)
		TIM2->EGR = TIM_EGR_CC1G;

	if(sr & TIM_SR_CC4IF)
		TIM4->EGR = TIM_EGR_CC1G;

	CH_IRQ_EPILOGUE();
}

IN_CCM void TIM8_UP_IRQHandler()
{
	TIM8->SR = 0;

#ifdef MOT_RIGHT
	updateADC(pRear);	// M4
#else
	updateADC(pFront);	// M2
#endif
}

IN_CCM void TIM1_UP_TIM16_IRQHandler()
{
	TIM1->SR = 0;

#ifdef MOT_RIGHT
	updateADC(pFront);	// M1
#else
	updateADC(pRear);	// M3
#endif
}

#define M13HallMask (EXTI_PR_PR10 | EXTI_PR_PR11 | EXTI_PR_PR12)

IN_CCM void EXTI3_IRQHandler()
{
	EXTI->PR = EXTI_PR_PR3;

	uint16_t pos24 = ((GPIOA->IDR & GPIO_PIN_3) >> 3) | ((GPIOC->IDR & (GPIO_PIN_14 | GPIO_PIN_15)) >> 13);

#ifdef MOT_RIGHT
	updateHall(pRear, pos24);
#else
	updateHall(pFront, pos24);
#endif
}

IN_CCM void EXTI15_10_IRQHandler()
{
	if(EXTI->PR & M13HallMask)
	{
		EXTI->PR = M13HallMask;

		uint16_t pos13 = (GPIOB->IDR & M13HallMask) >> 10;

#ifdef MOT_RIGHT
		updateHall(pFront, pos13);
#else
		updateHall(pRear, pos13);
#endif
	}
	else
	{
		EXTI->PR = (EXTI_PR_PR14 | EXTI_PR_PR15);

		uint16_t pos24 = ((GPIOA->IDR & GPIO_PIN_3) >> 3) | ((GPIOC->IDR & (GPIO_PIN_14 | GPIO_PIN_15)) >> 13);

#ifdef MOT_RIGHT
		updateHall(pRear, pos24);
#else
		updateHall(pFront, pos24);
#endif
	}
}

static void initHw()
{
	GPIOInitData gpioInit;

	// data structures
#ifdef MOT_RIGHT
	pRear->pTIMEncoder = TIM2;
	pRear->pTIMMotor = TIM8;
	pFront->pTIMEncoder = TIM4;
	pFront->pTIMMotor = TIM1;
#else
	pRear->pTIMEncoder = TIM4;
	pRear->pTIMMotor = TIM1;
	pFront->pTIMEncoder = TIM2;
	pFront->pTIMMotor = TIM8;
#endif

	pFront->curBat = 12.0f;
	pRear->curBat = 12.0f;

	// ### DMA configuration (ADC)
	// peripheral-to-memory, half-words, medium priority, memory increment, circular mode
	DMA1_Channel1->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA1_Channel1->CNDTR = DRIVE_ADC_SAMPLES;
	DMA1_Channel1->CPAR = (uint32_t)&ADC1->DR;
#ifdef MOT_RIGHT
	DMA1_Channel1->CMAR = (uint32_t)pRear->adc[0];
#else
	DMA1_Channel1->CMAR = (uint32_t)pFront->adc[0];
#endif

	DMA2_Channel1->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel1->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel1->CPAR = (uint32_t)&ADC2->DR;
#ifdef MOT_RIGHT
	DMA2_Channel1->CMAR = (uint32_t)pRear->adc[1];
#else
	DMA2_Channel1->CMAR = (uint32_t)pFront->adc[1];
#endif

	DMA2_Channel5->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel5->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel5->CPAR = (uint32_t)&ADC3->DR;
#ifdef MOT_RIGHT
	DMA2_Channel5->CMAR = (uint32_t)pFront->adc[0];
#else
	DMA2_Channel5->CMAR = (uint32_t)pRear->adc[0];
#endif

	DMA2_Channel2->CCR = DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_PL_MEDIUM | DMA_CCR_MINC | DMA_CCR_CIRC;
	DMA2_Channel2->CNDTR = DRIVE_ADC_SAMPLES;
	DMA2_Channel2->CPAR = (uint32_t)&ADC4->DR;
#ifdef MOT_RIGHT
	DMA2_Channel2->CMAR = (uint32_t)pFront->adc[1];
#else
	DMA2_Channel2->CMAR = (uint32_t)pRear->adc[1];
#endif

	// ### ADC
	RCC->AHBENR |= RCC_AHBENR_ADC12EN | RCC_AHBENR_ADC34EN;

	gpioInit.mode = GPIO_MODE_ANALOG;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_2, &gpioInit); // A2
	GPIOInit(GPIOB, GPIO_PIN_2 | GPIO_PIN_13 | GPIO_PIN_14, &gpioInit); // B2, B13, B14

	// independent, synchronous clock from AHB
	ADC1_2->CCR = ADC12_CCR_CKMODE_0;
	ADC3_4->CCR = ADC34_CCR_CKMODE_0;

	// enable voltage regulator for each ADC
	ADC1->CR = ADC_CR_ADVREGEN_0;
	ADC2->CR = ADC_CR_ADVREGEN_0;
	ADC3->CR = ADC_CR_ADVREGEN_0;
	ADC4->CR = ADC_CR_ADVREGEN_0;
	chThdSleep(MS2ST(4));	// stabilize regulator delay

	// start single-ended calibration
	ADC1->CR |= ADC_CR_ADCAL;
	ADC2->CR |= ADC_CR_ADCAL;
	ADC3->CR |= ADC_CR_ADCAL;
	ADC4->CR |= ADC_CR_ADCAL;
	chThdSleep(MS2ST(4));	// calibration delay

#ifdef MOT_RIGHT
	pRear->adcCal[0]  = ADC1->CALFACT;
	pRear->adcCal[1]  = ADC2->CALFACT;
	pFront->adcCal[0] = ADC3->CALFACT;
	pFront->adcCal[1] = ADC4->CALFACT;
#else
	pFront->adcCal[0] = ADC1->CALFACT;
	pFront->adcCal[1] = ADC2->CALFACT;
	pRear->adcCal[0]  = ADC3->CALFACT;
	pRear->adcCal[1]  = ADC4->CALFACT;
#endif

	// continuous conversion, 12Bit, trigger TIM8_TRGO2
	ADC1->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (8 << 6) | ADC_CFGR_DMACFG;
	ADC1->SQR1 = DRIVE_ADC_SAMPLES-1;

	ADC2->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (8 << 6) | ADC_CFGR_DMACFG;
	ADC2->SQR1 = DRIVE_ADC_SAMPLES-1;

	// continuous conversion, 12Bit, trigger TIM1_TRGO2
	ADC3->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC3->SQR1 = DRIVE_ADC_SAMPLES-1;

	ADC4->CFGR = ADC_CFGR_CONT | ADC_CFGR_EXTEN_0 | (10 << 6) | ADC_CFGR_DMACFG;
	ADC4->SQR1 = DRIVE_ADC_SAMPLES-1;

	for(uint8_t i = 1; i <= DRIVE_ADC_SAMPLES; i++)
	{
		ADCRegularChannelConfig(ADC1,  3, i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC2, 12, i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC3,  5, i, ADC_SampleTime_181Cycles5);
		ADCRegularChannelConfig(ADC4,  4, i, ADC_SampleTime_181Cycles5);
	}

	// configure offsets
	ADC1->OFR1 = ADC_OFR1_OFFSET1_EN | ( 3 << 26) | 1706;
	ADC2->OFR1 = ADC_OFR1_OFFSET1_EN | (12 << 26) | 1706;
	ADC3->OFR1 = ADC_OFR1_OFFSET1_EN | ( 5 << 26) | 1706;
	ADC4->OFR1 = ADC_OFR1_OFFSET1_EN | ( 4 << 26) | 1706;

	// Enable the ADC DMA
	ADC1->CFGR |= ADC_CFGR_DMAEN;
	ADC2->CFGR |= ADC_CFGR_DMAEN;
	ADC3->CFGR |= ADC_CFGR_DMAEN;
	ADC4->CFGR |= ADC_CFGR_DMAEN;

	ADC1->CR |= ADC_CR_ADEN;
	ADC2->CR |= ADC_CR_ADEN;
	ADC3->CR |= ADC_CR_ADEN;
	ADC4->CR |= ADC_CR_ADEN;
	chThdSleep(MS2ST(4));	// ADC ready delay

	// Enable the DMA channel
	DMA1_Channel1->CCR |= DMA_CCR_EN;
	DMA2_Channel1->CCR |= DMA_CCR_EN;
	DMA2_Channel5->CCR |= DMA_CCR_EN;
	DMA2_Channel2->CCR |= DMA_CCR_EN;

	// ### Encoder
	RCC->APB1ENR |= RCC_APB1ENR_TIM2EN | RCC_APB1ENR_TIM4EN;

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 1;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_10MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_0 | GPIO_PIN_1, &gpioInit);	// TIM2_CH1, TIM2_CH2

	gpioInit.alternate = 10;
	GPIOInit(GPIOA, GPIO_PIN_11, &gpioInit);				// TIM4_CH1

	gpioInit.alternate = 2;
	GPIOInit(GPIOB, GPIO_PIN_7, &gpioInit);					// TIM4_CH2

	//Timer Init - Encoder inputs
	TIM2->CR1 = 0;
	TIM2->CR2 = 0;
	TIM2->PSC = 0;
	TIM2->ARR = 2047;
	TIM2->RCR = 0;

	// IC1 mapped to TI1, IC2 mapped to TI2, filter = 0x0F on both, no prescaler
	TIM2->CCMR1 = TIM_CCMR1_CC1S_0 | (0x0F << 4) | TIM_CCMR1_CC2S_0 | (0x0F << 12);
	TIM2->CCER = TIM_CCER_CC1E | TIM_CCER_CC2E;	// enable input capture, rising polarity
	TIM2->DIER = TIM_DIER_CC1IE;
	TIM2->SMCR = 3;	// encoder mode 3, count all flanks
	TIM2->CR1 |= TIM_CR1_CEN;

	TIM4->CR1 = 0;
	TIM4->CR2 = 0;
	TIM4->PSC = 0;
	TIM4->ARR = 2047;
	TIM4->RCR = 0;

	// IC1 mapped to TI1, IC2 mapped to TI2, filter = 0x0F on both, no prescaler
	TIM4->CCMR1 = TIM_CCMR1_CC1S_0 | (0x0F << 4) | TIM_CCMR1_CC2S_0 | (0x0F << 12);
	TIM4->CCER = TIM_CCER_CC1E | TIM_CCER_CC2E;	// enable input capture, rising polarity
	TIM4->DIER = TIM_DIER_CC1IE;
	TIM4->SMCR = 3;	// encoder mode 3, count all flanks
	TIM4->CR1 |= TIM_CR1_CEN;

	NVICEnableIRQ(TIM2_IRQn, 1);	// warning: this is out of kernel IRQ level range. Do not call any kernel functions from IRQ!
	NVICEnableIRQ(TIM4_IRQn, 1);

	// ### Encoder Velocity Measurement Timer
	RCC->APB1ENR |= RCC_APB1ENR_TIM3EN;

	TIM3->CR1 = 0;
	TIM3->CR2 = 0;
	TIM3->PSC = 17;	// = 250ns = 4MHz
	TIM3->ARR = 0xFFFF;	// 65535
	TIM3->RCR = 0;

	TIM3->CCMR1 = 0;
	TIM3->CCR1 = 4000;
	TIM3->CCR3 = 64000;
	TIM3->CCR4 = 64000;

	TIM3->CR1 |= TIM_CR1_CEN;

	TIM3->DIER = TIM_DIER_CC1IE | TIM_DIER_CC3IE | TIM_DIER_CC4IE;

	NVICEnableIRQ(TIM3_IRQn, IRQL_SPEED_UPDATE);

	// ### Motors
	RCC->APB2ENR |= RCC_APB2ENR_TIM1EN | RCC_APB2ENR_TIM8EN;
	RCC->CFGR3 |= RCC_CFGR3_TIM1SW | RCC_CFGR3_TIM8SW;	// TIM1 + TIM8 clocks now running at 144MHz

	// Timer config
	// TIM1
	TIM1->CR1 = TIM_CR1_CMS_0;	// center-aligned mode 1
	TIM1->PSC = 0;
	TIM1->ARR = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	TIM1->RCR = 1;				// update event only every second count (e.g. downcounting) = 20kHz

	// PWM mode 2, preload enable on CCR(1-3) registers
	TIM1->CCMR1 = TIM_CCMR1_OC1PE | (7 << 4) | TIM_CCMR1_OC2PE | (7 << 12);
	TIM1->CCMR2 = TIM_CCMR2_OC3PE | (7 << 4) | (7 << 12);
	TIM1->CCER = TIM_CCER_CC1E | TIM_CCER_CC1NE | TIM_CCER_CC2E | TIM_CCER_CC2NE | TIM_CCER_CC3E | TIM_CCER_CC3NE | TIM_CCER_CC4E;
	TIM1->CR2 = TIM_CR2_OIS1 | TIM_CR2_OIS2 | TIM_CR2_OIS3 | (7 << 20);	// output idle state = high, OC4 on TRGO2 (connected to ADC)
	TIM1->CCR1 = 0;
	TIM1->CCR2 = 0;
	TIM1->CCR3 = 0;
	TIM1->CCR4 = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM1->EGR = TIM_EGR_UG;

	TIM1->BDTR = TIM_BDTR_AOE | TIM_BDTR_BKP | TIM_BDTR_OSSR;
	TIM1->DIER = TIM_DIER_UIE;

	TIM1->CR1 |= TIM_CR1_CEN;

	ADC1->CR |= ADC_CR_ADSTART;
	ADC2->CR |= ADC_CR_ADSTART;

	TIM1->BDTR |= TIM_BDTR_MOE;

	// TIM8
	TIM8->CR1 = TIM_CR1_CMS_0;	// center-aligned mode 1
	TIM8->PSC = 0;
	TIM8->ARR = PWM_MAX_VAL;	// = 3599 = 20kHz PWM (2*10kHz due to center aligned mode)
	TIM8->RCR = 1;				// update event only every second count (e.g. downcounting) = 20kHz

	// PWM mode 2, preload enable on CCR(1-3) registers
	TIM8->CCMR1 = TIM_CCMR1_OC1PE | (7 << 4) | TIM_CCMR1_OC2PE | (7 << 12);
	TIM8->CCMR2 = TIM_CCMR2_OC3PE | (7 << 4) | (7 << 12);
	TIM8->CCER = TIM_CCER_CC1E | TIM_CCER_CC1NE | TIM_CCER_CC2E | TIM_CCER_CC2NE | TIM_CCER_CC3E | TIM_CCER_CC3NE | TIM_CCER_CC4E;
	TIM8->CR2 = TIM_CR2_OIS1 | TIM_CR2_OIS2 | TIM_CR2_OIS3 | (7 << 20);	// output idle state = high, OC4 on TRGO2 (connected to ADC)
	TIM8->CCR1 = 0;
	TIM8->CCR2 = 0;
	TIM8->CCR3 = 0;
	TIM8->CCR4 = 864;	// trigger ADC conversion 6us after update event (allow IRQ to sum up previous value)
	TIM8->EGR = TIM_EGR_UG;

	TIM8->BDTR = TIM_BDTR_AOE | TIM_BDTR_BKP | TIM_BDTR_OSSR;
	TIM8->DIER = TIM_DIER_UIE;

	TIM8->CR1 |= TIM_CR1_CEN;

	ADC3->CR |= ADC_CR_ADSTART;
	ADC4->CR |= ADC_CR_ADSTART;

	TIM8->BDTR |= TIM_BDTR_MOE;

	// pin config - drive outputs
	GPIOA->ODR |= GPIO_PIN_7 | GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_12;
	GPIOB->ODR |= GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_6 | GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_15;
	GPIOC->ODR |= GPIO_PIN_13;

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 4;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_50MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_7, &gpioInit);
	GPIOInit(GPIOB, GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_15, &gpioInit);
	GPIOInit(GPIOC, GPIO_PIN_13, &gpioInit);

	gpioInit.alternate = 6;
	GPIOInit(GPIOA, GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_12, &gpioInit);

	gpioInit.alternate = 5;
	GPIOInit(GPIOB, GPIO_PIN_6, &gpioInit);

	gpioInit.alternate = 10;
	GPIOInit(GPIOB, GPIO_PIN_8 | GPIO_PIN_9, &gpioInit);

	// ### Hall Sensors
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING_FALLING;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_3, &gpioInit);
	GPIOInit(GPIOB, GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12, &gpioInit);
	GPIOInit(GPIOC, GPIO_PIN_14 | GPIO_PIN_15, &gpioInit);

	NVICEnableIRQ(TIM8_UP_IRQn, 2);
	NVICEnableIRQ(TIM1_UP_TIM16_IRQn, 2);
	NVICEnableIRQ(EXTI3_IRQn, 3);
	NVICEnableIRQ(EXTI15_10_IRQn, 3);
}

/**
 * PRIVATE functions
 */
IN_CCM static float min3(float a, float b, float c)
{
	if(a < b && a < c)
		return a;

	if(b < a && b < c)
		return b;

	return c;
}

typedef union _u32u16
{
	uint32_t u32;
	uint16_t u16[2];
} u32u16_t;

typedef union _i32i16
{
	int32_t i32;
	int16_t i16[2];
} i32i16_t;

IN_CCM static void updateADC(Drive* pDrive)
{
	int16_t i016, i116;
	float32_t i0f, i1f;
	i32i16_t i0i, i1i;

	i0i.i32 = pDrive->adc32[0][0];
	i1i.i32 = pDrive->adc32[1][0];

	for(uint8_t i = 1; i < DRIVE_ADC_SAMPLES/2; i++)
	{
		i0i.i32 = __QADD16(i0i.i32, pDrive->adc32[0][i]);
		i1i.i32 = __QADD16(i1i.i32, pDrive->adc32[1][i]);
	}

	i016 = i0i.i16[0]+i0i.i16[1];
	i116 = i1i.i16[0]+i1i.i16[1];

	i0f = i016*(1.0f/DRIVE_ADC_SAMPLES);
	i1f = i116*(1.0f/DRIVE_ADC_SAMPLES);

	i0f = i0f*-MOT_ADC_TO_CURRENT;
	i1f = i1f*-MOT_ADC_TO_CURRENT;

	pDrive->i[0] = pDrive->currentMapping[0][0]*i0f + pDrive->currentMapping[0][1]*i1f;
	pDrive->i[1] = pDrive->currentMapping[1][0]*i0f + pDrive->currentMapping[1][1]*i1f;
	pDrive->i[2] = pDrive->currentMapping[2][0]*i0f + pDrive->currentMapping[2][1]*i1f;

	pDrive->iAlpha = pDrive->i[0];
	pDrive->iBeta = 0.57735026918963f*pDrive->i[0] + 1.154700538379252f*pDrive->i[1];

	const float alpha = 0.98f;
	float rms = sqrtf(pDrive->iAlpha*pDrive->iAlpha + pDrive->iBeta*pDrive->iBeta);
	pDrive->rmsCurrent = (alpha*pDrive->rmsCurrent + (1.0f-alpha)*rms);

	pDrive->rmsCurrentTimestamp = SysTimeUSec();

	// call something on current update?
	// Rate: 20kHz, above kernel IRQ level
}

IN_CCM static void updateSpeed(Drive* pDrive)
{
	float vel;

	uint16_t dir = (pDrive->pTIMEncoder->CR1 & TIM_CR1_DIR);
	uint32_t encDt = 0;
	uint16_t steps = 0;
	for(uint8_t i = 0; i < DRIVE_ENC_SAMPLES; i++)
	{
		uint16_t dt = pDrive->encDtA[i];
		encDt += dt;
		if(dt < 64000)
			steps++;
	}

	if(encDt > steps*64000 || encDt < 80)
	{
		vel = 0;
	}
	else if(dir != pDrive->lastDir)
	{
		vel = 0;
		memset(pDrive->encDtA, 0xFF, DRIVE_ENC_SAMPLES*sizeof(uint16_t));
	}
	else
	{
		vel = (steps*1e6f)/(encDt*128);

		if(dir)
			vel *= -1;
	}

	pDrive->lastDir = dir;

	if(vel < 10000)
	{
		pDrive->curVelocity = vel;
		pDrive->curVelocityTimestamp = SysTimeUSec();
	}

	chSysLockFromIsr();
	chMBPostI(&pDrive->eventQueue, 0);
	chSysUnlockFromIsr();
}

IN_CCM static void updateHall(Drive* pDrive, uint16_t hallPos)
{
	pDrive->halPos = hallPos;

	if(pDrive->commutationMode == DRIVE_COM_MODE_BLOCK)
		updateOutput(pDrive);
}

IN_CCM static void updateOutput(Drive* pDrive)
{
	int32_t out = pDrive->outLevel;

	switch(pDrive->commutationMode)
	{
		case DRIVE_COM_MODE_BLOCK:
		{
			uint16_t hallPos = pDrive->halPos;

			if(out < 0)
			{
				out *= -1;
				hallPos ^= 0x7;
			}

			enablePhase(pDrive, ccerLUT[hallPos]);

			pDrive->pTIMMotor->CCR1 = (hallMod[hallPos][0]*out)/100000;
			pDrive->pTIMMotor->CCR2 = (hallMod[hallPos][1]*out)/100000;
			pDrive->pTIMMotor->CCR3 = (hallMod[hallPos][2]*out)/100000;
		}
		break;
		case DRIVE_COM_MODE_SINUSOIDAL:
		{
			uint16_t pos = pDrive->pTIMEncoder->CNT;

			enablePhase(pDrive, 0x555);

			const int32_t* pLUT;
			if(out < 0)
			{
				out *= -1;
				pos += SINUSOIDAL_ANGLE;
				pos %= 256;
				pLUT = sinTableRev[pos];
			}
			else
			{
				pos -= SINUSOIDAL_ANGLE;
				pos %= 256;
				pLUT = sinTableFwd[255-pos];
			}

			pDrive->pTIMMotor->CCR1 = (pLUT[0]*out)/100000;
			pDrive->pTIMMotor->CCR2 = (pLUT[1]*out)/100000;
			pDrive->pTIMMotor->CCR3 = (pLUT[2]*out)/100000;
		}
		break;
		default:
		{
			enablePhase(pDrive, 0x555);

			pDrive->pTIMMotor->CCR1 = 5000;
			pDrive->pTIMMotor->CCR2 = 5000;
			pDrive->pTIMMotor->CCR3 = 5000;
		}
	}
}

IN_CCM static void enablePhase(Drive* pDrive, uint32_t ccer)
{
	uint32_t oldCCER = pDrive->pTIMMotor->CCER & (TIM_CCER_CC1P | TIM_CCER_CC1NP | TIM_CCER_CC2P | TIM_CCER_CC2NP | TIM_CCER_CC3P | TIM_CCER_CC3NP);
	pDrive->pTIMMotor->CCER = oldCCER | ccer;
}

IN_CCM static void setCCR(Drive* pDrive, uint16_t a, uint16_t b, uint16_t c)
{
	pDrive->pTIMMotor->CCR1 = a;
	pDrive->pTIMMotor->CCR2 = b;
	pDrive->pTIMMotor->CCR3 = c;
}

msg_t DriveTask(void* params)
{
	Drive* pDrive = (Drive*)params;

	if(pDrive == pFront)
		chRegSetThreadName("D-Front");
	else
		chRegSetThreadName("D-Rear");

	msg_t event;

	while(1)
	{
		if(chMBFetch(&pDrive->eventQueue, &event, TIME_INFINITE) != RDY_OK)
			continue;

		// Luenberger-Observer update
		LuenbergerSISO2SUpdate(&pDrive->lbObserver, pDrive->setVoltage, pDrive->curVelocity*2*M_PI);
		pDrive->estVelocity = pDrive->lbObserver.x[1]*1/(2*M_PI);
		pDrive->estCurrent = pDrive->lbObserver.x[0];
		pDrive->estTorque = pDrive->lbObserver.x[0]*MOTOR_KT;

		pDrive->estTimestamp = SysTimeUSec();

/*		// Kalman-Filter update
		MAT_ELEMENT(pDrive->kf.u, 0, 0) = pDrive->setVoltage;
		MAT_ELEMENT(pDrive->kf.z, 0, 0) = pDrive->curVelocity*2*M_PI;
		if(pDrive->curVelocity > 0)
			MAT_ELEMENT(pDrive->kf.z, 1, 0) = pDrive->rmsCurrent;
		else
			MAT_ELEMENT(pDrive->kf.z, 1, 0) = -pDrive->rmsCurrent;

		KFPredict(&pDrive->kf);
		KFUpdate(&pDrive->kf);

		arm_mat_clip_f32(&pDrive->kf.Sigma, -100.0f, 100.0f);

		pDrive->estVelocity = MAT_ELEMENT(pDrive->kf.x, 0, 0)*1/(2*M_PI);
		pDrive->estCurrent = MAT_ELEMENT(pDrive->kf.x, 1, 0);
		pDrive->estTorque = MAT_ELEMENT(pDrive->kf.x, 1, 0)*MOTOR_KT;
//		pDrive->estLoadTorque = MAT_ELEMENT(pDrive->kf.x, 2, 0);
*/
		switch(pDrive->ctrlMode)
		{
			case DRIVE_CTRL_MODE_PID_SPEED:
			{
				// control
				pDrive->speed.pid.setpoint = pDrive->speed.setVelocity;
				PIDUpdate(&pDrive->speed.pid, pDrive->curVelocity);
//				PIDUpdate(&pDrive->speed.pid, pDrive->estVelocity);

				DriveSetVoltage(pDrive, pDrive->speed.pid.output);
			}
			break;
			case DRIVE_CTRL_MODE_PID_TORQUE:
			{
				pDrive->torque.pid.setpoint = pDrive->torque.setCurrent;
				PIDUpdate(&pDrive->torque.pid, pDrive->estCurrent);

//				float out = pDrive->estVelocity*(2*M_PI)*MOTOR_KE + (MOTOR_RESISTANCE/MOTOR_KT)*pDrive->torque.pid.output;
				float out = pDrive->torque.pid.output;

				DriveSetVoltage(pDrive, out);
			}
			break;
		}
	}

	return 0;
}

/**
 * PUBLIC functions
 */
void DriveInit()
{
	for(uint8_t i = 0; i < 2; i++)
	{
		Drive* pDrive = &drive[i];

		chMBInit(&pDrive->eventQueue, pDrive->eventQueueData, DRIVE_EVENT_QUEUE_SIZE);

		KFInit(&pDrive->kf, 2, 1, 2, pDrive->kfData);

		memcpy(pDrive->kf.A.pData, A, sizeof(A));
		memcpy(pDrive->kf.B.pData, B, sizeof(B));

		MAT_ELEMENT(pDrive->kf.C, 0, 0) = 1.0f;	// C matrix is [1 0 0; 0 1 0]
		MAT_ELEMENT(pDrive->kf.C, 1, 1) = 0.0f;

		DriveSetStateUncertainty(pDrive, 0.001f);
		DriveSetSensorUncertainty(pDrive, 0.000001f);

		LuenbergerSISO2SInit(&pDrive->lbObserver, lbA, lbB, lbC, lbL);

		PIDParams speedPIDParams;
		speedPIDParams.kp = 0.25f;
		speedPIDParams.ki = 25.0f;
		speedPIDParams.kd = 0.0f;
		speedPIDParams.outputMin = -10.0f;	// [V]
		speedPIDParams.outputMax = 10.0f;	// [V]

		PIDInit(&pDrive->speed.pid, &speedPIDParams, 0.001f, 1);

		PIDParams torquePIDParams;
		torquePIDParams.kp = 0.8f;
		torquePIDParams.ki = 1000.0f;
		torquePIDParams.kd = 0.0f;
		torquePIDParams.outputMin = -10.0f;	// [V]
		torquePIDParams.outputMax = 10.0f;	// [V]

		PIDInit(&pDrive->torque.pid, &torquePIDParams, 0.001f, 1);
	}

	initHw();

	// trigger initial hall IRQ to get a valid hall position
	EXTI->SWIER |= (EXTI_SWIER_SWIER10 | EXTI_SWIER_SWIER14);

	DriveSetCommutationMode(pFront, DRIVE_COM_MODE_BLOCK);
	DriveSetCommutationMode(pRear, DRIVE_COM_MODE_BLOCK);

	DriveSetCtrlMode(pFront, DRIVE_CTRL_MODE_PID_SPEED);
	DriveSetCtrlMode(pRear, DRIVE_CTRL_MODE_PID_SPEED);
}

void DriveSetStateUncertainty(Drive* pDrive, float noise)
{
	MAT_ELEMENT(pDrive->kf.Ex, 0, 0) = noise;
	MAT_ELEMENT(pDrive->kf.Ex, 1, 1) = noise;
//	MAT_ELEMENT(pDrive->kf.Ex, 2, 2) = noise*10.0f;
}

void DriveSetSensorUncertainty(Drive* pDrive, float noise)
{
	MAT_ELEMENT(pDrive->kf.Ez, 0, 0) = noise;
	MAT_ELEMENT(pDrive->kf.Ez, 1, 1) = noise*10.0f;
}

void DriveEnable(Drive* pDrive, uint8_t enable)
{
	if(enable)
	{
		if(pDrive->commutationMode == DRIVE_COM_MODE_NONE)
		{
			pDrive->commutationMode = pDrive->commutationModeBackup;
			updateOutput(pDrive);
		}
	}
	else
	{
		if(pDrive->commutationMode != DRIVE_COM_MODE_NONE)
		{
			pDrive->commutationModeBackup = pDrive->commutationMode;
			pDrive->commutationMode = DRIVE_COM_MODE_NONE;
			pDrive->outLevel = 0;
			updateOutput(pDrive);
		}
	}
}

void DriveSetCommutationMode(Drive* pDrive, uint8_t mode)
{
	if(pDrive->commutationMode == mode)
		return;

	uint8_t ctrlMode = pDrive->ctrlMode;
	DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);

	pDrive->commutationMode = DRIVE_COM_MODE_NONE;

	switch(mode)
	{
		case DRIVE_COM_MODE_NONE:
		{
		}
		break;
		case DRIVE_COM_MODE_BLOCK:
		{
		}
		break;
		case DRIVE_COM_MODE_SINUSOIDAL:
		{
			chThdSleepMilliseconds(100);
			DriveSetVoltageAngle(pDrive, 1.0, 0);
			chThdSleepMilliseconds(200);
			pDrive->pTIMEncoder->CNT = 0;
		}
		break;
		default:
			return;
	}

	pDrive->outLevel = 0;
	pDrive->commutationMode = mode;
	pDrive->commutationModeBackup = mode;
	updateOutput(pDrive);

	DriveSetCtrlMode(pDrive, ctrlMode);
}

void DriveSetCtrlMode(Drive* pDrive, uint8_t mode)
{
	if(pDrive->ctrlMode == mode)
		return;

	switch(mode)
	{
		case DRIVE_CTRL_MODE_NONE:
		{
			PIDEnable(&pDrive->speed.pid, 0);
		}
		break;
		case DRIVE_CTRL_MODE_PID_SPEED:
		{
			PIDEnable(&pDrive->speed.pid, 1);
		}
		break;
		case DRIVE_CTRL_MODE_PID_TORQUE:
			break;
		default:
			return;
	}

	pDrive->ctrlMode = mode;
}

#define VOLTAGE_SLOPE 1000

IN_CCM void DriveSetVoltage(Drive* pDrive, float voltage)
{
	pDrive->setVoltage = voltage;
	pDrive->setVoltageTimestamp = SysTimeUSec();

	int32_t newOutLevel = (int32_t)(pDrive->setVoltage/pDrive->curBat*10000.0f);
	if(newOutLevel > 10000)
		newOutLevel = 10000;

	if(newOutLevel < -10000)
		newOutLevel = -10000;

	if(newOutLevel - pDrive->outLevel > VOLTAGE_SLOPE)
		newOutLevel = pDrive->outLevel+VOLTAGE_SLOPE;

	if(newOutLevel - pDrive->outLevel < -VOLTAGE_SLOPE)
		newOutLevel = pDrive->outLevel-VOLTAGE_SLOPE;

	pDrive->outLevel = newOutLevel;
	updateOutput(pDrive);
}

void DriveSpeedSetPIDParams(Drive* pDrive, float p, float i, float d)
{
	PIDSetParams(&pDrive->speed.pid, p, i, d);
}

void DriveSpeedSetVelocity(Drive* pDrive, float vel)
{
	pDrive->speed.setVelocity = vel;
}

void DriveTorqueSetPIDParams(Drive* pDrive, float p, float i, float d)
{
	PIDSetParams(&pDrive->torque.pid, p, i, d);
}

void DriveTorqueSetTorque(Drive* pDrive, float torque)
{
	pDrive->torque.setCurrent = torque*1/MOTOR_KT;
}

void DriveSetPhaseVoltagesSVPWM(Drive* pDrive, float Va, float Vb, float Vc)
{
	float Vd = pDrive->curBat;
	if(Vd < 9.0f || Vd > 17.0f)
		Vd = 11.1f;

	float Vcom = -min3(Va, Vb, Vc);

	pDrive->u[0] = Va+Vcom;
	pDrive->u[1] = Vb+Vcom;
	pDrive->u[2] = Vc+Vcom;

	float da = 1/Vd*(Va+Vcom);
	float db = 1/Vd*(Vb+Vcom);
	float dc = 1/Vd*(Vc+Vcom);

	if(da > 1.0f)
		da = 1.0f;
	if(da < 0.0f)
		da = 0.0f;

	if(db > 1.0f)
		db = 1.0f;
	if(db < 0.0f)
		db = 0.0f;

	if(dc > 1.0f)
		dc = 1.0f;
	if(dc < 0.0f)
		dc = 0.0f;

	setCCR(pDrive, da*PWM_MAX_VAL, db*PWM_MAX_VAL, dc*PWM_MAX_VAL);
}

void DriveSetAlphaBeta(Drive* pDrive, float Valpha, float Vbeta)
{
	float Va = Valpha;
	float Vb = -0.5f*Valpha + 0.86602540378444f*Vbeta;
	float Vc = -0.5f*Valpha - 0.86602540378444f*Vbeta;

	DriveSetPhaseVoltagesSVPWM(pDrive, Va, Vb, Vc);
}

void DriveSetVoltageAngle(Drive* pDrive, float Vref, float angle)
{
	float sin, cos;
	while(angle > 180.0f)
		angle -= 360.0f;

	arm_sin_cos_f32(angle, &sin, &cos);

	DriveSetAlphaBeta(pDrive, cos*Vref, sin*Vref);
}

// multiply with max 10.000 and divide by 100000 to get motor TIM CCR values
static const int32_t sinTableFwd[256][3] = {
{     0,  31177,      0}, {   883,  31609,      0}, {  1766,  32023,      0}, {  2648,  32417,      0},
{  3529,  32791,      0}, {  4407,  33146,      0}, {  5282,  33481,      0}, {  6155,  33795,      0},
{  7023,  34089,      0}, {  7888,  34363,      0}, {  8747,  34616,      0}, {  9602,  34848,      0},
{ 10450,  35060,      0}, { 11293,  35250,      0}, { 12128,  35418,      0}, { 12956,  35566,      0},
{ 13777,  35692,      0}, { 14589,  35797,      0}, { 15392,  35880,      0}, { 16186,  35941,      0},
{ 16970,  35981,      0}, { 17744,  35999,      0}, { 18508,  35995,      0}, { 19260,  35970,      0},
{ 20001,  35923,      0}, { 20729,  35854,      0}, { 21445,  35764,      0}, { 22148,  35652,      0},
{ 22838,  35519,      0}, { 23514,  35365,      0}, { 24176,  35189,      0}, { 24823,  34992,      0},
{ 25456,  34773,      0}, { 26073,  34534,      0}, { 26674,  34274,      0}, { 27260,  33994,      0},
{ 27828,  33693,      0}, { 28380,  33371,      0}, { 28915,  33030,      0}, { 29433,  32668,      0},
{ 29933,  32287,      0}, { 30415,  31887,      0}, { 30878,  31467,      0}, { 31323,  31029,      0},
{ 31749,  30571,      0}, { 32156,  30096,      0}, { 32544,  29602,      0}, { 32912,  29090,      0},
{ 33260,  28561,      0}, { 33588,  28014,      0}, { 33896,  27451,      0}, { 34183,  26871,      0},
{ 34450,  26275,      0}, { 34696,  25663,      0}, { 34921,  25036,      0}, { 35125,  24394,      0},
{ 35308,  23736,      0}, { 35470,  23065,      0}, { 35610,  22380,      0}, { 35729,  21681,      0},
{ 35827,  20969,      0}, { 35902,  20245,      0}, { 35957,  19508,      0}, { 35989,  18760,      0},
{ 36000,  18000,      0}, { 35989,  17229,      0}, { 35957,  16449,      0}, { 35902,  15658,      0},
{ 35827,  14857,      0}, { 35729,  14048,      0}, { 35610,  13231,      0}, { 35470,  12405,      0},
{ 35308,  11572,      0}, { 35125,  10732,      0}, { 34921,   9885,      0}, { 34696,   9033,      0},
{ 34450,   8175,      0}, { 34183,   7312,      0}, { 33896,   6445,      0}, { 33588,   5573,      0},
{ 33260,   4699,      0}, { 32912,   3822,      0}, { 32544,   2942,      0}, { 32156,   2061,      0},
{ 31749,   1178,      0}, { 31323,    295,      0}, { 31467,      0,    589}, { 31887,      0,   1472},
{ 32287,      0,   2355}, { 32668,      0,   3235}, { 33030,      0,   4114}, { 33371,      0,   4991},
{ 33693,      0,   5864}, { 33994,      0,   6734}, { 34274,      0,   7600}, { 34534,      0,   8461},
{ 34773,      0,   9317}, { 34992,      0,  10168}, { 35189,      0,  11013}, { 35365,      0,  11850},
{ 35519,      0,  12681}, { 35652,      0,  13504}, { 35764,      0,  14319}, { 35854,      0,  15125},
{ 35923,      0,  15922}, { 35970,      0,  16710}, { 35995,      0,  17487}, { 35999,      0,  18254},
{ 35981,      0,  19010}, { 35941,      0,  19755}, { 35880,      0,  20488}, { 35797,      0,  21208},
{ 35692,      0,  21915}, { 35566,      0,  22610}, { 35418,      0,  23290}, { 35250,      0,  23957},
{ 35060,      0,  24609}, { 34848,      0,  25247}, { 34616,      0,  25869}, { 34363,      0,  26476},
{ 34089,      0,  27066}, { 33795,      0,  27641}, { 33481,      0,  28198}, { 33146,      0,  28739},
{ 32791,      0,  29262}, { 32417,      0,  29768}, { 32023,      0,  30256}, { 31609,      0,  30726},
{ 31177,      0,  31177}, { 30726,      0,  31609}, { 30256,      0,  32023}, { 29768,      0,  32417},
{ 29262,      0,  32791}, { 28739,      0,  33146}, { 28198,      0,  33481}, { 27641,      0,  33795},
{ 27066,      0,  34089}, { 26476,      0,  34363}, { 25869,      0,  34616}, { 25247,      0,  34848},
{ 24609,      0,  35060}, { 23957,      0,  35250}, { 23290,      0,  35418}, { 22610,      0,  35566},
{ 21915,      0,  35692}, { 21208,      0,  35797}, { 20488,      0,  35880}, { 19755,      0,  35941},
{ 19010,      0,  35981}, { 18254,      0,  35999}, { 17487,      0,  35995}, { 16710,      0,  35970},
{ 15922,      0,  35923}, { 15125,      0,  35854}, { 14319,      0,  35764}, { 13504,      0,  35652},
{ 12681,      0,  35519}, { 11850,      0,  35365}, { 11013,      0,  35189}, { 10168,      0,  34992},
{  9317,      0,  34773}, {  8461,      0,  34534}, {  7600,      0,  34274}, {  6734,      0,  33994},
{  5864,      0,  33693}, {  4991,      0,  33371}, {  4114,      0,  33030}, {  3235,      0,  32668},
{  2355,      0,  32287}, {  1472,      0,  31887}, {   589,      0,  31467}, {     0,    295,  31323},
{     0,   1178,  31749}, {     0,   2061,  32156}, {     0,   2942,  32544}, {     0,   3822,  32912},
{     0,   4699,  33260}, {     0,   5573,  33588}, {     0,   6445,  33896}, {     0,   7312,  34183},
{     0,   8175,  34450}, {     0,   9033,  34696}, {     0,   9885,  34921}, {     0,  10732,  35125},
{     0,  11572,  35308}, {     0,  12405,  35470}, {     0,  13231,  35610}, {     0,  14048,  35729},
{     0,  14857,  35827}, {     0,  15658,  35902}, {     0,  16449,  35957}, {     0,  17229,  35989},
{     0,  18000,  36000}, {     0,  18760,  35989}, {     0,  19508,  35957}, {     0,  20245,  35902},
{     0,  20969,  35827}, {     0,  21681,  35729}, {     0,  22380,  35610}, {     0,  23065,  35470},
{     0,  23736,  35308}, {     0,  24394,  35125}, {     0,  25036,  34921}, {     0,  25663,  34696},
{     0,  26275,  34450}, {     0,  26871,  34183}, {     0,  27451,  33896}, {     0,  28014,  33588},
{     0,  28561,  33260}, {     0,  29090,  32912}, {     0,  29602,  32544}, {     0,  30096,  32156},
{     0,  30571,  31749}, {     0,  31029,  31323}, {     0,  31467,  30878}, {     0,  31887,  30415},
{     0,  32287,  29933}, {     0,  32668,  29433}, {     0,  33030,  28915}, {     0,  33371,  28380},
{     0,  33693,  27828}, {     0,  33994,  27260}, {     0,  34274,  26674}, {     0,  34534,  26073},
{     0,  34773,  25456}, {     0,  34992,  24823}, {     0,  35189,  24176}, {     0,  35365,  23514},
{     0,  35519,  22838}, {     0,  35652,  22148}, {     0,  35764,  21445}, {     0,  35854,  20729},
{     0,  35923,  20001}, {     0,  35970,  19260}, {     0,  35995,  18508}, {     0,  35999,  17744},
{     0,  35981,  16970}, {     0,  35941,  16186}, {     0,  35880,  15392}, {     0,  35797,  14589},
{     0,  35692,  13777}, {     0,  35566,  12956}, {     0,  35418,  12128}, {     0,  35250,  11293},
{     0,  35060,  10450}, {     0,  34848,   9602}, {     0,  34616,   8747}, {     0,  34363,   7888},
{     0,  34089,   7023}, {     0,  33795,   6155}, {     0,  33481,   5282}, {     0,  33146,   4407},
{     0,  32791,   3529}, {     0,  32417,   2648}, {     0,  32023,   1766}, {     0,  31609,    883}, };

static const int32_t sinTableRev[256][3] = {
	{     0,      0,  31177}, {   883,      0,  31609}, {  1766,      0,  32023}, {  2648,      0,  32417},
	{  3529,      0,  32791}, {  4407,      0,  33146}, {  5282,      0,  33481}, {  6155,      0,  33795},
	{  7023,      0,  34089}, {  7888,      0,  34363}, {  8747,      0,  34616}, {  9602,      0,  34848},
	{ 10450,      0,  35060}, { 11293,      0,  35250}, { 12128,      0,  35418}, { 12956,      0,  35566},
	{ 13777,      0,  35692}, { 14589,      0,  35797}, { 15392,      0,  35880}, { 16186,      0,  35941},
	{ 16970,      0,  35981}, { 17744,      0,  35999}, { 18508,      0,  35995}, { 19260,      0,  35970},
	{ 20001,      0,  35923}, { 20729,      0,  35854}, { 21445,      0,  35764}, { 22148,      0,  35652},
	{ 22838,      0,  35519}, { 23514,      0,  35365}, { 24176,      0,  35189}, { 24823,      0,  34992},
	{ 25456,      0,  34773}, { 26073,      0,  34534}, { 26674,      0,  34274}, { 27260,      0,  33994},
	{ 27828,      0,  33693}, { 28380,      0,  33371}, { 28915,      0,  33030}, { 29433,      0,  32668},
	{ 29933,      0,  32287}, { 30415,      0,  31887}, { 30878,      0,  31467}, { 31323,      0,  31029},
	{ 31749,      0,  30571}, { 32156,      0,  30096}, { 32544,      0,  29602}, { 32912,      0,  29090},
	{ 33260,      0,  28561}, { 33588,      0,  28014}, { 33896,      0,  27451}, { 34183,      0,  26871},
	{ 34450,      0,  26275}, { 34696,      0,  25663}, { 34921,      0,  25036}, { 35125,      0,  24394},
	{ 35308,      0,  23736}, { 35470,      0,  23065}, { 35610,      0,  22380}, { 35729,      0,  21681},
	{ 35827,      0,  20969}, { 35902,      0,  20245}, { 35957,      0,  19508}, { 35989,      0,  18760},
	{ 36000,      0,  18000}, { 35989,      0,  17229}, { 35957,      0,  16449}, { 35902,      0,  15658},
	{ 35827,      0,  14857}, { 35729,      0,  14048}, { 35610,      0,  13231}, { 35470,      0,  12405},
	{ 35308,      0,  11572}, { 35125,      0,  10732}, { 34921,      0,   9885}, { 34696,      0,   9033},
	{ 34450,      0,   8175}, { 34183,      0,   7312}, { 33896,      0,   6445}, { 33588,      0,   5573},
	{ 33260,      0,   4699}, { 32912,      0,   3822}, { 32544,      0,   2942}, { 32156,      0,   2061},
	{ 31749,      0,   1178}, { 31323,      0,    295}, { 31467,    589,      0}, { 31887,   1472,      0},
	{ 32287,   2355,      0}, { 32668,   3235,      0}, { 33030,   4114,      0}, { 33371,   4991,      0},
	{ 33693,   5864,      0}, { 33994,   6734,      0}, { 34274,   7600,      0}, { 34534,   8461,      0},
	{ 34773,   9317,      0}, { 34992,  10168,      0}, { 35189,  11013,      0}, { 35365,  11850,      0},
	{ 35519,  12681,      0}, { 35652,  13504,      0}, { 35764,  14319,      0}, { 35854,  15125,      0},
	{ 35923,  15922,      0}, { 35970,  16710,      0}, { 35995,  17487,      0}, { 35999,  18254,      0},
	{ 35981,  19010,      0}, { 35941,  19755,      0}, { 35880,  20488,      0}, { 35797,  21208,      0},
	{ 35692,  21915,      0}, { 35566,  22610,      0}, { 35418,  23290,      0}, { 35250,  23957,      0},
	{ 35060,  24609,      0}, { 34848,  25247,      0}, { 34616,  25869,      0}, { 34363,  26476,      0},
	{ 34089,  27066,      0}, { 33795,  27641,      0}, { 33481,  28198,      0}, { 33146,  28739,      0},
	{ 32791,  29262,      0}, { 32417,  29768,      0}, { 32023,  30256,      0}, { 31609,  30726,      0},
	{ 31177,  31177,      0}, { 30726,  31609,      0}, { 30256,  32023,      0}, { 29768,  32417,      0},
	{ 29262,  32791,      0}, { 28739,  33146,      0}, { 28198,  33481,      0}, { 27641,  33795,      0},
	{ 27066,  34089,      0}, { 26476,  34363,      0}, { 25869,  34616,      0}, { 25247,  34848,      0},
	{ 24609,  35060,      0}, { 23957,  35250,      0}, { 23290,  35418,      0}, { 22610,  35566,      0},
	{ 21915,  35692,      0}, { 21208,  35797,      0}, { 20488,  35880,      0}, { 19755,  35941,      0},
	{ 19010,  35981,      0}, { 18254,  35999,      0}, { 17487,  35995,      0}, { 16710,  35970,      0},
	{ 15922,  35923,      0}, { 15125,  35854,      0}, { 14319,  35764,      0}, { 13504,  35652,      0},
	{ 12681,  35519,      0}, { 11850,  35365,      0}, { 11013,  35189,      0}, { 10168,  34992,      0},
	{  9317,  34773,      0}, {  8461,  34534,      0}, {  7600,  34274,      0}, {  6734,  33994,      0},
	{  5864,  33693,      0}, {  4991,  33371,      0}, {  4114,  33030,      0}, {  3235,  32668,      0},
	{  2355,  32287,      0}, {  1472,  31887,      0}, {   589,  31467,      0}, {     0,  31323,    295},
	{     0,  31749,   1178}, {     0,  32156,   2061}, {     0,  32544,   2942}, {     0,  32912,   3822},
	{     0,  33260,   4699}, {     0,  33588,   5573}, {     0,  33896,   6445}, {     0,  34183,   7312},
	{     0,  34450,   8175}, {     0,  34696,   9033}, {     0,  34921,   9885}, {     0,  35125,  10732},
	{     0,  35308,  11572}, {     0,  35470,  12405}, {     0,  35610,  13231}, {     0,  35729,  14048},
	{     0,  35827,  14857}, {     0,  35902,  15658}, {     0,  35957,  16449}, {     0,  35989,  17229},
	{     0,  36000,  18000}, {     0,  35989,  18760}, {     0,  35957,  19508}, {     0,  35902,  20245},
	{     0,  35827,  20969}, {     0,  35729,  21681}, {     0,  35610,  22380}, {     0,  35470,  23065},
	{     0,  35308,  23736}, {     0,  35125,  24394}, {     0,  34921,  25036}, {     0,  34696,  25663},
	{     0,  34450,  26275}, {     0,  34183,  26871}, {     0,  33896,  27451}, {     0,  33588,  28014},
	{     0,  33260,  28561}, {     0,  32912,  29090}, {     0,  32544,  29602}, {     0,  32156,  30096},
	{     0,  31749,  30571}, {     0,  31323,  31029}, {     0,  30878,  31467}, {     0,  30415,  31887},
	{     0,  29933,  32287}, {     0,  29433,  32668}, {     0,  28915,  33030}, {     0,  28380,  33371},
	{     0,  27828,  33693}, {     0,  27260,  33994}, {     0,  26674,  34274}, {     0,  26073,  34534},
	{     0,  25456,  34773}, {     0,  24823,  34992}, {     0,  24176,  35189}, {     0,  23514,  35365},
	{     0,  22838,  35519}, {     0,  22148,  35652}, {     0,  21445,  35764}, {     0,  20729,  35854},
	{     0,  20001,  35923}, {     0,  19260,  35970}, {     0,  18508,  35995}, {     0,  17744,  35999},
	{     0,  16970,  35981}, {     0,  16186,  35941}, {     0,  15392,  35880}, {     0,  14589,  35797},
	{     0,  13777,  35692}, {     0,  12956,  35566}, {     0,  12128,  35418}, {     0,  11293,  35250},
	{     0,  10450,  35060}, {     0,   9602,  34848}, {     0,   8747,  34616}, {     0,   7888,  34363},
	{     0,   7023,  34089}, {     0,   6155,  33795}, {     0,   5282,  33481}, {     0,   4407,  33146},
	{     0,   3529,  32791}, {     0,   2648,  32417}, {     0,   1766,  32023}, {     0,    883,  31609}, };
