/*
 * drive.h
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

#ifndef DRIVE_H_
#define DRIVE_H_

#include <stdint.h>

#include "stm32f30x.h"
#include "ch.h"
#include "intercom_constants.h"
#include "util/flash_fs.h"
#include "arm_math.h"
#include "util/kf.h"
#include "util/luenberger.h"
#include "util/pid.h"

#define IN_CCM __attribute__((section(".ccmtext")))
#define IN_CCM_VAR __attribute__((section(".ccm")))

#define MOTOR_RESISTANCE	1.2f			// [Ohm]
#define MOTOR_KT			0.0255f			// [Nm/A]
#define MOTOR_KE			0.026222324720f // [V/rad]

#define DRIVE_EVENT_QUEUE_SIZE		4

#define DRIVE_ENC_SAMPLES 4
#define DRIVE_ADC_SAMPLES 16
#define DRIVE_MOVE_THRESHOLD 0.25f

#define DRIVE_COM_MODE_NONE			0
#define DRIVE_COM_MODE_BLOCK		1
#define DRIVE_COM_MODE_SINUSOIDAL	2

#define DRIVE_CTRL_MODE_NONE		0
#define DRIVE_CTRL_MODE_PID_SPEED	1
#define DRIVE_CTRL_MODE_PID_TORQUE	2

// W3  W2  W1
//  5   5   5
#define CCER_A 0x005
#define CCER_B 0x050
#define CCER_C 0x500
#define CCER_AB (CCER_A | CCER_B)
#define CCER_BC (CCER_B | CCER_C)
#define CCER_CA (CCER_C | CCER_A)
#define CCER_ABC (CCER_A | CCER_B | CCER_C)

typedef struct _Drive
{
	// ### Setup ###
	// Motor and encoder timer
	TIM_TypeDef* pTIMMotor;
	TIM_TypeDef* pTIMEncoder;

	const float currentMapping[3][2];

	// ADC calibration values
	uint32_t adcCal[2];

	// ### Measured values ###
	// ADC DMA storage area
	union
	{
		uint16_t adc[2][DRIVE_ADC_SAMPLES];
		uint32_t adc32[2][DRIVE_ADC_SAMPLES/2];
	};

	uint16_t encDtA[DRIVE_ENC_SAMPLES];
	uint8_t encIndex;
	uint16_t lastDir;

	float curVelocity;	// current velocity [rounds per second]
	float i[3];	// current on each phase [A]
	float iAlpha;	// current on stator-fixed alpha axis [A]
	float iBeta;	// current on stator-fixed beta axis [A]
	float rmsCurrent;

	uint32_t curVelocityTimestamp;
	uint32_t rmsCurrentTimestamp;

	uint16_t halPos;

	// ### control related ###
	struct _speed
	{
		PID pid;
		float setVelocity;
	} speed;

	struct _torque
	{
		PID pid;
		float setCurrent;
	} torque;

	float setVoltage;
	uint32_t setVoltageTimestamp;

	volatile uint8_t commutationModeBackup;
	volatile uint8_t commutationMode;
	volatile int32_t outLevel;	// [-10000 - 10000]

	volatile uint8_t ctrlMode;

	Mailbox eventQueue;
	msg_t eventQueueData[DRIVE_EVENT_QUEUE_SIZE];

	// ### State Estimation ###
	KF kf;
	float kfData[KF_DATA_SIZE(2, 1, 2, 2)];

	float estVelocity; // estimated velocity (from KF) [rounds per second]
	float estCurrent; // estimated motor current (from KF) [A]
	float estTorque; // estimated motor torque (from KF) [Nm]
	float estLoadTorque; // estimated load torque (from KF) [Nm]

	uint32_t estTimestamp;	// Timestamp of estimation update

	LuenbergerSISO2S lbObserver;

	// ### output/info values ###
	float u[3];

	// ### required input ###
	float curBat;	// [V]
} Drive;

extern Drive drive[2];
extern Drive* pRear;
extern Drive* pFront;

void DriveInit();
msg_t DriveTask(void* params);

// Low-level functions
void DriveEnable(Drive* pDrive, uint8_t enable);
void DriveSetCommutationMode(Drive* pDrive, uint8_t mode);
void DriveSetVoltage(Drive* pDrive, float voltage);

// Select controller
void DriveSetCtrlMode(Drive* pDrive, uint8_t mode);

// Speed controller
void DriveSpeedSetPIDParams(Drive* pDrive, float p, float i, float d);
void DriveSpeedSetVelocity(Drive* pDrive, float vel);

// Torque controller
void DriveTorqueSetPIDParams(Drive* pDrive, float p, float i, float d);
void DriveTorqueSetTorque(Drive* pDrive, float torque);

// SVPWM helper functions (mostly for tests)
void DriveSetPhaseVoltagesSVPWM(Drive* pDrive, float Va, float Vb, float Vc);
void DriveSetAlphaBeta(Drive* pDrive, float Valpha, float Vbeta);
void DriveSetVoltageAngle(Drive* pDrive, float Vref, float angle);

// State estimator functions - KF
void DriveSetStateUncertainty(Drive* pDrive, float noise);
void DriveSetSensorUncertainty(Drive* pDrive, float noise);

#endif /* DRIVE_H_ */
