/*
 * spi1.c
 *
 *  Created on: 15.01.2014
 *      Author: AndreR
 */

#include "spi1.h"

#include "ch.h"
#include <string.h>

#include "util/bit_banding.h"
#include "util/log.h"
#include "util/sys_time.h"
#include "util/boot.h"
#include "util/init_hal.h"
#include "util/cobs.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "motors.h"

SPI1Global spi1;

#define EVENT_RX_DONE		0
#define EVENT_TX_DATA_READY	1
#define EVENT_REGULAR		3

#define DMA_Channel_TX DMA1_Channel3
#define DMA_Channel_RX DMA1_Channel2

#define DMAClearTCIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CCR, 1)
#define DMASetTCIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CCR, 1)
#define DMAGetTCIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CCR, 1)

#define DMAClearTEIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CCR, 3)
#define DMASetTEIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CCR, 3)
#define DMAGetTEIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CCR, 3)

__attribute__((section(".ccmtext"))) void TIM1_BRK_TIM15_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM15->SR = 0;

	chSysLockFromIsr();
	chMBPostI(&spi1.eventQueue, EVENT_REGULAR);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

__attribute__((section(".ccmtext"))) void EXTI4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = (1 << 4);

	if(GPIOA->IDR & GPIO_PIN_4)	// pin high?
	{
		// => deselected
		SPI1->CR2 &= ~(SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN);

	    while(SPI1->SR & SPI_SR_FTLVL)
	    	asm volatile("nop");

	    while(SPI1->SR & SPI_SR_BSY)
	    	asm volatile("nop");

	    while(SPI1->SR & SPI_SR_FRLVL)
	    	SPI1->DR;

		SPI1->CR1 &= ~SPI_CR1_SPE;

		DMA_Channel_RX->CCR &= ~DMA_CCR_EN;
		DMA_Channel_TX->CCR &= ~DMA_CCR_EN;

		while(DMA_Channel_RX->CCR & DMA_CCR_EN);
		while(DMA_Channel_TX->CCR & DMA_CCR_EN);

		DMA1->IFCR = DMA_IFCR_CGIF2 | DMA_IFCR_CGIF3;

		SPI1->SR = 0;	// clear CRC flag
		SPI1->DR;
		SPI1->SR;

		chSysLockFromIsr();
		chMBPostAheadI(&spi1.eventQueue, EVENT_RX_DONE);
		chSysUnlockFromIsr();

		spi1.txData.bytesWritten = 0;

		spi1.deselectTime = SysTimeUSec();

		TIM15->CNT = 0;
		TIM15->CR1 |= TIM_CR1_CEN;
	}
	else
	{
		// => slave selected

		// re-enable transfer
		DMA_Channel_RX->CNDTR = sizeof(spi1.rxData)/2;
		DMA_Channel_TX->CNDTR = sizeof(spi1.txData)/2;
		DMA_Channel_RX->CCR |= DMA_CCR_EN;
		DMA_Channel_TX->CCR |= DMA_CCR_EN;

		SPI1->SR = 0;
		SPI1->CR2 |= SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;
		SPI1->CR1 |= SPI_CR1_SPE;
	}

	CH_IRQ_EPILOGUE();
}

void SPI1Init()
{
	// init event queue/variables
	chMBInit(&spi1.eventQueue, spi1.eventQueueData, SPI1_EVENT_QUEUE_SIZE);
	spi1.txData.bytesWritten = 0;

#ifdef DEBUG
	spi1.pTxDma = DMA1_Channel3;
	spi1.pRxDma = DMA1_Channel2;
	spi1.pSPI = SPI1;
	spi1.pDMA = DMA1;
#endif

	FifoLinInit(&spi1.txFifo, spi1.fifoDataTx, SPI1_FIFO_SIZE_TX);
	FifoLinInit(&spi1.rxFifo, spi1.fifoDataRx, SPI1_FIFO_SIZE_RX);

	RCC->APB2ENR |= RCC_APB2ENR_SPI1EN;

	// init GPIO
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F3_50MHZ;
	gpioInit.pupd = GPIO_PUPD_DOWN;
	GPIOInit(GPIOA, GPIO_PIN_5 | GPIO_PIN_6, &gpioInit);	// SCK, MISO
	GPIOInit(GPIOB, GPIO_PIN_5, &gpioInit);	// MOSI

	// init SPI1
	SPI1->CR1 = SPI_CR1_SSM;	// prescaler 2, full duplex, slave
	SPI1->CR2 = (0xF << 8);	// data size 16Bit

	SPI1->SR = 0;	// clear CRC flag

	// init DMA
	// TX
	DMA_Channel_TX->CCR = DMA_PL_LOW | DMA_DIR_MEM2PERIPH | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_CCR_MINC;
	DMA_Channel_TX->CNDTR = sizeof(spi1.txData)/2;
	DMA_Channel_TX->CMAR = (uint32_t)&spi1.txData;
	DMA_Channel_TX->CPAR = (uint32_t)&SPI1->DR;

	// RX
	DMA_Channel_RX->CCR = DMA_PL_LOW | DMA_DIR_PERIPH2MEM | DMA_MSIZE_2BYTE | DMA_PSIZE_2BYTE | DMA_CCR_MINC;
	DMA_Channel_RX->CNDTR = sizeof(spi1.rxData)/2;
	DMA_Channel_RX->CMAR = (uint32_t)&spi1.rxData;
	DMA_Channel_RX->CPAR = (uint32_t)&SPI1->DR;

	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_RISING_FALLING;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_4, &gpioInit);

	// Init precision timer for motor regular intercom
	RCC->APB2ENR |= RCC_APB2ENR_TIM15EN;

	TIM15->CR1 = TIM_CR1_OPM;
	TIM15->CR2 = 0;
	TIM15->PSC = 71;	// 1Mhz
	TIM15->ARR = 299;	// 300us
	TIM15->EGR = TIM_EGR_UG;
	TIM15->DIER = TIM_DIER_UIE;

	NVICEnableIRQ(EXTI4_IRQn, IRQL_SPI1);
	NVICEnableIRQ(TIM1_BRK_TIM15_IRQn, IRQL_SPI1);
}

void SPI1TxReady(void* params)
{
	(void)params;

	chMBPost(&spi1.eventQueue, EVENT_TX_DATA_READY, TIME_IMMEDIATE);
}

msg_t SPI1Task(void* params)
{
	(void)params;
	msg_t event;

	chRegSetThreadName("SPI1");

	while(1)
	{
		if(chMBFetch(&spi1.eventQueue, &event, MS2ST(500)) != RDY_OK)
			continue;

		switch(event)
		{
			case EVENT_RX_DONE:
			{
				// Decode data
				uint16_t bytesToRead = spi1.rxData.bytesWritten;
				uint8_t* pFifoData;
				uint32_t bytesWritten;
				uint32_t pktSize = 0;

				if(bytesToRead > MOT_EXCHANGE_DATA_MSG_SIZE)
				{
					LogWarnC("bytesToRead > blockSize", bytesToRead);
					bytesToRead = MOT_EXCHANGE_DATA_MSG_SIZE;
				}

				for(uint16_t bytesRead = 0; bytesRead < bytesToRead; bytesRead += pktSize+1)
				{
					uint8_t* pData = spi1.rxData.blockData+bytesRead;
					uint8_t* pPktEnd = memchr(pData, 0, bytesToRead-bytesRead);

					if(pPktEnd == 0)
						break;

					pktSize = pPktEnd-pData;

					if(FifoLinReserve(&spi1.rxFifo, SPI1_RESERVE_SIZE, &pFifoData) != 0)
					{
						LogWarnC("Data loss", pktSize);
						break;
					}

					if(COBSDecode(pData, pktSize, pFifoData, SPI1_RESERVE_SIZE, &bytesWritten) != 0)
					{
						LogWarnC("COBSDecode error", pktSize);
						break;
					}

					FifoLinCommit(&spi1.rxFifo, bytesWritten);
				}

				if(spi1.rxFifo.numPackets && spi1.callback)
					(*spi1.callback)(spi1.pCbData);

				chMBPost(&spi1.eventQueue, EVENT_TX_DATA_READY, TIME_IMMEDIATE);
			}
			break;
			case EVENT_TX_DATA_READY:
			{
				uint8_t* pTxData;
				uint32_t pktSize;
				uint32_t bytesWritten;

				while(FifoLinGet(&spi1.txFifo, &pTxData, &pktSize) == 0)
				{
					if(SysTimeUSec()-spi1.deselectTime > 400)
						break;

					uint8_t* pOutData = spi1.txData.blockData+spi1.txData.bytesWritten;
					uint16_t sizeOut = sizeof(spi1.txData.blockData)-spi1.txData.bytesWritten-1;	// -1 to always accommodate for trailing 0

					if(COBSEncode(pTxData, pktSize, pOutData, sizeOut, &bytesWritten) != 0)
						break;

					if(spi1.txData.bytesWritten+bytesWritten > sizeof(spi1.txData.blockData))
					{
						LogWarn("Encoded data + bytesUsed > blockData");
						break;
					}

					pOutData += bytesWritten;
					++bytesWritten;
					*pOutData = 0;

					spi1.txData.bytesWritten += bytesWritten;

					FifoLinDelete(&spi1.txFifo);
				}
			}
			break;
			case EVENT_REGULAR:
			{
				// get motor data for next transfer
				MotorsRegularIntercom();
			}
			break;
		}
	}

	return 0;
}
