/*
 * fault_handler.c
 *
 *  Created on: 12.07.2010
 *      Author: rYan
 */

#ifdef STM32F4XX
#include "stm32f407xx.h"
#include <stdio.h>
#include <stdarg.h>
#else
#include "stm32f30x.h"
#endif
#include "core_cm4.h"

#pragma GCC optimize ("O0")

void NMIVector() __attribute__((naked));
void HardFaultVector() __attribute__((naked));
void MemManageVector() __attribute__((naked));
void BusFaultVector() __attribute__((naked));
void UsageFaultVector() __attribute__((naked));
void DebugMonitorVector() __attribute__((naked));

typedef struct _SCBDebug
{
	struct _MMFSR
	{
		uint8_t IACCVIOL;
		uint8_t DACCVIOL;
		uint8_t MUNSTKERR;
		uint8_t MSTKERR;
		uint8_t MLSPERR;
		uint8_t MMARVALID;
	} MMFSR;

	struct _BFSR
	{
		uint8_t IBUSERR;
		uint8_t PRECISERR;
		uint8_t IMPRECISERR;
		uint8_t UNSTKERR;
		uint8_t STKERR;
		uint8_t LSPERR;
		uint8_t BFARVALID;
	} BFSR;

	struct _UFSR
	{
		uint8_t UNDEFINSTR;
		uint8_t INVSTATE;
		uint8_t INVPC;
		uint8_t NOCP;
		uint8_t UNALIGNED;
		uint8_t DIVBYZERO;
	} UFSR;

	struct _HFSR
	{
		uint8_t VECTTBL;
		uint8_t FORCED;
	} HFSR;

	uint32_t MMFAR;
	uint32_t BFAR;
	uint32_t AFSR;
} SCBDebug;

static inline SCBDebug getSCBDebug()
{
	uint32_t cfsr = SCB->CFSR;
	uint32_t hfsr = SCB->HFSR;
	SCBDebug res;

	res.MMFSR.IACCVIOL = (cfsr >> 0) & 0x01;
	res.MMFSR.DACCVIOL = (cfsr >> 1) & 0x01;
	res.MMFSR.MUNSTKERR = (cfsr >> 3) & 0x01;
	res.MMFSR.MSTKERR = (cfsr >> 4) & 0x01;
	res.MMFSR.MLSPERR = (cfsr >> 5) & 0x01;
	res.MMFSR.MMARVALID = (cfsr >> 7) & 0x01;

	res.BFSR.IBUSERR = (cfsr >> 8) & 0x01;
	res.BFSR.PRECISERR = (cfsr >> 9) & 0x01;
	res.BFSR.IMPRECISERR = (cfsr >> 10) & 0x01;
	res.BFSR.UNSTKERR = (cfsr >> 11) & 0x01;
	res.BFSR.STKERR = (cfsr >> 12) & 0x01;
	res.BFSR.LSPERR = (cfsr >> 13) & 0x01;
	res.BFSR.BFARVALID = (cfsr >> 15) & 0x01;

	res.UFSR.UNDEFINSTR = (cfsr >> 16) & 0x01;
	res.UFSR.INVSTATE = (cfsr >> 17) & 0x01;
	res.UFSR.INVPC = (cfsr >> 18) & 0x01;
	res.UFSR.NOCP = (cfsr >> 19) & 0x01;
	res.UFSR.UNALIGNED = (cfsr >> 24) & 0x01;
	res.UFSR.DIVBYZERO = (cfsr >> 25) & 0x01;

	res.HFSR.VECTTBL = (hfsr >> 1) & 0x01;
	res.HFSR.FORCED = (hfsr >> 30) & 0x01;

	res.MMFAR = SCB->MMFAR;
	res.BFAR = SCB->BFAR;
	res.AFSR = SCB->AFSR;

	return res;
}

#ifdef STM32F4XX
static void initFaultPrint()
{
	USART1->CR1 &= ~(USART_CR1_TXEIE | USART_CR1_RXNEIE);	// disable IRQs
}

static int32_t faultPrint(const char* fmt, ...)
{
	va_list args;
	int32_t bytesWritten;
	static char printBuf[128];

	va_start(args, fmt);
	bytesWritten = vsnprintf(printBuf, 128, fmt, args);
	va_end(args);

	for(int32_t i = 0; i < bytesWritten; i++)
	{
		while(!(USART1->SR & USART_SR_TXE))
			asm volatile("nop");

		USART1->DR = printBuf[i];
	}

	return bytesWritten;
}
#endif

void faultHandler(unsigned int* args) __attribute__((used));

void faultHandler(unsigned int* args)
{
	unsigned int r0 = args[0];
	unsigned int r1 = args[1];
	unsigned int r2 = args[2];
	unsigned int r3 = args[3];
	unsigned int r12 = args[4];
	unsigned int lr = args[5];
	unsigned int pc = args[6];
	unsigned int psr = args[7];

	SCBDebug cfsr = getSCBDebug();

#ifdef STM32F30X
	TIM8->CR1 &= ~TIM_CR1_CEN;
	TIM1->CR1 &= ~TIM_CR1_CEN;
#endif

#ifdef STM32F4XX
	initFaultPrint();

	while(1)
	{
		faultPrint("--- System Fault ---\r\n");
		faultPrint("R0:  0x%08X\r\n", r0);
		faultPrint("R1:  0x%08X\r\n", r1);
		faultPrint("R2:  0x%08X\r\n", r2);
		faultPrint("R3:  0x%08X\r\n", r3);
		faultPrint("R12: 0x%08X\r\n", r12);
		faultPrint("LR:  0x%08X\r\n", lr);
		faultPrint("PC:  0x%08X\r\n", pc);
		faultPrint("PSR: 0x%08X\r\n\r\n", psr);

		faultPrint("MMFAR: 0x%08X\r\n", cfsr.MMFAR);
		faultPrint("BFAR:  0x%08X\r\n", cfsr.BFAR);
		faultPrint("AFSR:  0x%08X\r\n\r\n", cfsr.AFSR);

		faultPrint("- MMFSR -\r\n");
		faultPrint("IACCVIOL:    0x%02hX\r\n", (uint16_t)cfsr.MMFSR.IACCVIOL);
		faultPrint("DACCVIOL:    0x%02hX\r\n", (uint16_t)cfsr.MMFSR.DACCVIOL);
		faultPrint("MUNSTKERR:   0x%02hX\r\n", (uint16_t)cfsr.MMFSR.MUNSTKERR);
		faultPrint("MSTKERR:     0x%02hX\r\n", (uint16_t)cfsr.MMFSR.MSTKERR);
		faultPrint("MLSPERR:     0x%02hX\r\n", (uint16_t)cfsr.MMFSR.MLSPERR);
		faultPrint("MMFARVALID:  0x%02hX\r\n\r\n", (uint16_t)cfsr.MMFSR.MMARVALID);

		faultPrint("- BFSR -\r\n");
		faultPrint("IBUSERR:     0x%02hX\r\n", (uint16_t)cfsr.BFSR.IBUSERR);
		faultPrint("PRECISERR:   0x%02hX\r\n", (uint16_t)cfsr.BFSR.PRECISERR);
		faultPrint("IMPRECISERR: 0x%02hX\r\n", (uint16_t)cfsr.BFSR.IMPRECISERR);
		faultPrint("UNSTKERR:    0x%02hX\r\n", (uint16_t)cfsr.BFSR.UNSTKERR);
		faultPrint("STKERR:      0x%02hX\r\n", (uint16_t)cfsr.BFSR.STKERR);
		faultPrint("LSPERR:      0x%02hX\r\n", (uint16_t)cfsr.BFSR.LSPERR);
		faultPrint("BFARVALID:   0x%02hX\r\n\r\n", (uint16_t)cfsr.BFSR.BFARVALID);

		faultPrint("- UFSR -\r\n");
		faultPrint("UNDEFINSTR:  0x%02hX\r\n", (uint16_t)cfsr.UFSR.UNDEFINSTR);
		faultPrint("INVSTATE:    0x%02hX\r\n", (uint16_t)cfsr.UFSR.INVSTATE);
		faultPrint("INVPC:       0x%02hX\r\n", (uint16_t)cfsr.UFSR.INVPC);
		faultPrint("NOCP:        0x%02hX\r\n", (uint16_t)cfsr.UFSR.NOCP);
		faultPrint("UNALIGNED:   0x%02hX\r\n", (uint16_t)cfsr.UFSR.UNALIGNED);
		faultPrint("DIVBYZERO:   0x%02hX\r\n\r\n", (uint16_t)cfsr.UFSR.DIVBYZERO);

		faultPrint("- HFSR -\r\n");
		faultPrint("VECTTBL:    0x%02hX\r\n", (uint16_t)cfsr.HFSR.VECTTBL);
		faultPrint("FORCED:     0x%02hX\r\n\r\n", (uint16_t)cfsr.HFSR.FORCED);

		faultPrint("Dump complete. System execution stopped...\r\n");

		while(!(USART1->SR & USART_SR_RXNE))
		{
			asm volatile("nop");
		}
		USART1->DR;
	}
#endif

	asm volatile("BKPT #01");
	while(1);

	(void)r0;
	(void)r1;
	(void)r2;
	(void)r3;
	(void)r12;
	(void)lr;
	(void)pc;
	(void)psr;
	(void)cfsr;
}

void NMIVector()
{
	asm volatile("TST LR, #4");
	asm volatile("ITE EQ");
	asm volatile("MRSEQ R0, MSP");
	asm volatile("MRSNE R0, PSP");
	asm volatile("B faultHandler");

//	SCB->AIRCR = 0x05FA0000 | (SCB->AIRCR & 0x700) | 0x4;	// System Reset, Keep priority grouping
}

void HardFaultVector()
{
	asm volatile("TST LR, #4");
	asm volatile("ITE EQ");
	asm volatile("MRSEQ R0, MSP");
	asm volatile("MRSNE R0, PSP");
	asm volatile("B faultHandler");
}

void MemManageVector()
{
	asm volatile("TST LR, #4");
	asm volatile("ITE EQ");
	asm volatile("MRSEQ R0, MSP");
	asm volatile("MRSNE R0, PSP");
	asm volatile("B faultHandler");
}

void BusFaultVector()
{
	asm volatile("TST LR, #4");
	asm volatile("ITE EQ");
	asm volatile("MRSEQ R0, MSP");
	asm volatile("MRSNE R0, PSP");
	asm volatile("B faultHandler");
}

void UsageFaultVector()
{
	asm volatile("TST LR, #4");
	asm volatile("ITE EQ");
	asm volatile("MRSEQ R0, MSP");
	asm volatile("MRSNE R0, PSP");
	asm volatile("B faultHandler");
}

void DebugMonitorVector()
{
	asm volatile("TST LR, #4");
	asm volatile("ITE EQ");
	asm volatile("MRSEQ R0, MSP");
	asm volatile("MRSNE R0, PSP");
	asm volatile("B faultHandler");
}
