/*
 * flash.c
 *
 *  Created on: 12.06.2014
 *      Author: AndreR
 */

#include "util/flash.h"
#include "errors.h"
#include <string.h>

#include "stm32f30x.h"

#include "ch.h"

static int16_t waitForLastOperation()
{
	do
	{
		asm volatile("nop");
	}
	while(FLASH->SR & FLASH_SR_BSY);

	uint32_t status = FLASH->SR & (FLASH_SR_PGERR | FLASH_SR_WRPERR | FLASH_SR_EOP);

	FLASH->SR = status;	// clear flags

	if(status & FLASH_SR_WRPERR)
		return ERROR_FLASH_WRITE_PROTECT;

	if(status & FLASH_SR_PGERR)
		return ERROR_FLASH_NOT_ERASED;

	return 0;
}

static int16_t erasePage(uint32_t pageAddress)
{
	uint16_t result;

	// Wait for last operation to be completed
	result = waitForLastOperation();

	if(result == 0)
	{
		// If the previous operation is completed, proceed to erase the page
		FLASH->CR |= FLASH_CR_PER;
		FLASH->AR = pageAddress;
		FLASH->CR |= FLASH_CR_STRT;

		// Wait for last operation to be completed
		result = waitForLastOperation();

		// Disable the PER Bit
		FLASH->CR &= ~FLASH_CR_PER;
	}

	return result;
}

int16_t FlashErase(uint32_t addr, uint32_t endAddr)
{
	int16_t result = 0;
	addr &= 0xFFFFF800;

	// unlock flash
	FLASH->KEYR = 0x45670123;
	FLASH->KEYR = 0xCDEF89AB;

	for(; addr < endAddr; addr += 0x800)
	{
		result = erasePage(addr);
		if(result)
			break;
	}

	// lock flash
	FLASH->CR |= FLASH_CR_LOCK;

	return result;
}

int16_t FlashProgram(uint32_t addr,  uint32_t* pData, uint32_t length)
{
	int16_t result = 0;

	if(memcmp((void*)addr, pData, length*sizeof(uint32_t)) == 0)
		return 0;

	// unlock flash
	FLASH->KEYR = 0x45670123;
	FLASH->KEYR = 0xCDEF89AB;

	result = waitForLastOperation();
	if(result)
		return result;

	for(uint32_t i = 0; i < length; i++)
	{
		// select flash programming
		FLASH->CR |= FLASH_CR_PG;

		// program low half-word
		*(volatile uint16_t*)addr = (uint16_t)pData[i];

		result = waitForLastOperation();
		if(result)
			break;

		addr += 2;

		// program high half-word
		*(volatile uint16_t*)addr = (uint16_t)(pData[i] >> 16);

		result = waitForLastOperation();
		if(result)
			break;

		addr += 2;

		// clear flash programming bit
		FLASH->CR &= ~FLASH_CR_PG;
	}

	// lock flash
	FLASH->CR |= FLASH_CR_LOCK;

	return result;
}
