/*
 * flash.c
 *
 *  Created on: 12.06.2014
 *      Author: AndreR
 */

#include "util/flash.h"
#include "errors.h"
#include "stm32f407xx.h"
#include "ch.h"
#include <string.h>

#define FLASH_Sector_0     ((uint16_t)0x0000) /*!< Sector Number 0 */
#define FLASH_Sector_1     ((uint16_t)0x0008) /*!< Sector Number 1 */
#define FLASH_Sector_2     ((uint16_t)0x0010) /*!< Sector Number 2 */
#define FLASH_Sector_3     ((uint16_t)0x0018) /*!< Sector Number 3 */
#define FLASH_Sector_4     ((uint16_t)0x0020) /*!< Sector Number 4 */
#define FLASH_Sector_5     ((uint16_t)0x0028) /*!< Sector Number 5 */
#define FLASH_Sector_6     ((uint16_t)0x0030) /*!< Sector Number 6 */
#define FLASH_Sector_7     ((uint16_t)0x0038) /*!< Sector Number 7 */
#define FLASH_Sector_8     ((uint16_t)0x0040) /*!< Sector Number 8 */
#define FLASH_Sector_9     ((uint16_t)0x0048) /*!< Sector Number 9 */
#define FLASH_Sector_10    ((uint16_t)0x0050) /*!< Sector Number 10 */
#define FLASH_Sector_11    ((uint16_t)0x0058) /*!< Sector Number 11 */

#define SECTOR_MASK ((uint32_t)0xFFFFFF07)

static int16_t waitForLastOperation()
{
	do
	{
		asm volatile("nop");
	}
	while(FLASH->SR & FLASH_SR_BSY);

	uint32_t status = FLASH->SR & (FLASH_SR_PGAERR | FLASH_SR_PGPERR | FLASH_SR_PGSERR | FLASH_SR_WRPERR | FLASH_SR_SOP);

	FLASH->SR = status;	// clear flags

	if(status & (FLASH_SR_PGAERR | FLASH_SR_PGPERR | FLASH_SR_PGSERR))
		return ERROR_FLASH_INVALID_OPERATION;

	if(status & FLASH_SR_WRPERR)
		return ERROR_FLASH_WRITE_PROTECT;

	return 0;
}

static int16_t eraseSector(uint32_t sector)
{
	int16_t result;

	// Wait for last operation to be completed
	result = waitForLastOperation();

	if(result == 0)
	{
		// if the previous operation is completed, proceed to erase the sector
		FLASH->CR &= FLASH_CR_PSIZE;
		FLASH->CR |= FLASH_CR_PSIZE_1;	// word size
		FLASH->CR &= SECTOR_MASK;
		FLASH->CR |= FLASH_CR_SER | sector;
		FLASH->CR |= FLASH_CR_STRT;

		// Wait for last operation to be completed
		result = waitForLastOperation();

		// if the erase operation is completed, disable the SER Bit
		FLASH->CR &= (~FLASH_CR_SER);
		FLASH->CR &= SECTOR_MASK;
	}

	return result;
}

int16_t FlashErase(uint32_t addr, uint32_t endAddr)
{
	int16_t result = 0;

	// unlock flash
	FLASH->KEYR = 0x45670123;
	FLASH->KEYR = 0xCDEF89AB;

	// Clear pending flags (if any)
	FLASH->SR = 0xF3;

	do
	{
		if(addr < 0x08004000)
			result = eraseSector(FLASH_Sector_0);	// 16k

		if(result)
			break;

		if(addr < 0x08008000 && endAddr > 0x08004000)
			result = eraseSector(FLASH_Sector_1);	// 16k

		if(result)
			break;

		if(addr < 0x0800C000 && endAddr > 0x08008000)
			result = eraseSector(FLASH_Sector_2);	// 16k

		if(result)
			break;

		if(addr < 0x08010000 && endAddr > 0x0800C000)
			result = eraseSector(FLASH_Sector_3);	// 16k

		if(result)
			break;

		if(addr < 0x08020000 && endAddr > 0x08010000)
			result = eraseSector(FLASH_Sector_4);	// 64k

		if(result)
			break;

		if(addr < 0x08040000 && endAddr > 0x08020000)
			result = eraseSector(FLASH_Sector_5);	// 128k

		if(result)
			break;

		if(addr < 0x08060000 && endAddr > 0x08040000)
			result = eraseSector(FLASH_Sector_6);	// 128k

		if(result)
			break;

		if(addr < 0x08080000 && endAddr > 0x08060000)
			result = eraseSector(FLASH_Sector_7);	// 128k
	}while(0);

	FLASH->CR |= FLASH_CR_LOCK;

	return result;
}

int16_t FlashProgram(uint32_t addr,  uint32_t* pData, uint32_t length)
{
	int16_t result;

	if(memcmp((void*)addr, pData, length*sizeof(uint32_t)) == 0)
		return 0;

	// unlock flash
	FLASH->KEYR = 0x45670123;
	FLASH->KEYR = 0xCDEF89AB;

	waitForLastOperation();

	// Clear pending flags (if any)
	FLASH->SR = 0xF3;

	for(uint32_t i = 0; i < length; i++)
	{
	    // if the previous operation is completed, proceed to program the new data
	    FLASH->CR &= FLASH_CR_PSIZE;
	    FLASH->CR |= FLASH_CR_PSIZE_1;
	    FLASH->CR |= FLASH_CR_PG;

	    *(volatile uint32_t*)addr = pData[i];
		addr += 4;

	    /* Wait for last operation to be completed */
	    result = waitForLastOperation();
	    if(result)
	    	break;

	    /* if the program operation is completed, disable the PG Bit */
	    FLASH->CR &= (~FLASH_CR_PG);
	}

	FLASH->CR |= FLASH_CR_LOCK;

	return result;
}
