/*
 * spi_sync.c
 *
 *  Created on: 28.09.2013
 *      Author: AndreR
 */

#include "spi_sync.h"
#include "errors.h"
#include "util/bit_banding.h"
#include "util/init_hal.h"
#include "util/log.h"

#define SPIClearSPE(pSPI)	BBPeriphReset(pSPI->base, 6)
#define SPISetSPE(pSPI)		BBPeriphSet(pSPI->base, 6)
#define SPIGetSPE(pSPI)		BBPeriphGet(pSPI->base, 6)

#define DMAClearEn(pDMA)	BBPeriphReset((uint32_t)&pDMA->CR, 0)
#define DMASetEn(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 0)
#define DMAGetEn(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 0)

#define DMAClearTCI(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 4)
#define DMASetTCI(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 4)
#define DMAGetTCI(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 4)

#define DMAClearFlags(pDMA, ch) BBPeriphSet((uint32_t)&pDMA->IFCR, (ch-1)*4)

static void stopTransfer(SPISync* pSPI)
{
	pSPI->data.pCSPort->BSRRL = pSPI->data.csPin;

	DMAClearEn(pSPI->pDmaRx);
	DMAClearEn(pSPI->pDmaTx);
	*pSPI->pDMARxIFCR |= pSPI->IFCRRxMask; // clear rx flags
	*pSPI->pDMATxIFCR |= pSPI->IFCRTxMask; // clear tx flags
}

// IRQ -> no task switch can occur
void SPISyncDmaRxInterrupt(SPISync* pSPI)
{
	CH_IRQ_PROLOGUE();

	stopTransfer(pSPI);

	chSysLockFromIsr();
	chBSemSignalI(&pSPI->transferSem);
	chSysUnlockFromIsr();

	CH_IRQ_EPILOGUE();
}

void SPISyncInit(SPISync* pSPI, SPISyncData* pInit)
{
	pSPI->data = *pInit;

	chBSemInit(&pSPI->transferSem, FALSE);
	chMtxInit(&pSPI->taskMutex);

	pSPI->pDmaRx = DMAGenerateStreamTypeDef(pSPI->data.pDma, pSPI->data.dmaChRx);
	pSPI->pDmaTx = DMAGenerateStreamTypeDef(pSPI->data.pDma, pSPI->data.dmaChTx);
	pSPI->pDMARxIFCR = DMAGenerateIFCR(pSPI->data.pDma, pSPI->data.dmaChRx);
	pSPI->pDMATxIFCR = DMAGenerateIFCR(pSPI->data.pDma, pSPI->data.dmaChTx);
	pSPI->IFCRRxMask = DMAGenerateMaskAll(pSPI->data.dmaChRx);
	pSPI->IFCRTxMask = DMAGenerateMaskAll(pSPI->data.dmaChTx);

	// Select correct channel
	if(pInit->pRegister == SPI1) // DMA2 Channel3
	{
		pSPI->pDmaRx->CR = DMA_SxCR_CHSEL_0 | DMA_SxCR_CHSEL_1;
		pSPI->pDmaTx->CR = DMA_SxCR_CHSEL_0 | DMA_SxCR_CHSEL_1;
	}
	if(pInit->pRegister == SPI2 || pInit->pRegister == SPI3) // DMA1 Channel0
	{
		pSPI->pDmaRx->CR = 0;
		pSPI->pDmaTx->CR = 0;
	}

	// configure DMA
	// Medium Priority, Memory Increment, Memory-to-peripheral
	pSPI->pDmaTx->CR |= (pInit->dmaPrio << 16) | DMA_SxCR_MINC | DMA_SxCR_DIR_0;
	pSPI->pDmaTx->FCR = DMA_SxFCR_FTH_0 | DMA_SxFCR_DMDIS | DMA_SxFCR_FEIE; // FIFO 1/2 threshold, direct mode disabled, FIFO error IRQ enabled
	pSPI->pDmaTx->NDTR = 0;	// Number of data to transfer
	pSPI->pDmaTx->PAR = (uint32_t)&pSPI->data.pRegister->DR;	// peripheral address
	pSPI->pDmaTx->M0AR = 0;	// configured on transfer

	// Medium Priority, Memory Increment, Peripheral-to-Memory, Transfer complete  interrupt enable
	pSPI->pDmaRx->CR |= (pInit->dmaPrio << 16) | DMA_SxCR_MINC | DMA_SxCR_TCIE;
	pSPI->pDmaRx->FCR = DMA_SxFCR_FTH_0 | DMA_SxFCR_DMDIS | DMA_SxFCR_FEIE; // FIFO 1/2 threshold, direct mode disabled, FIFO error IRQ enabled
	pSPI->pDmaRx->NDTR = 0;
	pSPI->pDmaRx->PAR = (uint32_t)&pSPI->data.pRegister->DR;
	pSPI->pDmaRx->M0AR = 0; // configured on transfer

	pSPI->data.pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI;
	pSPI->data.pRegister->CR2 = SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;

	*pSPI->pDMARxIFCR |= pSPI->IFCRRxMask; // clear rx flags
	*pSPI->pDMATxIFCR |= pSPI->IFCRTxMask; // clear tx flags

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_OUTPUT;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_F4_2MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(pSPI->data.pCSPort, pSPI->data.csPin, &gpioInit);

	pSPI->data.pCSPort->BSRRL = pSPI->data.csPin;
}

void SPISyncStartTransfer(SPISync* pSPI, uint8_t* pSrc, uint8_t* pDst, uint32_t size)
{
	// Block multiple tasks entering this function
	chMtxLock(&pSPI->taskMutex);

	// Clear semaphore which indicates transfer is complete
	chBSemWaitTimeout(&pSPI->transferSem, TIME_IMMEDIATE);

	// select slave
	pSPI->data.pCSPort->BSRRH = pSPI->data.csPin;

	// configure SPI
	pSPI->data.pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_SPE;
	pSPI->data.pRegister->CR1 |=  pSPI->data.prescaler | pSPI->data.cpol << 1 | pSPI->data.cpha;

	// configure DMA
	if(pSrc == 0)
	{
		pSPI->pDmaTx->CR &= ~DMA_SxCR_MINC;
		pSPI->pDmaTx->M0AR = (uint32_t)&pSPI->dummy;
	}
	else
	{
		pSPI->pDmaTx->CR |= DMA_SxCR_MINC;
		pSPI->pDmaTx->M0AR = (uint32_t)pSrc;
	}

	if(pDst == 0)
	{
		pSPI->pDmaRx->CR &= ~DMA_SxCR_MINC;
		pSPI->pDmaRx->M0AR = (uint32_t)&pSPI->dummy;
	}
	else
	{
		pSPI->pDmaRx->CR |= DMA_SxCR_MINC;
		pSPI->pDmaRx->M0AR = (uint32_t)pDst;
	}

	pSPI->pDmaTx->NDTR = size;
	pSPI->pDmaRx->NDTR = size;

	// enable DMA
	DMASetEn(pSPI->pDmaRx);
	DMASetEn(pSPI->pDmaTx);
}

int16_t SPISyncWait(SPISync* pSPI)
{
	// check if wait is called from the task which called start transfer
	if(pSPI->taskMutex.m_owner != chThdSelf())
		return ERROR_SPI_SYNC_INVALID_TASK;

	// wait for transfer complete semaphore
	if(chBSemWaitTimeout(&pSPI->transferSem, pSPI->data.timeoutTicks) != RDY_OK)
	{
		stopTransfer(pSPI);

		chBSemReset(&pSPI->transferSem, FALSE);

		chMtxUnlock();

		LogError("SPISyncTimeout");

		return ERROR_SPI_SYNC_TIMEOUT;
	}
	chBSemReset(&pSPI->transferSem, FALSE);	// TODO: <- required?

	// this task is done
	chMtxUnlock();	// taskMutex

	return 0;
}

int16_t SPISyncTransfer(SPISync* pSPI, uint8_t* pSrc, uint8_t* pDst, uint32_t size)
{
	SPISyncStartTransfer(pSPI, pSrc, pDst, size);

	return SPISyncWait(pSPI);
}
