/*
 * usart.c
 *
 *  Created on: 04.11.2012
 *      Author: Andre
 */

#include "util/usart.h"

#include "util/log.h"
#include "util/bit_banding.h"
#include "util/init_hal.h"
#include "errors.h"

#include <string.h>

// uses DMA_StreamTypeDef
#define DMAClearEn(pDMA)	BBPeriphReset((uint32_t)&pDMA->CR, 0)
#define DMASetEn(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 0)
#define DMAGetEn(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 0)

#define DMAClearTCIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 4)
#define DMASetTCIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CR, 4)
#define DMAGetTCIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CR, 4)

#define DMAClearTEIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 2)
#define DMASetTEIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->CR, 2)
#define DMAGetTEIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->CR, 2)

#define DMAClearFEIE(pDMA) 	BBPeriphReset((uint32_t)&pDMA->FCR, 7)
#define DMASetFEIE(pDMA)	BBPeriphSet((uint32_t)&pDMA->FCR, 7)
#define DMAGetFEIE(pDMA)	BBPeriphGet((uint32_t)&pDMA->FCR, 7)

// uses USART_TypeDef
#define USARTClearEIE(pRegister)	BBPeriphReset((uint32_t)&pRegister->CR3, 0)
#define USARTSetEIE(pRegister)		BBPeriphSet((uint32_t)&pRegister->CR3, 0)
#define USARTGetEIE(pRegister)		BBPeriphGet((uint32_t)&pRegister->CR3, 0)

#define USARTClearTCIE(pRegister)	BBPeriphReset((uint32_t)&pRegister->CR1, 6)
#define USARTSetTCIE(pRegister)		BBPeriphSet((uint32_t)&pRegister->CR1, 6)
#define USARTGetTCIE(pRegister)		BBPeriphGet((uint32_t)&pRegister->CR1, 6)

#define USARTClearTE(pRegister)		BBPeriphReset((uint32_t)&pRegister->CR1, 3)
#define USARTSetTE(pRegister)		BBPeriphSet((uint32_t)&pRegister->CR1, 3)

#define USARTClearRE(pRegister)		BBPeriphReset((uint32_t)&pRegister->CR1, 2)
#define USARTSetRE(pRegister)		BBPeriphSet((uint32_t)&pRegister->CR1, 2)

#define STATE_IDLE				1
#define STATE_ACTIVE			2
#define STATE_WAIT_START		4

#define EVENT_TX_DONE			0x0001
#define EVENT_TX_ERROR			0x0002
#define EVENT_TX_START			0x0004
#define EVENT_TX_TIMER_DONE		0x0008
#define EVENT_TX_TIMEOUT		0x0010
#define EVENT_RX_ERROR_ORE		0x0101
#define EVENT_RX_ERROR_NE		0x0102
#define EVENT_RX_FE				0x0104
#define EVENT_RX_START			0x0108
#define EVENT_RX_TIMER_DONE		0x0110

#define EVENT_IS_RX(e) (e & 0x0100)

static inline void stopTx(USART* pUSART)
{
	USARTClearTCIE(pUSART->pRegister);	// disable TC IRQ on USART
	DMAClearEn(pUSART->pTxStream);	// disable DMA
	while(pUSART->pTxStream->CR & DMA_SxCR_EN);	// wait until stream is disabled
	USARTClearTE(pUSART->pRegister);	// disable transmitter
}

static inline void stopRx(USART* pUSART)
{
	USARTClearEIE(pUSART->pRegister);	// disable NE/FE/ORE IRQ on USART
	DMAClearEn(pUSART->pRxStream);	// disable DMA
	while(pUSART->pRxStream->CR & DMA_SxCR_EN);	// wait until stream is disabled
	USARTClearRE(pUSART->pRegister);	// disable transmitter
}


void USARTIRQ(USART* pUSART)
{
	CH_IRQ_PROLOGUE();

	uint16_t sr = pUSART->pRegister->SR;

	msg_t event = 0;

	if((sr & USART_SR_TC) && USARTGetTCIE(pUSART->pRegister))	// transfer complete
	{
		USARTClearTCIE(pUSART->pRegister);	// prevent IRQ refiring due to TC

		chSysLockFromIsr();
		if(chMBPostI(&pUSART->eventQueue, EVENT_TX_DONE) == RDY_TIMEOUT)
			LogWarn("Post EVENT_TX_DONE timed out");
		chSysUnlockFromIsr();
	}

	if((sr & (USART_SR_ORE | USART_SR_FE | USART_SR_NE)) && USARTGetEIE(pUSART->pRegister))
	{
		// clear interrupt enable
		USARTClearEIE(pUSART->pRegister);

		if(sr & USART_SR_ORE)
			event |= EVENT_RX_ERROR_ORE;
		if(sr & USART_SR_FE)
			event |= EVENT_RX_FE;
		if(sr & USART_SR_NE)
			event |= EVENT_RX_ERROR_NE;

		chSysLockFromIsr();
		if(chMBPostI(&pUSART->eventQueue, event) == RDY_TIMEOUT)
			LogWarn("Post EVENT_RX_* timed out");
		chSysUnlockFromIsr();
	}

	CH_IRQ_EPILOGUE();
}

void USARTTimerIRQ(USART* pUSART)
{
	CH_IRQ_PROLOGUE();

	chSysLockFromIsr();
	if(chMBPostI(&pUSART->eventQueue, EVENT_TX_TIMER_DONE) == RDY_TIMEOUT)
		LogWarn("Post EVENT_TX_TIMER_DONE timed out");
	chSysUnlockFromIsr();

	LogDebugC("TimerIRQ", (uint32_t)pUSART);

	CH_IRQ_EPILOGUE();
}

void USARTInit(USART* pUSART, USART_TypeDef* pRegister, USARTData data)
{
	pUSART->data = data;
	pUSART->pRegister = pRegister;
	pUSART->callback = 0;
	pUSART->stats.rxInvalid = 0;
	pUSART->txState = STATE_IDLE;
	pUSART->rxState = STATE_IDLE;

	// Init action queue
	chMBInit(&pUSART->eventQueue, pUSART->eventQueueData, 4);

	// Fifo init
	FifoLinInit(&pUSART->rxFifo, pUSART->data.pDataRx, pUSART->data.rxSize);
	FifoLinInit(&pUSART->txFifo, pUSART->data.pDataTx, pUSART->data.txSize);

	// find interrupt flag clear register and generate mask
	pUSART->pRxStream = DMAGenerateStreamTypeDef(data.pDMA, data.rxStream);
	pUSART->pTxStream = DMAGenerateStreamTypeDef(data.pDMA, data.txStream);
	pUSART->pDMARxIFCR = DMAGenerateIFCR(data.pDMA, data.rxStream);
	pUSART->pDMATxIFCR = DMAGenerateIFCR(data.pDMA, data.txStream);
	pUSART->pDMARxISR = DMAGenerateISR(data.pDMA, data.rxStream);
	pUSART->pDMATxISR = DMAGenerateISR(data.pDMA, data.txStream);
	pUSART->IFCRRxMask = DMAGenerateMaskAll(data.rxStream);
	pUSART->IFCRTxMask = DMAGenerateMaskAll(data.txStream);

	// DMA TX setup
	pUSART->pTxStream->CR = DMA_CHSEL(data.dmaChannel) | DMA_DIR_MEM2PERIPH | DMA_SxCR_MINC |
			DMA_PSIZE_1BYTE | DMA_MSIZE_1BYTE | DMA_PL_MEDIUM | DMA_MBURST_SINGLE | DMA_PBURST_SINGLE;
	pUSART->pTxStream->FCR = DMA_SxFCR_DMDIS | DMA_FTH_FULL;
	pUSART->pTxStream->PAR = (uint32_t)&pRegister->DR;
	pUSART->pTxStream->NDTR = 0;
	pUSART->pTxStream->M0AR = 0;

	// DMA RX setup
	pUSART->pRxStream->CR = DMA_CHSEL(data.dmaChannel) | DMA_DIR_PERIPH2MEM | DMA_SxCR_MINC |
			DMA_PSIZE_1BYTE | DMA_MSIZE_1BYTE | DMA_PL_MEDIUM | DMA_MBURST_SINGLE | DMA_PBURST_SINGLE;
	pUSART->pRxStream->FCR = DMA_SxFCR_DMDIS | DMA_FTH_FULL;
	pUSART->pRxStream->PAR = (uint32_t)&pRegister->DR;
	pUSART->pRxStream->NDTR = 0;
	pUSART->pRxStream->M0AR = 0;

	// USART setup
	pRegister->CR1 = USART_CR1_OVER8;
	pRegister->CR2 = 0;
	pRegister->CR3 = USART_CR3_ONEBIT;
	pRegister->BRR = USARTCalcBRR(data.baudrate, pRegister, 1);

	pRegister->SR = 0; // clear all flags
	USARTClearEIE(pRegister);
	USARTClearTCIE(pRegister);

	// NVIC setup
	NVICEnableIRQ(data.usartIRQn, data.IRQL-1);

	// clear DMA flags
	*pUSART->pDMARxIFCR = pUSART->IFCRRxMask;
	*pUSART->pDMATxIFCR = pUSART->IFCRTxMask;

	// finally enable USART
	pRegister->CR1 |= USART_CR1_UE;

	// enable DMA for TX and RX
	pRegister->CR3 |= USART_CR3_DMAT | USART_CR3_DMAR;

	// post rx start to enable receiver
	chMBPost(&pUSART->eventQueue, EVENT_RX_START, TIME_IMMEDIATE);
}

int16_t USARTSend(USART* pUSART, uint16_t length, uint8_t* pData)
{
	uint8_t* pDst;
	int16_t result;

	result = FifoLinReserve(&pUSART->txFifo, length, &pDst);
	if(result)
		return result;

	memcpy(pDst, pData, length);

	result = FifoLinCommit(&pUSART->txFifo, length);
	if(result)
		return result;

	if(chMBPost(&pUSART->eventQueue, EVENT_TX_START, MS2ST(500)) != RDY_OK)
		return ERROR_USART_QUEUE_FULL;

	return 0;
}

int16_t USARTSendConcat(USART* pUSART, uint16_t length, uint8_t* pData, uint16_t length2, uint8_t* pData2)
{
	uint8_t* pDst;
	int16_t result;

	result = FifoLinReserve(&pUSART->txFifo, length+length2, &pDst);
	if(result)
		return result;

	memcpy(pDst, pData, length);
	memcpy(pDst+length, pData2, length2);

	result = FifoLinCommit(&pUSART->txFifo, length+length2);
	if(result)
		return result;

	if(chMBPost(&pUSART->eventQueue, EVENT_TX_START, MS2ST(500)) != RDY_OK)
		return ERROR_USART_QUEUE_FULL;

	return 0;
}

int16_t USARTStartTransfer(USART* pUSART)
{
	if(pUSART->txState != STATE_IDLE) // already transmitting?
		return 0;

	if(chMBPost(&pUSART->eventQueue, EVENT_TX_START, MS2ST(500)) != RDY_OK)
		return ERROR_USART_QUEUE_FULL;

	return 0;
}

static void txStatemachine(USART* pUSART, msg_t event)
{
	int16_t result;
	uint8_t* pTx;

	if(event == EVENT_TX_ERROR)
	{
		LogDebugC("EVENT_TX_ERROR", event | (((uint32_t)pUSART->txState) << 16));

		return;
	}

	switch(pUSART->txState)
	{
		case STATE_IDLE:
		{
			if(event & EVENT_TX_START)
			{
				uint32_t txSize;

				result = FifoLinGet(&pUSART->txFifo, &pTx, &txSize);
				if(result)
					return;

				pUSART->txState = STATE_ACTIVE;

				pUSART->txTime = chTimeNow();

				pUSART->pTxStream->M0AR = (uint32_t)pTx;
				pUSART->pTxStream->NDTR = (uint16_t)txSize;
				*pUSART->pDMATxIFCR = pUSART->IFCRTxMask;	// clear flags
				pUSART->pRegister->SR &= ~USART_SR_TC;	// clear TC flag in USART
				DMASetEn(pUSART->pTxStream);	// enable DMA
				USARTSetTCIE(pUSART->pRegister);
				USARTSetTE(pUSART->pRegister);	// enable transmitter
			}
		}
		break;
		case STATE_ACTIVE:
		{
			if(event & EVENT_TX_DONE)
			{
				pUSART->pRegister->CR1 |= USART_CR1_SBK; // send break

				(*pUSART->data.pStartTimer)(100);

				pUSART->txState = STATE_WAIT_START;

				FifoLinDelete(&pUSART->txFifo);

				LogDebug("EVENT_TX_DONE + STATE_ACTIVE");
			}

			if(event & EVENT_TX_TIMEOUT)
			{
				//disable transmitter
				stopTx(pUSART);

				pUSART->txState = STATE_IDLE;

				FifoLinDelete(&pUSART->txFifo);

				LogDebug("EVENT_TX_TIMEOUT + STATE_ACTIVE");

				if(chMBPost(&pUSART->eventQueue, EVENT_TX_START, MS2ST(500)) != RDY_OK)
					LogError("Post EVENT_TX_START failed");
			}
		}
		break;
		case STATE_WAIT_START:
		{
			if((event & EVENT_TX_TIMER_DONE) || (event & EVENT_TX_TIMEOUT))
			{
				stopTx(pUSART);

				pUSART->txState = STATE_IDLE;

				if(chMBPost(&pUSART->eventQueue, EVENT_TX_START, MS2ST(500)) != RDY_OK)
					LogError("Post EVENT_TX_START failed");
			}
		}
		break;
		default: break;
	}
}

static void rxStatemachine(USART* pUSART, msg_t event)
{
	int16_t result;
	uint8_t* pRx;

	switch(pUSART->rxState)
	{
		case STATE_IDLE:
		{
			if(event & EVENT_RX_START)
			{
				// prepare new RX
				result = FifoLinReserve(&pUSART->rxFifo, pUSART->data.maxPacketSize, &pRx);
				if(result)
				{
					LogWarn("RX FIFO full, reception temporarily deactivated");
					pUSART->rxWaitStartTime = chTimeNow();
					pUSART->rxState = STATE_WAIT_START;
					return;
				}

				pUSART->rxState = STATE_ACTIVE;

				pUSART->pRxStream->NDTR = pUSART->data.maxPacketSize;
				pUSART->pRxStream->M0AR = (uint32_t)pRx;

				// clear flags
				pUSART->pRegister->SR &= ~(USART_SR_ORE | USART_SR_FE | USART_SR_NE | USART_SR_RXNE);
				pUSART->pRegister->DR;
				*pUSART->pDMARxIFCR = pUSART->IFCRRxMask;

				DMASetEn(pUSART->pRxStream);
				USARTSetEIE(pUSART->pRegister);
				USARTSetRE(pUSART->pRegister);
			}
		}
		break;
		case STATE_ACTIVE:
		{
			if(event & 0x07)	// NE,FE or ORE
			{
				stopRx(pUSART);
			}

			if(event == EVENT_RX_FE && DMAGetEn(pUSART->pRxStream))
			{
				LogWarn("DMA RX EN set + RX done");
			}

			if(event == EVENT_RX_FE)
			{
				uint16_t dataLength = pUSART->data.maxPacketSize-pUSART->pRxStream->NDTR-1; // -1 because errornous break frame is saved too

				result = FifoLinCommit(&pUSART->rxFifo, dataLength);
				if(result)
				{
					// that does not fit :(
					++pUSART->stats.rxInvalid;

					LogWarnC("FifoLinCommit failed for RX", result);
				}
				else
				{
					if(pUSART->callback)
						(*pUSART->callback)(pUSART->pUserData);
				}
			}

			if((event & EVENT_RX_ERROR_NE) == EVENT_RX_ERROR_NE || (event & EVENT_RX_ERROR_ORE) == EVENT_RX_ERROR_ORE)
			{
				LogWarnC("RX noise or overrun error", event);
			}

			pUSART->rxState = STATE_IDLE;

			if(chMBPost(&pUSART->eventQueue, EVENT_RX_START, MS2ST(500)) != RDY_OK)
				LogWarn("Post EVENT_RX_START failed");
		}
		break;
		case STATE_WAIT_START:
		{
			if(event & EVENT_RX_TIMER_DONE)
			{
				if(chMBPost(&pUSART->eventQueue, EVENT_RX_START, MS2ST(500)) != RDY_OK)
					LogWarn("Post EVENT_RX_START failed");
				else
					pUSART->rxState = STATE_IDLE;
			}
		}
		break;
		default: break;
	}
}

void USARTProcess(USART* pUSART)
{
	msg_t event;
	int32_t osResult;

	osResult = chMBFetch(&pUSART->eventQueue, &event, MS2ST(100));
	if(osResult == RDY_OK)
	{
		LogDebugC("Received event", event);
		if(EVENT_IS_RX(event))
			rxStatemachine(pUSART, event);
		else
			txStatemachine(pUSART, event);
	}

	if((pUSART->txState == STATE_ACTIVE) && chTimeElapsedSince(pUSART->txTime) > MS2ST(100))
	{
		if(chMBPost(&pUSART->eventQueue, EVENT_TX_TIMEOUT, MS2ST(10)) != RDY_OK)
			LogWarn("Post EVENT_TX_TIMEOUT failed");
	}

	if((pUSART->txState == STATE_WAIT_START) && chTimeElapsedSince(pUSART->txTime) > MS2ST(100))
	{
		if(chMBPost(&pUSART->eventQueue, EVENT_TX_TIMEOUT, MS2ST(10)) != RDY_OK)
			LogWarn("Post EVENT_TX_TIMEOUT failed");
	}

	if((pUSART->rxState == STATE_WAIT_START) && chTimeElapsedSince(pUSART->rxWaitStartTime) > MS2ST(10))
	{
		if(chMBPost(&pUSART->eventQueue, EVENT_RX_TIMER_DONE, MS2ST(10)) != RDY_OK)
			LogWarn("Post EVENT_RX_TIMER_DONE failed");
	}
}

msg_t USARTTask(void* params)
{
	USART* pUSART = (USART*)params;

	chRegSetThreadName("USART");

	while(42)
	{
		USARTProcess(pUSART);
	}

	return 0;
}
