/*
 * log_msgs.h
 *
 *  Created on: 21.10.2014
 *      Author: AndreR
 */

#ifndef LOG_MSGS_H_
#define LOG_MSGS_H_

#include <stdint.h>
#include "util/pid.h"

#ifndef PACKED
#define PACKED __attribute__((packed))
#endif

enum LogChunkType
{
	CTYPE_GYRO_RAW = 1,
	CTYPE_ACC_RAW,
	CTYPE_CTRL_INPUT,
	CTYPE_CTRL_STATE,
	CTYPE_CTRL_OUTPUT,
	CTYPE_CTRL_TIGGA_PIDS,

	CTYPE_MESSAGE_DESC = 0x8000,
	CTYPE_ELEMENT_DESC,
};

// fundamental header for all entries in the log file
typedef struct PACKED _LogChunkHeader
{
	uint16_t type;
	uint16_t length;
} LogChunkHeader;

// the header for all log entries, contains the LogChunkHeader
typedef struct PACKED _LogEntryHeader
{
	uint16_t type;
	uint16_t length;
	uint32_t timestamp;
} LogEntryHeader;

// --- LOG DEFINITIONS ---
typedef struct PACKED _LogRawI16XYZ
{
	LogEntryHeader header;
	int16_t xyz[3];
} LogRawI16XYZ;

#define MOT_FR 0
#define MOT_FL 1
#define MOT_RL 2
#define MOT_RR 3

typedef struct PACKED _LogCtrlInput
{
	LogEntryHeader header;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time[4];
		float cur[4];			// [A]
	} motorCur;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time[4];
		float vel[4];			// angular motor velocity [rad/s]
	} enc;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time;
		float rotVel[3];	// around X, Y, Z [rad/s]
	} gyr;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time;
		float linAcc[3];	// X, Y, Z [m^2/s]
	} acc;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time;
		float pos[3];		// X, Y, Z [m]
	} pos;

	struct PACKED
	{
		uint8_t updated;	// battery voltage [V]
		float voltage;
	} power;

	struct PACKED
	{
		uint32_t time[4];
		float vol[4];			// [V]
	} motorVol;

	struct PACKED
	{
		uint32_t time[4];
		float M[4];			// motor torque [Nm]
	} motorTorque;
} LogCtrlInput;

typedef struct PACKED _LogCtrlState
{
	LogEntryHeader header;

	uint8_t posUpdated;
	float pos[3];	// [m]

	uint8_t velUpdated;
	float vel[3];	// [m/s]

	uint8_t accUpdated;
	float acc[3];	// [m/s^2]
} LogCtrlState;

typedef struct PACKED _LogCtrlOutput
{
	LogEntryHeader header;

	float motorVel[4];	// [rad/s]
	uint8_t ctrlMode;
	float motorVoltage[4];	// Motor voltage [V], Range: -10 (CCW) - 10 (CW)
	float motorTorque[4];	// Motor torque [Nm]
} LogCtrlOutput;

typedef struct PACKED _LogCtrlTiggaPIDs
{
	LogEntryHeader header;

	uint32_t newTrajTime;
	float trajPos[3];
	float trajVel[3];
	float trajAcc[3];

	float accFwd[3];

	PIDLog pidPos[3];
	PIDLog pidVel[3];

	uint8_t trajActive;
	uint8_t accFeedFwd;
} LogCtrlTiggaPIDs;

#endif /* LOG_MSGS_H_ */
