/*
 * config.c
 *
 *  Created on: 16.05.2015
 *      Author: AndreR
 */

#include "config.h"
#include "intercom_reliable.h"
#include "intercom_constants.h"
#include "commands.h"
#include "log.h"
#include <string.h>

Config config;

static void icCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	switch(pHeader->command)
	{
		case CMD_CONFIG_QUERY_FILE_LIST:
		{
			ConfigInternalInjectEvent(CONFIG_EVENT_SEND_FILE_LIST);
		}
		break;
		case CMD_CONFIG_ITEM_DESC:
		{
			ConfigItemDesc* pDesc = (ConfigItemDesc*)pData;
			msg_t event = CONFIG_EVENT_SEND_ITEM_DESC | (((msg_t)pDesc->element) << 16) | pDesc->id;

			ConfigInternalInjectEvent(event);
		}
		break;
		case CMD_CONFIG_READ:
		{
			ConfigReadWrite* pRead = (ConfigReadWrite*)pData;
			msg_t event = CONFIG_EVENT_SEND_FILE | pRead->id;

			ConfigInternalInjectEvent(event);
		}
		break;
		case CMD_CONFIG_WRITE:
		{
			ConfigReadWrite* pWrite = (ConfigReadWrite*)pData;

			if(dataLength < sizeof(ConfigReadWrite))
			{
				LogErrorC("Invalid datalength", dataLength);
				return;
			}

			uint16_t size = dataLength - sizeof(ConfigReadWrite);
			pData += sizeof(ConfigReadWrite);

			ConfigInternalWriteFile(pWrite->id, pData, size);
		}
		break;
	}
}

void ConfigInit(ConfigOutputFunc outFunc)
{
	chMBInit(&config.eventQueue, config.eventQueueData, CONFIG_EVENT_QUEUE_SIZE);

	config.filesUsed = 0;
	config.outputFunction = outFunc;
}

void ConfigIntercomInit()
{
	IntercomClient iClient;
	iClient.pUserData = 0;
	iClient.rxCallback = &icCallback;
	iClient.section = INTERCOM_SECTION_CONFIG;

	IntercomAddClient(iClient);
}

ConfigFile* ConfigOpenOrCreate(const ConfigFileDesc* pConfigDesc, void* pData, uint16_t size, ConfigUpdateCallback updateCb)
{
	for(uint16_t i = 0; i < config.filesUsed; i++)
	{
		if(config.files[i].pDesc->cfgId == pConfigDesc->cfgId)
			return &config.files[i];
	}

	if(config.filesUsed == CONFIG_MAX_FILES)
		return 0;

	ConfigFile* pCfgFile = &config.files[config.filesUsed++];

	if(FlashFSOpenOrCreate(pConfigDesc->pName, pConfigDesc->version, pData, size, &pCfgFile->pFlashFile))
	{
		--config.filesUsed;
		return 0;
	}

	pCfgFile->pDesc = pConfigDesc;
	pCfgFile->pData = pData;
	pCfgFile->size = size;
	pCfgFile->updateCb = updateCb;

	return pCfgFile;
}

void ConfigNotifyUpdate(ConfigFile* pFile)
{
	FlashFSWrite(&pFile->pFlashFile, pFile->pData, pFile->size);

	if(pFile->updateCb)
		(*pFile->updateCb)(pFile->pDesc->cfgId);
}

void ConfigInternalWriteFile(uint16_t id, void* pData, uint16_t size)
{
	ConfigFile* pCfgFile = 0;

	for(uint16_t i = 0; i < config.filesUsed; i++)
	{
		if(config.files[i].pDesc->cfgId == id)
		{
			pCfgFile = &config.files[i];
			break;
		}
	}

	if(pCfgFile == 0)
	{
		LogErrorC("File does not exist", id);
		return;
	}

	if(pCfgFile->size != size)
	{
		LogErrorC("Size mismatch", (pCfgFile->size << 16 | size));
		return;
	}

	memcpy(pCfgFile->pData, pData, size);

	ConfigNotifyUpdate(pCfgFile);
}

void ConfigInternalInjectEvent(msg_t event)
{
	chMBPost(&config.eventQueue, event, TIME_IMMEDIATE);
}

void ConfigOutputIntercom(void* pData, uint16_t dataLength)
{
	IntercomHeader iHeader;
	iHeader.section = INTERCOM_SECTION_NETWORK;
	iHeader.command = IC_NETWORK_RELIABLE;
	iHeader.address = INTERCOM_ADDRESS_MAIN;

	IntercomReliableSend(&iHeader, dataLength, pData);
}

msg_t ConfigTask(void* params)
{
	(void)params;

	msg_t event;

	chRegSetThreadName("Config");

	while(1)
	{
		if(chMBFetch(&config.eventQueue, &event, TIME_INFINITE) != RDY_OK)
			continue;

		msg_t internalEvent = (event & 0xF0000000);

		switch(internalEvent)
		{
			case CONFIG_EVENT_SEND_FILE_LIST:
			{
				for(uint16_t i = 0; i < config.filesUsed; i++)
				{
					ConfigFile* pFile = &config.files[i];
					PacketHeader* pHeader = (PacketHeader*)config.procData;
					pHeader->section = SECTION_CONFIG;
					pHeader->cmd = CMD_CONFIG_FILE_STRUCTURE;

					ConfigFileStructure* pStruct = (ConfigFileStructure*)(config.procData+sizeof(PacketHeader));
					pStruct->id = pFile->pDesc->cfgId | (((uint16_t)intercom.ownId) << 12);
					pStruct->version = pFile->pDesc->version;

					uint8_t* pElements = config.procData+sizeof(ConfigFileStructure)+sizeof(PacketHeader);

					for(uint16_t ele = 0; ele < pFile->pDesc->numElements; ele++)
					{
						*pElements = pFile->pDesc->elements[ele].type;
						++pElements;
					}

					(*config.outputFunction)(config.procData, sizeof(PacketHeader)+sizeof(ConfigFileStructure)+pFile->pDesc->numElements);

					chThdSleepMilliseconds(50);
				}
			}
			break;
			case CONFIG_EVENT_SEND_FILE:
			{
				uint16_t cfgId = event & 0xFFFF;

				ConfigFile* pFile = 0;

				for(uint16_t i = 0; i < config.filesUsed; i++)
				{
					if(config.files[i].pDesc->cfgId == cfgId)
					{
						pFile = &config.files[i];
						break;
					}
				}

				if(pFile == 0)
				{
					LogErrorC("Unknown config", cfgId);
					continue;
				}

				PacketHeader* pHeader = (PacketHeader*)config.procData;
				pHeader->section = SECTION_CONFIG;
				pHeader->cmd = CMD_CONFIG_READ;

				ConfigReadWrite* pRead = (ConfigReadWrite*)(config.procData+sizeof(PacketHeader));
				pRead->id = cfgId | (((uint16_t)intercom.ownId) << 12);

				memcpy(config.procData+sizeof(PacketHeader)+sizeof(ConfigReadWrite), pFile->pData, pFile->size);

				(*config.outputFunction)(config.procData, sizeof(PacketHeader)+sizeof(ConfigReadWrite)+pFile->size);
				chThdSleepMilliseconds(20);
			}
			break;
			case CONFIG_EVENT_SEND_ITEM_DESC:
			{
				uint16_t cfgId = event & 0xFFFF;
				uint16_t element = (event & 0xFF0000) >> 16;

				const ConfigFileDesc* pFileDesc = 0;

				for(uint16_t i = 0; i < config.filesUsed; i++)
				{
					if(config.files[i].pDesc->cfgId == cfgId)
					{
						pFileDesc = config.files[i].pDesc;
						break;
					}
				}

				if(pFileDesc == 0)
				{
					LogErrorC("Unknown config", cfgId);
					continue;
				}

				PacketHeader* pHeader = (PacketHeader*)config.procData;
				pHeader->section = SECTION_CONFIG;
				pHeader->cmd = CMD_CONFIG_ITEM_DESC;

				ConfigItemDesc* pDesc = (ConfigItemDesc*)(config.procData+sizeof(PacketHeader));
				pDesc->id = cfgId | (((uint16_t)intercom.ownId) << 12);
				pDesc->element = element;

				if(element == 0xFF)
				{
					// file name
					size_t nameLength = strlen(pFileDesc->pName);
					memcpy(config.procData+sizeof(PacketHeader)+sizeof(ConfigItemDesc), pFileDesc->pName, nameLength);

					(*config.outputFunction)(config.procData, sizeof(PacketHeader)+sizeof(ConfigItemDesc)+nameLength);
					chThdSleepMilliseconds(20);

					continue;
				}

				if(element >= pFileDesc->numElements)
				{
					LogErrorC("Unknown element", (element << 8) | pFileDesc->numElements);
					continue;
				}

				// element name
				size_t nameLength = strlen(pFileDesc->elements[element].pName);
				memcpy(config.procData+sizeof(PacketHeader)+sizeof(ConfigItemDesc), pFileDesc->elements[element].pName, nameLength);

				(*config.outputFunction)(config.procData, sizeof(PacketHeader)+sizeof(ConfigItemDesc)+nameLength);
				chThdSleepMilliseconds(20);
			}
			break;
		}
	}

	return 0;
}

