/*
 * flash_fs.c
 *
 *  Created on: 25.06.2014
 *      Author: AndreR
 */

#include "flash_fs.h"
#include "util/flash.h"
#include "util/log.h"
#include "util/console.h"
#include <string.h>

FlashFS flashFS;

#define FLASH_FS_VERSION 0x2001

void FlashFSInit(uint32_t flashAddr, uint32_t flashSize)
{
	chMtxInit(&flashFS.writeMutex);

	flashSize -= 4;

	flashFS.version = *((uint32_t*)(flashAddr+flashSize));	// version info is the last word in FlashFS

	if(flashFS.version != FLASH_FS_VERSION)	// is this an old FlashFS v1 or info not set?
	{
		// yes, clear it
		FlashErase(flashAddr, flashAddr+flashSize);

		flashFS.version = FLASH_FS_VERSION;
		FlashProgram(flashAddr+flashSize, &flashFS.version, 1);
	}

	flashFS.bytesDel = 0;
	flashFS.bytesUsed = 0;
	flashFS.flashAddr = flashAddr;
	flashFS.flashSize = flashSize;
	flashFS.pWrite = (uint8_t*)flashAddr;
	flashFS.pEnd = flashFS.pWrite+flashSize;

	FlashFile* pFH = (FlashFile*)flashFS.pWrite;

	while(pFH->blockSize != 0xFFFF && (uint8_t*)pFH < flashFS.pEnd)
	{
		if(pFH->del != 0xFFFFFFFF)
			flashFS.bytesDel += pFH->blockSize;
		else
			flashFS.bytesUsed += pFH->blockSize;

		flashFS.pWrite += pFH->blockSize;
		pFH = (FlashFile*)flashFS.pWrite;
	}
}

FlashFile* FlashFSOpen(const char* pFilename, uint16_t version)
{
	const char* pFSFilename;

	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	while(pFH->blockSize != 0xFFFF && (uint8_t*)pFH < flashFS.pEnd)
	{
		pFSFilename = (const char*)(pRead+sizeof(FlashFile));

		if(pFH->del == 0xFFFFFFFF && strcmp(pFilename, pFSFilename) == 0 && pFH->version == version)
		{
			return pFH;
		}

		pRead += pFH->blockSize;
		pFH = (FlashFile*)pRead;
	}

	return 0;
}

FlashFile* FlashFSCreate(const char* pFilename, uint16_t version, uint16_t size)
{
	FlashFile* pFH = FlashFSOpen(pFilename, version);
	if(pFH)
		return pFH;

	// in case the file exists and just the version changed, we need to delete it
	FlashFSDelete(pFilename);

	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	pFH = (FlashFile*)pRead;
	uint16_t filenameLength = (uint16_t)strlen(pFilename)+1;	// include terminating null

	// construct header with aligned sizes
	FlashFile header;
	header.headerSize = sizeof(FlashFile) + filenameLength;
	header.headerSize += (4 - (header.headerSize%4))%4;
	header.blockSize = size + header.headerSize;
	header.blockSize += (4 - (header.blockSize%4))%4;	// align to 4 byte
	header.dataSize = size;
	header.del = 0xFFFFFFFF;
	header.version = version;

	// check free space
	if(flashFS.bytesUsed + header.blockSize > FLASH_TEMP_SIZE)
	{
		LogErrorC("No memory to create file", (uint32_t)pFilename);
		return 0;
	}

	// need to compact?
	if(flashFS.pWrite + header.blockSize > flashFS.pEnd)
	{
		int16_t result = FlashFSCompact();
		if(result)
		{
			LogErrorC("Compact failed", result);
			return 0;
		}
	}

	chMtxLock(&flashFS.writeMutex);

	// copy header to temporary buffer
	memcpy(flashFS.temp, &header, sizeof(FlashFile));

	// copy filename to temporary buffer after the header
	memcpy(flashFS.temp+sizeof(FlashFile), pFilename, filenameLength);

	// write header and filename
	FlashProgram((uint32_t)flashFS.pWrite, (uint32_t*)flashFS.temp, header.headerSize/4);

	pFH = (FlashFile*)flashFS.pWrite;

	flashFS.pWrite += pFH->blockSize;
	flashFS.bytesUsed += pFH->blockSize;

	chMtxUnlock();

	return pFH;
}

void FlashFSRead(FlashFile* pFH, void* pDst)
{
	uint8_t* pRead = (uint8_t*)pFH;
	pRead += pFH->headerSize;

	memcpy(pDst, pRead, pFH->dataSize);
}

int16_t FlashFSOpenOrCreate(const char* pFilename, uint16_t version, void* pData, uint16_t size, FlashFile** ppFH)
{
	*ppFH = FlashFSOpen(pFilename, version);
	if(!*ppFH)
	{
		*ppFH = FlashFSCreate(pFilename, version, size);
		if(!*ppFH)
			return FLASH_FS_NOMEM;

		FlashFSWrite(ppFH, pData, size);
		return FLASH_FS_CREATED;
	}
	else
	{
		FlashFSRead(*ppFH, pData);
		return FLASH_FS_OPEN;
	}
}

int16_t FlashFSWrite(FlashFile** ppFH, void* pData, uint16_t size)
{
	int16_t result;
	uint32_t nul = 0;
	FlashFile* pFH = *ppFH;

	// check size
	if(pFH->dataSize != size)
		return FLASH_FS_SIZE_MISMATCH;

	// compare content
	uint8_t* pRead = (uint8_t*)pFH;
	pRead += pFH->headerSize;
	if(memcmp(pRead, pData, size) == 0)
		return 0;

	// check free space
	if(flashFS.pWrite + pFH->blockSize > flashFS.pEnd)
	{
		result = FlashFSCompact();
		if(result)
		{
			LogErrorC("Compact failed", result);
			return result;
		}
	}

	chMtxLock(&flashFS.writeMutex);

	*ppFH = (FlashFile*)flashFS.pWrite;

	// copy header
	FlashProgram((uint32_t)flashFS.pWrite, (uint32_t*)pFH, pFH->headerSize/4);

	// mark old entry as deleted
	FlashProgram((uint32_t)&pFH->del, &nul, 1);

	// copy data
	memcpy(flashFS.temp, pData, size);
	FlashProgram((uint32_t)(flashFS.pWrite+pFH->headerSize), (uint32_t*)flashFS.temp, (pFH->blockSize-pFH->headerSize)/4);

	flashFS.pWrite += pFH->blockSize;
	flashFS.bytesDel += pFH->blockSize;

	chMtxUnlock();

	return 0;
}

void FlashFSDelete(const char* pFilename)
{
	uint32_t nul = 0;
	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	chMtxLock(&flashFS.writeMutex);

	while(pFH->blockSize != 0xFFFF && (uint8_t*)pFH < flashFS.pEnd)
	{
		const char* pFSFilename = (const char*)(pRead+sizeof(FlashFile));

		if(pFH->del == 0xFFFFFFFF && strcmp(pFilename, pFSFilename) == 0)
		{
			FlashProgram((uint32_t)&pFH->del, &nul, 1);
			flashFS.bytesDel += pFH->blockSize;
			flashFS.bytesUsed -= pFH->blockSize;
		}

		pRead += pFH->blockSize;
		pFH = (FlashFile*)pRead;
	}

	chMtxUnlock();
}

int16_t FlashFSCompact()
{
	if(flashFS.bytesUsed > FLASH_TEMP_SIZE)
	{
		LogErrorC("bytesUsed > FLASH_TEMP_SIZE", flashFS.bytesUsed);
		return FLASH_FS_NOMEM;
	}

	if(flashFS.bytesDel + flashFS.bytesUsed < flashFS.flashSize/2)
		return 0;	// no need to compact already

	chMtxLock(&flashFS.writeMutex);

	uint8_t* pTempWrite = flashFS.temp;

	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	while(pFH->blockSize != 0xFFFF && (uint8_t*)pFH < flashFS.pEnd)
	{
		if(pFH->del == 0xFFFFFFFF)
		{
			memcpy(pTempWrite, pRead, pFH->blockSize);
			pTempWrite += pFH->blockSize;
		}

		pRead += pFH->blockSize;
		pFH = (FlashFile*)pRead;
	}

	FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);

	FlashProgram(flashFS.flashAddr+flashFS.flashSize, &flashFS.version, 1);			// program version info
	FlashProgram(flashFS.flashAddr, (uint32_t*)flashFS.temp, flashFS.bytesUsed/4);	// program used data

	flashFS.bytesDel = 0;
	flashFS.pWrite = (uint8_t*)(flashFS.flashAddr+flashFS.bytesUsed);

	chMtxUnlock();

	return 0;
}

void FlashFSListFiles()
{
	uint8_t* pRead = (uint8_t*)flashFS.flashAddr;
	FlashFile* pFH = (FlashFile*)pRead;

	ConsolePrint("--- Flash FS ---\r\n");
	ConsolePrint("Version: 0x%08X\r\n", flashFS.version);
	ConsolePrint("Used: %8u of %8u Bytes\r\n", flashFS.bytesUsed, FLASH_TEMP_SIZE);
	ConsolePrint("Stat: %8u of %8u Bytes\r\n", flashFS.bytesUsed+flashFS.bytesDel, flashFS.flashSize);

	ConsolePrint("File                              Version     Size   OnDisk\r\n");

	chThdSleep(MS2ST(5));

	while(pFH->blockSize != 0xFFFF && (uint8_t*)pFH < flashFS.pEnd)
	{
		if(pFH->del == 0xFFFFFFFF)
		{
			const char* pFSFilename = (const char*)(pRead+sizeof(FlashFile));

			ConsolePrint("%-32s %8hu %8hu %8hu\r\n", pFSFilename, pFH->version, pFH->dataSize, pFH->blockSize);
		}

		pRead += pFH->blockSize;
		pFH = (FlashFile*)pRead;

		chThdSleep(MS2ST(2));
	}
}
