/*
 * intercom.c
 *
 *  Created on: 10.11.2012
 *      Author: AndreR
 */

#include "intercom.h"
#include "errors.h"
#include "crc.h"
#include "log.h"
#include <string.h>

Intercom intercom;

void IntercomRxReady(void* pUserData)
{
	(void)pUserData;

	chBSemSignal(&intercom.rxReady);
}

void IntercomInit(uint8_t ownAddr, uint8_t hub)
{
	intercom.numClientsUsed = 0;
	intercom.numMediaUsed = 0;
	intercom.ownAddress = ownAddr;
	intercom.hub = hub;
	chBSemInit(&intercom.rxReady, 1);

	for(uint8_t i = 0; i < 8; i++)
	{
		if(ownAddr & (1 << i))
		{
			intercom.ownId = i;
			break;
		}
	}
}

int16_t IntercomAddMedium(IntercomMedium medium)
{
	if(intercom.numMediaUsed == INTERCOM_MAX_MEDIA)
		return ERROR_INTERCOM_NO_FREE_SLOT;

	intercom.media[intercom.numMediaUsed] = medium;
	intercom.numMediaUsed++;

	return 0;
}

int16_t IntercomSend(const IntercomHeader* pHeader, uint16_t dataLength, void* _pData)
{
	uint8_t* pDst;
	int16_t result;
	uint8_t* pData = (uint8_t*)_pData;

	const uint32_t length = INTERCOM_HEADER_SIZE + dataLength + INTERCOM_TRAILER_SIZE;

	IntercomExecRxCallback(pHeader, pData, dataLength);

	for(uint8_t m = 0; m < intercom.numMediaUsed; m++)
	{
		IntercomMedium* pMedium = &intercom.media[m];

		if(pMedium->reachableAddresses & pHeader->address)
		{
			result = FifoLinReserve(pMedium->pTxFifo, length, &pDst);
			if(result)
				return result;

			memcpy(pDst, pHeader->_data, INTERCOM_HEADER_SIZE);
			memcpy(pDst+INTERCOM_HEADER_SIZE, pData, dataLength);
			*((uint32_t*)(pDst+dataLength+INTERCOM_HEADER_SIZE)) = CRC32CalcChecksumAlt(pDst, INTERCOM_HEADER_SIZE+dataLength);

			result = FifoLinCommit(pMedium->pTxFifo, length);
			if(result)
				return result;

			if(pMedium->txReady)
				(*pMedium->txReady)(pMedium->pTxReadyData);
		}
	}

	return 0;
}

int16_t IntercomAddClient(IntercomClient client)
{
	if(intercom.numClientsUsed == INTERCOM_MAX_CLIENTS)
		return ERROR_INTERCOM_NO_FREE_SLOT;

	intercom.clients[intercom.numClientsUsed] = client;
	intercom.numClientsUsed++;

	return 0;
}

void IntercomExecRxCallback(const IntercomHeader* pHeader, uint8_t* pData, uint32_t dataLength)
{
	if((pHeader->address & intercom.ownAddress))
	{
		for(uint8_t i = 0; i < intercom.numClientsUsed; i++)
		{
			if(intercom.clients[i].section == pHeader->section)
				(*intercom.clients[i].rxCallback)(pHeader, dataLength, pData, intercom.clients[i].pUserData);
		}
	}
}

msg_t IntercomTask(void* params)
{
	uint32_t length;
	uint8_t* pFifoData;
	int16_t result;
	(void)params;

	chRegSetThreadName("Intercom");

	while(1)
	{
		if(chBSemWaitTimeout(&intercom.rxReady, TIME_INFINITE) != RDY_OK)
			continue;

		for(uint8_t m = 0; m < intercom.numMediaUsed; m++)
		{
			IntercomMedium* pMedium = &intercom.media[m];

			while(1)
			{
				result = FifoLinGet(pMedium->pRxFifo, &pFifoData, &length);
				if(result)
					break;

				if(length < INTERCOM_HEADER_SIZE+INTERCOM_TRAILER_SIZE)
				{
					LogWarnC("Short packet", length);

					FifoLinDelete(pMedium->pRxFifo);
					continue;
				}

				uint32_t dataLength = length - INTERCOM_HEADER_SIZE - INTERCOM_TRAILER_SIZE;

				IntercomHeader* pHeader = (IntercomHeader*)pFifoData;
				uint8_t* pData = pFifoData + INTERCOM_HEADER_SIZE;
				uint32_t* pCrc = (uint32_t*)(pFifoData + length - INTERCOM_TRAILER_SIZE);

				if(CRC32VerifyChecksumAlt(pFifoData, length-INTERCOM_TRAILER_SIZE, *pCrc) != CRC32_MAGIC_NUMBER)
				{
					LogWarn("Intercom CRC32 checksum mismatch");

					FifoLinDelete(pMedium->pRxFifo);
					continue;
				}

				IntercomExecRxCallback(pHeader, pData, dataLength);

				if((pHeader->address & ~intercom.ownAddress) && intercom.hub)	// any remote recipients?
				{
					IntercomSend(pHeader, dataLength, pData);
				}

				FifoLinDelete(pMedium->pRxFifo);
			}
		}
	}

	return 0;
}
