/*
 * intercom_reliable.c
 *
 *  Created on: 30.03.2015
 *      Author: AndreR
 */

#include "intercom_reliable.h"

#include "intercom_constants.h"
#include "errors.h"
#include "log.h"

#include <string.h>

IntercomReliable intercomReliable;

static void reliableCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData);

void IntercomReliableInit()
{
	chMtxInit(&intercomReliable.sendMtx);
	chBSemInit(&intercomReliable.ackSem, 0);

	memset(intercomReliable.lastRxSeq, 0x7E, sizeof(uint16_t)*INTERCOM_MAX_ADDRESSES);

	IntercomClient iClient;
	iClient.pUserData = 0;
	iClient.section = INTERCOM_SECTION_RELIABLE;
	iClient.rxCallback = &reliableCallback;

	IntercomAddClient(iClient);
}

static uint8_t getBitPosition(uint8_t in)
{
	for(uint8_t i = 0; i < 8; i++)
	{
		if(in & (1 << i))
			return i;
	}

	return 0xFF;
}

// this function must NOT be called from an intercom callback handler
int16_t IntercomReliableSend(const IntercomHeader* pHeader, uint16_t dataLength, void* _pData)
{
	if(pHeader->address == 0)
	{
		LogWarn("Address not set");
		return ERROR_INVALID_PARAMETER;
	}

	// prevent re-entering function from multiple tasks
	chMtxLock(&intercomReliable.sendMtx);

	uint16_t wrappedDataLength = sizeof(ReliableDataAck) + INTERCOM_HEADER_SIZE + dataLength;
	if(wrappedDataLength > INTERCOM_RELIABLE_DATA_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	ReliableDataAck* pDataAckPkt = (ReliableDataAck*)intercomReliable.procBuf;

	// assemble data to be wrapped
	memcpy(intercomReliable.procBuf+sizeof(ReliableDataAck), pHeader, INTERCOM_HEADER_SIZE);
	memcpy(intercomReliable.procBuf+sizeof(ReliableDataAck)+INTERCOM_HEADER_SIZE, _pData, dataLength);

	pDataAckPkt->seq = intercomReliable.seq;
	pDataAckPkt->srcAddr = intercom.ownAddress;

	// data now fully assembled, send and wait for ACKs

	intercomReliable.sendAddrMask = pHeader->address;

	while(intercomReliable.sendAddrMask)
	{
		IntercomHeader iRHeader = {{intercomReliable.sendAddrMask, INTERCOM_SECTION_RELIABLE, IC_RELIABLE_DATA}};

		chBSemReset(&intercomReliable.ackSem, 1);

		IntercomSend(&iRHeader, sizeof(ReliableDataAck)+INTERCOM_HEADER_SIZE+dataLength, intercomReliable.procBuf);

		if(chBSemWaitTimeout(&intercomReliable.ackSem, MS2ST(INTERCOM_RELIABLE_TIMEOUT)) != RDY_OK)
			LogWarnC("Data loss", intercomReliable.sendAddrMask);
	}

	++intercomReliable.seq;

	chMtxUnlock();

	return 0;
}

// this is called from the intercom task
static void reliableCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	ReliableDataAck* pDataAckPkt = (ReliableDataAck*)pData;

	if(pHeader->command == IC_RELIABLE_DATA)
	{
		// check sequence number
		// if now != expected => process data, if equal => duplicate packet
		uint8_t srcIndex = getBitPosition(pDataAckPkt->srcAddr);
		if(intercomReliable.lastRxSeq[srcIndex] != pDataAckPkt->seq)
		{
			intercomReliable.lastRxSeq[srcIndex] = pDataAckPkt->seq;

			IntercomExecRxCallback((IntercomHeader*)(pData+sizeof(ReliableDataAck)), pData+sizeof(ReliableDataAck)+INTERCOM_HEADER_SIZE,
					dataLength-sizeof(ReliableDataAck)-INTERCOM_HEADER_SIZE);
		}
		else
		{
			LogWarnC("Duplicate packet", pDataAckPkt->seq);
		}

		// send ACK in all cases
		IntercomHeader iHeader = {{pDataAckPkt->srcAddr, INTERCOM_SECTION_RELIABLE, IC_RELIABLE_ACK}};
		ReliableDataAck ack = {pDataAckPkt->seq, intercom.ownAddress};
		IntercomSend(&iHeader, sizeof(ReliableDataAck), &ack);
	}

	if(pHeader->command == IC_RELIABLE_ACK)
	{
		// check sequence number
		// if equal to one in progress => process, otherwise ignore
		// multiple ACKs are don't care (should not happen)

		if(pDataAckPkt->seq == intercomReliable.seq)
		{
			intercomReliable.sendAddrMask &= ~(pDataAckPkt->srcAddr);	// kick out the source from the missing ACK sources

			if(intercomReliable.sendAddrMask == 0)
				chBSemSignal(&intercomReliable.ackSem);
		}
	}
}

