/*
 * terminal.c
 *
 *  Created on: 26.10.2014
 *      Author: AndreR
 */

#include "terminal.h"

#include "util/console.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

typedef struct _VT100Sequence
{
	const char* pSeq;
	uint16_t code;
	uint16_t default1;
	uint16_t default2;
} VT100Sequence;

static VT100Sequence sequences[] = {
	{ "[\1A", CURSOR_UP, 1, 0},
	{ "[\1B", CURSOR_DOWN, 1, 0},
	{ "[\1C", CURSOR_FORWARD, 1, 0},
	{ "[\1D", CURSOR_BACKWARD, 1, 0},
	{ "[\1~", FUNCTION_KEY, 0xFF, 0},
};

Terminal terminal;

void TerminalInit(InputQueue* pInQueue, OutputQueue* pOutQueue)
{
	terminal.pInQueue = pInQueue;
	terminal.pOutQueue = pOutQueue;
}

static VT100Result compareSeq(char* pIn, VT100Sequence* pVT100Seq)
{
	const char* pSeq = pVT100Seq->pSeq;
	VT100Result result;
	result.code = 0;
	result.param1 = 0xFFFF;
	result.param2 = 0xFFFF;

	while(1)
	{
		if(*pIn == 0 && *pSeq == 0)
		{
			result.code = pVT100Seq->code;
			return result;
		}

		if(*pIn == 0)
			break;

		if(*pSeq == 0)
			break;

		if(*pSeq == 1)	// (optional) parameter 1 to read
		{
			char* pNewIn;

			result.param1 = strtol(pIn, &pNewIn, 10);
			if(pIn == pNewIn)
			{
				// parameter not found
				result.param1 = pVT100Seq->default1;
			}

			pIn = pNewIn;
		}
		else if(*pSeq == 2)	// (optional) parameter 2 to read
		{
			char* pNewIn;

			result.param2 = strtol(pIn, &pNewIn, 10);
			if(pIn == pNewIn)
			{
				// parameter not found
				result.param2 = pVT100Seq->default2;
			}

			pIn = pNewIn;
		}
		else
		{
			if(*pSeq != *pIn)
				return result;

			++pIn;
		}

		++pSeq;
	}

	return result;
}

static uint8_t findEscapeSequence(VT100Result* pResult)
{
	VT100Result result;
	uint16_t numSequences = sizeof(sequences)/sizeof(sequences[0]);

	for(uint16_t i = 0; i < numSequences; i++)
	{
		result = compareSeq(terminal.inputBuffer, sequences+i);
		if(result.code != 0)
		{
			*pResult = result;
			return 1;
		}
	}

	return 0;
}

msg_t TerminalTask(void* params)
{
	(void)params;

	msg_t data;
	uint8_t escapeMode = 0;
	VT100Result vtResult;

	chRegSetThreadName("Terminal");

	chThdSleepMilliseconds(4000);

	while(1)
	{
		data = chIQGetTimeout(terminal.pInQueue, TIME_INFINITE);
		if(data < 0)
			continue;

		if(data > 31 && data < 127)	// readable ASCII char?
		{
			if(terminal.inBufPos < TERMINAL_INPUT_BUFFER_SIZE)
			{
				terminal.inputBuffer[terminal.inBufPos++] = data;
			}

			if(escapeMode)
			{
				if(findEscapeSequence(&vtResult))
				{
					// sequence complete
					if(terminal.escCb)
						(*terminal.escCb)(&vtResult);

					terminal.inBufPos = 0;
					memset(terminal.inputBuffer, 0, TERMINAL_INPUT_BUFFER_SIZE);

					escapeMode = 0;
				}
			}
			else
			{
				chOQPutTimeout(terminal.pOutQueue, data, TIME_IMMEDIATE);
			}
		}

		if(data == 127)	// backspace
		{
			if(terminal.inBufPos > 0)
			{
				terminal.inputBuffer[terminal.inBufPos] = 0;
				--terminal.inBufPos;
				chOQPutTimeout(terminal.pOutQueue, data, TIME_IMMEDIATE);
			}
		}

		if(data == '\n' || data == '\r')
		{
			// command complete, process
			if(terminal.cmdCb)
				(*terminal.cmdCb)(terminal.inputBuffer, terminal.inBufPos);

			terminal.inBufPos = 0;
			memset(terminal.inputBuffer, 0, TERMINAL_INPUT_BUFFER_SIZE);

			escapeMode = 0;
		}

		if(data == 27)	// escape character?
		{
			terminal.inBufPos = 0;
			memset(terminal.inputBuffer, 0, TERMINAL_INPUT_BUFFER_SIZE);

			escapeMode = 1;
		}
	}

	return 0;
}

void TerminalWrite(uint8_t* pData, uint32_t dataLength)
{
	for(uint32_t i = 0; i < dataLength; i++)
	{
		chOQPutTimeout(terminal.pOutQueue, pData[i], TIME_IMMEDIATE);
	}
}
