function [ in, state, out, sampleTimes ] = processCtrlData( logData )
%LOADCTRLDATA Summary of this function goes here
%   Detailed explanation goes here

% define fixed parameters
Fs = 1000;
Ts = 1/Fs;

front = 30*pi/180;
back = 45*pi/180;
botRadius = 0.076;
wheelRadius = 0.0245;
% friction = [1.35,1.02];
friction = [1,1];
theta = [front, pi-front, pi+back, 2*pi-back];
D = [-sin(theta')*friction(1), cos(theta')*friction(2), ones(4,1)*botRadius];
Dx = pinv(D);
DW = [-cos(theta'), sin(theta'), zeros(4,1)];
DWx = pinv(DW);
v_mot_none = null(Dx); % paper description f_k

% extract start and end time
tStart = roundn(logData.ctrl_input.data(1,1)+0.1, -1);
tEnd = roundn(logData.ctrl_input.data(end,1)-0.1, -1);

sampleTimes = (tStart:Ts:tEnd)';
numSamples = length(sampleTimes);

[in.pos.gyr.local, in.vel.gyr.local] = procGyr();
[in.pos.acc.local, in.vel.acc.local, in.acc.acc.local] = procAcc();
[in.mot.vel, in.enc.vel, in.slip.enc, in.vel.enc.local, in.pos.enc.local] = procMotorEncIn();
[in.pos.vis.global, in.pos.vis.local, in.vel.vis.local, in.acc.vis.local] = procVision();
[out.mot.vel, out.enc.vel, out.slip.enc, out.vel.enc.local] = procMotorEncOut();
[out.mot.vol] = procMotorVol();
[in.mot.cur] = procMotorCur();
[state.pos.global, state.vel.global, state.acc.global, state.pos.local, ...
    state.vel.local, state.acc.local] = procState();

function [ orientation, velocity ] = procGyr()
    % col 20 is gyr_updated
    gyrVelData = logData.ctrl_input.data(:,21:24);
    velocity = zeros(numSamples, 3);
    orientation = zeros(numSamples, 3);
    
    [~, ia] = unique(gyrVelData(:,1));
    gyrVelData = gyrVelData(ia, :);
    gyrVelData(:,1) = gyrVelData(:,1)*1e-6;
    velocity(:,1) = interp1(gyrVelData(:,1), gyrVelData(:,2), sampleTimes);
    velocity(:,2) = interp1(gyrVelData(:,1), gyrVelData(:,3), sampleTimes);
    velocity(:,3) = interp1(gyrVelData(:,1), gyrVelData(:,4), sampleTimes);
    
    orientation(:,1) = cumtrapz(sampleTimes, velocity(:,1));
    orientation(:,2) = cumtrapz(sampleTimes, velocity(:,2));
    orientation(:,3) = cumtrapz(sampleTimes, velocity(:,3));
end

function [posGlobal, velGlobal, accGlobal, posLocal, velLocal, accLocal] = procState()
    time = logData.ctrl_state.data(:,1);
    posData = logData.ctrl_state.data(:,3:5);
    velData = logData.ctrl_state.data(:,7:9);
    accData = logData.ctrl_state.data(:,11:13);
    
    [~, ia] = unique(time);	% find unqiue entries
    time = time(ia, :);
    posData = posData(ia, :);	% delete duplicates
    velData = velData(ia, :);	% delete duplicates
    accData = accData(ia, :);	% delete duplicates

    velLocalData = velData;
    accLocalData = accData;
    
    	% Multiturn angle correction for vision data
	correctionTurns = 0;
	newOrient = zeros(length(posData), 1);
	newOrient(1) = posData(1, 3);
	for i = 2:length(posData)
		angleChange = posData(i-1,3)-posData(i,3);
		if angleChange > 1.5*pi
			correctionTurns = correctionTurns + 1;
		end
		if angleChange < -1.5*pi
			correctionTurns = correctionTurns - 1;
		end
		
		newOrient(i) = posData(i,3)+correctionTurns*2*pi;
	end
	posData(:,3) = newOrient;
	
	% Align positions with local frame
	angle = pi/2-posData(1,3);
	rotMat = [cos(angle) -sin(angle); sin(angle) cos(angle)];

	posLocalData = posData - repmat(posData(1,1:3), length(posData), 1); % remove offset
	posLocalData(:,1:2) = (rotMat*(posLocalData(:,1:2)'))';
    
    posGlobal = zeros(numSamples, 3);
    velGlobal = zeros(numSamples, 4);
    accGlobal = zeros(numSamples, 4);
    
    posGlobal(:,1) = interp1(time, posData(:,1), sampleTimes);
    posGlobal(:,2) = interp1(time, posData(:,2), sampleTimes);
    posGlobal(:,3) = interp1(time, posData(:,3), sampleTimes);
    
    velGlobal(:,1) = interp1(time, velData(:,1), sampleTimes);
    velGlobal(:,2) = interp1(time, velData(:,2), sampleTimes);
    velGlobal(:,3) = interp1(time, velData(:,3), sampleTimes);
    
    velGlobal(:,4) = sqrt(velGlobal(:,1).^2+velGlobal(:,2).^2);

    accGlobal(:,1) = interp1(time, accData(:,1), sampleTimes);
    accGlobal(:,2) = interp1(time, accData(:,2), sampleTimes);
    accGlobal(:,3) = interp1(time, accData(:,3), sampleTimes);
    
    accGlobal(:,4) = sqrt(accGlobal(:,1).^2+accGlobal(:,2).^2);
    
    posLocal = zeros(size(posGlobal));
    velLocal = velGlobal;
    accLocal = accGlobal;

    posLocal(:,1) = interp1(time, posLocalData(:,1), sampleTimes);
    posLocal(:,2) = interp1(time, posLocalData(:,2), sampleTimes);
    posLocal(:,3) = interp1(time, posLocalData(:,3), sampleTimes);
    
    velLocal(:,1) = interp1(time, velLocalData(:,1), sampleTimes);
    velLocal(:,2) = interp1(time, velLocalData(:,2), sampleTimes);
    velLocal(:,3) = interp1(time, velLocalData(:,3), sampleTimes);

    accLocal(:,1) = interp1(time, accLocalData(:,1), sampleTimes);
    accLocal(:,2) = interp1(time, accLocalData(:,2), sampleTimes);
    accLocal(:,3) = interp1(time, accLocalData(:,3), sampleTimes);
    
    for i = 1:numSamples
        angle = pi/2-posGlobal(i,3);
		rotMat = [cos(angle) -sin(angle); sin(angle) cos(angle)];
		velLocal(i,1:2) = (rotMat*(velLocal(i,1:2)'))';
        accLocal(i,1:2) = (rotMat*(accLocal(i,1:2)'))';
    end
end

function [position, velocity, accel] = procAcc()
    % col 25 is acc_updated
    accData = logData.ctrl_input.data(:,26:29);
    position = zeros(numSamples, 3);
    velocity = zeros(numSamples, 4);
    accel = zeros(numSamples, 3);
    
    [~, ia] = unique(accData(:,1));
    accData = accData(ia, :);
    accData(:,1) = accData(:,1)*1e-6;
    accel(:,1) = interp1(accData(:,1), accData(:,2), sampleTimes);
    accel(:,2) = interp1(accData(:,1), accData(:,3), sampleTimes);
    accel(:,3) = interp1(accData(:,1), accData(:,4), sampleTimes);
	
	offsets = mean(accel(1:1000,:));
	accel(:,1) = accel(:,1) - offsets(1);
	accel(:,2) = accel(:,2) - offsets(2);
	accel(:,3) = accel(:,3) - offsets(3);
    
    for i = 1:numSamples
		angle = in.pos.gyr.local(i,3);
		rotMat = [cos(angle) -sin(angle); sin(angle) cos(angle)];
		accel(i,1:2) = (rotMat*(accel(i,1:2)'))';
    end
    
    velocity(:,1) = cumtrapz(sampleTimes, accel(:,1));
    velocity(:,2) = cumtrapz(sampleTimes, accel(:,2));
    velocity(:,3) = cumtrapz(sampleTimes, accel(:,3));
    
    velocity(:,4) = sqrt(velocity(:,1).^2+velocity(:,2).^2);
    
    position(:,1) = cumtrapz(sampleTimes, velocity(:,1));
    position(:,2) = cumtrapz(sampleTimes, velocity(:,2));
    position(:,3) = cumtrapz(sampleTimes, velocity(:,3));
	
	accel(:,1) = smooth(accel(:,1), 100);
	accel(:,2) = smooth(accel(:,2), 100);
	accel(:,3) = smooth(accel(:,3), 100);
end

function [motVel, encVel, encSlip, localVel, posLocal] = procMotorEncIn()
    % col 11 is vel_updated
    motorVelCurData = logData.ctrl_input.data(:,12:19);
    motVel = zeros(numSamples, 4);

    for m = 1:4
        motorVelCurData(:,m) = motorVelCurData(:,m)*1e-6;
        [~, ia] = unique(motorVelCurData(:,m));
        mOrig = [motorVelCurData(ia, m) motorVelCurData(ia, m+4)];
        motVel(:,m) = interp1(mOrig(:,1), mOrig(:,2), sampleTimes);
    end
    
    encVel = motVel*1/(2*pi)*15/50*2*wheelRadius*pi;
    
    encSlip = zeros(numSamples,1);
    for i = 1:numSamples
        encSlip(i) = dot(encVel(i,:), v_mot_none);
    end
    
    localVel = zeros(numSamples, 4);
    for i = 1:numSamples
        localVel(i,1:3) = Dx*encVel(i,:)';
    end
    
    localVel(:,4) = sqrt(localVel(:,1).^2+localVel(:,2).^2);
    
    posLocal = zeros(numSamples, 3);
	posLocal(:,3) = cumtrapz(sampleTimes, localVel(:,3));
	
	fixedVel = zeros(numSamples, 2);
	for i = 1:numSamples
		angle = in.pos.gyr.local(i,3);
		rotMat = [cos(angle) -sin(angle); sin(angle) cos(angle)];
		fixedVel(i,:) = (rotMat*(localVel(i,1:2)'))';
	end
	
    posLocal(:,1) = cumtrapz(sampleTimes, fixedVel(:,1));
    posLocal(:,2) = cumtrapz(sampleTimes, fixedVel(:,2));
end

function [motVelSet, encVelSet, encSlip, xyw] = procMotorEncOut()
    motorVelSetData = logData.ctrl_output.data(:,2:5);
    motVelSet = zeros(numSamples, 4);

    if size(logData.ctrl_output.data,1) > 0
      for m = 1:4
          motVelSet(:, m) = interp1(logData.ctrl_output.data(:,1), motorVelSetData(:,m), sampleTimes);
      end
    end
    
    encVelSet = motVelSet*1/(2*pi)*15/50*2*wheelRadius*pi;

    encSlip = zeros(numSamples,1);
    for i = 1:numSamples
        encSlip(i) = dot(encVelSet(i,:), v_mot_none);
    end
    
    xyw = zeros(numSamples, 4);
    for i = 1:numSamples
        xyw(i,1:3) = Dx*encVelSet(i,:)';
    end
    
    xyw(:,4) = sqrt(xyw(:,1).^2+xyw(:,2).^2);
end

function [voltage] = procMotorVol()
    % col 37 is vol_m1_t
    motorVolSetData = logData.ctrl_input.data(:,37:44);
    voltage = zeros(numSamples, 4);

    for m  = 1:4
        [~, ia] = unique(motorVolSetData(:,m));
        mOrig = [motorVolSetData(ia, m) motorVolSetData(ia, m+4)];
        mOrig(:,1) = mOrig(:,1)*1e-6;
%        motorVolSetData(:,m) = motorVolSetData(:,m)*1e-6;
        voltage(:,m) = interp1(mOrig(:,1), mOrig(:,2), sampleTimes);
    end
end

function [current] = procMotorCur()
    % col 2 is cur_updated
    motorCurData = logData.ctrl_input.data(:,3:10);
    current = zeros(numSamples, 4);

    for m = 1:4
        [~, ia] = unique(motorCurData(:,m));
        mOrig = [motorCurData(ia, m) motorCurData(ia, m+4)];
        mOrig(:,1) = mOrig(:,1)*1e-6;
        current(:,m) = interp1(mOrig(:,1), mOrig(:,2), sampleTimes);
    end
end

function [posGlobal, posLocal, velLocal, accLocal] = procVision()
    % col 30 is pos_updated
    visionPosData = logData.ctrl_input.data(:,31:34);	% input
    posGlobal = zeros(numSamples, 3);	% output
    posLocal = zeros(numSamples, 3);	% output
    velLocal = zeros(numSamples, 4);	% output
	accLocal = zeros(numSamples, 4);	% output
    [~, ia] = unique(visionPosData(:,1));	% find unqiue entries
    visionPosData = visionPosData(ia, :);	% delete duplicates
    visionPosData(:,1) = visionPosData(:,1)*1e-6;	% scale time to [s]
    visionPosLocalData = zeros(size(visionPosData));
    
	if length(visionPosData) < 10
        return;
	end
	
	% Multiturn angle correction for vision data
	correctionTurns = 0;
	newOrient = zeros(length(visionPosData), 1);
	newOrient(1) = visionPosData(1, 4);
	for i = 2:length(visionPosData)
		angleChange = visionPosData(i-1,4)-visionPosData(i,4);
		if angleChange > 1.5*pi
			correctionTurns = correctionTurns + 1;
		end
		if angleChange < -1.5*pi
			correctionTurns = correctionTurns - 1;
		end
		
		newOrient(i) = visionPosData(i,4)+correctionTurns*2*pi;
	end
	visionPosData(:,4) = newOrient;
	
	% Align positions with local frame
	angle = pi/2-visionPosData(1,4);
	rotMat = [cos(angle) -sin(angle); sin(angle) cos(angle)];

	visionPosLocalData(:,1) = visionPosData(:,1);
	visionPosLocalData(:,2:4) = visionPosData(:,2:4) - repmat(visionPosData(1,2:4), length(visionPosData), 1);
	visionPosLocalData(:,2:3) = (rotMat*(visionPosLocalData(:,2:3)'))';
    
    for i = 1:3
        posGlobal(:,i) = interp1(visionPosData(:,1), visionPosData(:,i+1), sampleTimes);
        posLocal(:,i) = interp1(visionPosLocalData(:,1), visionPosLocalData(:,i+1), sampleTimes);
        velLocal(:,i) = [0; diff(posGlobal(:,i))/Ts];
        velLocal(:,i) = smooth(velLocal(:,i), 50);
        accLocal(:,i) = [0; diff(velLocal(:,i))/Ts];
        accLocal(:,i) = smooth(accLocal(:,i), 100);
    end
    
    velLocal(:,4) = sqrt(velLocal(:,1).^2+velLocal(:,2).^2);
	accLocal(:,4) = sqrt(accLocal(:,1).^2+accLocal(:,2).^2);
    
    for i = 1:numSamples
        angle = pi/2-posGlobal(i,3);
		rotMat = [cos(angle) -sin(angle); sin(angle) cos(angle)];
		velLocal(i,1:2) = (rotMat*(velLocal(i,1:2)'))';
        accLocal(i,1:2) = (rotMat*(accLocal(i,1:2)'))';
    end
end

end